/*******************************************************************************
+
+  LEDA 5.0  
+
+
+  agd_server_old.c
+
+
+  Copyright (c) 1995-2004
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-----------------------------------------------------------------------------
//
//    agd_server.c
//
//    authors: G. Koch & S. Naeher
//    last modified:  Feb 2000
//
//-----------------------------------------------------------------------------

#include <LEDA/graphics/graphwin.h>
#include <LEDA/system/file.h>
#include <LEDA/system/stream.h>
#include <LEDA/internal/std/ctype.h>
#include <LEDA/internal/std/stdio.h>

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cerr;
using std::endl;
using std::ostream;
using std::istream;
#endif


#include <AGD/FPPLayout.h>
#include <AGD/SchnyderLayout.h>
#include <AGD/ConvexLayout.h>
#include <AGD/ConvexDrawLayout.h>
#include <AGD/PlanarStraightLayout.h>
#include <AGD/PlanarDrawLayout.h>
#include <AGD/MixedModelLayout.h>
#include <AGD/TreeLayout.h>
#include <AGD/PlanarizationLayout.h>
#include <AGD/SugiyamaLayout.h>
#include <AGD/SpringLayout.h>
#include <AGD/TutteLayout.h>
#include <AGD/PureOrthogonalLayout.h>
#include <AGD/QuasiOrthogonalLayout.h>
#include <AGD/OrthogonalLayout.h>
#include <AGD/VisibilityRepresentation.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/MaxAllMapper.h>

#include <AGD/FastHierarchyLayout.h>
#include <AGD/DfsRanking.h>
#include <AGD/LongestPathRanking.h>
#include <AGD/BarycenterHeuristic.h>
#include <AGD/MedianHeuristic.h>
#include <AGD/GreedyInsertHeuristic.h>
#include <AGD/GreedySwitchHeuristic.h>
#include <AGD/SplitHeuristic.h>

#include <AGD/SubgraphPlanarizer.h>
#include <AGD/PlanarSubgraph.h>
#include <AGD/ShortestPathInserter.h>
//#include <AGD/MaximumPlanarSubgraph.h>


  struct agd_serv_intern_struct_def {
   	string string_name;
   	union {
           int int_wert;
           bool bool_wert;
           double double_wert;
           float float_wert;
        } wert;      
        string  string_wert;
   	list <string> enum_list;
  };

inline ostream& operator<<(ostream& strm,const agd_serv_intern_struct_def&) 
{ return strm; }
      
inline istream& operator>>(istream& strm,const agd_serv_intern_struct_def&) 
{ return strm; }

  struct agd_serv_struct_def {
   	list <agd_serv_intern_struct_def > optenum_list;
   	list <agd_serv_intern_struct_def > optbool_list;
   	list <agd_serv_intern_struct_def > optint_list;
   	list <agd_serv_intern_struct_def > optdouble_list;
   	list <agd_serv_intern_struct_def > optfloat_list;
   	list <agd_serv_intern_struct_def > optstring_list;
   	list <agd_serv_intern_struct_def > optinfo_list;        
   	string fkt;
   	int integer;
  };

inline ostream& operator << (ostream& strm,const agd_serv_struct_def&) 
{ return strm; }
      
inline istream& operator >> (istream& strm,const agd_serv_struct_def&) 
{ return strm; }

class agd_serv_base {


protected :

  //GraphWin * gw;

  list <agd_serv_struct_def > fkt_opt_list;
  list <string> agd_function_list;
  list <string> agd_option_list;
  string optfile;

  bool read_string(ifstream& opt,string& s);
  int  read_element(ifstream& opt);
  bool read_enum(ifstream& opt,agd_serv_struct_def& agd_serv_struct,string& s);
  bool parser_opt(ifstream& opt);
  bool read_optfile(string optfile);

public : 
  
  agd_serv_base() {}
 ~agd_serv_base();

};


agd_serv_base::~agd_serv_base() {}

int agd_serv_base::read_element(ifstream& opt) 
{ 
  char c;
  do opt.get(c); while (opt && isspace(c));

  if (!opt.fail())
  { opt.putback(c);
    if (c == '[')   return 1; // list or structure 
    if (c == '"')   return 2; // string
    if (isdigit(c)) return 3; // number (int or float)
    if (c == ']')   return 4; // end
   }

  return 0;
}

   
bool agd_serv_base::read_string(ifstream& opt,string& s) 
{
  string s1;
  opt >> s1;
  while (s1.pos("\"")==-1) return false;
  s = s1.del("\"");
  while ((s.pos("\"") == -1) && (!opt.fail())) {
     opt >> s1;
     s = s+ " " + s1;
  }
  s = s.del_all("\"");
  if (opt.fail()) return false;
  return true;
}
   

bool agd_serv_base::read_enum(ifstream& opt,
                              agd_serv_struct_def& agd_serv_struct,string& s)
{
  string s1;

  agd_serv_intern_struct_def agd_intern_struct;
  agd_intern_struct.string_name = s;
  opt >> s;					//bezeichner
  opt >> s;
  opt >> agd_intern_struct.wert.int_wert; 	//vorauswahl
  opt >> s;
  opt >> s;
  if (s.pos("[") == -1) return false;
  opt >> s;	
  while (s.pos("]") == -1 && !opt.fail()) 
  { // auswahl
    if (read_string(opt,s))
       agd_intern_struct.enum_list.append(s);
    else 
       return false;
    opt >> s;
  }
  opt >> s;
  agd_serv_struct.optenum_list.append(agd_intern_struct);
  if (opt.fail()) return false;
  return true;   
}	
  
   
bool agd_serv_base::parser_opt(ifstream& opt) 
{
 string s,s1,s2,s3; 
 agd_serv_intern_struct_def agd_intern_struct;
  
 opt >> s;
 if (s.pos("[")== -1) return false;	
 opt >> s;	
 while (s.pos("]") == -1)
 { agd_serv_struct_def agd_serv_struct;   
   read_string (opt,s1);
   agd_serv_struct.fkt = s1;		//fktname	
   opt >> s;  				
   if (s.pos('[') == -1 || opt.fail()) return false; // struktur anfang	
   opt >> s; 	
   while (s.pos("]") == -1)
   { if (s == "infos") 
     { opt >> s3;
       if (s3.pos("[")== -1) return false;
       opt >> s3;
       while ((s3.pos("]") == -1) && (!opt.fail())) 
       { agd_intern_struct.string_name = s3;        
         if (! read_string(opt,s3)) return false;
         agd_intern_struct.string_wert = s3;
         agd_serv_struct.optinfo_list.append(agd_intern_struct);
         opt >> s3;
        }
      } 
      else 
      { 
        switch (read_element(opt)) {

         case 0: return 0;
      
         case 1: { if (!read_enum(opt,agd_serv_struct,s)) return 0;
                   break; 
                  }

         case 2: { if (! read_string(opt,s3)) return false;
                   agd_intern_struct.string_name = s;
                   if (s3=="true") {
                      agd_intern_struct.wert.bool_wert = true;
                      agd_serv_struct.optbool_list.append(agd_intern_struct);
                   } else {
                      if (s3 == "false") {
                         agd_intern_struct.wert.bool_wert = false;
                         agd_serv_struct.optbool_list.append(agd_intern_struct);
                      } else {
                         if (! read_string(opt,s3)) return 0;
                         agd_intern_struct.string_wert = s3;
                         agd_serv_struct.optstring_list.append(agd_intern_struct);
                      }
                   }
                   break;
                 }
           
         case 3: { // number
                   agd_intern_struct.string_name = s;
                   string num_str;
                   opt >> num_str;
                   if (num_str.pos('.') > -1)
             	     agd_intern_struct.wert.double_wert = atof(num_str);
                   else
             	     agd_intern_struct.wert.int_wert = atoi(num_str);
                   agd_serv_struct.optdouble_list.append(agd_intern_struct);
                   break; 
                  }	      
        }
      }
     opt >> s;
   }
  opt >> s;
  fkt_opt_list.append(agd_serv_struct);
 }

 if (opt.fail()) return false;
 return true;
}

bool agd_serv_base::read_optfile(string ofile) {
ifstream opt (ofile);
if (!opt) return false;
string s;
   opt >> s;
   while ((s != "AGD_option_list") && (!opt.fail()))  opt >> s;
   if (opt.fail()) return false;
   return parser_opt(opt);
}





class agd_serv_class : agd_serv_base {

private:

   FPPLayout	 		fpp;
   SchnyderLayout   		schnyder;
   ConvexLayout     		convex;
   ConvexDrawLayout 		convexdraw;
   PlanarStraightLayout 	planarstraight;
   PlanarDrawLayout     	planardraw;
   MixedModelLayout  		mixedmodel;
   TreeLayout 			tree;
   PlanarizationLayout	        planarization;
   SugiyamaLayout		sugiyama;
   SpringLayout 		spring;
   TutteLayout 			tutte;
   VisibilityRepresentation 	visibility;
   PureOrthogonalLayout         pureortho;
   QuasiOrthogonalLayout        quasiortho;
   OrthogonalLayout             orthogonal;
   FastHierarchyLayout 		hierarchy;


   int    argc;
   char** argv;

   string in_file_type;
   string out_file_type;

   bool err;
   bool infos;

   GraphWin gw;

   void make_std_infos (ostream& opt,LayoutModule& F);
   void make_infos_only (ostream& opt,LayoutModule& F);

   void make_fpp (ostream& opt);
   void make_schnyder (ostream& opt);
   void make_convex (ostream& opt);
   void make_convexdraw (ostream& opt);        
   void make_planarstraight (ostream& opt);  
   void make_planardraw (ostream& opt);  
   void make_mixedmodel (ostream& opt);
   void make_tree (ostream& opt);
   void make_sugiyama (ostream& opt);
   void make_spring (ostream& opt);
   void make_tutte (ostream& opt);
   void make_visibility (ostream& opt);
   void make_pureortho (ostream& opt);
   void make_quasiortho (ostream& opt);
   void make_orthogonal (ostream& opt);
   void make_planarization (ostream& opt);

   void search_struct(LayoutModule& F,agd_serv_struct_def& opt_fkt_struct);

   bool call_function (LayoutModule& L);
   bool call_grid_function(GridLayoutModule& F);

   bool call_fpp(GridLayoutModule& F); 
   bool call_schnyder(GridLayoutModule& F);
   bool call_mixedmodel(GridLayoutModule& F);
   bool call_planarstraight(GridLayoutModule& F);
   bool call_sugiyama(LayoutModule& F);
   bool call_convex(GridLayoutModule& F);
   bool call_convexdrawraw(GridLayoutModule& F);
   bool call_planardraw(GridLayoutModule& F);
   bool call_tree(LayoutModule& F);
   bool call_spring(LayoutModule& F);
   bool call_tutte(LayoutModule& F);
   bool call_pureortho(LayoutModule& F);
   bool call_quasiortho(LayoutModule& F);
   bool call_orthogonal(LayoutModule& F);
   bool call_visibility(GridLayoutModule& F);
   bool call_planarization(LayoutModule& F);

   void list_functions(); 
   void list_functions(ostream&);

   bool error_msg(string  msg);
   
   string bool_str(bool b) 
   { if (b) return "\"true\"";
     else return "\"false\"" ;
    }
   
   bool read_graph();
   bool write_graph(node_array<int>& n_index,
                    edge_array<int>& e_index, graph& G);

public:

agd_serv_class(int argc0, char* argv0[]): gw(string("%s(%s)",argv0[0],argv0[1]))
{ 
  if (argc0 > 4) 
     in_file_type = argv0[4];
  else
     in_file_type = "gw";

  if (argc0 > 5) 
    out_file_type = argv0[5];
  else
    out_file_type = "gw";

  argc = argc0 - 1; 
  argv = argv0;
  infos = false;
}


~agd_serv_class() {}

void init(int argc0, char* argv0[]) 
{ argc = argc0 - 1; 
  argv = argv0;
}   
   
void init(int argc0, char* argv0[], string i_type, string o_type) 
{ in_file_type  = i_type;
  out_file_type = o_type;
  argc = argc0 - 1; 
  argv = argv0;
} 

void file_typ(string i_type, string o_type) 
{ in_file_type  = i_type;
  out_file_type = o_type;   
}

string name()         const { return "agd_server";    }
string long_name()    const { return "agd_server";    }
string author()       const { return "Gunter Koch and Stefan Naeher"; }
string impl_author()  const { return "Gunter Koch";   }
string impl_version() const { return "1.4";           }
string impl_date()    const { return "December 1998"; }
   
bool agd_serv();

};




/**************************** make_std_infos ****************************/

void agd_serv_class::make_std_infos (ostream& opt,LayoutModule& F)
{ opt << "\t\twork_infos " << bool_str(infos) << endl;
   opt << "\t\tinfos [" << endl;
      opt << "\t\t\tname: \""    << F.long_name()    << "\"" <<  endl;
      opt << "\t\t\tauthor: \""  << F.author()       << "\"" << endl;
      opt << "\t\t\tImplementation: \"" << F.impl_author() << "\"" << endl;
      opt << "\t\t\tdate: \""    << F.impl_date()    << "\"" << endl;
      opt << "\t\t\tversion: \"" << F.impl_version() << "\"" << endl;
    opt << "\t\t]" << endl;
}


void agd_serv_class::make_infos_only (ostream& opt,LayoutModule& F) 
{ opt << "\tfunction \"" <<  F.name() << "\" ["<< endl;
  make_std_infos (opt,F);       
  opt << "\t]" << endl;     
}



void agd_serv_class::list_functions() 
{ if (argc > 1) 
  { ofstream out(argv[2]);
    if (out) list_functions(out);
    return;
   }
  list_functions(cout);
}

   

void agd_serv_class::list_functions(ostream& out) 
{
  out << "Creator \"AGD_SERVER\""<< endl;
  out << "Version " << impl_version() << endl;
  out << "AGD_function_list [" << endl;
  out << "\tfunction \"" << fpp.name()            << "\"" << endl;
  out << "\tfunction \"" << schnyder.name()       << "\"" << endl;
  out << "\tfunction \"" << convex.name()         << "\"" << endl;
  out << "\tfunction \"" << convexdraw.name()     << "\"" << endl;
  out << "\tfunction \"" << planarstraight.name() << "\"" << endl;
  out << "\tfunction \"" << planardraw.name()     << "\"" << endl;
  out << "\tfunction \"" << mixedmodel.name()     << "\"" << endl;
  out << "\tfunction \"" << pureortho.name()      << "\"" << endl;
  out << "\tfunction \"" << quasiortho.name()     << "\"" << endl;
  out << "\tfunction \"" << orthogonal.name()     << "\"" << endl;
  out << "\tfunction \"" << tree.name()           << "\"" << endl;
  out << "\tfunction \"" << sugiyama.name()       << "\"" << endl;
  out << "\tfunction \"" << planarization.name()  << "\"" << endl;
  out << "\tfunction \"" << spring.name()         << "\"" << endl;
  out << "\tfunction \"" << tutte.name()          << "\"" << endl;
  out << "\tfunction \"" << visibility.name()     << "\"" << endl;
  out << "]" << endl;
  
  out.setf(ios::showpoint);

  out << "AGD_option_list [" << endl;    

  make_fpp (out);
  make_schnyder (out);
  make_convex (out);
  make_convexdraw (out);        
  make_planarstraight (out);  
  make_planardraw (out);  
  make_mixedmodel (out);
  make_pureortho (out);
  make_quasiortho (out);
  make_orthogonal (out);
  make_tree (out);
  make_sugiyama (out);
  make_planarization(out);
  make_spring (out);
  make_tutte (out);
  make_visibility (out);
  out << "]" << endl;
}



//****************************** call_function ******************************/


bool agd_serv_class::read_graph() 
{
 if (in_file_type == "gml") 
 { if (gw.read_gml(argv[2])) 
   { error_msg("error reading gml file.");
     return false;
    }
   return true;
 }

 if (in_file_type == "gw") 
 { if (gw.read (argv[2])) 
   { error_msg("error reading gw file.");
     return false;
    } 
   return true;
 }

 if (in_file_type == "layout") 
 { ifstream in (argv[2]);
   if(gw.load_layout (in)) 
   { error_msg("error reading gw layout.");
     return false;
    } 
   return true;
  }

  error_msg("illegal file format."); 
  return false;
}


bool agd_serv_class::write_graph(node_array<int>& n_index,
                                 edge_array<int>& e_index , graph& G) 
{
  if (out_file_type == "gml") return gw.save_gml(argv[2]);
  if (out_file_type == "gw")  return gw.save_gw(argv[2]);
  if (out_file_type == "ps")  return gw.save_ps(argv[2]);

  if (out_file_type == "layout") 
  { ofstream out (argv[2]);
    G.sort_nodes(n_index);
    G.sort_edges(e_index);
    gw.save_layout(out);
    return true;
   }

  error_msg("illegal file format."); 
  return false;
}
        



bool agd_serv_class::call_function (LayoutModule& F) 
{
  if (!read_graph()) return false;

  delete_file(argv[2]);
  
  graph& G = gw.get_graph();

  if (&F == (LayoutModule*)&orthogonal || &F == (LayoutModule*)&visibility)
  { node v;
    forall_nodes(v,G) gw.set_shape(v,rectangle_node);
   }

  node_array<int> n_index(G);
  edge_array<int> e_index(G);

  int i = 0;
  node v;
  forall_nodes(v,G) 
  { n_index[v] = index(v);
    edge e;
    forall_adj_edges(e,v) e_index[e] = i++;
   }  

  GraphWinInterface I (gw);

  AgdKey agd_key;

  if (!F.check(gw.get_graph(),agd_key)) 
  { error_msg("Unsatisfied graph property: " + agd_key.info());
    return false;
   }

#if defined(__unix__) && defined(__GNUC__)

  if (infos && argc > 2 && is_file(argv[3]))
    { // write work infos into file argv[3] 
      ofstream ans (argv[3], ios::trunc | ios::app);
      ans << "AGD_Information\t\"";
      streambuf* str_buf = cout.rdbuf();
      cout.rdbuf(ans.rdbuf()); 
      F.call(gw.get_graph(),I); 
      ans << endl; 
     }
  else

#endif

  F.call(gw.get_graph(),I); 

  return write_graph(n_index,e_index,G);
}


bool agd_serv_class::call_grid_function(GridLayoutModule& F) 
{ MaxAllMapper mapper (40,2,20,DPoint(10,10));
  F.set_mapper(mapper);
  return call_function(F);
}



void agd_serv_class::search_struct(LayoutModule& F, agd_serv_struct_def& result)
{ agd_serv_struct_def x;
  forall(x,fkt_opt_list) 
     if(x.fkt == F.name()) result = x;
}



bool agd_serv_class::error_msg(string  msg) 
{ if (argc > 1) 
  { ofstream out (argv[2]);
    if (!out)  return false;
    out << "AGD_error_file\t\"";
    out << msg; 
    out << "\"\n";
    out.close(); 
    return true;
  } 
  return false;
}



bool agd_serv_class::agd_serv() 
{
  if (argc < 1 ) return false;
  string func = argv[1];

  infos = false;
  
  if (func == "list" || func == "check") 
  { list_functions();
    return true;
   }
  
  if (argc < 3 ) 
  { error_msg("error in parameterlist");
    return false;
  }

  bool ok = true;
	
       if (func == fpp.name())        ok = call_fpp(fpp);	     
  else if (func == schnyder.name())   ok = call_schnyder(schnyder);
  else if (func == convex.name())     ok = call_convex(convex);
  else if (func == convexdraw.name()) ok = call_convexdrawraw(convexdraw);
  else if (func == planarstraight.name())  
                                      ok = call_planarstraight(planarstraight);
  else if (func == planardraw.name()) ok = call_planardraw(planardraw);
  else if (func == mixedmodel.name()) ok = call_mixedmodel(mixedmodel);
  else if (func == tree.name())       ok = call_tree(tree);
  else if (func == planarization.name())
                                      ok = call_planarization(planarization);
  else if (func == sugiyama.name())   ok = call_sugiyama(sugiyama);
  else if (func == spring.name())     ok = call_spring(spring);
  else if (func == tutte.name())      ok = call_tutte(tutte);
  else if (func == visibility.name()) ok = call_visibility(visibility);
  else if (func == pureortho.name())  ok = call_pureortho(pureortho);
  else if (func == quasiortho.name()) ok = call_quasiortho(quasiortho);
  else if (func == orthogonal.name()) ok = call_orthogonal(orthogonal);
  else error_msg("unknown algorithm.");

  ofstream opt (argv[3],ios::app);

  opt << endl;
  opt << "Creator " << "\"AGD_SERVER\"" << endl;
  opt << "Version " << impl_version()   << endl;
  opt.close();

  return ok;  
}




/************************* de Fraysseix/Pach/Pollack**********************/

void agd_serv_class::make_fpp(ostream& opt) {
   	make_infos_only (opt,fpp);  
}


bool agd_serv_class::call_fpp(GridLayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    infos = opt_fkt_struct.optbool_list[it].wert.bool_wert;
  }
  return call_grid_function(F);
}


                       
/************************* SchnyderLayout *******************************/

void agd_serv_class::make_schnyder (ostream& opt) {
   	make_infos_only (opt,schnyder);  
}


bool agd_serv_class::call_schnyder(GridLayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it = opt_fkt_struct.optbool_list.first();
    infos = opt_fkt_struct.optbool_list[it].wert.bool_wert;
   }
  return call_grid_function(F);
}


/************************* ConvexLayout *******************************/

void agd_serv_class::make_convex(ostream& opt) 
{ opt << "\tfunction \"" <<  convex.name() << "\" ["<< endl;
  opt << "\t\tsize_opt " << bool_str(convex.size_opt()) << endl;  
  opt << "\t\tbase_ratio " << convexdraw.base_ratio() << endl;                     
  make_std_infos (opt,convex); 
  opt << "\t]" << endl;
}


bool agd_serv_class::call_convex(GridLayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it = opt_fkt_struct.optbool_list.first();
    convex.size_opt (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optdouble_list.first();
    convex.base_ratio (opt_fkt_struct.optdouble_list[it].wert.double_wert);
   }
  return call_grid_function(F);
}

/************************* ConvexDrawLayout *****************************/

void agd_serv_class::make_convexdraw(ostream& opt) {                

   opt << "\tfunction \"" << convexdraw.name() << "\" [" << endl;
      opt << "\t\tsize_opt " << bool_str(convexdraw.size_opt()) << endl;      
      opt << "\t\ttside_opt " << bool_str(convexdraw.side_opt()) << endl;
      opt << "\t\tbase_ratio " << convexdraw.base_ratio() << endl;       
        make_std_infos (opt,convexdraw); 
   opt << "\t]" << endl;        
 
}


bool agd_serv_class::call_convexdrawraw(GridLayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    convexdraw.size_opt (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    convexdraw.side_opt (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optdouble_list.first();
    convexdraw.base_ratio (opt_fkt_struct.optdouble_list[it].wert.double_wert);
   }
  return call_grid_function(F);
}


/*********************** PlanarStraightLayout **************************/

void agd_serv_class::make_planarstraight(ostream& opt) 
{ opt << "\tfunction \"" << planarstraight.name() << "\" [" << endl;
    opt << "\t\tsize_opt " << bool_str(planarstraight.size_opt()) << endl; 
    opt << "\t\tmax_outer " << bool_str(planarstraight.max_outer()) << endl; 
    opt << "\t\tbase_ratio " << planarstraight.base_ratio() << endl;
    make_std_infos (opt,planarstraight); 
  opt << "\t]" << endl;          
}


bool agd_serv_class::call_planarstraight(GridLayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    planarstraight.size_opt (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    planarstraight.max_outer (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optdouble_list.first();
    planarstraight.base_ratio (opt_fkt_struct.optdouble_list[it].wert.double_wert);
   }
   return call_grid_function(F);
}

/************************ PlanarDrawLayout ****************************/

void agd_serv_class::make_planardraw(ostream& opt)
{ opt << "\tfunction \"" << planardraw.name() << "\" [" << endl;
    opt << "\t\tsize_opt " << bool_str(planardraw.size_opt()) << endl;
    opt << "\t\tside_opt " << bool_str(planardraw.side_opt()) << endl;
    opt << "\t\tmax_outer " << bool_str (planardraw.max_outer()) << endl;
    opt << "\t\tbase_ratio " << planardraw.base_ratio() << endl;
    make_std_infos (opt,planardraw); 
  opt << "\t]" << endl;          
}


bool agd_serv_class::call_planardraw(GridLayoutModule& F) 
{
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    planardraw.size_opt (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    planardraw.side_opt (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    planardraw.max_outer (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optdouble_list.first();
    planardraw.base_ratio (opt_fkt_struct.optdouble_list[it].wert.double_wert);
   }
   return call_grid_function(F);
}

/************************ MixedModelLayout *****************************/

void agd_serv_class::make_mixedmodel(ostream& opt)
{ opt << "\tfunction \"" << mixedmodel.name() << "\" [" << endl;
    opt << "\t\tbase_ratio " << mixedmodel.base_ratio() << endl;
    make_std_infos (opt,mixedmodel); 
  opt << "\t]" << endl;          
}


bool agd_serv_class::call_mixedmodel(GridLayoutModule& F) 
{
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optdouble_list.first();
    mixedmodel.base_ratio ((opt_fkt_struct.optdouble_list[it]).wert.double_wert);
   }
  return call_grid_function(F);
}

/************************* TreeLayout *******************************/

void agd_serv_class::make_tree(ostream& opt) 
{ opt << "\tfunction \"" << tree.name() << "\" [" << endl; 
    opt << "\t\tsib_separation " << tree.sib_separation() << endl;
    opt << "\t\tsubtree_separation " << tree.subtree_separation() << endl;
    opt << "\t\tlevel_separation " << tree.level_separation() << endl;
    opt << "\t\tleft_margin " << tree.left_margin() << endl;
    opt << "\t\tupper_margin " << tree.upper_margin() << endl;
    make_std_infos (opt,tree); 
  opt << "\t]" << endl;          
}


bool agd_serv_class::call_tree(LayoutModule& F) 
{
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
        
    search_struct(F, opt_fkt_struct);
        
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    infos = opt_fkt_struct.optbool_list[it].wert.bool_wert;

    it = opt_fkt_struct.optdouble_list.first();
    tree.sib_separation(opt_fkt_struct.optdouble_list[it].wert.double_wert); 

    it = opt_fkt_struct.optdouble_list.succ(it);
    tree.subtree_separation(opt_fkt_struct.optdouble_list[it].wert.double_wert);

    it = opt_fkt_struct.optdouble_list.succ(it);
    tree.level_separation (opt_fkt_struct.optdouble_list[it].wert.double_wert);

    it = opt_fkt_struct.optdouble_list.succ(it);
    tree.left_margin (opt_fkt_struct.optdouble_list[it].wert.double_wert);

    it = opt_fkt_struct.optdouble_list.succ(it);
    tree.upper_margin (opt_fkt_struct.optdouble_list[it].wert.double_wert);
   }
  return call_function(F);
}

/************************* SugiyamaLayout ******************************/


void agd_serv_class::make_sugiyama(ostream& opt)
{
   int  heuristic = 0; 
   bool sugiyama_infos = false;
   int  ranking = 0;
   int  cross_min = 1;

   opt << "\tfunction \"" << sugiyama.name() << "\" [" << endl;

      opt << "\t\tranking [" << endl;
         opt << "\t\t\tranking_wahl " << ranking << endl;
         opt << "\t\t\tranking_lst [" << endl;
            opt << "\t\t\t\tlst \"LongestPath\"" << endl;
            opt << "\t\t\t\tlst \"DFS\"" << endl;
         opt <<"\t\t\t]" << endl;
      opt << "\t\t]" << endl;    

      opt << "\t\theuristic [" << endl;
         opt << "\t\t\theuristic_wahl " << heuristic << endl;
         opt << "\t\t\theuristic_lst [" << endl;                  
            opt << "\t\t\t\tlst \"Split\"" << endl;
            opt << "\t\t\t\tlst \"Barycenter\"" << endl;
            opt << "\t\t\t\tlst \"Median\"" << endl;
            opt << "\t\t\t\tlst \"Greedy Insert\"" << endl;
            opt << "\t\t\t\tlst \"Greedy Switch\"" << endl;
         opt <<"\t\t\t]" << endl;
      opt << "\t\t]" << endl;    

      opt << "\t\tfixed_level_distance " << 
                	bool_str(hierarchy.fixed_layer_distance()) << endl; 
      opt << "\t\tnode_distance " <<  hierarchy.node_distance() << endl;
      opt << "\t\tlevel_distance " << hierarchy.layer_distance() << endl;
      make_std_infos (opt,sugiyama); 
   opt << "\t]" << endl;          
}


bool agd_serv_class::call_sugiyama(LayoutModule& F) 
{
   int ranking;
   int heuristic;
   int yu;

   if (argc > 2 && read_optfile(argv[3]))
   {
      agd_serv_struct_def opt_fkt_struct;
      search_struct(F, opt_fkt_struct);

      list_item it;
      it = opt_fkt_struct.optbool_list.first();
      hierarchy.fixed_layer_distance(opt_fkt_struct.optbool_list[it].wert.bool_wert);
      it = opt_fkt_struct.optbool_list.succ(it);
      infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);

      it = opt_fkt_struct.optdouble_list.first();
      hierarchy.node_distance(opt_fkt_struct.optdouble_list[it].wert.int_wert);

      it = opt_fkt_struct.optdouble_list.succ(it);
      hierarchy.layer_distance(opt_fkt_struct.optdouble_list[it].wert.int_wert);

      sugiyama.set_comp_coord(hierarchy);


      it = opt_fkt_struct.optenum_list.first();
      ranking = (opt_fkt_struct.optenum_list[it]).wert.int_wert;

      switch (ranking) { 
          case 0: sugiyama.set_ranking(LongestPathRanking());
		  break;
	  case 1: sugiyama.set_ranking(DfsRanking());
		  break;
       }

      it = opt_fkt_struct.optenum_list.succ(it);
      heuristic = opt_fkt_struct.optenum_list[it].wert.int_wert;

      switch (heuristic) {

	case 0: sugiyama.set_cross_min(SplitHeuristic());
		break;
	case 1: sugiyama.set_cross_min(BarycenterHeuristic());
		break;
	case 2: sugiyama.set_cross_min(MedianHeuristic());
		break;
	case 3: sugiyama.set_cross_min(GreedyInsertHeuristic());
		break;
	case 4: sugiyama.set_cross_min(GreedySwitchHeuristic());
		break;
      }
   }             

   return call_function(F); 
}


/************************* SpringLayout *******************************/     

void agd_serv_class::make_spring (ostream& opt) 
{ opt << "\tfunction \"" << spring.name() << "\" [" << endl;
    opt << "\t\tnoise " << bool_str(spring.noise()) << endl;
    opt << "\t\titerations " << spring.iterations() << endl;
    make_std_infos (opt,spring); 
  opt << "\t]" << endl;          
}

bool agd_serv_class::call_spring(LayoutModule& F) 
{
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    spring.noise (opt_fkt_struct.optbool_list[it].wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optint_list.first();
    spring.iterations (opt_fkt_struct.optint_list[it].wert.int_wert);
   }
  return call_function(F);
}

/************************* TutteLayout *******************************/

void agd_serv_class::make_tutte (ostream& opt) {
     make_infos_only (opt,tutte);
}


bool agd_serv_class::call_tutte(LayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    infos = opt_fkt_struct.optbool_list[it].wert.bool_wert;
   }
  return call_function(F);
}


/*********************** VisibilityLayout *****************************/

void agd_serv_class::make_visibility (ostream& opt) {
     make_infos_only (opt,visibility);
}    


bool agd_serv_class::call_visibility(GridLayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    infos = opt_fkt_struct.optbool_list[it].wert.bool_wert;
   }
  return call_grid_function(F);
}



/********************** PureOrthogonalLayout ****************************/     

void agd_serv_class::make_pureortho (ostream& opt) 
{ opt << "\tfunction \"" << pureortho.name() << "\" [" << endl;
    opt << "\t\tcheck " << bool_str(pureortho.verify()) << endl; 
    make_std_infos (opt,pureortho); 
  opt << "\t]" << endl;          
}

bool agd_serv_class::call_pureortho(LayoutModule& F) 
{ 
  if (argc > 2 && read_optfile(argv[3]))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    pureortho.verify((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
   }
  return call_function(F);
}

/********************** QuasiOrthogonalLayout **************************/     

void agd_serv_class::make_quasiortho (ostream& opt) 
{ opt << "\tfunction \"" << quasiortho.name() << "\" [" << endl;
     opt << "\t\tcheck " << bool_str(quasiortho.verify()) << endl;
     make_std_infos (opt,quasiortho); 
  opt << "\t]" << endl;          
}


bool agd_serv_class::call_quasiortho(LayoutModule& F) 
{ 
  if (argc > 2 && (read_optfile(argv[3])))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    quasiortho.verify((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
   }
  return call_function(F);
}


/************************ OrthogonalLayout ***************************/     

void agd_serv_class::make_orthogonal (ostream& opt) 
{ opt << "\tfunction \"" << orthogonal.name() << "\" [" << endl;
     opt << "\t\tcheck " << bool_str(orthogonal.verify()) << endl;
     make_std_infos (opt,orthogonal); 
  opt << "\t]" << endl;          
}


bool agd_serv_class::call_orthogonal(LayoutModule& F) 
{ 
  if (argc > 2 && (read_optfile(argv[3])))
  { agd_serv_struct_def opt_fkt_struct;
    search_struct(F, opt_fkt_struct);
    list_item it;
    it = opt_fkt_struct.optbool_list.first();
    orthogonal.verify((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
    it = opt_fkt_struct.optbool_list.succ(it);
    infos = ((opt_fkt_struct.optbool_list[it]).wert.bool_wert);
   }
  return call_function(F);
}



/************************ PlanarizationLayout *****************************/

void agd_serv_class::make_planarization(ostream& opt)
{
  int layouter = 8;  // quasiortho 

  opt << "\tfunction \"" << planarization.name() << "\" [" << endl;
      opt << "\t\tlayouter [" << endl;
         opt << "\t\t\tlayouter_wahl " << layouter << endl;     
         opt << "\t\t\tlayouter_lst [" << endl;
           opt << "\t\t\t\tlst \"" << fpp.name()      << "\"" << endl;
           opt << "\t\t\t\tlst \"" << schnyder.name()     << "\"" << endl;
           opt << "\t\t\t\tlst \"" << convex.name()   << "\"" << endl;
           opt << "\t\t\t\tlst \"" << convexdraw.name()  << "\"" << endl;
           opt << "\t\t\t\tlst \"" << planarstraight.name() << "\"" << endl;
           opt << "\t\t\t\tlst \"" << planardraw.name() << "\"" << endl;
           opt << "\t\t\t\tlst \"" << mixedmodel.name()      << "\"" << endl;
           opt << "\t\t\t\tlst \"" << pureortho.name()  << "\"" << endl;
           opt << "\t\t\t\tlst \"" << quasiortho.name()  << "\"" << endl;
           opt << "\t\t\t\tlst \"" << orthogonal.name()   << "\"" << endl;
         opt <<"\t\t\t]" << endl;
      opt << "\t\t]" << endl;  

  make_std_infos (opt,planarization); 
  opt << "\t]" << endl;          


/*    
     int planarizer = 1;     
     SubgraphPlanarizer 	subgraph;
     PlanarSubgraph	  	planarsubgraph;
   //MaximumPlanarSubgraph 	maxplanarsubgrah;   //Abacus & Cplex
     ShortestPathInserter	shortestpathinserter;

    opt << "\t\tplanarizer [" << endl;
      opt "\t\t\planarizer_wahl " << planarizer << endl;
      opt "\t\t\planarizer_lst [" << endl;
        opt << "\t\t\t\tlst \"" << subgraph.name() << "\"" << endl;
        opt << "\t\t\t\tlst \"" << planarsubgraph.name() << "\"" << endl;
        opt << "\t\t\t\tlst \"" << shortestpathinserter.name() << "\"" << endl;
//      opt << "\t\t\t\tlst \"" << maxplanarsubgraph.name() << "\"" << endl;
    opt <<"\t\t\t]" << endl;
*/

}


bool agd_serv_class::call_planarization(LayoutModule& F) 
{
   int layouter; 
   int planarizer;     

/*
	SubgraphPlanarizer 	subgraph;
	PlanarSubgraph	  	planarsubgraph;
//	MaximumPlanarSubgraph 	maxplanarsubgrah;    		Abacus & Cplex
	ShortestPathInserter	shortestpathinserter;
*/


   if (argc > 2 && read_optfile(argv[3])) 
   { 
     agd_serv_struct_def opt_fkt_struct;
        
     search_struct(F,opt_fkt_struct);

     list_item it = opt_fkt_struct.optenum_list.first();
     layouter = opt_fkt_struct.optenum_list[it].wert.int_wert;

     switch (layouter) {
	case 0: planarization.set_planar_layouter(fpp);
		break;
	case 1: planarization.set_planar_layouter(schnyder);
		break;
	case 2: planarization.set_planar_layouter(convex);
		break;
	case 3: planarization.set_planar_layouter(convexdraw);
		break;
	case 4: planarization.set_planar_layouter(planarstraight);
		break;
	case 5: planarization.set_planar_layouter(planardraw);
		break;
	case 6: planarization.set_planar_layouter(mixedmodel);
		break;
	case 7: planarization.set_planar_layouter(pureortho);
		break;
	case 8: planarization.set_planar_layouter(quasiortho);
		break;
	case 9: planarization.set_planar_layouter(orthogonal);
		break;
     }

/*	
	switch (planarizer) {
	case 0: planarization.set_planarizer(subgraph);
		break;
	case 1: planarization.set_planarizer(planarsubgraph);
		break;
	case 2: planarization.set_planarizer(shortestpathinserter);
		break;
  	case 3: planarization.set_planarizer(maxplanarsubgrah);
  		break;

	}
*/

        it = opt_fkt_struct.optbool_list.first();
        infos = opt_fkt_struct.optbool_list[it].wert.bool_wert;
   }

   return call_function(F);
}


void print_usage() {

cerr<< endl;
cerr<< "usage:" << endl;
cerr<< endl;

cerr<< "   agd_server list <fname>" << endl; 
cerr<< endl;
cerr<< "\t writes the list of available AGD algorithms and their" << endl;
cerr<< "\t default options in GML format to file <fname>. If <fname>" << endl;
cerr<< "\t is omitted the list is written to standard output instead." << endl;
cerr<< endl;
cerr<< endl;

cerr<< "   agd_server <alg> <gfile> <optfile> <iform> <oform>" << endl;
cerr<< endl;
cerr<< "\t reads a graph in format <iform> from file <gfile>, applies"   <<endl;
cerr<< "\t algorithm <alg> to it using the options from file <optfile>," <<endl;
cerr<< "\t and writes the result in format <oform> back to <gfile>."     <<endl;
cerr<< "\t There are two possible file formats for <iform> and <oform>"  <<endl;
cerr<< "\t \"gw\": the GraphWin file format and \"gml\": the GML format."<<endl;
cerr<< "\t A default option file is created by calling \"agd_server list"<<endl;
cerr<< "\t <optfile>\" (see above) and then can be edited as needed."    <<endl;
cerr<< endl;
}




#if defined(WINMAIN)

int main() {
  int    argc;
  char** argv;
  get_command_line(argc,argv);

#else

int main(int argc, char *argv[]) {

#endif


  if (argc < 2)
  { print_usage();
    return 1;
   }

  agd_serv_class* serv = new agd_serv_class(argc,argv);

  if (getenv("AGD_SERVER_TRACE"))
  { cerr << "agd_server " << serv->impl_version() << "  "; 
    for(int i=1; i<argc; i++) cerr << argv[i] << " ";
    cerr << endl;
   }

  bool agd_result = serv->agd_serv();

  delete serv;

  return agd_result ? 0 : 1;
}


