/*******************************************************************************
+
+  LEDA 5.0  
+
+
+  cluster_demo.c
+
+
+  Copyright (c) 1995-2004
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <LEDA/graphics/graphwin.h>

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::cerr;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ios;
using std::endl;
#endif


#include <AGD/CGraphWin.h>
#include <AGD/ClusterOrthogonalLayout.h>

#include <AGD/PlanAug.h>
#include <AGD/PlanarSubgraph.h>
#include <AGD/LEDAMakeBiconnected.h>

#include <AGD/generators.h>
#include <AGD/CGLayout.h>


class GraphWinAGD : public CGraphWin
{
public:

	GraphWinAGD();
	GraphWinAGD(const char* winlabel);

	virtual ~GraphWinAGD() { }

protected:
	void optionsGlobal();
	void optionsPlanSub();
	void optionsPlanAug();
	void optionsCOrtho();

	ClusterOrthogonalLayout colt;

	PlanAug plan_aug;
	LEDAMakeBiconnected make_bic;
	PlanarSubgraph plan_sub;

public:
	// call function for cluster drawings
	static void callColtLayout(GraphWin& GW) {
	        GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		callClusterGridLayout(gwAGD, gwAGD.colt);
	}

	static void callRandomPlanarGraph (GraphWin& GW) {
		gen_random_graph (GW, random_planar_graph,
			"Create Random Planar Graph");
	}
	static void callRandomPlanarTricGraph (GraphWin& GW) {                                      
	  gen_random_graph (GW, planar_triconnected_graph,                                    
			    "Create Triconnected Planar Graph");                                        
        }
	static void callRandomPlanarBicGraph (GraphWin& GW);

	static void callPlanAug (GraphWin& GW) {
		GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		callAugment (gwAGD, gwAGD.plan_aug);
	}
	static void callMakeBic (GraphWin& GW) {
		GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		callAugment (gwAGD, gwAGD.make_bic);
	}

	static void callPlanSub (GraphWin& GW) {
		GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		callSubgraph (gwAGD, gwAGD.plan_sub);
	}

	static void callOptionsGlobal (GraphWin& GW) {
		GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		gwAGD.optionsGlobal();
	}

	static void callOptionsPlanSub (GraphWin& GW) {
		GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		gwAGD.optionsPlanSub();
	}

	static void callOptionsPlanAug (GraphWin& GW) {
		GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		gwAGD.optionsPlanAug();
	}
	static void callOptionsCOrtho (GraphWin& GW) {
	        GraphWinAGD &gwAGD = (GraphWinAGD&)GW;
		gwAGD.optionsCOrtho();
	}


	static bool callAugment (GraphWinAGD& GW, AugmentationModule& A);
	static bool callSubgraph (GraphWinAGD& GW, SubgraphModule& S);
	// declaration of call function for cluster drawings 
	static bool callClusterGridLayout (GraphWinAGD& GW, CGGridLayoutModule& CGGL);

protected:
	static string get_error_message(AgdKey p);

	int  the_base_algo, the_compactor;
	bool the_pres_embedding;

	int the_mapper;
	double the_min_over_x, the_min_over_y;
	double the_sep_x, the_sep_y;

	static bool rbg_multi, rbg_map;

	int layout_menu, generators_menu, options_menu,
		graph_menu, augment_menu, subgraph_menu;

	void gen_pre(int n, int m);
	void gen_post();

	int openPanel(panel &P);
	static void embed_random(const graph& G, node_array<double>& x,
		node_array<double>& y);
	static void gen_random_graph(GraphWin& gw, void (*f)(graph&,int,int), string s);
	static void gen_random_graph1(GraphWin& gw, void (*f)(graph&,int), string s);

	static void random_bic_graph (graph& G, int num_nodes);
};


//-------------------------------------------------------------------
//
//                         CGraphWinAGD.cpp
//
// This demo program shows, how to use cluster layout algorithms 
// of the AGD-Library with LEDA's GraphWin.
//
//    authors: Carsten Gutwenger, Dirk Luettke-Huettman
//    date:    octobre 1999
//-------------------------------------------------------------------

#include <AGD/CGraphWinInterface.h>
#include <AGD/GraphWinInterface.h>

#include <AGD/QuasiOrthogonalLayout.h>
#include <AGD/OrthogonalLayout.h>
#include <AGD/LongestPathCompaction.h>
#include <AGD/FlowCompaction.h>

#include <AGD/AGDLayout.h>
#include <AGD/GraphGLI.h>
#include <AGD/CGGridLayoutInfo.h>

#include <AGD/LeveledGraph.h>

#include <AGD/MaxAllMapper.h>
#include <AGD/MaxEachMapper.h>
#include <AGD/FixedGridMapper.h>
#include <AGD/FixedSizeMapper.h>

#include <AGD/TopologyModule.h>

#include <AGD/ClusterGraph.h>
#include <AGD/CGGridLayoutInfo.h>
#include <AGD/ClusterGraphGLI.h>

#include <AGD/PlaneClusterGraphCopy.h>
#include <AGD/CPlanarEmbed.h>
#include <AGD/PlanarSubgraph.h>


//-------------------------------------------------------------------
//                 main()-function for agd_demo
//-------------------------------------------------------------------

int main()
{
	GraphWinAGD GW("ClusterGraphWin");

	GW.open();

	return 0;
}


//-------------------------------------------------------------------
//                            constructor
//-------------------------------------------------------------------

GraphWinAGD::GraphWinAGD(const char *win_label) : CGraphWin(win_label)
{
	// set some  GraphWin options to appropriate values
    set_node_shape (rectangle_node);
	set_edge_direction(undirected_edge);
	set_zoom_objects(true);
	set_zoom_labels(true);
	set_node_width(20);
	set_node_height(20);
	set_node_label_type(index_label);

	// create menues for AGD
	layout_menu = add_menu ("Layout ");
	gw_add_simple_call (*this, callColtLayout, colt.name(), layout_menu);
	
	// create menu entries for graph generators
	generators_menu = add_menu ("Generators ");
	add_simple_call (callRandomPlanarGraph, "planar", generators_menu);
	add_simple_call (callRandomPlanarBicGraph, "planar biconnected",
		generators_menu);
	add_simple_call (callRandomPlanarTricGraph, "planar triconnected",
		generators_menu);

	// create menu entries for augmentation and subgraph algorithms
	graph_menu      = add_menu ("Graph ");
	augment_menu    = add_menu ("Augmentation", graph_menu);
	subgraph_menu   = add_menu ("Subgraph", graph_menu);
	add_simple_call (callMakeBic, "biconnected", augment_menu);
	add_simple_call (callPlanAug, "planar biconnected", augment_menu);
	add_simple_call (callPlanSub, "planar", subgraph_menu);

	// create menu entries for setting optional parameters
	options_menu    = add_menu ("Options ");
	gw_add_simple_call (*this, callOptionsCOrtho, "ClusterQuasiOrthogonalLayout", options_menu );
	add_separator(options_menu);
	add_simple_call (callOptionsPlanAug,"planar biconnected",options_menu );
	add_simple_call (callOptionsPlanSub,"planar",options_menu );
	add_separator(options_menu);
	add_simple_call (callOptionsGlobal,"Global",options_menu );

	the_mapper = 0;
	the_min_over_x = 3;  the_min_over_y = 3;
	the_sep_x      = 10; the_sep_y      = 10;
	the_pres_embedding = true;

	the_base_algo = 0;
	the_compactor = 1;
}


//-------------------------------------------------------------------
//            build error message for violated property
//-------------------------------------------------------------------

string GraphWinAGD::get_error_message(AgdKey p)
{
	if (p == key::planar)
		return "The graph is not planar.";
	if (p == key::simple)
		return "The graph ist not simple.";
	if (p == key::connected)
		return "The graph is not connected.";
	if (p == key::biconnected)
		return "The graph is not biconnected.";
	if (p == key::triconnected)
		return "The graph is not triconnected.";
	if (p == key::four_graph)
		return "The graph is not a four-graph.";
	if (p == key::no_self_loops)
		return "The graph has self loops.";
	if (p == key::directed)
		return "The graph is not directed.";
	if (p == key::undirected)
		return "The graph is not undirected.";
	if (p == key::no_reversals)
		return "The graph has reversal edges.";
	if (p == key::acyclic)
		return "The graph is not acyclic.";
	if (p == key::tree)
		return "The graph does not represent a tree.";
	if (p == key::forest)
		return "The graph does not represent a forest.";
	if (p == key::single_source)
		return "The graph has not a single source.";
	if (p == key::single_sink)
		return "The graph has not a single sink.";
	if (p == key::bipartite)
		return "The graph is not bipartite.";
	if (p == key::c_planar)
	    return "The graph is not c-planar.";

	return "property <" + p.info() + "> is not satisfied.";
}


//-------------------------------------------------------------------
//                 generic calling functions
//-------------------------------------------------------------------

bool GraphWinAGD::callClusterGridLayout(GraphWinAGD& GW, 
					CGGridLayoutModule& L) 
{
	ClusterOrthogonalLayout  &colt  = (ClusterOrthogonalLayout &)L;
	ClusterGraph &CG = GW.get_clustergraph();

	TopologyModule T;  
	CPlanarEmbed CEmbed;

	CGraphWinInterface gwi(GW);
	CGGridLayout cggl(CG);
/*
	Cluster* cl;
	edge e;
*/

	AgdKey p;

	leda_face f_0 = nil;
	list<edge> add_edges;

	if (L.check(CG, p)) {
		cout << "\ncalling " << L.name() << endl;  
		if(GW.the_pres_embedding == false) {
			CEmbed.call(CG, add_edges);
			//CG.make_planar_map();
			colt.call_grid(CG, f_0, cggl);
		} else {
			f_0 = T.call(CG, gwi, add_edges);
			if (f_0 == nil) {
				CG.del_edges(add_edges);
				CEmbed.call(CG, add_edges);
			}
			colt.call_grid(CG, f_0, cggl);
		}

		CG.del_edges(add_edges);

		IRect ir = L.bounding_box();
		cout << "grid size: " << ir.width() << " * " << ir.height() << " = " 
			<< ir.width() * ir.height() << endl;

		ClusterGraphGLI cggli(cggl);

		int gs = 2*GW.get_grid_dist();
		if (GW.the_mapper == 3 && gs > 0) {
			FixedSizeMapper M(gs,gs,
				GW.the_min_over_x,GW.the_min_over_y,
				GW.the_sep_x, GW.the_sep_y);

			M.call(cggli,gwi);

		} else if (GW.the_mapper == 2 && gs > 0) {
			FixedGridMapper M(gs,gs,
				GW.the_min_over_x,GW.the_min_over_y,
				GW.the_sep_x, GW.the_sep_y);

			M.call(cggli,gwi);

		} else if (GW.the_mapper == 1) {
			MaxEachMapper M(40,40,GW.the_min_over_x,GW.the_min_over_y,
				GW.the_sep_x, GW.the_sep_y);

			M.call(cggli,gwi);

		} else {
			MaxAllMapper M(40,GW.the_min_over_x,GW.the_sep_x);

			M.call(cggli,gwi);
		}

		return true;	
	} else {
		GW.acknowledge (get_error_message(p)); // no, so we display the error message
	}
	return false;
}


bool GraphWinAGD::callAugment (GraphWinAGD& GW, AugmentationModule& A) {
	AgdKey p;

	// can A be called for the graph displayed in GW ?
	if (A.check(GW.get_graph(), p)) {
		cout << "\ncalling " << A.name() << endl;

		list<edge> e_list;
		edge e;

		A.call (GW.get_graph(), e_list); // yes, we can call A

		GW.update_graph();

		GW.deselect_all_edges();
		forall (e, e_list)
			GW.select(e);

		cout << A.num_added_edges() << " edge(s) added." << endl;

		return true;

	} else {
		GW.acknowledge (get_error_message(p)); // no, so we display the error message
	}

	return false;
}


bool GraphWinAGD::callSubgraph (GraphWinAGD& GW, SubgraphModule& S)
{
	graph &G = GW.get_graph();
	AgdKey p;

	// can S be called for the graph displayed in GW ?
	if (S.check(G, p)) {
		cout << "\ncalling " << S.name() << endl;
		S.call (G); // yes, we can call S
		GW.update_graph();

		cout << S.num_del_edges() << " edge(s) deleted." << endl;

		return true;

	} else {
		GW.acknowledge (get_error_message(p)); // no, so we display the error message
	}

	return false;
}


//-------------------------------------------------------------------
//                 setting options of algorithms
//-------------------------------------------------------------------

void GraphWinAGD::optionsCOrtho()
{
	int base_algo = the_base_algo, compactor = the_compactor;

	panel P("Options for ClusterQuod");
	P.choice_item ("base algorithm:", base_algo, "QuasiOrthogonal", "Orthogonal");
	P.choice_item ("compaction:", compactor, "LongestPath", "Flow");

	P.button ("Ok");
	int call   = P.button ("Call");
	int cancel = P.button ("Cancel");

	int result = P.open (get_window());

	if (result != cancel) {
		the_compactor = compactor;
		CompactionModule *comp;
		switch(compactor) {
		case 0:
			comp = new LongestPathCompaction; break;
		default:
			comp = new FlowCompaction;
		}

		the_base_algo = base_algo;
		Ovid *layouter;
		switch (base_algo) {
		case 0:
			layouter = new QuasiOrthogonalLayout;
			break;
		default:
			layouter = new OrthogonalLayout;
			break;
		}
		layouter->set_compactor(*comp);
		colt.set_base_algorithm(*layouter);

		delete layouter;
		delete comp;
	}
	if (result == call) {
		callColtLayout(*this);
	}
}

int GraphWinAGD::openPanel(panel &P)
{
	int xleft,xright,ytop,ybot;

	get_window().frame_box(xleft,ytop,xright,ybot);

	if (xleft+xright > window::screen_width())
		return P.open(-(xleft-1),ytop);
	else
		return P.open(xright+1,ytop);
}

void GraphWinAGD::optionsPlanSub()
{
	int runs = plan_sub.runs();

	panel P ("Options for Planar Subgraph");

	P.int_item ("runs:", runs);

	P.button ("Ok");
	int call   = P.button ("Call");
	int cancel = P.button ("Cancel");

	int result = openPanel(P);
	if (result != cancel) {
		plan_sub.runs(runs);
	}
	if (result == call) {
		callPlanSub(*this);
	}
}

void GraphWinAGD::optionsPlanAug()
{
	bool size_strategy = plan_aug.size_strategy();
	bool use_blabels = plan_aug.use_blabels();

	panel P ("Options for Planar Biconnected");

	P.bool_item ("size strategy:", size_strategy);
	P.bool_item ("use b-labels:", use_blabels);

	P.button ("Ok");
	int call   = P.button ("Call");
	int cancel = P.button ("Cancel");

	int result = openPanel(P);
	if (result != cancel) {
		plan_aug.size_strategy(size_strategy);
		plan_aug.size_strategy(use_blabels);
	}
	if (result == call) {
		callPlanSub(*this);
	}
}

void GraphWinAGD::optionsGlobal()
{
	bool   pres_embedding = the_pres_embedding;
	int    mapper         = the_mapper;
	double min_over_x     = the_min_over_x;
	double min_over_y     = the_min_over_y;
	double sep_x          = the_sep_x;
	double sep_y          = the_sep_y;

	panel P ("Global Options");

	P.text_item("\\bf\\blue Grid Mapper");
	P.choice_item ("mapper:", mapper, "MaxAll", "MaxEach", "FixedGrid", "FixedSize");
	P.real_item ("min. overhang (x-coord.):", min_over_x);
	P.real_item ("min. overhang (y-coord.):", min_over_y);
	P.real_item ("separation (x-coord.):", sep_x);
	P.real_item ("separation (y-coord.):", sep_y);

	P.text_item("\\bf\\blue Plane Algorithms");
	P.bool_item ("preserve embedding:",  pres_embedding);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	if (P.open() != cancel) {
		the_pres_embedding = pres_embedding;
		the_mapper         = mapper;
		the_min_over_x     = min_over_x;
		the_min_over_y     = min_over_y;
		the_sep_x          = sep_x;
		the_sep_y          = sep_y;
	}
}


//-------------------------------------------------------------------
//                      creating random graphs
//-------------------------------------------------------------------

bool GraphWinAGD::rbg_multi = false, GraphWinAGD::rbg_map = false;

void GraphWinAGD::gen_pre(int n, int m)
{
	set_gen_nodes (n);
	set_gen_edges (m);
	graph& G = get_graph();

	G.clear();
	update_graph();
}

void GraphWinAGD::gen_post()
{
	graph& G=get_graph();
	bool old_flush = set_flush(false); 

	update_graph();

	node_array<double> x(G);
	node_array<double> y(G);
	embed_random(G,x,y);
	update_graph();
	adjust_coords_to_win(x,y);
	set_layout(x,y);
	embed_edges();
	set_flush(old_flush);
	redraw();
	set_flush(false);
	call_init_graph_handler();
	set_flush(old_flush);
}

void GraphWinAGD::callRandomPlanarBicGraph(GraphWin &gw1)
{
	GraphWinAGD& gw = (GraphWinAGD&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P("Create Biconnected Planar Graph");
	int n = gw.get_gen_nodes(),
		m = gw.get_gen_edges();
	
	P.int_item ("nodes", n);
	P.int_item ("edges", m);
	P.bool_item ("multi-edges", rbg_multi);
	P.bool_item ("planar map", rbg_map);

	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.openPanel(P) == 0 || n < 0 || m < 0) return;

	gw.gen_pre(n,m);
	planar_biconnected_graph (gw.get_graph(), n, m, rbg_multi, rbg_map);
	gw.gen_post();
}

void GraphWinAGD::gen_random_graph(GraphWin& gw1, void (*f)(graph&,int,int), string s)
{
	GraphWinAGD& gw = (GraphWinAGD&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P(s);
	int n = gw.get_gen_nodes(),
		m = gw.get_gen_edges();

	P.int_item ("nodes", n);
	P.int_item ("edges", m);
	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.openPanel(P) == 0 || n < 0 || m < 0) return;

	gw.gen_pre(n,m);
	f (gw.get_graph(), n, m);
	gw.gen_post();
}

void GraphWinAGD::gen_random_graph1(GraphWin& gw1, void (*f)(graph&,int), string s)
{
	GraphWinAGD& gw = (GraphWinAGD&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P(s);
	int n = gw.get_gen_nodes();

	P.int_item ("nodes", n);
	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.openPanel(P) == 0 || n < 0) return;

	gw.gen_pre(n,gw.get_gen_edges());

	f (gw.get_graph(), n);

	gw.gen_post();
}


void GraphWinAGD::embed_random(const graph& G, node_array<double>& x,
	node_array<double>& y) 
{
	random_source ran;
	node v;
	forall_nodes(v,G)  ran >> x[v] >> y[v];
}
