/*******************************************************************************
+
+  AGD 1.2
+
+
+  CGGridLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         CGGridLayout.h
//
//
//    author: Dirk Luetke-Huettmann
//    date:   may 1999
//-------------------------------------------------------------------

#ifndef _CGGRID_LAYOUT_H
#define _CGGRID_LAYOUT_H

#include <LEDA/core/map.h>
#include <LEDA/graph/graph.h>

#include <AGD/GridLayout.h>
#include <AGD/ClusterGraph.h>
#include <AGD/PlaneClusterGraphCopy.h>

/*{\Manpage {CGGridLayout} {} {Grid Layouts for Cluster Graphs} {cggl} }*/

/*{\Mbaseclasses

\begin{hierarchy}{7}{2.5}
\hdclass{0}{1.5}{6}{GridLayout}
\arhook{6}{2}{7}
\hdclass{1}{0}{6.5}{CGGridLayout}
\end{hierarchy}
}*/

class __exportAGD CGGridLayout : public GridLayout {

/*{\Mdefinition
A cluster graph grid layout |\Mvar| extends its base class
|GridLayout| by functions to store the additional coordinates
for the position of the cluster regions.}*/
  
private:
  //const Cluster* rc_p;
  const ClusterGraph *_ccgp;

  leda::map<Cluster*, int> _clx;
  leda::map<Cluster*, int> _cly;
  leda::map<Cluster*, int> _clwidth;
  leda::map<Cluster*, int> _clheight;

  //leda::list<Cluster*> clusters;

public:

  /*{\Mcreation}*/
  
  CGGridLayout();
  /*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	the empty layout with no associated graph.}*/
  
  CGGridLayout(const ClusterGraph &GC) : GridLayout(GC), _ccgp(&GC) { }
  /*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
    a grid layout for cluster graph |CG|. Initially, all x- and
    y-coordinates are $0$, all nodes have width and height $0$, and
    all bend point lists are empty.}*/ 

  CGGridLayout(const PlaneClusterGraphCopy &GC) : GridLayout(GC) {
    _ccgp = GC.get_clustergraph();
  }
  
  /*{\Moperations}*/

  const int& x(leda::node v) const {
    return GridLayout::x(v);
  }

  int& x(leda::node v) {
    return GridLayout::x(v);
  }

  const int& y(leda::node v) const {
    return GridLayout::y(v);
  }
  int& y(leda::node v) {
    return GridLayout::y(v);
  }
  
  int width(leda::node v) const {
    return GridLayout::width(v);
  }
  
  void width(leda::node v, int width) {
    GridLayout::width(v, width);
  }
  
  int height(leda::node v) const {
    return GridLayout::height(v);
  }
  
  void height(leda::node v, int height) {
    GridLayout::height(v, height);
  }

  const ClusterGraph &get_graph() const {
	  return *_ccgp;
  }
  /*{\Mop returns the associated cluster graph.}*/

  const int& x(Cluster* cl) const {
    return _clx[cl];
  }
  /*{\Mop returns a reference to the x-coordinate of cluster |cl|.}*/ 

  const int& y(Cluster* cl) const {
    return _cly[cl];
  }
  /*{\Mop returns a reference to the y-coordinate of cluster |cl|.}*/

  const int& width(Cluster* cl) const {
    return _clwidth[cl];
  }
  /*{\Mop returns a reference to the width of cluster |cl|.}*/

  const int& height(Cluster* cl) const {
    return _clheight[cl];
  }
  /*{\Mop returns a reference to the height of cluster |cl|.}*/

  leda::map<Cluster*, int> &clx() {
    return _clx;
  }

  leda::map<Cluster*, int> &cly() {
    return _cly;
  }

  const IPolyline &bends(leda::edge e) const {
    return GridLayout::bends(e);
  }

  IPolyline &bends(leda::edge e) {
    return GridLayout::bends(e); 
  }

  void x(Cluster* cl, int x) {
    _clx[cl] = x;
  }
  /*{\Mop sets the x-coordinate of cluster |cl| to |x|.}*/

  void y(Cluster* cl, int y) {
    _cly[cl] = y;
  }
  /*{\Mop sets the y-coordinate of cluster |cl| to |y|.}*/

  void width(Cluster* cl, int width) {
    _clwidth[cl] = width;
  }
  /*{\Mop sets the height of cluster |cl| to |height|.}*/

  void height(Cluster* cl, int height) {
    _clheight[cl] = height;
  }
  /*{\Mop sets the width of cluster |cl| to |width|.}*/

  const IRect bounding_box(void);

};

#endif
