/*******************************************************************************
+
+  AGD 1.2
+
+
+  CondBase.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//-------------------------------------------------------------------
//
//                           CondBase.h
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_COND_BASE_H
#define _AGD_COND_BASE_H

#include <AGD/KeySet.h>
#include <LEDA/core/array.h>




/*{\Manpage {CondBase} {} {Base Class for Pre- and Postconditions} {P} }*/

class __exportAGD CondBase : private leda::array<KeySet> {

/*{\Mdefinition
|\Mname| is the common base class for classes |PreCond| and |PostCond| and
provides operations for pre- and postconditions. It maintains a collection
of key sets $S_{low},\ldots,S_{high}$.
}*/

typedef leda::array<KeySet> lak;

public:

	CondBase() { }
	CondBase(int low, int high) : leda::array<KeySet>(low,high) { }
	CondBase(const CondBase& p) : leda::array<KeySet>(p) { }

	virtual ~CondBase() { }

	/*{\Moperations}*/

	int low() const {
		return lak::low();
	}
	/*{\Mop returns |low|.}*/

	int num_param() const {
		return high();
	}
	/*{\Mop returns the number of parameters for which |\Mvar|
	is defined; more exactly it returns |high|.}*/

	void insert (int i, const AgdKey &k) {
		access(i).insert(k);
	}
	/*{\Mop inserts key $k$ into the $i$th key set of |\Mvar|.}*/

	void del (int i, const AgdKey &k) {
		access(i).del(k);
	}
	/*{\Mop deletes key $k$ from the $i$th key set of |\Mvar|.}*/

	bool member (int i, const AgdKey &k) const {
		return access(i).member(k);
	}
	/*{\Mop returns |true| iff key $k$ is contained in the $i$th
	key set.}*/

	const KeySet &operator[](int i) const {
		return access(i);
	}

	KeySet &operator[](int i) {
		return access(i);
	}
	/*{\Marrop returns a reference to the $i$th key set.}*/

	CondBase &operator=(const CondBase &c) {
		lak::operator=(c);
		return *this;
	}

protected:
	const KeySet &access(int i) const {
		return lak::operator[](i);
	}

	KeySet &access(int i) {
		return lak::operator[](i);
	}
};



#endif
