/*******************************************************************************
+
+  AGD 1.2
+
+
+  ConstLedaGraphGridConcept.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                     ConstLedaGraphGridConcept.h
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#ifndef _AGD_CLGG_CONCEPT_H
#define _AGD_CLGG_CONCEPT_H


#include <AGD/impl/GridConceptBase.h>
#include <LEDA/graph/graph.h>

class LayoutInterface;
class PlaneGraphCopy;
class GridCoordinateMapper;


/*{\Manpage {ConstLedaGraphGridConcept} {} {Grid Concept for LEDA Graphs} {C} }*/

class __exportAGD ConstLedaGraphGridConcept : public GridConceptBase {

/*{\Mdefinition
The class |\Mname| represents an implementation concept for grid layout
algorithms which work on a const LEDA graph |G| and access the grid
layout via an object of type |GridLayout|. The graph can not be modified
by the algorithm. The update() functions remove unnecessary bend points.

A computed bounding box can be passed to the grid concept using
set\_bbox(). Otherwise, the bounding box is computed on demand.
}*/

	const leda::graph *_cgp;

	bool new_layout;

	// forbid copying
	ConstLedaGraphGridConcept (const ConstLedaGraphGridConcept &) { }
	ConstLedaGraphGridConcept &operator=(const ConstLedaGraphGridConcept &) {
		return *this;
	}

public:

	ConstLedaGraphGridConcept (const leda::graph &G, LayoutInterface &A);
		
	ConstLedaGraphGridConcept (const leda::graph &G, GridLayout &gl) :
		_cgp(&G), new_layout(false)
	{
		_glp = &gl;
	}

	// never called
	ConstLedaGraphGridConcept (const leda::graph &G, leda::face f, LayoutInterface &A) { }
	ConstLedaGraphGridConcept (PlaneGraphCopy &PG, leda::face f, LayoutInterface &A) { }
	ConstLedaGraphGridConcept (const leda::graph &G, leda::face , GridLayout &gl) { }
	ConstLedaGraphGridConcept (PlaneGraphCopy &PG, leda::face , GridLayout &gl) { }


	~ConstLedaGraphGridConcept();


	/*{\Moperations}*/
	/*{\Mtext |\Mname| implements the constructors and operations which are
	required for template argument |GRID_CONCEPT| of |GridLayoutModuleImpl|.
	Furthermore, the following operations are provided:}*/

	static void set_postcondition(leda::list<AgdKey> &post) { }

	void get_node_grid_size (GridCoordinateMapper &gcm,
		const LayoutInterface &A);
	void get_node_grid_size (const GridLayout &) { }

	void update (GridCoordinateMapper &gcm, LayoutInterface &A);
	void update (GridLayout &gl);


	const leda::graph &get_graph() const {
		return *_cgp;
	}
	/*{\Mop returns the const LEDA graph |G|.}*/

	/*{\Moptions warnings=no}*/

	/*
	GridLayout &get_layout() const;
	*/
	/*{\Mop returns the grid layout of |G|.}*/

	/*
	void set_zero_size();
	*/
	/*{\Mop sets the grid width and height of all nodes to $0$.}*/

	/*
	void set_bbox (const IRect &bbox);
	*/
	/*{\Mop sets the bounding box of the grid layout to |bbox|.}*/
};


#endif
