/*******************************************************************************
+
+  AGD 1.2
+
+
+  DRect.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           DRect.h
//
//    author: Carsten Gutwenger
//    date:   june 1997
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_DRECT_H
#define _AGD_DRECT_H

#include <AGD/DPoint.h>


/*{\Manpage {DRect} {} {Double Rectangles} {dr} }*/

class __exportAGD DRect {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is a rectangle
$(left,right,bottom,top)$ in the real plane $\real^2$.
That is $left\leq right$ and $bottom\leq top$.}*/

public:

	/*{\Mcreation}*/

	DRect () {
		l = r = t = b = 0.0;
	}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	the rectangle $(0,0,0,0)$.}*/

	DRect (double xmin, double xmax, double ymin, double ymax) {
		l = xmin; r = xmax;
		b = ymin; t = ymax;
	}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	the rectangle $(xmin,xmax,ymin,ymax)$.}*/

	DRect (const DPoint& ll, const DPoint& ur) {
		l = ll.xcoord();
		r = ur.xcoord();
		b = ll.ycoord();
		t = ur.ycoord();
	}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	the rectangle with lower left corner |ll| and upper right corner |ur|.}*/


	/*{\Moperations}*/

	DPoint lower_left () const {
		return DPoint (l, b);
	}
	/*{\Mop returns the lower left corner of |\Mvar|.}*/

	DPoint upper_right () const {
		return DPoint (r, t);
	}
	/*{\Mop returns the upper right corner of |\Mvar|.}*/

	/*{\Moptions nextwarning=no }*/
	const double& left () const {
		return l;
	}
	double& left () {
		return l;
	}
	/*{\Mop returns the x-coordinate of the lower left corner
	of |\Mvar|.}*/

	/*{\Moptions nextwarning=no }*/
	const double& right () const {
		return r;
	}
	double& right () {
		return r;
	}
	/*{\Mop returns the x-coordinate of the upper right corner
	of |\Mvar|.}*/

	/*{\Moptions nextwarning=no }*/
	const double& top () const {
		return t;
	}
	double& top () {
		return t;
	}
	/*{\Mop returns the y-coordinate of the upper right corner
	of |\Mvar|.}*/

	/*{\Moptions nextwarning=no }*/
	const double& bottom () const {
		return b;
	}
	double& bottom () {
		return b;
	}
	/*{\Mop returns the y-coordinate of the lower left corner
	of |\Mvar|.}*/

	double width () const {
		return r-l;
	}
	/*{\Mop returns the width $(= |right|-|left|)$ of |\Mvar|.}*/

	double height () const {
		return t-b;
	}
	/*{\Mop returns the height $(= |bottom|-|top|)$ of |\Mvar|.}*/

	DPoint center () const {
		return DPoint ((l+r)/2,(b+t)/2);
	}
	/*{\Mop returns the center point of |\Mvar|.}*/


private:
	double l, r, t, b;
};


#endif
