/*******************************************************************************
+
+  AGD 1.2
+
+
+  DfsRanking.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                          DfsRanking.h
//
//
//    author: Carsten Gutwenger
//    date:   february 1998
//-------------------------------------------------------------------


#ifndef _AGD_DFS_RANKING_H
#define _AGD_DFS_RANKING_H


#include <AGD/RankAssignment.h>


/*{\Manpage {DfsRanking} {} {Ranking by Depth First Search} {R} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1.5}{1.5}{5.5}{RankAssignment}
\arhook{7}{2}{8}
\hdclass{3}{0}{5.5}{DfsRanking}
\end{hierarchy}
}*/

class __exportAGD DfsRanking : public RankAssignment {

/*{\Mdefinition
The |\Mname| module computes the following ranking of a graph $G$:
We build a DFS forest $T$ of the underlying undirected graph
of $G$ and set
\[rank(v) = \left\{\begin{array}{cl}
0 & \mbox{if }v\mbox{ is a root of }T\\
rank(pred(v))+1 & \mbox{otherwise}
\end{array}\right.\]
where $pred(v)$ is the predecessor of $v$ in $T$. Notice, that |\Mname|
only considers the underlying undirected graph and therefore does not
take into account edge directions.

\horizline
\commoninfo{DFS-Ranking}{DFS-Ranking}
{}{C. Gutwenger}{February 1998}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{directed} \,\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation}*/

	DfsRanking();
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~DfsRanking() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "DFS-Ranking";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "February 1998";
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|RankAssignment|).
	}*/

	/* inherited
	bool check (const leda::graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void call (const leda::graph& G, leda::node_array<int>& rank);
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:
	void dfs(leda::node v, int r, leda::node_array<int>& rank);
};


/*{\Mimplementation
The computation of the ranking takes time $\BigO(\Labs{V}+\Labs{E})$.
}*/


#endif
