/*******************************************************************************
+
+  AGD 1.2
+
+
+  GraphCopyConcept.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       GraphCopyConcept.h
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#ifndef _AGD_GC_CONCEPT_H
#define _AGD_GC_CONCEPT_H


#include <LEDA/graph/graph.h>
class GraphCopy;
class PlaneGraphCopy;
class GraphCopyLI;


/*{\Manpage {GraphCopyConcept} {} {Concept for Graph Copies} {C} }*/

class __exportAGD GraphCopyConcept {

/*{\Mdefinition
The class |\Mname| represents an implementation concept for layout
algorithms which work on a graph copy |GC| and access the layout
via a layout interface $A$ of type |GraphCopyLI|. The graph copy can be modified by
the algorithm. The following transformations are performed by the update()
function:
\begin{itemize}
\item The polylines of reversed edges are reversed.
\item The polylines of splitted edges are concatenated.
\end{itemize}
}*/

	GraphCopy   *_gcp;
	GraphCopyLI *_gclip;

	// forbid copying
	GraphCopyConcept (const GraphCopyConcept &) { }
	GraphCopyConcept &operator=(const GraphCopyConcept &) {
		return *this;
	}

public:

	GraphCopyConcept (const leda::graph &G, LayoutInterface &A);

	// never called
	GraphCopyConcept (const leda::graph &G, leda::face f, LayoutInterface &A) { }
	GraphCopyConcept (PlaneGraphCopy &PG, leda::face f, LayoutInterface &A) { }


	~GraphCopyConcept();


	/*{\Moperations}*/
	/*{\Mtext |\Mname| implements the constructors and operations which are
	required for template argument |CONCEPT| of |LayoutModuleImpl|.
	Furthermore, the following operations are provided:}*/

	void update();


	GraphCopy &get_graph() const {
		return *_gcp;
	}
	/*{\Mop returns the graph copy |GC|.}*/

	GraphCopyLI &get_layout() {
		return *_gclip;
	}
	/*{\Mop returns the layout interface $A$ for |GC|.}*/

	static void set_postcondition(leda::list<AgdKey> &post) { }
};


#endif
