/*******************************************************************************
+
+  AGD 1.2
+
+
+  GraphCopyGLI.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        GraphCopyGLI.h
//
//    author: Carsten Gutwenger
//    date:   july 1999
//-------------------------------------------------------------------


#ifndef _AGD_GRAPH_COPY_GLI_H
#define _AGD_GRAPH_COPY_GLI_H


#include <AGD/GridLayoutInfo.h>
#include <AGD/GraphCopy.h>
#include <AGD/GridLayout.h>
#include <AGD/impl/GCGridConceptBase.h>


/*{\Manpage {GraphCopyGLI} {} {GLI's for Graph Copies} {gli} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{GridLayoutInfo}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{GraphCopyGLI}
\end{hierarchy}
}*/

class __exportAGD GraphCopyGLI : public GridLayoutInfo, protected GCGridConceptBase {

/*{\Mdefinition
An instance of type |\Mname| represents a grid layout info for a graph
G. The grid layout is given by a grid layout |gl| of a copy |GC| of |G|,
which is passed as a grid implementation concept of type |GraphCopyGridConcept|
or |PlaneGraphCopyGridConcept|. The access methods of |\Mvar| implement
the following transformations:
\begin{itemize}
\item The polylines of reversed edges are reversed.
\item The polylines of splitted edges are concatenated.
\end{itemize}
The bounding box of the grid layout can be passed to the constructor of |\Mname|.
If not, the bounding box is explicitly computed in the constructor.
}*/

public:

	/*{\Mcreation}*/

	GraphCopyGLI(const GCGridConceptBase &C) :
		GCGridConceptBase(C) { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout info for |C|.}*/

	GraphCopyGLI(GridLayout &gl, GraphCopy &GC) :
		GCGridConceptBase(gl,GC) { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout info for |gl| and |GC|. \precond |gl| is a grid layout
	for |GC|.}*/

	GraphCopyGLI(const GCGridConceptBase &C, const IRect &bbox) :
		GCGridConceptBase(C) {
			set_bbox(bbox);
		}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout info for |C| with bounding box |bbox|.}*/

	GraphCopyGLI(GridLayout &gl, GraphCopy &GC, const IRect &bbox) :
		GCGridConceptBase(gl,GC) {
			set_bbox(bbox);
		}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout info for |gl| and |GC| with bounding box |bbox|. \precond
	|gl| is a grid layout for |GC|.}*/

	GraphCopyGLI(const GraphCopyGLI &gliA) : GCGridConceptBase(gliA) { }

	~GraphCopyGLI() { }


	/*{\Moperations}*/
	/*{\Mtext
	|\Mname| implements the access methods of its base class |GridLayoutInfo|.
	}*/

	int x (leda::node v) const {
		return _glp->x(_gcp->copy(v));
	}

	int y (leda::node v) const {
		return _glp->y(_gcp->copy(v));
	}

	IPolyline bends (leda::edge e) const {
		return get_bends(e);
	}

	int grid_width (leda::node v) const {
		return _glp->width(_gcp->copy(v));
	}

	int grid_height (leda::node v) const {
		return _glp->height(_gcp->copy(v));
	}

	IRect get_gbbox() {
		return GCGridConceptBase::get_bbox();
	}

	void grid_width (leda::node v, int w) const {
		_glp->width(_gcp->copy(v),w);
	}

	void grid_height (leda::node v, int h) const {
		_glp->height(_gcp->copy(v),h);
	}
};


#endif
