/*******************************************************************************
+
+  AGD 1.2
+
+
+  GraphGLI.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         GraphGLI.h
//
//    author: Carsten Gutwenger
//    date:   july 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_GRAPH_GLI_H
#define _AGD_GRAPH_GLI_H


#include <AGD/GridLayoutInfo.h>
#include <AGD/GridLayout.h>


/*{\Manpage {GraphGLI} {} {GLI's for Graphs} {gli} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{GridLayoutInfo}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{GraphGLI}
\end{hierarchy}
}*/

class __exportAGD GraphGLI : public GridLayoutInfo {

/*{\Mdefinition
An instance of type |\Mname| represents a grid layout info for a graph
G. The grid layout is simply given by a grid layout |gl| of |G|. Thus,
the access methods of |\Mvar| directly call the respective functions of |gl|.
The bounding box of |gl| can be passed to the constructor of |\Mname|.
If not, the bounding box is explicitly computed in the constructor.
}*/

	GridLayout *_glp;
	IRect      _bbox;

public:

	/*{\Mcreation}*/

	GraphGLI(GridLayout &gl) : _glp(&gl) {
		_bbox = gl.bounding_box();
	}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout info for grid layout |gl|.}*/

	GraphGLI(GridLayout &gl, const IRect &bbox) : _glp(&gl), _bbox(bbox) { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout info for grid layout |gl| with bounding box |bbox|.}*/

	GraphGLI(const GraphGLI &gliA) : _glp(gliA._glp), _bbox(gliA._bbox) { }
	/*{\Mcreate copy constructor.}*/

	~GraphGLI() { }


	/*{\Moperations}*/

	GraphGLI& operator=(const GraphGLI &gliA) {
		_glp = gliA._glp;
		_bbox = gliA._bbox;
		return *this;
	}
	/*{\Mbinop assigns |gliA| to |\Mvar|.}*/

	/*{\Mtext\medskip
	|\Mname| implements the access methods of its base class |GridLayoutInfo|.
	}*/

	int x (leda::node v) const {
		return _glp->x(v);
	}

	int y (leda::node v) const {
		return _glp->y(v);
	}

	IPolyline bends (leda::edge e) const {
		return _glp->get_compact_bends(e);
	}

	int grid_width (leda::node v) const {
		return _glp->width(v);
	}

	int grid_height (leda::node v) const {
		return _glp->height(v);
	}

	IRect get_gbbox() {
		return _bbox;
	}

	void grid_width (leda::node v, int w) const {
		_glp->width(v,w);
	}

	void grid_height (leda::node v, int h) const {
		_glp->height(v,h);
	}
};


#endif
