/*******************************************************************************
+
+  AGD 1.2
+
+
+  LongestPathCompaction.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 189 "Lweb/compaction.lw"
#ifndef _AGD_LONGEST_PATH_COMPACTION_H 
#define _AGD_LONGEST_PATH_COMPACTION_H

#include <AGD/CompactionModule.h>


/*{\Moptions
outfile=LongestPathCompaction.man
}*/

/*{\Manpage {LongestPathCompaction} {} {Longest Path--Based One--dim. Compaction} {C}
  }*/

/*{\Mbaseclasses

  \begin{hierarchy}{8}{4}
  \hdclass{0}{3}{5.5}{AGDModule} \arhook{5.5}{3.5}{6.5}
  \hdclass{0.5}{1.5}{6.5}{CompactionModule} \arhook{7}{2}{8}
  \hdclass{1}{0}{7.5}{LongestPathCompaction}
  \end{hierarchy}
  }*/

class __exportAGD LongestPathCompaction : public CompactionModule {
  
  /*{\Mdefinition
The class |\Mname| represents the one--dimensional compaction module based on longest path computations in the so called constraint graphs \cite{KM98b}.
}*/
  
public:
  
  /*{\Mcreation C 5}*/

  LongestPathCompaction() { }
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

  ~LongestPathCompaction() {}
  
  /*{\Moperations 1.5 2.5}*/
  
  leda::string name() const { 
    return leda::string("1_dim Compaction (long. paths)"); 
  }
  
  leda::string long_name() const { 
    return leda::string("One-Dimensional Compaction Module Based on Longest Paths"); 
  }
  
  leda::string author() const { return leda::string("?"); }
  
  virtual leda::string impl_author() const { 
    return leda::string("G. W. Klau"); 
  }
  
  virtual leda::string impl_version() const { return
    leda::string("1.0"); }
  
  virtual leda::string impl_date() const { return
                                            leda::string("July 1998"); }
  
  /*{\Mtext \headerline{Standard Interface} }*/

  
  double call(GridLayout& L); 
  /*{\Mop Performs a visibility--based compaction on the layout. The compaction
    is realized by an alternating sequence of longest path computations in the
    constraint graphs for $x$ and $y$--direction}*/
  
  int call(const leda::graph& G, leda::node_map<int>& x, leda::node_map<int>& y); 

  
  int call(const leda::graph& G, LayoutInterface& I);
  
  void call(const OrthogonalRepresentation& H, leda::node_map<int>& x, bool
            horizontally); 
  
  int call(const OrthogonalRepresentation& H, GridLayout& L); 
  /*{\Mop Performs a visibility--based compaction on the constraint graphs
    defined by the shape of the drawing. To do this, $H$ must be equivalent
    to a complete shape description.}*/

  //  int call(const OrthogonalRepresentation& H, GridLayout& L, 
  //   leda::list<Label>& Lambda); 
  
  AGDModule *clone() const;
  
protected:
  
  void longest_paths(const leda::graph& G, leda::node_array<int>& l);
  
  void assign_gridlines(leda::graph& G, const
                        leda::node_array<double>& x, 
                        leda::node_map<int>& x_grid);
  
};

#endif

