/*******************************************************************************
+
+  AGD 1.2
+
+
+  OptCrossMin.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         OptCrossMin.h
//
//
//    author: Matthias Elf
//    date:   September 1998
//-------------------------------------------------------------------


#ifndef _AGD_CROSS_MIN_OPT_H
#define _AGD_CROSS_MIN_OPT_H

#include <AGD/TwoLayerCrossMin.h>
#include <AGD/Hierarchy.h>


/*{\Manpage {OptCrossMin} {} {Optimal Crossing Minimization} {M} }*/

/*{\Mbaseclasses
\abacus
\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{6}{OptCrossMin}
\end{hierarchy}
}*/

class __exportAGDopt OptCrossMin : public TwoLayerCrossMin {

/*{\Mdefinition
The class |\Mname| implements an optimal two layer crossing minimization 
based on a branch \& cut algorithm for the \emph{linear ordering 
problem} \cite{GJR84}. It reorders the nodes on the changeable layer such that the number 
of crossings is minimum.

\horizline
\commoninfo{Opt. Cross. Min.}{Optimal Crossing Minimization}{M. Gr\"otschel, M. J\"unger, G. Reinelt}{M. Elf}
{September 1998}{}

\headerline{Pre- and Postcondition}
\begin{tabular}{l@{ = }l}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{tabular}

\horizline
}*/

public:

	/*{\Mcreation M 4}*/

	OptCrossMin() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~OptCrossMin() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Opt. Cross. Min.";
	}

	leda::string long_name () const {
		return "Optimal Crossing Minimization";
	}

	leda::string author () const {
		return "M. Gr\224tschel, M. J\201nger, G. Reinelt";
	}

	leda::string impl_author () const {
		return "M. Elf";
	}

	leda::string impl_date () const {
		return "September 1998";
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/*
	void init (const leda::graph& G, const Hierarchy& H);
	*/
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	/*
	void cleanup();
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


	AGDModule *clone() const {
		return new OptCrossMin();
	}

private:
	void setupCostMatrix( const Layer &L );
};


#endif
