/*******************************************************************************
+
+  AGD 1.2
+
+
+  PSInterface.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      PSInterface.h
//
//    author: David Alberts
//    date:   December 1997
//-------------------------------------------------------------------


#include <AGD/LayoutInterface.h>


#ifndef _AGD_PS_INTERFACE_H
#define _AGD_PS_INTERFACE_H


/*{\Manpage {PSInterface} {} {Interface to PostScript} {A} }*/

/*{\Mbaseclasses

\begin{hierarchy}{7}{2.5}
\hdclass{0}{1.5}{5.5}{LayoutInterface}
\arhook{5.5}{2}{6.5}
\hdclass{1}{0}{6}{PSInterface}
\end{hierarchy}
}*/

class __exportAGD PSInterface : public LayoutInterface {

/*{\Mdefinition
  An instance |\Mvar| of type |\Mname| is the interface to the
  graphical attributes in a PostScript$^{TM}$ file. When a layout
  algorithm is called with an instance of |\Mname| as a parameter,
  then the drawing is saved to a file in PostScript$^{TM}$ format.
  The file name can be specified in the constructor of |\Mname|.
  The generated file can be used as Encapsulated PostScript$^{TM}$,
  too.
}*/

public:

  /*{\Mcreation}*/

  PSInterface(const leda::graph &G, leda::string name="unnamed.ps");
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
       to an interface to file |name|.}*/

  ~PSInterface() { }


  /*{\Moperations}*/
	/*{\Mtext
	|\Mname| inherits the operations from its base class |LayoutInterface|
	(see Sect.~\ref{Layout Interfaces})
	}*/

  double get_x(leda::node v) const { return xcoord[v]; }

  void set_x(leda::node v, double new_x) { xcoord[v] = new_x; }

  double get_y(leda::node v) const { return ycoord[v]; }

  void set_y(leda::node v, double new_y) { ycoord[v] = new_y; }

  DPolyline get_bend_points(leda::edge e) const { return bends[e]; }

  void set_bend_points(leda::edge e, const DPolyline& l) { bends[e] = l; }

  double get_width(leda::node v) const { return node_width[v]; }

  void set_width(leda::node v, double new_w) { node_width[v] = new_w; }

  double get_height(leda::node v) const { return node_height[v]; }

  void set_height(leda::node v, double new_h) { node_height[v] = new_h; }

  DPoint get_source_anchor(leda::edge e) const { return sanch[e]; }

  void set_source_anchor(leda::edge e, const DPoint& pos) { sanch[e] = pos; }

  DPoint get_target_anchor(leda::edge e) const { return tanch[e]; }

  void set_target_anchor(leda::edge e, const DPoint& pos) { tanch[e] = pos; }

  void init(const leda::graph &G, leda::string name);
  void update();

private:

  void _init(const leda::graph &G, const leda::string &name);

  leda::string              filename;
  leda::node_map<double>    xcoord;      // node coordinates
  leda::node_map<double>    ycoord;
  leda::node_map<double>    node_width;
  leda::node_map<double>    node_height; 
  leda::edge_map<DPolyline> bends;       // edge bends
  leda::edge_map<DPoint>    sanch;       // source and target anchors
  leda::edge_map<DPoint>    tanch;

};

#endif
