//-------------------------------------------------------------------
//
//                       ParallelEdgeRanking.h
//
//
//    author: Sven Thiel
//    date:   may 2003
//-------------------------------------------------------------------


#ifndef _AGD_P_E_RANKING_H
#define _AGD_P_E_RANKING_H


#include <AGD/RankAssignment.h>
#include <AGD/AGDModuleOption.h>


/*{\Manpage {ParallelEdgeRanking} {} {Ranking for Graphs with Parallel Edges} {R} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1.5}{1.5}{5.5}{RankAssignment}
\arhook{7}{2}{8}
\hdclass{2}{0}{6.5}{ParallelEdgeRanking}
\end{hierarchy}
}*/

class __exportAGD ParallelEdgeRanking : public RankAssignment {

/*{\Mdefinition
The |\Mname| module computes a ranking of a graph $G=(V,E)$ 
(with parallel edges):

\horizline
\commoninfo{Parallel Edge Ranking}{Parallel Edge Ranking}
{}{S. Thiel}{March 2003}{1.0}

\begin{preposttab}
precondition & $\{\, |directed| \,\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf basic ranking}\\
a |RankAssignment| option used for computing an initial ranking.
\end{options}
}*/

public:

	/*{\Mcreation}*/

	ParallelEdgeRanking();
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~ParallelEdgeRanking() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Parallel Edge Ranking";
	}

	leda::string impl_author () const {
		return "S. Thiel";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "March 2003";
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|RankAssignment|).
	}*/

	/* inherited
	bool check (const leda::graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void call (const leda::graph& G, leda::node_array<int>& rank);
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_subgraph (const SubgraphModule& M);
	*/
	/*{\Mop}*/

	AGD_DECLARE_MODULE (RankAssignment, basic_ranking)

	/*{\Mtext \stdend}*/


private:
};

/*{\Mimplementation
The computation of the ranking takes time $\BigO(\Labs{V})$ 
plus the time required for the basic ranking.
}*/


#endif
