/*******************************************************************************
+
+  AGD 1.2
+
+
+  PostCond.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           PostCond.h
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_POST_COND_H
#define _AGD_POST_COND_H

#include <AGD/CondBase.h>


/*{\Manpage {PostCond} {} {Postconditions} {P} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{CondBase}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{PostCond}
\end{hierarchy}
}*/

class __exportAGD PostCond : public CondBase {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is a postcondition of a module $M$.
It is represented by a collection of key sets $S_0,\ldots,S_m$.
The key set $S_i$, $i\geq1$ corresponds to the $i$th output parameter of $M$,
i.e., $m$ is the number of output parameters. The set $S_0$ has a
special role. It represents the general postcondition of $M$.
}*/

public:

	/*{\Mcreation}*/

	PostCond(int m = 1) : CondBase(0,m) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|
	initialized to a postcondition for $m$ output parameters.
	Each key set is initialized to the empty set.}*/

	PostCond(const PostCond& p) : CondBase(p) { }

	~PostCond() { }

	/*{\Moperations}*/
	/*{\Mtext |\Mname| inherits the operations of its base class
	|CondBase|.}*/

	PostCond &operator=(const PostCond &c) {
		CondBase::operator=(c);
		return *this;
	}
};



#endif
