/*******************************************************************************
+
+  AGD 1.2
+
+
+  PostCondRule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//-------------------------------------------------------------------
//
//                        PostCondRule.h
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_PRECONDITION_H
#define _AGD_PRECONDITION_H

#include <AGD/TaggedRule.h>
#include <LEDA/core/list.h>
#include <LEDA/core/array.h>



/*{\Manpage {PostCondRule} {} {Postcondition Rules} {P} }*/

class __exportAGD PostCondRule : private leda::array<leda::list<TaggedRule> > {
typedef leda::array<leda::list<TaggedRule> > lalt;
/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is a rule for the postcondition
of a module $M$. It is represented by a collection of lists of tagged rules 
$R_0,\ldots,R_m$. The list $r_i$, $i\geq1$ corresponds to the $i$th output
parameter of $M$, i.e., $m$ is the number of output parameters.
}*/

public:

	/*{\Mcreation}*/

	PostCondRule(int m = 1) : leda::array<leda::list<TaggedRule> >(0,m) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|
	initialized to a postcondition rule for $m$ output parameters.
	Each list is initialized to the empty list.}*/

	PostCondRule(const PostCondRule &p) : leda::array<leda::list<TaggedRule> >(p)
	{ }

	~PostCondRule() { }


	/*{\Moperations}*/

	int num_param() const {
		return high();
	}
	/*{\Mop returns the number of parameters for which |\Mvar|
	is defined; more exactly it returns |m|.}*/

	const leda::list<TaggedRule> &operator[](int i) const {
		return access(i);
	}

	leda::list<TaggedRule> &operator[](int i) {
		return access(i);
	}
	/*{\Marrop returns the list $R_i$ of rules corresponding to the $i$th
	parameter.}*/

	const TaggedRule &rule(int i, leda::list_item it) const {
		return access(i)[it];
	}

	TaggedRule &rule(int i, leda::list_item it) {
		return access(i)[it];
	}
	/*{\Mop returns the rule at list item |it| in $R_i$.}*/

	AgdKey right(int i, leda::list_item it) const {
		return access(i)[it].right();
	}
	/*{\Mop returns the right side of the rule at
	list item |it| in $R_i$.}*/

	const leda::list<TRLiteral>& left(int i, leda::list_item it) const {
		return access(i)[it].left();
	}
	/*{\Mop returns the left side of the rule at
	list item |it| in $R_i$.}*/

	void del(int i, leda::list_item it) {
		access(i).del_item(it);
	}
	/*{\Mop deletes the item |it| in $R_i$.}*/

	PostCondRule &operator=(const PostCondRule &p) {
		lalt::operator=(p);
		return *this;
	}

private:
	const leda::list<TaggedRule> &access(int i) const {
		return lalt::operator[](i);
	}

	leda::list<TaggedRule> &access(int i) {
		return lalt::operator[](i);
	}

};


#endif
