/*******************************************************************************
+
+  AGD 1.2
+
+
+  generators.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        generators.h
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#ifndef _AGD_GENERATORS_H
#define _AGD_GENERATORS_H


#include <AGD/basic.h>
#include <LEDA/graph/graph.h>

//class leda::graph;

class LeveledGraph;



/*{\Manpage {generators} {} {Graph Generators} }*/


void __exportAGD planar_biconnected_graph(leda::graph &G, int n, int m,
	bool multiEdges = false, bool planarMap = false);
/*{\Mfuncl   creates a planar, biconnected graph
$G$ with $n$ vertices and $m$ edges. If |multiEdges| is set to
|false|, $G$ will not contain multi-edges. If |planarMap|
is set to |true|, $G$ will represent a planar map, that is,
all reversal edges will be present, otherwise not.

The generator creates always planar biconnected graphs with $n'$ nodes
and $m'$ edges, such that $n'\geq3$ and $n'\leq m'\leq 3n'-6$. The
parameters $n$ and $m$ are adapted if they violate one of the
equations above.}*/

void __exportAGD planar_triconnected_graph (leda::graph &G, int n, int m);
/*{\Mfuncl   creates a planar, triconnected and simple graph $G$ with $n$
nodes and $m$ edges. The number of edges can differ slightly from $m$.
Note: For a triconnected planar graph with $n$ nodes $(n\geq 4)$ and $m$ edges
always holds $\left\lceil\frac{3}{2}n\right\rceil \leq m \leq 3n-6.$}*/

void __exportAGD random_tree(leda::graph &G, int n, int max_deg=0, int max_width=0);
/*{\Mfuncl creates a rooted tree $G$ with $n$ nodes,
maximal node degree |max_deg|, and maximal level width |max_width|.
If |max_deg| is $0$, the is maximal degree is not restricted,
if |max_width| is $0$, the level width is not restricted.}*/

void __exportAGD st_planar_graph (leda::graph &G, int n, int m, bool multi_edges, double st_prob);
/*{\Mfuncl   creates an st-planar graph
$G$ with $n$ vertices and $m$ edges. If |multi_edges| is set to
|false|, $G$ will not contain multi-edges. The parameter |st_prob|
determines the propability that the edge $(s,t)$ is contained
in $G$.}*/

void __exportAGD random_hierarchy(LeveledGraph& LG, int n,
	int m, bool planar = false,
	bool singleSource = false, bool longEdges = true);
/*{\Mfuncl creates a random leveled graph $LG$ with $n$ nodes
and $m$ edges. The graph is leveled planar, if |planar| is set to |true|.
It contains a single source, if |singleSource| is set to |true|. If
|longEdges| is set to |false|, then adjacent nodes are on neighbored
levels.}*/


#include <LEDA/graph/graph.h>

leda::node __exportAGD ST_NUMBERING1(const leda::graph& G, leda::node_array<int>& st_num,
	leda::list<leda::node>& st_list, leda::edge e_st = nil);


#endif
