/************************************************************************

         Filename   :   PQTree.h
         Version    :   1.0.3  1997
         Author     :   Sebastian Leipert
         Language   :   C++

************************************************************************/

#ifndef PQTREE_H
#define PQTREE_H
#include <LEDA/system/basic.h>
#include <LEDA/system/stream.h>
#include <LEDA/internal/std/string.h>

#include "PQTreeRoot.h"
#include "PQTree.inc"

#include "pqNode.h"
#include "internalNode.h"
#include "leaf.h"
#include "pqQueue.h" 
#include "pqStack.h"
#include "leafKey.h"
#include "internalKey.h"
#include "pqNodeKey.h"



template<class T,class X,class Y>

class PQTree: public PQTreeRoot {

public:

           PQTree();

   virtual ~PQTree();

   int     addNewLeavesToTree(internalNode<T,X,Y>* father,
                              leafKey<T,X,Y>** ArrayOfElements,
                              int numberOfElements);

   int     addNewLeavesToTree(internalNode<T,X,Y>* father,
                              leafKey<T,X,Y>** ArrayOfElements,
                              int numberOfElements,
                              pqNode<T,X,Y>* leftSibling,
                              pqNode<T,X,Y>* rightSibling);

   void    emptyNode(pqNode<T,X,Y>* node_ptr);

   int     front(pqNode<T,X,Y>* node_ptr,leafKey<T,X,Y>** ArrayOfElements);

   int     frontExcept(pqNode<T,X,Y>* node_ptr,
                       leafKey<T,X,Y>** ArrayOfElements,
                       int exceptNumber,
                       int* Exceptions);

   void    printOutCurrentTree(char *filename,char* ending);

         
   
      virtual void CleanNode(pqNode<T,X,Y>** nodePtr);

      virtual void Cleanup();

      virtual void clientDefinedEmptyNode(pqNode<T,X,Y>* node_ptr);

      virtual void emptyAllPertinentNodes();

      virtual int Initialize(int numberOfElements,
                             leafKey<T,X,Y>** ArrayOfElements);

      virtual int Reduction(int numberOfConsecutiveElements,
                            leafKey<T,X,Y>** ArrayOfElements,
                            int redNumber = 0);



protected:

      pqNode<T,X,Y>*   _root;                     
      pqNode<T,X,Y>*   _pertinentRoot;
      pqNode<T,X,Y>*   _pseudoRoot;   

      int            _identificationNumber;      
      int            _numberOfLeaves;

      pqStack<pqNode<T,X,Y>*>  *_pertinentNodes; 



   

   #ifdef DEBUG
   virtual   int   Bubble(int numberOfConsecutiveElements,
                          leafKey<T,X,Y>** ArrayOfElements,
                          int redNumber);
   #else
   virtual   int   Bubble(int numberOfConsecutiveElements,
                          leafKey<T,X,Y>** ArrayOfElements);
   #endif

   virtual   int   Reduce(int numberOfConsecutiveElements,
                          leafKey<T,X,Y>** ArrayOfElements,int printYes = 0);

   virtual   int   template_L1(pqNode<T,X,Y> *node_ptr,
                               int isRoot);

   virtual   int   template_P1(pqNode<T,X,Y> *node_ptr,
                               int isRoot);

   virtual   int   template_P2(pqNode<T,X,Y> **node_ptr);

   virtual   int   template_P3(pqNode<T,X,Y> *node_ptr);

   virtual   int   template_P4(pqNode<T,X,Y> **node_ptr);
      
   virtual   int   template_P5(pqNode<T,X,Y> *node_ptr);

   virtual   int   template_P6(pqNode<T,X,Y> **node_ptr);

   virtual   int   template_Q1(pqNode<T,X,Y> *node_ptr,
                               int isRoot); 
      
   virtual   int   template_Q2(pqNode<T,X,Y> *node_ptr,
                               int isRoot);

   virtual   int   template_Q3(pqNode<T,X,Y> *node_ptr);
      

   
   virtual int    addNodeToNewParent(pqNode<T,X,Y>* parent,
                                     pqNode<T,X,Y>* child);

   virtual int    addNodeToNewParent(pqNode<T,X,Y>* parent,
                                     pqNode<T,X,Y>* child,
                                     pqNode<T,X,Y>* leftBrother,
                                     pqNode<T,X,Y>* rightBrother);

   virtual int    checkIfOnlyChild(pqNode<T,X,Y> *child,
                                   pqNode<T,X,Y> *parent);

   virtual void   destroyNode(pqNode<T,X,Y> *node_ptr);

   virtual void   exchangeNodes(pqNode<T,X,Y> *oldNode,
                                pqNode<T,X,Y> *newNode);

   virtual void   linkChildrenOfQnode(pqNode<T,X,Y> *installed,
                                      pqNode<T,X,Y> *newChild);

   virtual void   removeChildFromSiblings(pqNode<T,X,Y>* node_ptr);

   virtual int    removeNodeFromTree(pqNode<T,X,Y>* parent,pqNode<T,X,Y>* child);




   
      pqStack<pqNode<T,X,Y>*>*  fullChildrenStack(pqNode<T,X,Y>* nodePtr);

      pqStack<pqNode<T,X,Y>*>*  partialChildrenStack(pqNode<T,X,Y>* nodePtr); 


   
      virtual pqNode<T,X,Y>* clientLeftEndmost(pqNode<T,X,Y>* nodePtr) const;

      virtual pqNode<T,X,Y>* clientRightEndmost(pqNode<T,X,Y>* nodePtr) const;

      virtual pqNode<T,X,Y>* clientNextSib(pqNode<T,X,Y>* nodePtr,
                                         pqNode<T,X,Y>* other) const;

      virtual pqNode<T,X,Y>* clientSibLeft(pqNode<T,X,Y>* nodePtr) const;

      virtual pqNode<T,X,Y>* clientSibRight(pqNode<T,X,Y>* nodePtr) const;


   
      virtual int   clientPrintNodeCategorie(pqNode<T,X,Y>* nodePtr);

      virtual char* clientPrintStatus(pqNode<T,X,Y>* nodePtr);

      virtual char* clientPrintType(pqNode<T,X,Y>* nodePtr);


private:

      
   void          addLeavesRunThrough(internalNode<T,X,Y>* parent,
                                     leafKey<T,X,Y>** ArrayOfElements,
                                     int numberOfElements,
                                     leaf<T,X,Y>* _actualSon);
                              
   int           checkChain(pqNode<T,X,Y> *node_ptr,
                            pqNode<T,X,Y> *firstFull,
                            pqNode<T,X,Y> **seqStart,
                            pqNode<T,X,Y> **seqEnd);

   void          copyFullChildrenToPartial(pqNode<T,X,Y> *node_ptr,
                                           pqNode<T,X,Y> *partialChild);

   pqNode<T,X,Y>*  createNodeAndCopyFullChildren(pqStack<pqNode<T,X,Y>*> *fullNodes);

   int           inExceptions(int searchKey,
                              int Exceptions[],
                              int arraySize);

   void          printNode(char *filename,
                           int number,
                           pqNode<T,X,Y>* father,
                           pqNode<T,X,Y>* son);  

   void          removeBlock(pqNode<T,X,Y> *node_ptr,
                             int isRoot);

   void          sortExceptions(int Exceptions[],
                                int arraySize);



};



/************************************************************************
                      addNewLeavesToTree
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::addNewLeavesToTree(internalNode<T,X,Y>* father,
                                  leafKey<T,X,Y>** ArrayOfElements,
                                  int numberOfElements)
{
   leafKey<T,X,Y>*        _newKey = NULL;
   leaf<T,X,Y>*       _aktualSon = NULL;
   pqNode<T,X,Y>*       _oldSon = NULL;
   pqNode<T,X,Y>*       _firstSon = NULL;

   int             i = 0;

   if (numberOfElements > 0)
   {
      if (father->_childCount > 0)   
      {
         printf("ERROR when adding new leaves: Brothers expected.\n");
         return FALSE;
      }
      else
      {
         _newKey = ArrayOfElements[0];
	 _aktualSon = new leaf<T,X,Y>(_identificationNumber++,EMPTY,_newKey);
	 _firstSon = _aktualSon;
	 _firstSon->_parent = father;
	 _firstSon->_parentType = father->type();
	 father->_childCount++;
	 _oldSon = _firstSon;


         for (i = 1; i <= (numberOfElements-1); i++)
	 {
	    _newKey = ArrayOfElements[i];
	    _aktualSon = new leaf<T,X,Y>(_identificationNumber++,EMPTY,_newKey);
	    _aktualSon->_parent = father;
	    _aktualSon->_parentType = father->type();
	    father->_childCount++;
	    _oldSon->_sibRight = _aktualSon;
	    _aktualSon->_sibLeft = _oldSon;
	    _oldSon = _aktualSon;
	 }


         if (father->type() == P_NODE)
	 {
	    _firstSon->_sibLeft = _oldSon;
	    _oldSon->_sibRight = _firstSon;
	    father->_referenceChild = _firstSon;
	    _firstSon->_referenceParent = father;
	 }
	 else if (father->type() == Q_NODE)
	 {
	    father->_leftEndmost = _firstSon;
	    father->_rightEndmost = _oldSon;
	 }

         
         return TRUE; 
      }
   }
   else return FALSE;
   
}



/************************************************************************
                      addNewLeavesToTree
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::addNewLeavesToTree(internalNode<T,X,Y>* parent,
                                  leafKey<T,X,Y>** ArrayOfElements,
                                  int numberOfElements,
                                  pqNode<T,X,Y>* leftBrother,
                                  pqNode<T,X,Y>* rightBrother)
{

   leafKey<T,X,Y>*      _newKey = NULL;
   leaf<T,X,Y>*     _actualSon = NULL;

   
   if ((leftBrother == NULL) && (rightBrother == NULL))
      return addNewLeavesToTree(parent,ArrayOfElements,numberOfElements);
   else if (numberOfElements > 0)
   {
      _newKey = ArrayOfElements[0];
      _actualSon = new leaf<T,X,Y>(_identificationNumber++,EMPTY,_newKey);
      _actualSon->_parent = parent;
      _actualSon->_parentType = parent->type();
      parent->_childCount++;


      if (parent->type() == P_NODE)
      {
         pqNode<T,X,Y>* _brother = NULL;
	 pqNode<T,X,Y>* _sister  = NULL;
	 if (leftBrother != NULL)
	    _brother = leftBrother;
	 else
	    _brother = rightBrother;
	 _sister = _brother->_sibRight;

	 _actualSon->_sibLeft = _brother;
	 _brother->_sibRight = _actualSon;

	 addLeavesRunThrough(parent,ArrayOfElements,numberOfElements,_actualSon);

	 _actualSon->_sibRight = _sister;
	 _sister->_sibLeft = _actualSon;
	 return TRUE;
	       

      }

      else if (leftBrother == NULL)
      {
         if (rightBrother == parent->_leftEndmost)
	 {
	    _actualSon->_sibRight = rightBrother;
	    if (!rightBrother->putSibling(_actualSon))
	       return FALSE;
	    else
	    {
	       addLeavesRunThrough(parent,ArrayOfElements,numberOfElements,_actualSon);
	       parent->_leftEndmost = _actualSon;
	       return TRUE;
	    }
	 }
	 else if (rightBrother == parent->_rightEndmost)
	 {
	    _actualSon->_sibLeft = rightBrother;
	    if (!rightBrother->putSibling(_actualSon))
	       return FALSE;
	    else
	    {
	       addLeavesRunThrough(parent,ArrayOfElements,numberOfElements,_actualSon);
	       parent->_rightEndmost = _actualSon;
	       return TRUE;
	    }
	 }
	 else
	 {
	    cout << "ERROR: PQTree: addNewLeavesToParent: "
	         << " missing second brother." << endl;
	    return FALSE;
	 }


      }

      else if (rightBrother == NULL)
      {
         if (leftBrother == parent->_rightEndmost)
	 {
	    _actualSon->_sibLeft = leftBrother;
	    if (!leftBrother->putSibling(_actualSon))
	       return FALSE;
	    else
	    {
	       addLeavesRunThrough(parent,ArrayOfElements,numberOfElements,_actualSon);
	       parent->_rightEndmost = _actualSon;
	       return TRUE;
	    }
	 }
	 else if (leftBrother == parent->_leftEndmost)
	 {
	    _actualSon->_sibRight = leftBrother;
	    if (!leftBrother->putSibling(_actualSon))
	       return FALSE;
	    else
	    {
	       addLeavesRunThrough(parent,ArrayOfElements,numberOfElements,_actualSon);
	       parent->_leftEndmost = _actualSon;
	       return TRUE;
	    }
	 }
	 else
	 {
	    cout << "ERROR: PQTree: addNewLeavesToParent: "
	         << " missing second brother." << endl;
	    return FALSE;
	 }


      }

      else
      {
         if (!leftBrother->changeSiblings(rightBrother,_actualSon))
	 {
	    cout << "ERROR: PQTree: addNewLeavesToParent: "
	         << " brothers are not immediate siblings." << endl;
	    return FALSE;
	 }
	 else
	 {
	    if (leftBrother->_sibRight == _actualSon)
	       _actualSon->_sibLeft = leftBrother;
	    else
	       _actualSon->_sibRight = leftBrother;
	    addLeavesRunThrough(parent,ArrayOfElements,numberOfElements,_actualSon);
	    _actualSon->putSibling(rightBrother);
	    return rightBrother->changeSiblings(leftBrother,_actualSon);
	 }


      }
   }
   else
      return TRUE;
}





/************************************************************************
                         CleanNode
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::CleanNode(pqNode<T,X,Y>** nodePtr)
{
  (*nodePtr) = (*nodePtr);
}





/************************************************************************
                         Cleanup
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::Cleanup()
{

   pqNode<T,X,Y>*  _checkNode = NULL;
   pqNode<T,X,Y>*  _firstSon  = NULL;
   pqNode<T,X,Y>*  _nextSon   = NULL;
   pqNode<T,X,Y>*  _lastSon   = NULL;
   pqNode<T,X,Y>*  _oldSib    = NULL;
   pqNode<T,X,Y>*  _holdSib   = NULL;

  
   pqQueue<pqNode<T,X,Y>*> *helpqueue = new pqQueue<pqNode<T,X,Y>*>;


   if (_root != NULL)
   {
      emptyAllPertinentNodes();
      if (_root->type() == P_NODE)
      {
         if (_root->_referenceChild != NULL) 
         {
            _firstSon = _root->_referenceChild;
            helpqueue->enQueue(_firstSon);

            if (_firstSon->_sibRight != NULL)
            _nextSon = _firstSon->_sibRight;
            while (_nextSon != _firstSon)
            {
               helpqueue->enQueue(_nextSon);
               _nextSon = _nextSon->_sibRight;
            }
         }
      }
      else if (_root->type() == Q_NODE)
      {  
         _firstSon = _root->_leftEndmost;
         helpqueue->enQueue(_firstSon);
            
         _lastSon = _root->_rightEndmost;
         helpqueue->enQueue(_lastSon);

         _nextSon = _lastSon->getNextSib(_oldSib);
         _oldSib = _lastSon;
         while (_nextSon != _firstSon)
         {
            helpqueue->enQueue(_nextSon);
            _holdSib = _nextSon->getNextSib(_oldSib);
            _oldSib = _nextSon;
            _nextSon = _holdSib;
         }
      }


      CleanNode(&_root);
      if (_root)
          delete _root;

      while (!helpqueue->queueEmpty())
      {
         _checkNode = helpqueue->deQueue();

	 if (_checkNode->type() == P_NODE)
	 {  
	    if (_checkNode->_referenceChild != NULL) 
	    {
	       _firstSon = _checkNode->_referenceChild;
	       helpqueue->enQueue(_firstSon);

	       if (_firstSon->_sibRight != NULL)
	       _nextSon = _firstSon->_sibRight;
	       while (_nextSon != _firstSon)
	       {
	          helpqueue->enQueue(_nextSon);
	          _nextSon = _nextSon->_sibRight;
	       }
	    }
	 }
	 else if (_checkNode->type() == Q_NODE)
	 {
	    _oldSib = NULL;
	    _holdSib = NULL;

	    _firstSon = _checkNode->_leftEndmost;
	    helpqueue->enQueue(_firstSon);

	    _lastSon = _checkNode->_rightEndmost;
	    helpqueue->enQueue(_lastSon);

	    _nextSon = _lastSon->getNextSib(_oldSib);
	    _oldSib = _lastSon;
	    while (_nextSon != _firstSon)
	    {
	       helpqueue->enQueue(_nextSon);
	       _holdSib = _nextSon->getNextSib(_oldSib);
	       _oldSib = _nextSon;
	       _nextSon = _holdSib;
	    }
	 }


         CleanNode(&_checkNode);
         if (_checkNode)
             delete _checkNode;
      }
   }
   delete helpqueue;

   CleanNode(&_pseudoRoot);
   if (_pseudoRoot)
       delete _pseudoRoot;

   if (_pertinentNodes)
       delete _pertinentNodes;

   _root = NULL;
   _pertinentRoot = NULL;
   _pseudoRoot = NULL;
   _pertinentNodes = NULL;

   _numberOfLeaves = 0;
   _identificationNumber = 0;
   
}



/************************************************************************
                 clientDefinedEmptyNode
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::clientDefinedEmptyNode(pqNode<T,X,Y>* nodePtr)
{
   emptyNode(nodePtr);
}



/************************************************************************
                         Constructor
************************************************************************/ 


template<class T,class X,class Y> PQTree<T,X,Y>::PQTree()
{
   _root = NULL;
   _pertinentRoot = NULL;
   _pseudoRoot = NULL;
   
   _numberOfLeaves = 0;
   _identificationNumber = 0;

   _pertinentNodes = NULL;
}



/************************************************************************
                         Destructor
************************************************************************/ 


template<class T,class X,class Y> PQTree<T,X,Y>::~PQTree()
{
   Cleanup();
}



/************************************************************************
                 emptyAllPertinetNodes
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::emptyAllPertinentNodes()
{
   pqNode<T,X,Y>   *node_ptr;


   while(!_pertinentNodes->stackEmpty())
   {
      node_ptr = _pertinentNodes->pop();
      switch (node_ptr->status())
      {
         case TO_BE_DELETED:
                               if (node_ptr == _root)
                                   _root = NULL;
                               CleanNode(&node_ptr);
                               if (node_ptr)
                                   delete node_ptr;
                               break;

         case FULL:
                               emptyNode(node_ptr);
                               break;
         case PARTIAL:
                               emptyNode(node_ptr);
                               break;

         default:
                               clientDefinedEmptyNode(node_ptr);
                               break;
      }
   }
   _pseudoRoot->_pertChildCount = 0;
   _pseudoRoot->_pertLeafCount = 0;
   _pseudoRoot->fullChildren->cleanupStack();
   _pseudoRoot->partialChildren->cleanupStack();
   _pseudoRoot->status(EMPTY);
   _pseudoRoot->mark(UNMARKED);
}



/************************************************************************
                        emptyNode
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::emptyNode(pqNode<T,X,Y> *node_ptr)
{
   node_ptr->status(EMPTY);
   node_ptr->_pertChildCount = 0;
   node_ptr->_pertLeafCount = 0;
   node_ptr->fullChildren->cleanupStack();
   node_ptr->partialChildren->cleanupStack();
   node_ptr->mark(UNMARKED);
}



/************************************************************************
                        front
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::front(pqNode<T,X,Y>* node_ptr,leafKey<T,X,Y>** ArrayOfElements)
{

   
   pqNode<T,X,Y>*  _checkNode = NULL;
   pqNode<T,X,Y>*  _firstSon  = NULL;
   pqNode<T,X,Y>*  _nextSon   = NULL;
   pqNode<T,X,Y>*  _lastSon   = NULL;
   pqNode<T,X,Y>*  _oldSib    = NULL;
   pqNode<T,X,Y>*  _holdSib   = NULL;

   int      _identificationNumber = 0;
  
   pqQueue<pqNode<T,X,Y>*> *helpqueue = new pqQueue<pqNode<T,X,Y>*>;
   helpqueue->enQueue(node_ptr);
  
   while (!helpqueue->queueEmpty())
   {
      _checkNode = helpqueue->deQueue();

      if (_checkNode->type() == P_NODE)
      {  
         if (_checkNode->_referenceChild != NULL) 
         {
            _firstSon = _checkNode->_referenceChild;
            helpqueue->enQueue(_firstSon);
 
            if (_firstSon->_sibRight != NULL)
               _nextSon = _firstSon->_sibRight;
            while (_nextSon != _firstSon)
            {
               helpqueue->enQueue(_nextSon);
               _nextSon = _nextSon->_sibRight;
            }
         }
      }
      else if (_checkNode->type() == Q_NODE)
      {
         _oldSib = NULL;
         _holdSib = NULL;

         _firstSon = _checkNode->_leftEndmost;
         helpqueue->enQueue(_firstSon);
      
         _lastSon = _checkNode->_rightEndmost;
         helpqueue->enQueue(_lastSon);

         _nextSon = _lastSon->getNextSib(_oldSib);
         _oldSib = _lastSon;
         while (_nextSon != _firstSon)
         {
            helpqueue->enQueue(_nextSon);
            _holdSib = _nextSon->getNextSib(_oldSib);
            _oldSib = _nextSon;
            _nextSon = _holdSib;
         }

      }

      else if (_checkNode->type() == LEAF)
         ArrayOfElements[_identificationNumber++] = _checkNode->getKey();      

   }

   delete helpqueue;

   return _identificationNumber;
}



/************************************************************************
                        frontExcept
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::frontExcept(pqNode<T,X,Y>* node_ptr,leafKey<T,X,Y>** ArrayOfElements,
                           int exceptNumber,int* Exceptions)
{

   
   pqNode<T,X,Y>*  _checkNode = NULL;
   pqNode<T,X,Y>*  _firstSon  = NULL;
   pqNode<T,X,Y>*  _nextSon   = NULL;
   pqNode<T,X,Y>*  _lastSon   = NULL;
   pqNode<T,X,Y>*  _oldSib    = NULL;
   pqNode<T,X,Y>*  _holdSib   = NULL;

   int      _identificationNumber = 0;

   if (exceptNumber <= 0)
      return front(node_ptr,ArrayOfElements);
   else
   {
      sortExceptions(Exceptions,exceptNumber);
      pqQueue<pqNode<T,X,Y>*> *helpqueue = new pqQueue<pqNode<T,X,Y>*>;
      if (!inExceptions(node_ptr->_identificationNumber,
                        Exceptions,exceptNumber))
         helpqueue->enQueue(node_ptr);
  
      while (!helpqueue->queueEmpty())
      {
         _checkNode = helpqueue->deQueue();

         if (_checkNode->type() == P_NODE)
         {  
            if (_checkNode->_referenceChild != NULL) 
            {
               _firstSon = _checkNode->_referenceChild;
               if (!inExceptions(_firstSon->_identificationNumber,
                                 Exceptions,exceptNumber))
                  helpqueue->enQueue(_firstSon);
 
               if (_firstSon->_sibRight != NULL)
                  _nextSon = _firstSon->_sibRight;
               while (_nextSon != _firstSon)
               {
                  if (!inExceptions(_nextSon->_identificationNumber,
                                    Exceptions,exceptNumber))
                     helpqueue->enQueue(_nextSon);
                  _nextSon = _nextSon->_sibRight;
               }
            }
         }
         else if (_checkNode->type() == Q_NODE)
         {
            _oldSib = NULL;
            _holdSib = NULL;

            _firstSon = _checkNode->_leftEndmost;
            if (!inExceptions(_firstSon->_identificationNumber,
                              Exceptions,exceptNumber))
               helpqueue->enQueue(_firstSon);
      
            _lastSon = _checkNode->_rightEndmost;
            if (!inExceptions(_lastSon->_identificationNumber,
                              Exceptions,exceptNumber))
               helpqueue->enQueue(_lastSon);

            _nextSon = _lastSon->getNextSib(_oldSib);
            _oldSib = _lastSon;
            while (_nextSon != _firstSon)
            {
               if (!inExceptions(_nextSon->_identificationNumber,
                                 Exceptions,exceptNumber))
                  helpqueue->enQueue(_nextSon);
               _holdSib = _nextSon->getNextSib(_oldSib);
               _oldSib = _nextSon;
               _nextSon = _holdSib;
            }

         }

         else if (_checkNode->type() == LEAF)
            ArrayOfElements[_identificationNumber++] = _checkNode->getKey();      

      }

      delete helpqueue;

      return _identificationNumber;
   }
}



/************************************************************************
                         Initialize
************************************************************************/ 


template<class T,class X,class Y> 
int PQTree<T,X,Y>::Initialize(int numberOfElements,leafKey<T,X,Y>** ArrayOfElements)
{
   internalNode<T,X,Y>   *_newNode   = NULL;
   internalNode<T,X,Y>   *_newNode2  = NULL;
   leaf<T,X,Y>     *_newLeaf   = NULL;

   _pertinentNodes = new pqStack<pqNode<T,X,Y>*>;


   if ( numberOfElements > 0)
   {
      _newNode2 = new internalNode<T,X,Y>(-1,Q_NODE,PARTIAL);
      _pseudoRoot = _newNode2;

       if ( numberOfElements > 1)
       {
           _newNode = new internalNode<T,X,Y>(_identificationNumber++,P_NODE,EMPTY);
           _root = _newNode;
           _root->_sibLeft = _root;
           _root->_sibRight = _root;
           return addNewLeavesToTree(_newNode,ArrayOfElements, numberOfElements);
       }
       else
       {
           _newLeaf = new leaf<T,X,Y>(_identificationNumber++,EMPTY,ArrayOfElements[0]);
           _root = _newLeaf;
           _root->_sibLeft = _root;
           _root->_sibRight = _root;
           return TRUE;
       }
   }    
   else
      return FALSE;
}



/************************************************************************
                 printOutCurrentTree
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::printOutCurrentTree(char *name,char *ending)
{
   char filename[INFOSIZE];

   strcpy(filename,name);
   strcat(filename,ending);

   ofstream output(filename,ios::out);
   output << "#TYPE: COMPLETE TREE" << endl
          << "#TIME: NOT" << endl
          << "#BOUNDS: NONE" << endl
          << "#INFORMATION: STANDARD" << endl
          << "#NODE_NUMBER: NONE" << endl;

   pqNode<T,X,Y>*  _checkNode = NULL;
   pqNode<T,X,Y>*  _firstSon  = NULL;
   pqNode<T,X,Y>*  _nextSon   = NULL;
   pqNode<T,X,Y>*  _lastSon   = NULL;
   pqNode<T,X,Y>*  _oldSib    = NULL;
   pqNode<T,X,Y>*  _holdSib   = NULL;
   
   int   _count = 1;
  
   pqQueue<pqNode<T,X,Y>*> *helpqueue = new pqQueue<pqNode<T,X,Y>*>;
   _root->_debugTreeNumber = _count;

   output << "n " << _count << ' ' << "\\iROOT: "
          << (_root->type() == Q_NODE ? " Q-NODE" :
             (_root->type() == P_NODE ? " P-NODE" : 
             (_root->type() == LEAF   ? " LEAF"   :
             (clientPrintType(_root) ) )))
          << ", IdentificationNumber: " << _root->_identificationNumber
          << ", Status: "
          << (_root->status() == FULL ? " FULL"           :
             (_root->status() == EMPTY ? " EMPTY"         :
             (_root->status() == PARTIAL ? " PARTIAL"     :
             (_root->status() == PERTINENT ? " PERTINENT" : 
             (clientPrintStatus(_root) ) ))))
          << (_root->getNodeInfo() != NULL ?
              _root->getNodeInfo()->print()   : " "  )
          << (_root->type() == LEAF ? " "                   :
             (_root->getInternal() != NULL ?
              _root->getInternal()->print()   : " " ))
          << (_root->type() != LEAF ? " "                   :
             (_root->getKey() != NULL ?
              _root->getKey()->print()   : " " ))
          << " Child Count: " << _root->_childCount
          << "\\i" << endl;

   output << "c " << _count++ << ' '
          << (_root->type() == Q_NODE  ? Q_NODE_CAT    : 
             (_root->type() == P_NODE  ? P_NODE_CAT    :
             (_root->status() == FULL  ? LEAF_FULL_CAT :
             (_root->status() == EMPTY ? LEAF_CAT      :
             (clientPrintNodeCategorie(_root) ) ))))
          << endl;
  
   helpqueue->enQueue(_root);
   
   while (!helpqueue->queueEmpty())
   {
      _checkNode = helpqueue->deQueue();

      if (_checkNode->type() == P_NODE)
      {  
         if (_checkNode->_referenceChild != NULL) 
         {
            _firstSon = _checkNode->_referenceChild;
            helpqueue->enQueue(_firstSon);
            printNode(filename,_count++,_checkNode,_firstSon);
 
            if (_firstSon->_sibRight != NULL)
               _nextSon = _firstSon->_sibRight;
            while (_nextSon != _firstSon)
            {
               helpqueue->enQueue(_nextSon);
               printNode(filename,_count++,_checkNode,_nextSon);
               _nextSon = _nextSon->_sibRight;
            }
         }
      }
      else if (_checkNode->type() == Q_NODE)
      {
         _oldSib = NULL;
         _holdSib = NULL;

         _firstSon = _checkNode->_leftEndmost;
         helpqueue->enQueue(_firstSon);
         printNode(filename,_count++,_checkNode,_firstSon);
      
         _lastSon = _checkNode->_rightEndmost;
         if ( _firstSon != _lastSon)
         {
             helpqueue->enQueue(_lastSon);
             printNode(filename,_count++,_checkNode,_lastSon);

             _nextSon = _lastSon->getNextSib(_oldSib);
             _oldSib = _lastSon;
             while (_nextSon != _firstSon)
             {
                 helpqueue->enQueue(_nextSon);
                 printNode(filename,_count++,_checkNode,_nextSon);
                 _holdSib = _nextSon->getNextSib(_oldSib);
                 _oldSib = _nextSon;
                 _nextSon = _holdSib;
             }
         }

      }

   }

   delete helpqueue;
}



/************************************************************************
                         Reduction
************************************************************************/ 

template<class T,class X,class Y> 
int PQTree<T,X,Y>::Reduction(int numberOfConsecutiveElements,
                             leafKey<T,X,Y>** ArrayOfElements,
                             int redNumber)
{

#ifdef DEBUG
   int success = Bubble(numberOfConsecutiveElements,ArrayOfElements,redNumber);
#else
   int success = Bubble(numberOfConsecutiveElements,ArrayOfElements);
#endif

   if (!success)
      return success;
   else
      return Reduce(numberOfConsecutiveElements,ArrayOfElements,redNumber);

}





/************************************************************************
                      addNodeToNewParent
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::addNodeToNewParent(pqNode<T,X,Y>* parent,
                                      pqNode<T,X,Y>* child)
{
   if (parent->type() != P_NODE && parent->type() != Q_NODE)
   {
       cout << "ERROR in addNodeToNewParent: parent type not valid" << endl;
       return FALSE;
   }
   if (child != NULL)
   {
      if (parent->_childCount > 0)   
      {
         printf("ERROR when adding new nodes: Brothers expected.\n");
         return FALSE;
      }
      else
      {
         child->_parent = parent;
         child->_parentType = parent->type();
         parent->_childCount++;

         if (parent->type() == P_NODE)
	 {
	    child->_sibLeft = child;
	    child->_sibRight = child;
	    parent->_referenceChild = child;
	    child->_referenceParent = parent;
	 }
	 else if (parent->type() == Q_NODE)
	 {
	    parent->_leftEndmost = child;
	    parent->_rightEndmost = child;
	 }


         
         return TRUE;
      }
   }
   else 
       return FALSE;
   
}



/************************************************************************
                      addNodeToNewParent
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::addNodeToNewParent(pqNode<T,X,Y>* parent,
                                      pqNode<T,X,Y>* child,
                                      pqNode<T,X,Y>* leftBrother,
                                      pqNode<T,X,Y>* rightBrother)
{
   
   if (parent != NULL)
   {
      if (parent->type() != P_NODE && parent->type() != Q_NODE)
      {
         cout << "ERROR in addNodeToNewParent: parent type not valid" << endl;
         return FALSE;
      }
      else if ((leftBrother == NULL) && (rightBrother == NULL))
         return addNodeToNewParent(parent,child);
      else if (child != NULL)
      {
         child->_parent = parent;
         child->_parentType = parent->type();
         parent->_childCount++;

         if (parent->type() == P_NODE)
         {
            pqNode<T,X,Y>* _brother = NULL;
	    pqNode<T,X,Y>* _sister  = NULL;
	    if (leftBrother != NULL)
	       _brother = leftBrother;
	    else
	       _brother = rightBrother;
	    _sister = _brother->_sibRight;
	    child->_sibLeft = _brother;
	    child->_sibRight = _sister;
	    _brother->_sibRight = child;
	    _sister->_sibLeft = child;
	    return TRUE;


         }
      
         else if (leftBrother == NULL)
         {
         if (rightBrother == parent->_leftEndmost)
	 {
	    parent->_leftEndmost = child;
	    child->_sibRight = rightBrother;
	    return rightBrother->putSibling(child,LEFT);
	 }
	 else if (rightBrother == parent->_rightEndmost)
	 {
	    parent->_rightEndmost = child;
	    child->_sibLeft = rightBrother;
	    return rightBrother->putSibling(child,LEFT);
	 }
	 else
	 {
	    cout << "ERROR: PQTree: addNodeToNewParent: "
	         << " missing second brother." << endl;
	    return FALSE;
	 }


         }

         else if (rightBrother == NULL)
         {
            if (leftBrother == parent->_rightEndmost)
	    {
	       parent->_rightEndmost = child;
	       child->_sibLeft = leftBrother;
	       return leftBrother->putSibling(child,RIGHT);
	    }
	    else if (leftBrother == parent->_leftEndmost)
	    {
	       parent->_leftEndmost = child;
	       child->_sibRight = leftBrother;
	       return leftBrother->putSibling(child,RIGHT);
	    }
	    else
	    {
	       cout << "ERROR: PQTree: addNodeToNewParent: "
	            << " missing second brother." << endl;
	       return FALSE;
	    }



         }

         else
         {
            if (rightBrother->changeSiblings(leftBrother,child) &&
	        leftBrother->changeSiblings(rightBrother,child))
	    {
	       if (leftBrother->_sibRight == child)
	       {
	          child->_sibLeft = leftBrother;
	          child->_sibRight = rightBrother;
	       }
	       else 
	       {
	          child->_sibLeft = rightBrother;
	          child->_sibRight = leftBrother;
	       }
	       return TRUE;
	    }
	    else
	    {
	       cout << "ERROR: PQTree: addNodeToNewParent: "
	            << " brothers are not siblings." << endl;
	       return FALSE;
	    }


	       






         }
      }
      else
          return FALSE;
   }
   else if (leftBrother != NULL && rightBrother != NULL)
   {
      if (rightBrother->changeSiblings(leftBrother,child) &&
          leftBrother->changeSiblings(rightBrother,child))
      {
         if (leftBrother->_sibRight == child)
         {
            child->_sibLeft = leftBrother;
            child->_sibRight = rightBrother;
         }
         else 
         {
            child->_sibLeft = rightBrother;
            child->_sibRight = leftBrother;
         }
         return TRUE;
      }
      else
      {
         cout << "ERROR: PQTree: addNodeToNewParent: "
              << " brothers are not siblings." << endl;
         return FALSE;
      }


         






   }       
   else
      return FALSE;

}





/************************************************************************
                           Bubble
************************************************************************/ 

#ifdef DEBUG
template<class T,class X,class Y>
int PQTree<T,X,Y>::Bubble(int number,
                          leafKey<T,X,Y>** ArrayOfElements,
                          int redNumber)
#else
template<class T,class X,class Y>
int PQTree<T,X,Y>::Bubble(int number,
                          leafKey<T,X,Y>** ArrayOfElements)
#endif
{
   pqQueue<pqNode<T,X,Y>*> *_processNodes = new pqQueue<pqNode<T,X,Y>*>;
   pqStack<pqNode<T,X,Y>*> *_blockedNodes = new pqStack<pqNode<T,X,Y>*>;

#ifdef DEBUG
   pqStack<pqNode<T,X,Y>*> *_processLeaves = new pqStack<pqNode<T,X,Y>*>;
#endif


   int         _blockCount       = 0;
   int         _numBlocked       = 0;
   int         _offTheTop        = 0;

   int         _blockedSiblings  = 0;
   int         i                 = 0;
   pqNode<T,X,Y>*   _checkLeaf   = NULL;
   pqNode<T,X,Y>*   _checkNode   = NULL;
   pqNode<T,X,Y>*   _checkSib    = NULL;
   pqNode<T,X,Y>*   _holdSib     = NULL;
   pqNode<T,X,Y>*   _oldSib      = NULL;
   pqNode<T,X,Y>*   _parent      = NULL;
   pqNode<T,X,Y>*   _lastBlocked = NULL;

   for (i = 0; i <= (number-1); i++)
   {
      _checkLeaf = ArrayOfElements[i]->nodePointer();
      _checkLeaf->mark(QUEUED);
      _processNodes->enQueue(_checkLeaf);
      _pertinentNodes->push(_checkLeaf);
   #ifdef DEBUG
      _checkLeaf->status(FULL);
      _processLeaves->push(_checkLeaf);
   #endif

   }

 


#ifdef DEBUG
   char nb[INFOSIZE];
   sprintf(nb,"%d",redNumber);
   printOutCurrentTree("PQ_B_Print",nb);
   while (!_processLeaves->stackEmpty())
   {
       _checkLeaf = _processLeaves->pop();
       _checkLeaf->status(EMPTY);
   }

#endif

   while ((_processNodes->queueSize() + _blockCount + _offTheTop) > 1)
   {
      if (_processNodes->queueSize() == 0)
         {
	    delete _processNodes;
	    return FALSE;
	 }



      _checkNode = _processNodes->deQueue();
      _blockedNodes->push(_checkNode);
      _checkNode->mark(BLOCKED);
      _blockedSiblings = 0;


      if ((_checkNode->_parentType != P_NODE) && (_checkNode != _root))
                                                // _checkNode is son of a Q_NODE.
                                                // Check if it is blocked.
      {
         if (clientSibLeft(_checkNode) == NULL)
                                                // _checkNode is endmost child of
                                                // a Q_NODE. It has a valid pointer
                                                // to its parent.
         {
            _checkNode->mark(UNBLOCKED);
            if (clientSibRight(_checkNode) && clientSibRight(_checkNode)->mark() == BLOCKED)
               _blockedSiblings++;
         }

         else if (clientSibRight(_checkNode) == NULL)
                                                // _checkNode is endmost child of
                                                // a Q_NODE. It has a valid pointer
                                                // to its parent.
         {
            _checkNode->mark(UNBLOCKED);
            if (clientSibLeft(_checkNode) && clientSibLeft(_checkNode)->mark() == BLOCKED)
               _blockedSiblings++;
         }
             

         else
                                                // _checkNode is not endmost child of
                                                // a Q_NODE. It has not a valid pointer
                                                // to its parent.
         {
            if (clientSibLeft(_checkNode)->mark() == UNBLOCKED)
                                                // _checkNode is adjacent to an
                                                // unblocked node. Take its parent.
            {
               _checkNode->mark(UNBLOCKED);
               _checkNode->_parent = clientSibLeft(_checkNode)->_parent;
            }
            else if (clientSibLeft(_checkNode)->mark() == BLOCKED)
               _blockedSiblings++;

            if (clientSibRight(_checkNode)->mark() == UNBLOCKED)
                                                // _checkNode is adjacent to an 
                                                // unblocked node. Take its parent.
            {
               _checkNode->mark(UNBLOCKED);
               _checkNode->_parent = clientSibRight(_checkNode)->_parent;
            }
            else if (clientSibRight(_checkNode)->mark() == BLOCKED)
                _blockedSiblings++;
         }
      }
         
      else                                      // _checkNode is son of a P_NODE
                                                // and children of P_NODEs
                                                // cannot be blocked.
            _checkNode->mark(UNBLOCKED);



      if (_checkNode->mark() == UNBLOCKED)
      {
         _parent = _checkNode->_parent;
         if (_blockedSiblings > 0)
	 {
	    if (clientSibLeft(_checkNode) != NULL)
	    {
	       _checkSib = clientSibLeft(_checkNode);
	       _oldSib = _checkNode;
	       _holdSib = NULL;
	       while (_checkSib->mark() == BLOCKED)
	       {
	          _checkSib->mark(UNBLOCKED);
	          _checkSib->_parent = _parent;
	          _numBlocked--;
	          _parent->_pertChildCount++;
	          _holdSib = clientNextSib(_checkSib,_oldSib);
	          _oldSib = _checkSib;
	          _checkSib = _holdSib;
	          if (_checkSib == NULL)
	          {
	             cout << "ERROR: PQTree->Bubble: Blocked node as "
	                  << " endmost child of a Q_NODE." << endl;
	             _checkSib = _oldSib;
	          }
	       }
	    }



	    if (clientSibRight(_checkNode) != NULL)
	    {
	       _checkSib = clientSibRight(_checkNode);
	       _oldSib = _checkNode;
	       _holdSib = NULL;
	       while (_checkSib->mark() == BLOCKED)
	       {
	          _checkSib->mark(UNBLOCKED);
	          _checkSib->_parent = _parent;
	          _numBlocked--;
	          _parent->_pertChildCount++;
	          _holdSib = clientNextSib(_checkSib,_oldSib);
	          _oldSib = _checkSib;
	          _checkSib = _holdSib;
	          if (_checkSib == NULL)
	          {
	             cout << "ERROR: PQTree->Bubble: Blocked node as "
	                  << " endmost child of a Q_NODE." << endl;
	             _checkSib = _oldSib;
	          }
	       }
	    }
	    
	 }


         if (_parent == NULL)
	                                           // _checkNode is root of the tree.
	    _offTheTop = 1;

	 else
	                                           // _checkNode is not the root.
	 {
	    _parent->_pertChildCount++;
	    if (_parent->mark() == UNMARKED)
	    {
	       _processNodes->enQueue(_parent);
	       _pertinentNodes->push(_parent);
	       _parent->mark(QUEUED);
	    }
	    
	 }

	 _blockCount = _blockCount - _blockedSiblings;
	 _blockedSiblings = 0;


      }
      else
      {
         _blockCount = _blockCount + 1 - _blockedSiblings;
	 _numBlocked++;
	 _lastBlocked = _checkNode;


      }
   }

   if (_blockCount == 1)
   {
      while (!_blockedNodes->stackEmpty())
      {
         _checkNode = _blockedNodes->pop();
         if (_checkNode->mark() == BLOCKED)
         {
            _checkNode->mark(UNBLOCKED);
            _checkNode->_parent = _pseudoRoot;
            _pseudoRoot->_pertChildCount++;              
            if (_checkNode->endmostChild())
            {
               cout << "ERROR: PQTree->Bubble: Blocked node as "
                    << " endmost child of a Q_NODE." << endl;
               _checkSib = _oldSib;
            }
         }
      }
   }





   delete _processNodes;
   delete _blockedNodes;
   
   return TRUE;
}



/************************************************************************
                          checkIfOnlyChild
************************************************************************/ 

template<class T,class X,class Y>
int PQTree<T,X,Y>::checkIfOnlyChild(pqNode<T,X,Y> *child,
                               pqNode<T,X,Y> *parent)

{
   if ((parent->type() == P_NODE && parent->_childCount == 1)
       || (parent->type() == Q_NODE && parent->_leftEndmost == child
           && parent->_rightEndmost == child))
   {
      removeChildFromSiblings(child);
      child->_parent = parent->_parent;
      if (parent->_parent != NULL)         // parent is not the root.
         exchangeNodes(parent,child);
      else
      {
         exchangeNodes(parent,child);
         _root = child;
      }
      destroyNode(parent);
      return TRUE;
   }
   else
      return FALSE;
}



/************************************************************************
                        clientLeftEndmost
************************************************************************/ 


template<class T,class X,class Y>
pqNode<T,X,Y>* PQTree<T,X,Y>::clientLeftEndmost(pqNode<T,X,Y>* nodePtr) const 

{
  return nodePtr->_leftEndmost;
}



/************************************************************************
                        clientRightEndmost
************************************************************************/ 


template<class T,class X,class Y>
pqNode<T,X,Y>* PQTree<T,X,Y>::clientRightEndmost(pqNode<T,X,Y>* nodePtr) const 

{
  return nodePtr->_rightEndmost;
}




/************************************************************************
                           clientNextSib
************************************************************************/ 


template<class T,class X,class Y>
pqNode<T,X,Y>* PQTree<T,X,Y>::clientNextSib(pqNode<T,X,Y>* nodePtr,
                                       pqNode<T,X,Y>* other) const
{
  return nodePtr->getNextSib(other);
}






/************************************************************************
                        clientPrintNodeCategorie
************************************************************************/ 



template<class T,class X,class Y>
int PQTree<T,X,Y>::clientPrintNodeCategorie(pqNode<T,X,Y>* nodePtr) 

{
  if (nodePtr != NULL)
     return 1;                     // 1 is the standard node categrie
                                   // in the Tree Interface.
  else
     return 0;
}





/************************************************************************
                        clientPrintStatus
************************************************************************/ 


template<class T,class X,class Y>
char*  PQTree<T,X,Y>::clientPrintStatus(pqNode<T,X,Y>* nodePtr) 

{
  if (nodePtr != NULL)
     return (char*)"ERROR";
  else
     return (char*)"ERROR: clientPrintStatus: NO NODE ACCESSED";            
}





/************************************************************************
                        clientPrintType
************************************************************************/ 


template<class T,class X,class Y>
char*  PQTree<T,X,Y>::clientPrintType(pqNode<T,X,Y>* nodePtr) 

{
  if (nodePtr != NULL)
     return (char*)"ERROR";
  else
     return (char*)"ERROR: clientPrintType: NO NODE ACCESSED";            

                 
}



/************************************************************************
                            clientSibLeft
************************************************************************/ 


template<class T,class X,class Y>
pqNode<T,X,Y>* PQTree<T,X,Y>::clientSibLeft(pqNode<T,X,Y>* nodePtr) const 

{
  return nodePtr->_sibLeft;
}




/************************************************************************
                           clientSibRight
************************************************************************/ 


template<class T,class X,class Y>
pqNode<T,X,Y>* PQTree<T,X,Y>::clientSibRight(pqNode<T,X,Y>* nodePtr) const 

{
  return nodePtr->_sibRight;
}



/************************************************************************
                          destroyNode
************************************************************************/ 

template<class T,class X,class Y>
void PQTree<T,X,Y>::destroyNode(pqNode<T,X,Y> *node_ptr)
{

   node_ptr->status(TO_BE_DELETED);
}



/************************************************************************
                     exchangeNodes
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::exchangeNodes(pqNode<T,X,Y> *oldNode,
                              pqNode<T,X,Y> *newNode)
{
   if (oldNode->_referenceParent != NULL)
   {
      oldNode->_referenceParent->_referenceChild = newNode;
      newNode->_referenceParent = oldNode->_referenceParent;
      oldNode->_referenceParent = NULL;


   }

   else if (oldNode->endmostChild())
   {
      if (oldNode->_parent->_leftEndmost == oldNode)
         oldNode->_parent->_leftEndmost = newNode;
      else if (oldNode->_parent->_rightEndmost == oldNode)
         oldNode->_parent->_rightEndmost = newNode;


   }

   if ((oldNode->_sibLeft == oldNode) && (oldNode->_sibRight == oldNode))
   {
      oldNode->_sibLeft = NULL;
      oldNode->_sibRight = NULL;
      if (oldNode->_parent == NULL)
      {
         newNode->_sibLeft = newNode;
         newNode->_sibRight = newNode;
      }
      else
      {
         newNode->_sibLeft = newNode;
         newNode->_sibRight = newNode;
      }


   }
   else if (oldNode->_sibLeft == oldNode)
      cout << "ERROR: PQTree: in exchange Nodes "
           << "sibling pointers of old node are not compatible. "
           << endl;
   else if (oldNode->_sibRight == oldNode)
      cout << "ERROR: PQTree: in exchange Nodes "
           << "sibling pointers of old node are not compatible. "
           << endl;

   if (oldNode->_sibLeft != NULL)
   {
      if (oldNode->_sibLeft->_sibRight == oldNode)
         oldNode->_sibLeft->_sibRight = newNode;
      else if (oldNode->_sibLeft->_sibLeft == oldNode)
         oldNode->_sibLeft->_sibLeft = newNode;
      else
         cout << "ERROR: exchangeNodes: "
              << "Sibling was not connected to child" << endl;
      newNode->_sibLeft = oldNode->_sibLeft;
      oldNode->_sibLeft = NULL;
   }

   if (oldNode->_sibRight != NULL)
   {
      if (oldNode->_sibRight->_sibLeft == oldNode)
         oldNode->_sibRight->_sibLeft = newNode;
      else if (oldNode->_sibRight->_sibRight == oldNode)
         oldNode->_sibRight->_sibRight = newNode;
      else
         cout << "ERROR: removeChildFromSiblings: "
              << "Sibling was not connected to child" << endl;
      newNode->_sibRight = oldNode->_sibRight;
      oldNode->_sibRight = NULL;
   }
      


   newNode->_parentType = oldNode->_parentType;
   newNode->_parent = oldNode->_parent;
}




/************************************************************************
                       fullChildrenStack
************************************************************************/ 


template<class T,class X,class Y>
pqStack<pqNode<T,X,Y>*>*  PQTree<T,X,Y>::fullChildrenStack(pqNode<T,X,Y>* nodePtr)

{
   return nodePtr->fullChildren;
}



/************************************************************************
                       linkChildrenOfQnode
************************************************************************/ 

template<class T,class X,class Y>
void PQTree<T,X,Y>::linkChildrenOfQnode(pqNode<T,X,Y> *installed, 
                                        pqNode<T,X,Y> *newChild)

{
   if ((installed != NULL) && (newChild != NULL))
   {
      if (installed->_sibLeft == NULL)
      {
         installed->_sibLeft = newChild;
         if (newChild->_sibRight == NULL)
            newChild->_sibRight = installed;
         else
            newChild->_sibLeft = installed;
      }
      else if (installed->_sibRight == NULL)
      {
         installed->_sibRight = newChild;
         if (newChild->_sibLeft == NULL)
            newChild->_sibLeft = installed;
         else
            newChild->_sibRight = installed;
      }
      else
         cout << "ERROR: linkChildrenOfQnode: "
              << "endmost child with 2 siblings encountered" << endl;
   }
}



/************************************************************************
                      partialChildrenStack
************************************************************************/ 

template<class T,class X,class Y> 
pqStack<pqNode<T,X,Y>*>*  PQTree<T,X,Y>::partialChildrenStack(pqNode<T,X,Y>* nodePtr)

{
   return nodePtr->partialChildren;
}



/************************************************************************
                            Reduce
************************************************************************/ 

template<class T,class X,class Y>
int PQTree<T,X,Y>::Reduce(int number,
                          leafKey<T,X,Y> **ArrayOfElements,int printYes)
{


   int                   i              = 0;
   pqNode<T,X,Y>*          _checkLeaf     = NULL;
   pqNode<T,X,Y>*          _checkNode     = NULL;
   int                   _pertLeafCount = 0;
   pqQueue<pqNode<T,X,Y>*>*  _processNodes  = new pqQueue<pqNode<T,X,Y>*>;

   // Needed for Printing Serice
   int                   _count = 1;
   char                  _number[8];

   
   for (i = 0; i <= (number-1); i++)
   {
      _checkLeaf = ArrayOfElements[i]->nodePointer();
      _checkLeaf->status(FULL);
      _checkLeaf->_pertLeafCount = 1;
      _processNodes->enQueue(_checkLeaf);
      _pertLeafCount++;
   }




   _checkNode = _processNodes->front();
   while ((_checkNode != NULL) && (_processNodes->queueSize() > 0))
   {
      _checkNode = _processNodes->deQueue();


      if (_checkNode->_pertLeafCount < _pertLeafCount)
      {
         
	 _checkNode->_parent->_pertLeafCount = _checkNode->_parent->_pertLeafCount
	                                     + _checkNode->_pertLeafCount;

	 _checkNode->_parent->_pertChildCount--;
	 if (!_checkNode->_parent->_pertChildCount)
	    _processNodes->enQueue(_checkNode->_parent);
	 if (!template_L1(_checkNode,FALSE))
	 if (!template_P1(_checkNode,FALSE))
	 if (!template_P3(_checkNode))
	 if (!template_P5(_checkNode))
	 if (!template_Q1(_checkNode,FALSE))
	 if (!template_Q2(_checkNode,FALSE))
	    _checkNode= NULL;

	 if (printYes)
	 {
	     sprintf(_number,"%d",_count++);
	     printOutCurrentTree((char*)"RED_",_number);
	 }

      }
      else
      {
         if (!template_L1(_checkNode,TRUE))
	 if (!template_P1(_checkNode,TRUE))
	 if (!template_P2(&_checkNode))
	 if (!template_P4(&_checkNode))
	 if (!template_P6(&_checkNode))
	 if (!template_Q1(_checkNode,TRUE))
	 if (!template_Q2(_checkNode,TRUE))
	 if (!template_Q3(_checkNode))
	 _checkNode = NULL;

	 if (printYes)
	 {
	     sprintf(_number,"%d",_count++);
	     printOutCurrentTree((char*)"RED_",_number);
	 }

      }
   }

   _pertinentRoot = _checkNode;

   delete _processNodes;

   if (_pertinentRoot == NULL)
      return FALSE;
   else
      return TRUE;
}



/************************************************************************
                     removeChildFromSiblings
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::removeChildFromSiblings(pqNode<T,X,Y>* node_ptr)
{
   pqNode<T,X,Y>* sibling = NULL;

   if (node_ptr->_referenceParent != NULL)
   {
      node_ptr->_referenceParent->_referenceChild = node_ptr->_sibRight;
      node_ptr->_sibRight->_referenceParent = node_ptr->_referenceParent;
      if (node_ptr->_referenceParent->_referenceChild == node_ptr)
         node_ptr->_referenceParent->_referenceChild = NULL;
      node_ptr->_referenceParent = NULL;


   }
   else if (node_ptr->endmostChild())
   {
      
      sibling = node_ptr->getNextSib(NULL);

      if (node_ptr->_parent->_leftEndmost == node_ptr)
         node_ptr->_parent->_leftEndmost = sibling;
      else if (node_ptr->_parent->_rightEndmost == node_ptr)
         node_ptr->_parent->_rightEndmost = sibling;
      if (sibling != NULL)
         sibling->_parent = node_ptr->_parent;
         


   }

   if ((node_ptr->_sibRight != NULL) && (node_ptr->_sibRight != node_ptr))
   {
      if (node_ptr->_sibRight->_sibLeft == node_ptr)
         node_ptr->_sibRight->_sibLeft = node_ptr->_sibLeft;
      else if (node_ptr->_sibRight->_sibRight == node_ptr)
         node_ptr->_sibRight->_sibRight = node_ptr->_sibLeft;
      else
         cout << "ERROR: removeChildFromSiblings: "
              << "Sibling was not connected to child" << endl;
   }
   if ((node_ptr->_sibLeft != NULL) && (node_ptr->_sibLeft != node_ptr))
   {
      if (node_ptr->_sibLeft->_sibRight == node_ptr)
         node_ptr->_sibLeft->_sibRight = node_ptr->_sibRight;
      else if (node_ptr->_sibLeft->_sibLeft == node_ptr)
         node_ptr->_sibLeft->_sibLeft = node_ptr->_sibRight;
      else
         cout << "ERROR: removeChildFromSiblings: "
              << "Sibling was not connected to child" << endl;
   }



   node_ptr->_sibRight = NULL;
   node_ptr->_sibLeft = NULL;
}



/************************************************************************
                 removeNodeFromTree
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::removeNodeFromTree(pqNode<T,X,Y>* parent,pqNode<T,X,Y>* child)
{
   if (parent !=NULL)
   {
      removeChildFromSiblings(child);
      parent->_childCount--;
      if ((child->status() == FULL) || (child->status() == PARTIAL))
         parent->_pertChildCount--;
      return parent->_childCount;
   }
   else
   {
      cout << "ERROR: PQTree: removeNodeFromTree: "
           << "parent is invalid NULL-pointer." << endl;
      return -1;
   }
   
}



/************************************************************************
                          template_L1
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_L1(pqNode<T,X,Y> *node_ptr,
                           int isRoot)
{
   if ((node_ptr->type() == LEAF) && (node_ptr->status() == FULL))
   {
      if (!isRoot)
         node_ptr->_parent->fullChildren->push(node_ptr);
      return TRUE;
   }
   else
      return FALSE;
}



/************************************************************************
                          template_P1
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_P1(pqNode<T,X,Y> *node_ptr,
                           int isRoot)
{
   if (node_ptr->type() != P_NODE ||
       node_ptr->fullChildren->count() != node_ptr->_childCount)
      return FALSE;
   else
   {
      node_ptr->status(FULL);
      if (!isRoot)
         node_ptr->_parent->fullChildren->push(node_ptr);

#ifdef DEBUG
      cout << "Template P1 succesfull." << endl; 
#endif      
      return TRUE;
   }
}



/************************************************************************
                          template_P2
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_P2(pqNode<T,X,Y> **node_ptr)
{
   pqNode<T,X,Y> *_newNode = NULL;

   if ((*node_ptr)->type() != P_NODE ||
       (*node_ptr)->partialChildren->count() > 0)
      return FALSE;

   else
   {
      (*node_ptr)->_childCount = (*node_ptr)->_childCount -
                                 (*node_ptr)->fullChildren->count() + 1;
                                       // Gather all full children of node_ptr
                                       // as children of the new P-node.
                                       // Delete them from node_ptr.

      _newNode = createNodeAndCopyFullChildren((*node_ptr)->fullChildren);
                                       // Correct parent-pointer and
                                       // sibling-pointers of the new P-node.
                                       
      _newNode->_parent = (*node_ptr);
      _newNode->_sibRight = (*node_ptr)->_referenceChild->_sibRight;
      _newNode->_sibLeft = _newNode->_sibRight->_sibLeft;
      _newNode->_sibLeft->_sibRight = _newNode;
      _newNode->_sibRight->_sibLeft = _newNode;
      _newNode->_parentType = P_NODE;
                                       // The new P-node now is the root of
                                       // the pertinent subtree.
      (*node_ptr) = _newNode;

#ifdef DEBUG
      cout << "Template P2 succesfull." << endl; 
#endif      


      return TRUE;
   }
   
}




/************************************************************************
                          template_P3
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_P3(pqNode<T,X,Y> *node_ptr)
{

   internalNode<T,X,Y>   *_newNode = NULL;
   pqNode<T,X,Y>     *_newQnode = NULL;
   pqNode<T,X,Y>     *_newPnode = NULL;
   pqNode<T,X,Y>     *_emptyNode = NULL;

   
   if (node_ptr->type() != P_NODE || node_ptr->partialChildren->count() > 0)
      return FALSE;

   else
   {
      _newNode = new internalNode<T,X,Y>(_identificationNumber++,Q_NODE,PARTIAL);
      _newQnode = _newNode;
      _pertinentNodes->push(_newQnode);

      exchangeNodes(node_ptr,_newQnode);
      node_ptr->_parent = _newQnode;
      node_ptr->_parentType = Q_NODE;

      _newQnode->_leftEndmost = (node_ptr);
      _newQnode->_childCount = 1;



      
      if (node_ptr->fullChildren->count() > 0)
      {
         node_ptr->_childCount = node_ptr->_childCount -
                                    node_ptr->fullChildren->count();
                                    
         _newPnode = createNodeAndCopyFullChildren(node_ptr->fullChildren);
         _newPnode->_parentType = Q_NODE;

                                           // Update _newQnode.
         _newQnode->_childCount++;
         _newQnode->fullChildren->push(_newPnode);
                                           // Update sibling pointers.
         node_ptr->_sibRight = _newPnode;
         _newPnode->_sibLeft = node_ptr;
         _newQnode->_rightEndmost = _newPnode;
         _newPnode->_parent = _newQnode;
      }



                                      // Check if node_ptr contains
                                      // only one son. If so, node_ptr
                                      // will be deleted from the tree.
      _emptyNode = node_ptr->_referenceChild;
      checkIfOnlyChild(_emptyNode,node_ptr);
                                      // Update partial_children stack of
                                      // the parent of the new Q-node.
      _newQnode->_parent->partialChildren->push(_newQnode);      

#ifdef DEBUG
      cout << "Template P3 succesfull." << endl; 
#endif      

      return TRUE;
   }
}



/************************************************************************
                          template_P4
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_P4(pqNode<T,X,Y> **node_ptr)
{
   pqNode<T,X,Y>  *_partialChild;           // pointer to the _partialChild

   
   if ((*node_ptr)->type() != P_NODE   ||
       (*node_ptr)->partialChildren->count() != 1)
      return FALSE;
      
   else
   {
      _partialChild = (*node_ptr)->partialChildren->pop();
      copyFullChildrenToPartial(*node_ptr,_partialChild);
                                      // If node_ptr does not have any
                                      // empty children, then it has to
                                      // be deleted and the partial node
                                      // is occupying its place in the tree.
      checkIfOnlyChild(_partialChild,*node_ptr);
                                      // The partial child now is
                                      // root of the pertinent subtree.
      *node_ptr = _partialChild;


#ifdef DEBUG
      cout << "Template P4 succesfull." << endl; 
#endif      

      return TRUE;
   }
}



/************************************************************************
                          template_P5
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_P5(pqNode<T,X,Y> *node_ptr)
{
   pqNode<T,X,Y>  *_partialChild    = NULL;
   pqNode<T,X,Y>  *_checkNode       = NULL;   
   pqNode<T,X,Y>  *_emptyNode       = NULL;
   int           _emptyChildCount = 0;  
   

   if ((node_ptr->type() != P_NODE) ||
       (node_ptr->partialChildren->count() != 1))
      return FALSE;
      
   else   
   {
      _emptyChildCount = node_ptr->_childCount - node_ptr->fullChildren->count() - 1;
      _partialChild = node_ptr->partialChildren->pop();
      node_ptr->_parent->partialChildren->push(_partialChild);

      removeChildFromSiblings(_partialChild);
      exchangeNodes(node_ptr,_partialChild);

      copyFullChildrenToPartial(node_ptr,_partialChild);



      if (_emptyChildCount > 0)
      {
         if (_emptyChildCount == 1)
	 {
	    _emptyNode = node_ptr->_referenceChild;
	    removeChildFromSiblings(_emptyNode);
	 }
	              
	 else
	 {
	    _emptyNode = node_ptr;
	    _emptyNode->_childCount = _emptyChildCount;
	 }


         if (clientLeftEndmost(_partialChild)->status() == EMPTY)
	 {
	    _checkNode = _partialChild->_leftEndmost;
	    _partialChild->_leftEndmost = _emptyNode;
	 }
	 else if (clientRightEndmost(_partialChild)->status() == EMPTY)
	 {
	    _checkNode = _partialChild->_rightEndmost;
	    _partialChild->_rightEndmost = _emptyNode;
	 }
	 else
	    printf("template_P5: ERROR Endmost_child not found\n");

	 linkChildrenOfQnode(_checkNode,_emptyNode);
	 _emptyNode->_parent = _partialChild;
	 _emptyNode->_parentType = Q_NODE;
	 _partialChild->_childCount++;


      }
      
                                       // If node_ptr did not have any empty
                                       // children it has to be deleted.
      if (_emptyChildCount <= 1)
         destroyNode(node_ptr);

#ifdef DEBUG
      cout << "Template P5 succesfull." << endl; 
#endif      

      return TRUE;
   }
}



/************************************************************************
                          template_P6
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_P6(pqNode<T,X,Y> **node_ptr)

{
   pqNode<T,X,Y>  *_partial_1       = NULL;
   pqNode<T,X,Y>  *_partial_2       = NULL;
   pqNode<T,X,Y>  *_fullEnd_1       = NULL;
   pqNode<T,X,Y>  *_fullEnd_2       = NULL;
   pqNode<T,X,Y>  *_emptyEnd_2      = NULL;
   pqNode<T,X,Y>  *_realEmptyEnd_2  = NULL;   

   if ((*node_ptr)->type() != P_NODE   ||
       (*node_ptr)->partialChildren->count() != 2)
      return FALSE;
      
   else
   {
      _partial_1 = (*node_ptr)->partialChildren->pop();
      _partial_2 = (*node_ptr)->partialChildren->pop();

      removeChildFromSiblings(_partial_2);
      (*node_ptr)->_childCount--;

      copyFullChildrenToPartial(*node_ptr,_partial_1);



      if (clientLeftEndmost(_partial_1)->status() == FULL)
         _fullEnd_1 = _partial_1->_leftEndmost;
      else if (clientRightEndmost(_partial_1)->status() == FULL)
         _fullEnd_1 = _partial_1->_rightEndmost;
      else
         cout << "ERROR: template_5: "
              << "partial child with no FULL endmost child detected."
              << endl;

      if (clientLeftEndmost(_partial_2)->status() == FULL)
         _fullEnd_2 = _partial_2->_leftEndmost;
      else if(clientLeftEndmost(_partial_2)->status() == EMPTY)
      {
         _emptyEnd_2 = _partial_2->_leftEndmost;
         _realEmptyEnd_2 = clientLeftEndmost(_partial_2);
      }
      else
         cout << "ERROR: template_5: partial child with "
              << "no FULL or EMPTY endmost child detected" << endl;


      if (clientRightEndmost(_partial_2)->status() == FULL)
         _fullEnd_2 = _partial_2->_rightEndmost;
      else if(clientRightEndmost(_partial_2)->status() == EMPTY)
      {
         _emptyEnd_2 = _partial_2->_rightEndmost;
         _realEmptyEnd_2 = clientRightEndmost(_partial_2);
      }
      else
         cout << "ERROR: template_5: partial child with "
              << "no FULL or EMPTY endmost child detected" << endl;


      if (_fullEnd_2 == _emptyEnd_2)
         cout << "ERROR: template_5: partial child with "
              << "same type of endmost child detected" << endl;



      while (!_partial_2->fullChildren->stackEmpty())
         _partial_1->fullChildren->push(_partial_2->fullChildren->pop());

      linkChildrenOfQnode(_fullEnd_1,_fullEnd_2);
      if (_partial_1->_leftEndmost == _fullEnd_1)
         _partial_1->_leftEndmost = _emptyEnd_2;
      else
         _partial_1->_rightEndmost = _emptyEnd_2;

      _emptyEnd_2->_parent = _partial_1;
      _emptyEnd_2->_parentType = Q_NODE;

      _realEmptyEnd_2->_parent = _partial_1;
      _realEmptyEnd_2->_parentType = Q_NODE;

      _partial_1->_childCount = _partial_1->_childCount +
                                _partial_2->_childCount;
      destroyNode(_partial_2);



                                      // If node_ptr does not have any
                                      // empty children, then it has to
                                      // be deleted and the partial node
                                      // is occupying its place in the tree.
      checkIfOnlyChild(_partial_1,*node_ptr);
                                    // _partial_1 is now root of the
                                    // pertinent subtree.
      *node_ptr = _partial_1;

#ifdef DEBUG
      cout << "Template P6 succesfull." << endl; 
#endif      


      return TRUE;
   }
}



/************************************************************************
                          template_Q1
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_Q1(pqNode<T,X,Y> *node_ptr,
                           int isRoot)
{

   pqNode<T,X,Y>*   _seqStart = NULL;
   pqNode<T,X,Y>*   _seqEnd = NULL;

   if (node_ptr->type() == Q_NODE &&
       node_ptr != _pseudoRoot &&
       clientLeftEndmost(node_ptr)->status() == FULL &&
       clientRightEndmost(node_ptr)->status() == FULL)
   {

#ifdef DEBUG
      cout << "Template Q1 entered." << endl; 
#endif      

      if (checkChain(node_ptr,clientLeftEndmost(node_ptr),&_seqStart,&_seqEnd))
      {
         node_ptr->status(FULL);
         if (!isRoot)
            node_ptr->_parent->fullChildren->push(node_ptr);
         return TRUE;
      }
      else 
         return FALSE;
   }
   else
   {
      return FALSE;
   }
}



/************************************************************************
                          template_Q2
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_Q2(pqNode<T,X,Y> *node_ptr,
                           int isRoot)
{
   pqNode<T,X,Y>   *_fullNode      = NULL;
   pqNode<T,X,Y>   *_sequenceBegin = NULL;
   pqNode<T,X,Y>   *_sequenceEnd   = NULL;
   pqNode<T,X,Y>   *_partialChild  = NULL;
   
   int       _sequenceCons   = FALSE;
   

   if (node_ptr->type() != Q_NODE ||
       node_ptr->partialChildren->count() > 1)
      return FALSE;
      
   else
   {
#ifdef DEBUG
      cout << "Template Q2 entered." << endl; 
#endif      
      if (node_ptr->fullChildren->count() > 0)
      {
         if (node_ptr->_leftEndmost != NULL)
	 {
	    _fullNode = clientLeftEndmost(node_ptr);
	     if (_fullNode->status() != FULL) 
	        _fullNode = NULL;
	 }
	 if (node_ptr->_rightEndmost != NULL  && _fullNode == NULL)
	 {
	    _fullNode = clientRightEndmost(node_ptr);
	    if (_fullNode->status() != FULL)
	       _fullNode = NULL;
	 }


         
         if (_fullNode != NULL)
	    _sequenceCons = checkChain(node_ptr,_fullNode,
	                               &_sequenceBegin,&_sequenceEnd);

	 if (_sequenceCons && (node_ptr->partialChildren->count() == 1))
	 {
	    node_ptr->partialChildren->startAtBottom();
	    _partialChild = node_ptr->partialChildren->readNext();
	    _sequenceCons = FALSE;

	    if (clientSibLeft(_sequenceEnd) == _partialChild ||
	        clientSibRight(_sequenceEnd) == _partialChild)
	       _sequenceCons = TRUE;
	 }


      }

      
      else
      {
         if (!node_ptr->partialChildren->stackEmpty())
         {
            node_ptr->partialChildren->startAtBottom();
	    _partialChild = node_ptr->partialChildren->readNext();
	    if ((clientLeftEndmost(node_ptr) == _partialChild) ||
	        (clientRightEndmost(node_ptr) == _partialChild))
	       _sequenceCons = TRUE; 


         }
      }

      if (_sequenceCons)
         removeBlock(node_ptr,isRoot);
      return _sequenceCons;
   }
}



/************************************************************************
                          template_Q3
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::template_Q3(pqNode<T,X,Y> *node_ptr)                      

{
   pqNode<T,X,Y>   *_fullChild    = NULL;
   pqNode<T,X,Y>   *_fullStart    = NULL;
   pqNode<T,X,Y>   *_fullEnd      = NULL;  
   pqNode<T,X,Y>   *partial_1     = NULL;
   pqNode<T,X,Y>   *partial_2     = NULL;
   int            cons_sequence = FALSE; 
   int            found         = FALSE;
   

   if (node_ptr->type() != Q_NODE || node_ptr->partialChildren->count() >= 3)
      return FALSE;
      
   else
   {
#ifdef DEBUG
      cout << "Template Q3 ." << endl; 
#endif      
      if (!node_ptr->fullChildren->stackEmpty())
      {
         node_ptr->fullChildren->startAtBottom();
         _fullChild = node_ptr->fullChildren->readNext();
         cons_sequence = checkChain(node_ptr,_fullChild,&_fullStart,&_fullEnd);
         if (cons_sequence)
         {
            node_ptr->partialChildren->startAtBottom();
	    while (!node_ptr->partialChildren->readLast())
	    {
	       partial_1 = node_ptr->partialChildren->readNext();
	       found = FALSE;
	       if ( (clientSibLeft(_fullStart) == partial_1)  ||
	            (clientSibRight(_fullStart) == partial_1) ||
	            (clientSibLeft(_fullEnd) == partial_1)    ||
	            (clientSibRight(_fullEnd) == partial_1)     )
	          found = TRUE;
	       if (!found)
	          cons_sequence = found;
	          
	    }


  
         }
      }

      else if (node_ptr->partialChildren->count() == 2)
      {
         node_ptr->partialChildren->startAtBottom();
	 partial_1 = node_ptr->partialChildren->readNext();
	 partial_2 = node_ptr->partialChildren->readNext();
	 if ( (clientSibLeft(partial_1) == partial_2)  ||
	      (clientSibRight(partial_1) == partial_2)   )
	    found = TRUE;
	 cons_sequence = found;

      }

 

      int isRoot = TRUE;
      if (cons_sequence)
         removeBlock(node_ptr,isRoot);
      return cons_sequence;
   }
}





/************************************************************************
                      addLeavesRunThrough
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::addLeavesRunThrough(internalNode<T,X,Y>* parent,
                                   leafKey<T,X,Y>** ArrayOfElements,
                                   int numberOfElements,
                                   leaf<T,X,Y>* _actualSon)
{
   leafKey<T,X,Y>*   _newKey = NULL;
   int       i      = 0;

   for (i = 1; i <= (numberOfElements-1); i++)
   {
      _newKey = ArrayOfElements[i];
      _actualSon = new leaf<T,X,Y>(_identificationNumber++,EMPTY,_newKey);
      _actualSon->_parent = parent;
      _actualSon->_parentType = parent->type();
      parent->_childCount++;
      if (ArrayOfElements[i-1]->nodePointer()->putSibling(_actualSon)
          == LEFT)
         _actualSon->_sibRight = ArrayOfElements[i-1]->nodePointer();
      else
         _actualSon->_sibLeft = ArrayOfElements[i-1]->nodePointer();
   }
}
   


/************************************************************************
                          checkChain
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::checkChain(pqNode<T,X,Y> *node_ptr,pqNode<T,X,Y> *firstFull,
                       pqNode<T,X,Y> **seqStart, pqNode<T,X,Y> **seqEnd)

{
   int           _fullCount   = 0;
   int           _notFull     = FALSE;      
   pqNode<T,X,Y>  *_checkNode;       
   pqNode<T,X,Y>  *_leftNext    = NULL;
   pqNode<T,X,Y>  *_leftOld     = NULL;
   pqNode<T,X,Y>  *_rightNext   = NULL;
   pqNode<T,X,Y>  *_rightOld    = NULL;
   
   _fullCount = node_ptr->fullChildren->count();
   _fullCount--;                    // firstFull does have a FULL label.

   _leftNext = clientSibLeft(firstFull);
   (*seqEnd) = firstFull;
   if (_leftNext != NULL)
   {
      if (_leftNext->status() == FULL)
         _fullCount--;
      else
         _notFull = TRUE;
      if (_notFull)
         (*seqEnd) = firstFull;
      else
      {
         _leftOld = firstFull;
         _checkNode = _leftNext;
    
         while (_fullCount > 0 && !_notFull)
                                       // There are still full children to be
                                       // counted, and no empty child has been
                                       // encountered on this side.
         {
            _leftNext = clientNextSib(_checkNode,_leftOld);
            if (_leftNext != NULL && _leftNext->status() == FULL)
               _fullCount--;
            else
               _notFull = TRUE;
            _leftOld = _checkNode;
            _checkNode = _leftNext;
         }
         if (_checkNode != NULL && _checkNode->status() == FULL)
            (*seqEnd)  = _checkNode;
         else if (_leftOld != NULL && _leftOld->status() == FULL)
            (*seqEnd) = _leftOld;
         else
            cout << "ERROR: checkChain: "
                 << "searching consecutive sequence in Q2 or Q3."
                 << endl; 
      }
   }

    
   _notFull = FALSE;
   _rightNext = clientSibRight(firstFull);
   (*seqStart) = firstFull;
   if(_rightNext != NULL)
   {
      if (_rightNext->status() == FULL)
         _fullCount--;
      else
         _notFull = TRUE;
      if (_notFull)
         (*seqStart) = firstFull;
      else
      {
         _rightOld = firstFull;
         _checkNode = _rightNext;

         while (_fullCount > 0 && !_notFull)
                                       // There are still full children to be
                                       // counted, and no empty child has been
                                       // encountered on this side.
         {
            _rightNext = clientNextSib(_checkNode,_rightOld);
            if (_rightNext != NULL && _rightNext->status() == FULL)
               _fullCount--;
            else
               _notFull = TRUE;
            _rightOld = _checkNode;
            _checkNode = _rightNext;
         }
         if (_checkNode != NULL && _checkNode->status() == FULL)
            (*seqStart) = _checkNode;
         else if (_rightOld != NULL && _rightOld->status() == FULL)
            (*seqStart) = _rightOld;
         else
               cout << "ERROR: checkChain: "
                    << "searching consecutive seqeuence in Q2 or Q3."
                    << endl;
      }
   }

    

   if (firstFull == (*seqEnd))
   {
       _checkNode = (*seqEnd);
       (*seqEnd) = (*seqStart);
       (*seqStart) = _checkNode;
   }
      
   if (_fullCount == 0)
                                    // All full children occupy a consecutive
                                    // sequence.
      return TRUE;

   else
      return FALSE;
}



/************************************************************************
                     copyFullChildrenToPartial
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::copyFullChildrenToPartial(pqNode<T,X,Y> *node_ptr,
                                        pqNode<T,X,Y> *partialChild)
{
   pqNode<T,X,Y>  *_newNode = NULL;
   pqNode<T,X,Y>  *_checkNode = NULL;

   if (node_ptr->fullChildren->count() > 0)
                                        // There are some full children to
                                        // be copied.
   {
      node_ptr->_childCount = node_ptr->_childCount -
                              node_ptr->fullChildren->count();

      _newNode = createNodeAndCopyFullChildren(node_ptr->fullChildren);

                                         // Introduce _newNode as endmost
                                         // child of the partial Q-node.
      partialChild->_childCount++;
      partialChild->fullChildren->push(_newNode);

      if (clientLeftEndmost(partialChild)->status() == FULL)
      {
         _checkNode = partialChild->_leftEndmost;
         partialChild->_leftEndmost = _newNode;
         linkChildrenOfQnode(_checkNode,_newNode);
         
      }
      else if (clientRightEndmost(partialChild)->status() == FULL)
      {
         _checkNode = partialChild->_rightEndmost;
         partialChild->_rightEndmost = _newNode;
         linkChildrenOfQnode(_checkNode,_newNode);
      }
      else
         cout << "ERROR: copyFullChildrenToPartial: "
              << "Endmost_child not found." << endl;

      
      _newNode->_parent = partialChild;
      _newNode->_parentType = Q_NODE;
   }
}



/************************************************************************
                     createNodeAndCopyFullChildren
************************************************************************/ 


template<class T,class X,class Y>
pqNode<T,X,Y>* PQTree<T,X,Y>::createNodeAndCopyFullChildren(pqStack<pqNode<T,X,Y>*> *fullNodes)
{
   pqNode<T,X,Y>  *_newNode    = NULL;
   pqNode<T,X,Y>  *_oldSon     = NULL;
   pqNode<T,X,Y>  *_firstSon   = NULL;
   pqNode<T,X,Y>  *_checkSon   = NULL;

   internalNode<T,X,Y> *_newPQnode = NULL;

   if (fullNodes->count() == 1)
   {
      _newNode = fullNodes->pop();
      removeChildFromSiblings(_newNode);


   }
   
   else
   {
      _newPQnode = new internalNode<T,X,Y>(_identificationNumber++,P_NODE,FULL);
      _newNode = _newPQnode;
      _pertinentNodes->push(_newNode);

      _newNode->_pertChildCount = fullNodes->count();
      _newNode->_childCount = fullNodes->count();

_firstSon = fullNodes->pop();
removeChildFromSiblings(_firstSon);
_newNode->fullChildren->push(_firstSon);
_firstSon->_parent = _newNode;
_firstSon->_parentType = _newNode->type();
_oldSon = _firstSon;



while (!fullNodes->stackEmpty())
{
   _checkSon = fullNodes->pop();
   removeChildFromSiblings(_checkSon);
   _newNode->fullChildren->push(_checkSon);
   _oldSon->_sibRight = _checkSon;
   _checkSon->_sibLeft = _oldSon;
   _checkSon->_parent = _newNode;
   _checkSon->_parentType = _newNode->type();
   _oldSon = _checkSon;
}
_firstSon->_sibLeft = _oldSon;
_oldSon->_sibRight = _firstSon;
_newNode->_referenceChild = _firstSon;
_firstSon->_referenceParent = _newNode;





   }
   
   return _newNode;
}



/************************************************************************
                       inExceptions
************************************************************************/ 


template<class T,class X,class Y>
int PQTree<T,X,Y>::inExceptions(int searchKey,int Exceptions[],int arraySize)

{
   int  low = 0;
   int  high = arraySize - 1;

   int  middle = 0;
   
   while (low <= high)
   {
      middle = (low + high) / 2;

      if (searchKey == Exceptions[middle])
         return TRUE;
      else if (searchKey < Exceptions[middle])
         high = middle - 1;
      else
         low = middle + 1;
   }

   return FALSE;
}



/************************************************************************
                           printNode
************************************************************************/ 

template<class T,class X,class Y> 
void PQTree<T,X,Y>::printNode(char *filename, int number,
                              pqNode<T,X,Y>* parent, pqNode<T,X,Y>* son)
{
   ofstream output(filename,ios::app);


   output << "e " << parent->_debugTreeNumber
          << ' ' <<number << endl;

   output << "n " << number << ' ' << "\\i"
          << (son->type() == Q_NODE ? " Q-NODE" :
             (son->type() == P_NODE ? " P-NODE" : 
             (son->type() == LEAF   ? " LEAF"   :
             (clientPrintType(son) ) )))
          << " IdentificationNumber: " << son->_identificationNumber
          << ", Status: "
          << (son->status() == FULL ? " FULL"           :
             (son->status() == EMPTY ? " EMPTY"         :
             (son->status() == PARTIAL ? " PARTIAL"     :
             (son->status() == PERTINENT ? " PERTINENT" : 
             (clientPrintStatus(son) ) ))))
          << (son->getNodeInfo() != NULL ?
              son->getNodeInfo()->print()   : " "  )
          << (son->getInternal() != NULL ?
              son->getInternal()->print()   : " " )
          << (son->getKey() != NULL ?
              son->getKey()->print()   : " " )
          << " Child Count: " << son->_childCount
          << "\\i" << endl;

   output << "c " << number << ' '
          << (son->type() == Q_NODE  ? Q_NODE_CAT     :
             (son->type() == P_NODE  ? P_NODE_CAT     :
             (son->status() == FULL  ? LEAF_FULL_CAT  : 
             (son->status() == EMPTY ? LEAF_CAT       :
             (clientPrintNodeCategorie(son) ) ))))
          << endl;
  
   son->_debugTreeNumber = number;
}



/************************************************************************
                          removeBlock
************************************************************************/ 


template<class T,class X,class Y>
void PQTree<T,X,Y>::removeBlock(pqNode<T,X,Y> *node_ptr, int isRoot)



{
   pqNode<T,X,Y>  *partial_1        = NULL;
   pqNode<T,X,Y>  *end_full_1       = NULL;
   pqNode<T,X,Y>  *end_empty_1      = NULL;
   pqNode<T,X,Y>  *real_full_1      = NULL;
   pqNode<T,X,Y>  *real_empty_1     = NULL;

   pqNode<T,X,Y>  *partial_2        = NULL; 
   pqNode<T,X,Y>  *end_full_2       = NULL;
   pqNode<T,X,Y>  *end_empty_2      = NULL;
   pqNode<T,X,Y>  *real_full_2      = NULL;
   pqNode<T,X,Y>  *real_empty_2     = NULL;
      
   pqNode<T,X,Y>  *sib_full_1       = NULL; 
   pqNode<T,X,Y>  *sib_partial_1    = NULL;
   pqNode<T,X,Y>  *sib_empty_1      = NULL;  
   pqNode<T,X,Y>  *non_status_sib_1 = NULL;

   pqNode<T,X,Y>  *sib_full_2       = NULL;   
   pqNode<T,X,Y>  *sib_partial_2    = NULL;
   pqNode<T,X,Y>  *sib_empty_2      = NULL;  
   pqNode<T,X,Y>  *non_status_sib_2 = NULL;

   pqNode<T,X,Y>  *helpptr          = NULL;
   pqNode<T,X,Y>  *_checkVarLeft    = NULL; 
   pqNode<T,X,Y>  *_checkVarRight   = NULL;

   int          endmost_check     = FALSE;      
   
   
   node_ptr->status(PARTIAL);
   if (!isRoot)
      node_ptr->_parent->partialChildren->push(node_ptr);
      
   if (!node_ptr->partialChildren->stackEmpty())
                                        // Get a partial child.
   {
      partial_1 = node_ptr->partialChildren->pop();
      _checkVarLeft = clientLeftEndmost(partial_1);
      _checkVarRight = clientRightEndmost(partial_1);
      if (_checkVarLeft->status() == FULL)
      {
         end_full_1 = partial_1->_leftEndmost;
         real_full_1 = _checkVarLeft;
      }
      else if (_checkVarRight->status() == FULL)
      {
         end_full_1 = partial_1->_rightEndmost;
         real_full_1 = _checkVarRight;
      }
      else
         cout << "ERROR: removeBlock: "
              << "partial child with no full endmost child detected."
              << endl;

      if (_checkVarLeft->status() == EMPTY)
      {
         end_empty_1 = partial_1->_leftEndmost;
         real_empty_1 = _checkVarLeft;
      }
      else if (_checkVarRight->status() == EMPTY)
      {
         end_empty_1 = partial_1->_rightEndmost;
         real_empty_1 = _checkVarRight;
      }
      else
         cout << "ERROR: removeBlock: "
              << "partial child with no empty endmost child detected."
              << endl;



      if (clientSibLeft(partial_1) != NULL)
      {
         if (clientSibLeft(partial_1)->status() == FULL)
            sib_full_1 = partial_1->_sibLeft;
         else if (clientSibLeft(partial_1)->status() == EMPTY)
            sib_empty_1 = partial_1->_sibLeft;
         else if (clientSibLeft(partial_1)->status() == PARTIAL)
            sib_partial_1 = partial_1->_sibLeft;
      }
      else 
         non_status_sib_1 = partial_1->_sibLeft;

      if (clientSibRight(partial_1) != NULL)
      {
         if (clientSibRight(partial_1)->status() == FULL)
            sib_full_1 = partial_1->_sibRight;
         else if (clientSibRight(partial_1)->status() == EMPTY)
            sib_empty_1 = partial_1->_sibRight;
         else if (clientSibRight(partial_1)->status() == PARTIAL)
            sib_partial_1 = partial_1->_sibRight;
      }
      else if (non_status_sib_1 == NULL)
         non_status_sib_1 = partial_1->_sibRight;
      else 
        cout << "ERROR 1: PQTree: removeBlock: partial child detected "
             << "with no siblings of valid status."
             << endl;


   }

   if (!node_ptr->partialChildren->stackEmpty())
                                       // There is a second partial child.
   {
      partial_2 = node_ptr->partialChildren->pop();
      _checkVarLeft = clientLeftEndmost(partial_2);
      _checkVarRight = clientRightEndmost(partial_2);
      if (_checkVarLeft->status() == FULL)
      {
         end_full_2 = partial_2->_leftEndmost;
         real_full_2 = _checkVarLeft;
      }
      else if (_checkVarRight->status() == FULL)
      {
         end_full_2 = partial_2->_rightEndmost;
         real_full_2 = _checkVarRight;
      }
      else
         cout << "ERROR: removeBlock: "
              << "partial child with no full endmost child detected."
              << endl;

      if (_checkVarLeft->status() == EMPTY)
      {
         end_empty_2 = partial_2->_leftEndmost;
         real_empty_2 = _checkVarLeft;
      }
      else if (_checkVarRight->status() == EMPTY)
      {
         end_empty_2 = partial_2->_rightEndmost;
         real_empty_2 = _checkVarRight;
      }
      else
         cout << "ERROR: removeBlock: "
              << " partial child with no empty endmost child detected."
              << endl;


      if (clientSibLeft(partial_2) != NULL)
      {
         if (clientSibLeft(partial_2)->status() == FULL)
            sib_full_2 = partial_2->_sibLeft;
         else if (clientSibLeft(partial_2)->status() == EMPTY)
            sib_empty_2 = partial_2->_sibLeft;
         else if (clientSibLeft(partial_2)->status() == PARTIAL)
            sib_partial_2 = partial_2->_sibLeft;
      }
      else 
         non_status_sib_2 = partial_2->_sibLeft;


      if (clientSibRight(partial_2) != NULL)
      {
         if (clientSibRight(partial_2)->status() == FULL)
            sib_full_2 = partial_2->_sibRight;
         else if (clientSibRight(partial_2)->status() == EMPTY)
            sib_empty_2 = partial_2->_sibRight;
         else if (clientSibRight(partial_2)->status() == PARTIAL)
            sib_partial_2 = partial_2->_sibRight;
      }
      else if (non_status_sib_2 == NULL)
         non_status_sib_2 = partial_2->_sibRight;
      else 
        cout << "ERROR 2: PQTree: removeBlock: partial child detected "
             << "with no siblings of valis status." 
             << endl;


   }

   if (partial_1 != NULL && partial_2 != NULL)
   {
      if (sib_full_1 != NULL && sib_full_2 != NULL)
                                             // There are full children  between
                                             // the 2 partial nodes.
                                             // Connect the full children
                                             // between the 2 partial children
                                             // with the full endmost children
                                             // of the 2 partial nodes.
      {
         sib_full_1->changeSiblings(partial_1,end_full_1);
         end_full_1->putSibling(sib_full_1);
         sib_full_2->changeSiblings(partial_2,end_full_2);
         end_full_2->putSibling(sib_full_2);
      }

      else if (sib_partial_1 != NULL && sib_partial_2 != NULL)
                                             // There are no full children between
                                             // the 2 partial nodes. Connect the
                                             // full endmost children of the
                                             // partial nodes as siblings.
      {
         if (partial_1 == sib_partial_2 && partial_2 == sib_partial_1)
                                             // Regular Case.
         {
            end_full_1->putSibling(end_full_2);
            end_full_2->putSibling(end_full_1);
         }
                                            // Only ignored children between 
                                            // partial_1 and partial_2. 
         else
         {
            end_full_1->putSibling(sib_partial_1);
            sib_partial_1->changeSiblings(partial_1,end_full_1);
            end_full_2->putSibling(sib_partial_2);
            sib_partial_2->changeSiblings(partial_2,end_full_2);
         }
         
      }
                                            // Include the children of the 
                                            // partial children with their
                                            // full nodes inbetween into
                                            // the sequence of the children of
                                            // Q-node node_ptr.
      if (sib_empty_1 == NULL)
                                            // partial_1 is endmost child of
                                            // node_ptr. Make the empty endmost
                                            // child of partial_1 be the new endmost
                                            // child of node_ptr.
      {
         if (non_status_sib_1 == NULL)
                                            // Regular case.
         {
            node_ptr->changeEndmost(partial_1,end_empty_1);
         }
         else
                                            // Only ignored children between 
                                            // partial_1 and one end of node_ptr. 
         {
            non_status_sib_1->changeSiblings(partial_1,end_empty_1);
            end_empty_1->putSibling(non_status_sib_1);
         }
         end_empty_1->_parent = node_ptr;
         real_empty_1->_parent = node_ptr;
      }
                                     
      else
                                            // partial_1 is not endmost child.
      {
         sib_empty_1->changeSiblings(partial_1,end_empty_1);
         end_empty_1->putSibling(sib_empty_1);
      }


      if (sib_empty_2 == NULL)
                                            // partial_2 is endmost child of
                                            // node_ptr. Make the empty endmost
                                            // child of partial_2 be the new endmost
                                            // child of node_ptr.
      {
         if (non_status_sib_2 == NULL)
                                            // Regular case.
         {
            node_ptr->changeEndmost(partial_2,end_empty_2);
         }
         else
                                            // Only ignored children between 
                                            // partial_1 and one end of node_ptr. 
         {
            non_status_sib_2->changeSiblings(partial_2,end_empty_2);
            end_empty_2->putSibling(non_status_sib_2);
         }
         end_empty_2->_parent = node_ptr;
         real_empty_2->_parent = node_ptr;
      }
                                     
      else
                                            // partial_2 is not endmost child.
      {
         sib_empty_2->changeSiblings(partial_2,end_empty_2);
         end_empty_2->putSibling(sib_empty_2);
      }


                                            // Copy the full children of partial_1
                                            // and partial_2 to node_ptr.
      while (!partial_2->fullChildren->stackEmpty())
      {
         helpptr = partial_2->fullChildren->pop();
         node_ptr->fullChildren->push(helpptr);
      }
         node_ptr->_childCount = node_ptr->_childCount +partial_2->_childCount - 1;
                                    
      destroyNode(partial_2);

      while (!partial_1->fullChildren->stackEmpty())
      {
         helpptr = partial_1->fullChildren->pop();
         node_ptr->fullChildren->push(helpptr);
      }
      node_ptr->_childCount = node_ptr->_childCount +partial_1->_childCount - 1;

      destroyNode(partial_1);



   }

   else if (partial_1 != NULL)
   {
      if ((clientLeftEndmost(node_ptr) == partial_1) ||
         (clientRightEndmost(node_ptr) == partial_1))
                                            // partial_1 is endmost child.
         endmost_check = TRUE;
               
      if (sib_full_1 != NULL)
                                            // There are full children on one
                                            // side of the partial node.
                                            // Connect the full children with
                                            // the full endmost child of partial_1.
      {
         sib_full_1->changeSiblings(partial_1,end_full_1);
         end_full_1->putSibling(sib_full_1);
      }

      else if (!endmost_check)
                                            // There are not any full children
                                            // and partial_1 is not endmost.
                                            // So get the 2nd empty sibling
                                            // of partial_1 and connect it
                                            // to the full endmost child
                                            // of partial_1.
      {
         if (partial_1->_sibLeft != sib_empty_1)
            sib_empty_2 = partial_1->_sibLeft;
         else
            sib_empty_2 = partial_1->_sibRight;

         sib_empty_2->changeSiblings(partial_1,end_full_1);
         end_full_1->putSibling(sib_empty_2);
      }

      else
                                            // partial_1 is endmost child
                                            // and there are no full children.
                                            // Make the full endmost child of 
                                            // partial_1 be the endmost_child
                                            // of node_ptr.
      {

         if (non_status_sib_1 == NULL)
                                            // Regular case.
         {
            node_ptr->changeEndmost(partial_1,end_full_1);
         }
         else
                                            // Only ignored children between 
                                            // partial_1 and one end of node_ptr. 
         {
            non_status_sib_1->changeSiblings(partial_1,end_full_1);
            end_full_1->putSibling(non_status_sib_1);
         }
         end_full_1->_parent = node_ptr;
         real_full_1->_parent = node_ptr;

      }

      if (sib_empty_1 == NULL)
                                            // There are no empty children.
                                            // partial_1 is endmost child of
                                            // node_ptr. Make the empty endmost
                                            // child of partial_1 be the new endmost
                                            // child of node_ptr.
      {
         if (non_status_sib_1 == NULL)
                                            // Regular case.
         {
            node_ptr->changeEndmost(partial_1,end_empty_1);
         }
         else
                                            // Only ignored children between 
                                            // partial_1 and one end of node_ptr. 
         {
            non_status_sib_1->changeSiblings(partial_1,end_empty_1);
            end_empty_1->putSibling(non_status_sib_1);
         }
         end_empty_1->_parent = node_ptr;
         real_empty_1->_parent = node_ptr;
      }
                                     
      else
                                            // There are empty children. So connect
                                            // the empty endmost child of partial_1
                                            // with sib_empty_1.
      {
         sib_empty_1->changeSiblings(partial_1,end_empty_1);
         end_empty_1->putSibling(sib_empty_1);
      }

      while (!partial_1->fullChildren->stackEmpty())
      {
         helpptr = partial_1->fullChildren->pop();
         node_ptr->fullChildren->push(helpptr);
      }

      node_ptr->_childCount = node_ptr->_childCount + partial_1->_childCount - 1;

      destroyNode(partial_1);


   }

   // else node_ptr does not have partial children. Then nothing is to do.
   
}



/************************************************************************
                       sortExceptions
************************************************************************/ 

template<class T,class X,class Y>
void PQTree<T,X,Y>::sortExceptions(int Exceptions[],int arraySize)

{
   int  i = 0;
   int  _changed = TRUE;
   int  _item = 0;


   while (_changed)
   {
      _changed = FALSE;
      for (i = 0; i < (arraySize-1); i++)
      {
         if (Exceptions[i] > Exceptions[i+1])
         {
            _item = Exceptions[i];
            Exceptions[i] = Exceptions[i+1];
            Exceptions[i+1] = _item;
            _changed = TRUE;
         }

      }
   }
}




#endif


