/************************************************************************

         Filename   :   pqNode.h
         Version    :   1.0  1997
         Author     :   Sebastian Leipert
         Language   :   C++

************************************************************************/
#ifndef PQNODE_H
#define PQNODE_H
#include <LEDA/system/basic.h>
#include "def.inc"
#include "pqNode.inc"
#include "pqStack.h"
#include "nodeRoot.h"


template<class T,class X,class Y> class PQTree;
template<class T,class X,class Y> class leafKey;
template<class T,class X,class Y> class pqNodeKey;
template<class T,class X,class Y> class internalKey;


template<class T,class X,class Y> class pqNode: public nodeRoot {

   friend class PQTree<T,X,Y>;

public:
   
   virtual leafKey<T,X,Y>*      getKey() const = 0;
   virtual int              setKey(leafKey<T,X,Y>* pointerToKey) = 0;

   virtual internalKey<T,X,Y>* getInternal() const = 0;
   virtual int              setInternal(internalKey<T,X,Y>* pointerToInternal) = 0;

   virtual int              mark() const = 0;          
   virtual void             mark(int) = 0;

   virtual int              status() const = 0;        
   virtual void             status(int) = 0;

   virtual int              type() const = 0;          
   virtual void             type(int) = 0;
      


protected:

   
      int           _childCount;
      int           _debugTreeNumber;
      int           _identificationNumber;
      int           _parentType;    
      int           _pertChildCount; 
      int           _pertLeafCount; 
      int           _userDefinedCount;

                                   
   pqNode<T,X,Y>        *_firstFull;
   pqNode<T,X,Y>        *_leftEndmost;     
   pqNode<T,X,Y>        *_parent;         
   pqNode<T,X,Y>        *_referenceChild;
   pqNode<T,X,Y>        *_referenceParent;
   pqNode<T,X,Y>        *_rightEndmost;    
   pqNode<T,X,Y>        *_sibLeft;
   pqNode<T,X,Y>        *_sibRight;
   pqNodeKey<T,X,Y>    *_pointerToInfo;


   pqStack<pqNode<T,X,Y>*>  *fullChildren; 
   pqStack<pqNode<T,X,Y>*>  *partialChildren;




public:



int changeEndmost(pqNode<T,X,Y>* oldEnd, pqNode<T,X,Y>* newEnd)
{
   if (_leftEndmost == oldEnd)
   {
      _leftEndmost = newEnd;
      return TRUE;
   }
   else if (_rightEndmost == oldEnd)
   {
      _rightEndmost = newEnd;
      return TRUE;
   }
   else
   {
      cout << "ERROR: pqNode::changeEndmost: old Endmost not found."
           << endl;
      return FALSE;

   }
}

         
virtual int changeSiblings(pqNode<T,X,Y>* oldSib, pqNode<T,X,Y>* newSib)
{
   if (_sibLeft == oldSib)
   {
      _sibLeft = newSib;
      return TRUE;
   }
   else if (_sibRight == oldSib)
   {
      _sibRight = newSib;
      return TRUE;
   }
   else
   {
      printf("ERROR: pqNode::changeSiblings: old Sibling not found.\n");
      return FALSE;
   }
}
   
   
virtual int childCount() const
{
   return _childCount;
}


virtual int childCount(int count) 
{
   return _childCount = count;
}


pqNode(int count,pqNodeKey<T,X,Y>* infoPtr)
{

   _identificationNumber = count;
   _childCount = 0;
   _pertChildCount = 0;
   _pertLeafCount = 0;
   _debugTreeNumber = 0;
   _userDefinedCount = 0;
   _parentType = 0;

   _parent = NULL;
   _firstFull = NULL;
   _sibLeft = NULL;
   _sibRight = NULL;
   _referenceChild = NULL;
   _referenceParent = NULL;
   _leftEndmost = NULL;
   _rightEndmost = NULL;   

   fullChildren = new pqStack<pqNode<T,X,Y>*>;
   partialChildren = new pqStack<pqNode<T,X,Y>*>;

   _pointerToInfo = infoPtr;
   infoPtr->setNodePointer(this);
   

}


pqNode(int count)
{

   _identificationNumber = count;
   _childCount = 0;
   _pertChildCount = 0;
   _pertLeafCount = 0;
   _debugTreeNumber = 0;
   _userDefinedCount = 0;
   _parentType = 0;

   _parent = NULL;
   _firstFull = NULL;
   _sibLeft = NULL;
   _sibRight = NULL;
   _referenceChild = NULL;
   _referenceParent = NULL;
   _leftEndmost = NULL;
   _rightEndmost = NULL;   

   fullChildren = new pqStack<pqNode<T,X,Y>*>;
   partialChildren = new pqStack<pqNode<T,X,Y>*>;

   _pointerToInfo = NULL;

}


virtual ~pqNode()
{
  delete fullChildren;
  delete partialChildren;
}


virtual int endmostChild()
{
   if (_sibLeft == NULL || _sibRight == NULL)
      return TRUE;
   else
      return FALSE;
}
     

virtual pqNode<T,X,Y>* getEndmost(pqNode<T,X,Y>* other) const
{
   if (_leftEndmost != other)
      return _leftEndmost;
   else if (_rightEndmost != other)
      return _rightEndmost;
   else
      return NULL;
}


virtual pqNode<T,X,Y>* getEndmost(int side) const
{
   if (side == LEFT || side == 0)
      return _leftEndmost;
   else if(side == RIGHT) 
      return _rightEndmost;
   else
      return NULL;
}


virtual pqNode<T,X,Y>* getSib(int side) const
{

   if (side == LEFT)
      return _sibLeft;
   else if (side == RIGHT) 
      return _sibRight;
   else
      return NULL;
}


virtual pqNode<T,X,Y>* getNextSib(pqNode<T,X,Y>* other) const
{

   if (_sibLeft != other)
      return _sibLeft;
   else if (_sibRight != other)
      return _sibRight;
   else
      return NULL;
}


virtual pqNodeKey<T,X,Y>* getNodeInfo() const
{
   return _pointerToInfo;
}


virtual int identificationNumber() const
{
   return _identificationNumber;
}


virtual pqNode<T,X,Y>* parent() const
{
   return _parent;
}


virtual pqNode<T,X,Y>* parent(pqNode<T,X,Y>* newParent)
{
   return _parent = newParent;
}


virtual int  parentType() const
{
   return _parentType;
}


virtual int parentType(int newParentType)
{
   return _parentType = newParentType;
}


virtual int pertChildCount() const
{
   return _pertChildCount;
}


virtual int pertChildCount(int count) 
{
   return _pertChildCount = count;
}


virtual int putSibling(pqNode<T,X,Y>* newSib)
{
   if (_sibLeft == NULL)
   {
      _sibLeft = newSib;
      return LEFT;
   }
   else if (_sibRight == NULL)
   {
      _sibRight = newSib;
      return RIGHT;
   }
   else
   {
      cout << "ERROR: pqNode::putSibling: Two existing siblings detected"
           << endl
           << "IdentificationNumber: pqNode = " << _identificationNumber
           << "_sibLeft = " << _sibLeft->_identificationNumber
           << "_sibRight = " << _sibRight->_identificationNumber
           << endl;
      return FALSE;
   }
}


virtual int putSibling(pqNode<T,X,Y>* newSib,int preference)
{
   if (preference == LEFT)
      return putSibling(newSib);
   else if (preference == RIGHT)
   {
      if (_sibRight == NULL)
      {
         _sibRight = newSib;
         return RIGHT;
      }
      else if (_sibLeft == NULL)
      {
         _sibLeft = newSib;
         return LEFT;
      }
      else
      {
         cout << "ERROR: pqNode::putSibling: Two existing siblings detected"
              << endl
              << "IdentificationNumber: pqNode = " << _identificationNumber
              << "_sibLeft = " << _sibLeft->_identificationNumber
              << "_sibRight = " << _sibRight->_identificationNumber
              << endl;
         return FALSE;
      }
   }
   else
   {
      cout << "ERROR: pqNode::putSibling: Non existing direction specified."
           << endl;
      return FALSE;
   }
}


virtual pqNode<T,X,Y>* referenceChild() const
{
   return _referenceChild;
}


virtual pqNode<T,X,Y>* referenceParent() const
{
   return _referenceParent;
}


virtual int setNodeInfo(pqNodeKey<T,X,Y>* pointerToInfo) 
{
   _pointerToInfo = pointerToInfo;
   if (pointerToInfo != NULL)
   {    
      _pointerToInfo->setNodePointer(this);
      return TRUE;
   }
   else
      return FALSE;
}


virtual int userDefinedCount() const
{
   return _userDefinedCount;
}



virtual int userDefinedCount(int count)
{
   return _userDefinedCount = count;
}


virtual void userDefinedDecr(int decrement)
{
   _userDefinedCount = _userDefinedCount - decrement;
}


virtual void userDefinedIncr(int increment)
{
   _userDefinedCount = _userDefinedCount + increment;
}


};


#endif

