// label.h: interface for the label class.
//
//////////////////////////////////////////////////////////////////////

#ifndef _AGD_IMPL_LABEL_H
#define _AGD_IMPL_LABEL_H

#include <LEDA/core/list.h>
#include "bctree.h"

enum STOP_CAUSE {PLANARITY, C_DEGREE, B_DEGREE, ROOT};

class l_matcher;
class label_struct;
typedef label_struct* label;

class label_struct  {
friend class l_matcher;
public:
	label_struct (cutpoint cp, block p, STOP_CAUSE sc);
	label_struct (block p1, block p2, block b, cutpoint head);
	void stop_cause (STOP_CAUSE sc);
	bool is_blabel ();
	block get_pendant (int number);
	bool is_clabel();
	block first_pendant();
	block last_pendant ();
	int size();
	void remove_pendant (block p);
	STOP_CAUSE stop_cause ();
private:
	void delete_all_pendants ();
	void add_pendant (block b);
	bc_node parent ();
	cutpoint head ();
	void change2clabel (cutpoint new_head);

	block parent_;
	STOP_CAUSE sc_;
	cutpoint head_;
	list<block> pendants_;
};

LEDA_BEGIN_NAMESPACE
inline int compare (label const& l1, label const& l2)  {
	return compare ((long)l1, (long)l2);
}
LEDA_END_NAMESPACE

#endif //
