/*******************************************************************************
+
+  AGD 1.2
+
+
+  make_connected.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



#include "l_matcher.h"
#include <LEDA/core/sortseq.h>

using namespace leda;

inline int block_size_ord (block const& bl )
        { return number_of_nodes (bl); }

//***************************************************************************
// Diese Funktion macht einen evtl. nicht zusammenhaengenden Graphen zusammen-
// haengend. Der Unterschied zum LEDA-Programm besteht darin, dass hier versucht
// wird, Pendants miteinander zu verknuepfen. Insgesamt werden n - 1 Kanten
// addiert (n = #Zusammenhangskomponenten), jede Komponente bekommt hoechstens
// 2 Kanten, jeder Pendant - hoechstens eine. Wenn jede Zusammenhangs-
// komponente des Graphen auch 2-zusammenhaengend ist, dann wird der zusammenhaengender 
// Ausgabegraph auch 2-fach zusammenhaengend.
//***************************************************************************/
int MAKE_CONNECTED (graph& Gr, bctree& BC, list <edge> &added )   {
 if ( BC.number_of_components () == 1 )  return 0;

 blocklist isolated = BC.all_blocks_with_degree (0);
 blocklist pendants = BC.all_blocks_with_degree (1);
 block bl;
 int  nIsolated = isolated.size ();

//Sortiere Bloecke nach der Anzahl der Knoten
 pendants.bucket_sort (1, Gr.number_of_nodes (), &block_size_ord);
 
 node v, n, last_node;

 bool ONLY_ISOLATED = ( pendants.empty ()) ? true : false;
// ONLY_ISOLATED==true -> der Graph kann leicht 2-zusammenhaengend gemacht werden.
// Bilde eine Kette ueber alle isolierten Bloecke des Graphen. 
// Schiliesse die Kette mit einer Kante vom letzten zum ersten Block in 
// der Kette ab. last_node ist ein Knoten im ersten Block der Kette.
// Jede Komponente im Graphen bekommt zwei neue inzidente Kante, die
// Kanten muessen mit verschiedenen Knoten inzident sein (falls die Anzahl
// der Knoten in Komponente >1 ist).
 edge e;


 if ( ONLY_ISOLATED ) last_node = first_node (isolated.pop ());
 else last_node = BC.adj_to_cutpoint_node ( pendants.pop ());

 if ( ONLY_ISOLATED ) { 
   if ( indeg (last_node) + outdeg (last_node) == 0 )//ist isolierter Knoten
     v = last_node;
   else {
     if ( outdeg (last_node))
       e = Gr.first_adj_edge (last_node);
     else
       e = Gr.first_in_edge (last_node); 
     v = opposite (last_node, e);
   }

   n = first_node (isolated.tail ());
   if ( indeg (n) + outdeg (n) > 0 ) {
      e = ( Gr.first_adj_edge (n)) ? Gr.first_adj_edge (n) : Gr.first_in_edge (n);
      node tmp = opposite (n, e);
      n = tmp;
   }
   added.push ( Gr.new_edge (n, v));
 }//Ich habe jetzt letzte, abschliessende Kante hinzugefuegt. 

 list<bc_node> roots = BC.roots ();//Die Wurzel-bc-nodes der ZHK's

 //Verknuepfe die isolierten Bloecke miteinander
 while ( !isolated.empty () ) {
   bl = isolated.pop ();
   roots.del_item (roots.search (bl));

   n =  first_node(bl);

   if ( number_of_nodes (bl) == 1 ) e = nil;
   else e =(Gr.first_adj_edge (n)) ? Gr.first_adj_edge (n) : Gr.first_in_edge (n);

   added.push ( Gr.new_edge ( last_node, n));
   last_node = (e) ? opposite (n, e) : first_node(bl);

 }
 if ( ONLY_ISOLATED ) 
   return added.size ();

 if ( BC.number_of_components () - nIsolated == 1 )
   return added.size ();
   
 sortseq <int, bc_node> roots_nums;
 bc_node bcn;
//Mit Hilfe von roots_nums werde ich Pendants nach den ZHK zu denen sie
//gehoeren sortieren. Jede ZHK darf hoechstens 2 neue Kanten bekommen,
//die mit Pendants inzident sein muessen. 

 forall ( bcn, roots )
   if (bcn_degree (bcn))
     roots_nums.insert (dfs_number (bcn), bcn);

//In dieser Map hat jede Wurzel die Liste der Pendants, die zu ihrer
//ZHK gehoeren. Ein Pendant p gehoert zu der ZHK, derer Wurzel maximale
//DFS-Nummer hat, die kleiner als die Nummer von p ist (locate_pred).

 map <bc_node, blocklist> trees;
 
 forall ( bl, pendants ) {
   bcn = roots_nums.inf ( roots_nums.locate_pred (dfs_number (bl)));
   trees[bcn].push (bl);
 }
 
 block last_pendant = nil;

 while ( !roots.empty () ) {//Verknuepfe Komps
    bcn = roots.pop ();

    if ( trees[bcn].size () == 1 && last_pendant && roots.size () > 0 ) {
         roots.append (bcn);
	 bcn = roots.pop ();
    }
    
    bl = trees[bcn].pop ();
    if ( last_pendant ) 
       added.push ( Gr.new_edge ( BC.adj_to_cutpoint_node (last_pendant),
				  BC.adj_to_cutpoint_node (bl)));
    
    if ( !trees[bcn].empty ())
        last_pendant = trees[bcn].pop ();
    else 
        last_pendant = bl;
 }

 return added.size ();
}

