/*******************************************************************************
+
+  AGD 1.2
+
+
+  ReducedDigraph.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      ReducedDigraph.cpp
//
//    author: Carsten Gutwenger
//    date:   decembre 1998
//-------------------------------------------------------------------


#include <AGD/ReducedDigraph.h>

using namespace leda;


ReducedDigraph::ReducedDigraph()
{
	add_precondition (key::directed);
	add_precondition (key::acyclic);
	add_post_rule (key::reduced_acyclic);
}


AGDModule *ReducedDigraph::clone() const
{
	return new ReducedDigraph;
}

/*
node array mark[w]:
  bit 1: w in visited
  bit 2: w is neighbour of v 
  bit 3: (v,w) is transitive
*/

void ReducedDigraph::dfs(node v)
{
	visited->push(v);
	mark[v] |= 1;

	node w;
	forall_adj_nodes(w,v) {
		if (mark[w] & 2) mark[w] |= 4;

		if ((mark[w] & 1) == 0)
			dfs(w);
	}
}

void ReducedDigraph::do_call (const graph& G, list<edge>& transitives)
{
	node v, w;
	edge e;

	mark.init(G,0);
	visited = new b_stack<node>(G.number_of_nodes());

	forall_nodes(v,G) {
		forall_adj_nodes(w,v)
			mark[w] = 2;
		forall_adj_nodes(w,v)
			if ((mark[w] & 1) == 0) dfs(w);

		forall_out_edges(e,v)
			if (mark[target(e)] & 4)
				transitives.append(e);

		while (!visited->empty())
			mark[visited->pop()] = 0;
	}

	mark.init();
	delete visited;
}
