/*******************************************************************************
+
+  AGD 1.2
+
+
+  COrthoLayout.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <AGD/ClusterOrthogonalLayout.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/QuasiOrthogonalLayout.h>

using namespace leda;


ClusterOrthogonalLayout::ClusterOrthogonalLayout()
{
  add_precondition (key::c_planar);
  add_precondition (key::simple);
  add_precondition (key::no_self_loops);

  add_post_rule (key::no_crossings);

  PreCond pre;
  PostCond post;

  pre [1] << key::planar << key::simple << key::no_self_loops;
  post[1] << key::no_crossings;
  QuasiOrthogonalLayout L;
  L.traditional(true);
  base_algorithm.init(pre,post,L);
}
  
bool ClusterOrthogonalLayout::set_base_algorithm (const Ovid& M)
{
	if (base_algorithm.set(M)) {
		base_algorithm.get().traditional(true);
		return true;
	} else return false;
}

AGDModule* ClusterOrthogonalLayout::clone() const
{
  ClusterOrthogonalLayout *M = new ClusterOrthogonalLayout;
        
  M->set_base_algorithm(base_algorithm.get());

  return M;
}

  
void ClusterOrthogonalLayout::
     do_call(CGPlaneGraphCopyGridConcept& C) {

  
  PlaneClusterGraphCopy  &h    = C.get_clustergraph();
  CGGridLayout           &cggl = C.get_layout(); 

  list<Cluster*> lsubcl;

  Cluster* cl;
  const list<Cluster*> all_clusters = h.all_clusters();

  edge incident_edge=0; 
  edge e=0;

  list<node> cl_boundary, rel_nodes;
  Cluster* subcl;
  bool member=false;
  node v, target;

  node c_e, v_start;

  edge b_e, act_e;
  edge fe_start=0, fe_act=0, fe_last=0;

  list<edge> boundary_edges;

  list_item it = nil;
  int min_x, min_y , max_x, max_y, x, y;

  IPolyline bends;
  IPoint bend_pos;

  if((outer_face = C.get_exterior_face()) == nil) {
    lsubcl = h.get_rootcluster()->get_subclusters();
    if(lsubcl.size() != 0) {
      int sz_outer_face = 0;
      int sz_f_0;
      face f_0;
      forall(cl, lsubcl) {
        list<edge> cl_incident_edges = cl->get_edge_order();
        forall(e, cl_incident_edges) {
          f_0 = h.face_of(e);
          if((sz_f_0 = h.size(f_0)) > sz_outer_face) { 
            outer_face = f_0;
            sz_outer_face = sz_f_0;
          }
        }
      }
    }
  }

  forall(cl, all_clusters) {
    
  rel_nodes.clear();

  forall(v, cl->get_nodes()) rel_nodes.append(v);
 
  forall(subcl,  cl->get_subclusters()) {
    cl_boundary = cl_boundary_nodes[subcl];
    rel_nodes.conc(cl_boundary);
  }

  forall(v, rel_nodes) {
    forall_out_edges(e, v) {
      target = h.target(e);
      if(h.cluster_of(target) == cl) member = true;
      else {
        member  = is_boundary_member(cl, target);
        member |= h.is_member(cl, target);
      }
      if(!member) {
        incident_edge = e;
        break;
      }
    }
    if(!member) break;
  }

                                     
  act_e = incident_edge;
  v_start = nil;

  do {
    
    act_e = h.split(h.reversal(act_e));
    c_e = h.source(act_e);
    if(v_start == nil) {
      v_start = c_e;
      fe_start =  h.face_cycle_succ(h.reversal(act_e));
    }
    cl_boundary_nodes[cl].append(c_e);
//#line 322 "COrthoLayout.nw"
    
//#line 363 "COrthoLayout.nw"
    if(cl_boundary_nodes[cl].size()>1) {    

      fe_act = h.reversal(act_e);
      fe_act = h.face_cycle_succ(fe_act);
      
      if(h.face_of(fe_last) == outer_face) {
        b_e = h.split_face(fe_last, fe_act);
        outer_face = h.face_of(b_e);
      } else b_e = h.split_face(fe_last, fe_act);
      
      cl_boundary_edges[cl].append(b_e);
    }
    
//#line 323 "COrthoLayout.nw"
    
//#line 397 "COrthoLayout.nw"
    fe_last = act_e;
    cl_boundary_nodes[cl].append(c_e);
    
    do {
      // running around 'cluster-outerface'
      act_e = h.face_cycle_succ(act_e);
      target = h.target(act_e);
      
      if(h.cluster_of(target) == cl) member = true;
      else member = is_boundary_member(cl,target);

    } while (member);    

//#line 324 "COrthoLayout.nw"
  } while(target != v_start);
  
//#line 417 "COrthoLayout.nw"
  if(cl_boundary_nodes[cl].size()>1) {
    if(h.face_of(h.reversal(fe_last)) == outer_face) {
      b_e = h.split_face(fe_last, fe_start);
      outer_face = h.face_of(b_e);
    } else b_e = h.split_face(fe_last, fe_start);
    
    cl_boundary_edges[cl].append(b_e);
  } 



//#line 209 "COrthoLayout.nw"
  }

//#line 111 "COrthoLayout.nw"
  
//#line 447 "COrthoLayout.nw"
  forall(cl, all_clusters) {
    forall(e, cl_boundary_edges[cl]) 
      boundary_edges.append(e);
  }


//#line 463 "COrthoLayout.nw"
  Ovid &algo = base_algorithm.get();

  algo.forbid_concave_bends(boundary_edges);
  algo.call_grid((const graph&) h, outer_face, cggl);

/*
  if(base_algo == quasi_orthogonal) 
    quod.forbid_concave_bends(boundary_edges);
  else if(base_algo == orthogonal)
    pure_ortho.forbid_concave_bends(boundary_edges);


  AgdKey p;
  switch(base_algo) {
  case quasi_orthogonal:
    if (!quod.check(h, p)) cerr << p.info() << endl;
    quod.call_grid((const graph&) h, outer_face, cggl);
    break;
  case orthogonal:
  default:
    if (!pure_ortho.check(h, p)) cerr << p.info() << endl;
    pure_ortho.call_grid((const graph&) h, outer_face, cggl);
    break;
  }
*/
//#line 113 "COrthoLayout.nw"
  
//#line 547 "COrthoLayout.nw"
  forall(cl, all_clusters) {
	  bool first = true;
    min_x = max_x = 0;
    min_y = max_y = 0;
    forall(e, cl_boundary_edges[cl]) {
      
//#line 568 "COrthoLayout.nw"
    bends = cggl.bends(e);
    it = bends.first();

//#line 552 "COrthoLayout.nw"
      
//#line 580 "COrthoLayout.nw"
  //if(it != nil && min_x == 0 && max_x == 0) {
	if (it && first) {
		first = false;
    bend_pos = bends.inf(it);
    x = bend_pos.xcoord();
    y = bend_pos.ycoord();
    min_x = max_x = x;
    min_y = max_y = y;
    it = bends.succ(it);
  }

//#line 553 "COrthoLayout.nw"
      while(it != nil) {
        
//#line 599 "COrthoLayout.nw"
  bend_pos = bends.inf(it);
  x = bend_pos.xcoord();
  y = bend_pos.ycoord();
  if(max_x < x) max_x = x;
  if(min_x > x) min_x = x;
  if(max_y < y) max_y = y;
  if(min_y > y) min_y = y;
  it = bends.succ(it);

//#line 555 "COrthoLayout.nw"
      }
    }
    
//#line 617 "COrthoLayout.nw"
  cggl.x(cl, min_x); 
  cggl.y(cl, min_y);
  cggl.width(cl,(max_x - min_x));
  cggl.height(cl,(max_y - min_y));

//#line 558 "COrthoLayout.nw"
  }


//#line 115 "COrthoLayout.nw"
  cl_boundary_nodes.clear();
  cl_boundary_edges.clear();
}

//#line 19 "COrthoLayout.nw"
  
//#line 646 "COrthoLayout.nw"
  
//#line 655 "COrthoLayout.nw"
bool ClusterOrthogonalLayout::
     is_boundary_member(Cluster* cl, node v) {

  Cluster* subcl, *subcl2;

forall(subcl, cl->get_subclusters()) {
  node w = 0;
  forall(w, cl_boundary_nodes[subcl])  {
    if(v == w) return true;
    forall(subcl2, subcl->get_subclusters()) {
      node u = 0;
      forall(u, cl_boundary_nodes[subcl2])
        if(v == u) return true;
    }
  }
}
return false;

}



