/*******************************************************************************
+
+  AGD 1.2
+
+
+  C_planarConnectedC_Graph.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



/**********************************************************************

  FILENAME  : C_planarConnectedC_Graph.cc
  VERSION   : 1.0.0
  AUTHOR    : Christoph Liebel
  LANGUAGE  : C++

**********************************************************************/


#include <LEDA/graph/graph.h>

#include <AGD/impl/cplanarity/C_planarConnectedC_Graph.h>
#include <AGD/impl/cplanarity/extendPQdirected.h>

using namespace leda;

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** clusterData ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

clusterData::clusterData()
{
    number = nodes = edges = 0;
    oneVirtEdge = 0;
    originalToActual = actualToOriginal = NULL;
    nodeNumbers = edgeNumbers = storeNumbers = storeEdges = clusterEnd = NULL;
    saveNodeNumbers = saveEdgeNumbers = oldEdge = oldNode = graphEnd = NULL;
    cluster_graph = NULL;
    leafBKeys = NULL;
}

/*****************************************************************************/
/*****************************              **********************************/
/***************************** ~clusterData **********************************/
/*****************************              **********************************/
/*****************************************************************************/

clusterData::~clusterData()
{
    if (cluster_graph)    delete    cluster_graph;
    cluster_graph = 0;

    if (originalToActual) delete [] originalToActual;
    if (actualToOriginal) delete [] actualToOriginal;
    if (storeNumbers)     delete [] storeNumbers;
    if (storeEdges)       delete [] storeEdges;
    if (edgeNumbers)      delete [] edgeNumbers;
    if (nodeNumbers)      delete [] nodeNumbers;
    if (oldNode)          delete [] oldNode;
    if (oldEdge)          delete [] oldEdge;
    if (saveNodeNumbers)  delete [] saveNodeNumbers;
    if (saveEdgeNumbers)  delete [] saveEdgeNumbers;
    if (graphEnd)         delete [] graphEnd;
    if (clusterEnd)       delete [] clusterEnd;

    originalToActual = actualToOriginal = NULL;
    nodeNumbers = edgeNumbers = storeNumbers = storeEdges = clusterEnd = NULL;
    saveNodeNumbers = saveEdgeNumbers = oldEdge = oldNode = graphEnd = NULL;

    if (leafBKeys)
        {
            for (int j = 0; j < edges; j++)
                {    
                    delete leafBKeys[j];
                    leafBKeys[j] = NULL;
                }

            delete [] leafBKeys;
            leafBKeys = 0;
        }
}


/*****************************************************************************/
/*****************************                          **********************/
/***************************** C_planarConnectedC_Graph **********************/
/*****************************                          **********************/
/*****************************************************************************/

C_planarConnectedC_Graph::C_planarConnectedC_Graph()
    : clusterGraph()
{
    C_planar = 1;
    C_embedded = 1;
    actual_N = actual_M = 0;
    clusterNodes = NULL;
    updateAdjList = saveLoops = NULL;
    generatedHubs = hubs = NULL;
    embedding = NULL;
    S = NULL;
    rememberEdgeNumbers = NULL;

}


/*****************************************************************************/
/******************                                         ******************/
/****************** C_planarConnectedC_Graph(clusterGraph&) ******************/
/******************                                         ******************/
/*****************************************************************************/

C_planarConnectedC_Graph::C_planarConnectedC_Graph(const clusterGraph& cGraph)
    :clusterGraph(cGraph)
{
    C_planar = -1;
    C_embedded = -1;
    actual_N = actual_M = 0;
    clusterNodes = NULL;
    updateAdjList = saveLoops = NULL;
    generatedHubs = hubs = NULL;
    embedding = NULL;
    S = NULL;
    rememberEdgeNumbers = NULL;
}


/*****************************************************************************/
/************                                                     ************/
/************ C_planarConnectedC_Graph(C_planarConnectedC_Graph&) ************/
/************                                                     ************/
/*****************************************************************************/

C_planarConnectedC_Graph::C_planarConnectedC_Graph(
                               const C_planarConnectedC_Graph& cGraph)
: clusterGraph(cGraph),C_planar(cGraph.C_planar),C_embedded(cGraph.C_embedded)
{
    // Initialising the data that belongs only to the class
    // 'C_planarConnectedC_Graph' and not to the base-classes
    // 'basicGraph' and 'clusterGraph'

    // protected data
    embedding = NULL;

    if (cGraph.embedding)       // copy array 'embedding'
    {
        embedding = new pqQueue<int>*[clusterNumber];

		int i;
        for (i = 0; i < clusterNumber; i++)
            embedding[i] = new pqQueue<int>;

        for (i = 0; i < clusterNumber; i++)
        {
            if (!cGraph.embedding[i]->queueEmpty())
                {
                    int save_front = cGraph.embedding[i]->deQueue();

                    cGraph.embedding[i]->enQueue(save_front);
                    embedding[i]->enQueue(save_front);

                    while (cGraph.embedding[i]->front() != save_front)
                        {
                            int store_int = cGraph.embedding[i]->deQueue();

                            cGraph.embedding[i]->enQueue(store_int);
                            embedding[i]->enQueue(store_int);
                        }
                }
        }
    }        
    
    // private data
    actual_N = actual_M = 0;
    clusterNodes = NULL;
    updateAdjList = saveLoops = NULL;
    generatedHubs = hubs = NULL;
    S = NULL;
    rememberEdgeNumbers = NULL;
}

/*****************************************************************************/
/*****************************                           *********************/
/***************************** ~C_planarConnectedC_Graph *********************/
/*****************************                           *********************/
/*****************************************************************************/

C_planarConnectedC_Graph::~C_planarConnectedC_Graph()
{
    deleteGraph();
}

/*****************************************************************************/
/*****************************                ********************************/
/***************************** writeEmbedding ********************************/
/*****************************                ********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::writeEmbedding(ClusterGraph& C,
                                              list<edge>& newEdges)
{
/*    newEdges.clear();
    node *eS=new node[m];
    node *eT=new node[m];
    edge *eR=new edge[m];
    node *nR=new node[n];
    int no;
    edge e;
    node f;
    no=0;
    forall_nodes(f,C) nR[no++]=f;
    no=0; 
    forall_edges(e,C) {
        eR[no]=e; 
        eS[no]=C.source(e);
        eT[no]=C.target(e);
        no++;
    }
    outEdge oe;
    list<node> dummies;
    forall_edges(e,C) {
        f=C.new_node(C.get_rootcluster()); 
        C.move_edge(e,f,f);
        dummies.push(f);
    }
    C.make_map(newEdges);
    for(int i=0;i<n;i++) {
        adjacencyList[i].reset();
        while(adjacencyList[i].out(oe)) {
            if(eS[oe.getNumber()]==nR[i])
                C.move_edge(eR[oe.getNumber()],nR[i],nR[oe.getNode()]);
            if(eT[oe.getNumber()]==nR[i])
                C.move_edge(C.reversal(eR[oe.getNumber()]),nR[i],nR[oe.getNode()]);
        }
    }
    forall(f,dummies) C.del_node(f);
*/
//
    newEdges.clear();

    edge *eR=new edge[m];
    node *nR=new node[n];

    int num = 0;
	node v;
    forall_nodes(v,C)
		nR[num++]=v;

    num=0; 
	edge e;
    forall_edges(e,C) {
        eR[num]=e; 
        num++;
    }

    C.make_map(newEdges);

	ofstream os("output.txt");

	list<edge> allEdges;
    for(int i=0;i<n;i++) {
		os << index(nR[i]) << ": ";
        adjacencyList[i].reset();
	    outEdge oe;
        while(adjacencyList[i].out(oe)) {
			edge e = eR[oe.getNumber()];
			allEdges.append((source(e) == nR[i]) ? e : C.reversal(e));
			os << " " << index(C.opposite(nR[i],e));
		}
		os << endl;
	}

	C.sort_edges(allEdges);
//

    pqStack<Cluster*> clusterStack;
    pqStack<int> clusterN;
    Cluster *cl1,*cl2;
    int count=n,count1;
    clusterStack.push(C.get_rootcluster());
    clusterN.push(count++);
    list<edge> circ_edges;
    while(!clusterStack.stackEmpty()&&!clusterN.stackEmpty()) {
        cl1=clusterStack.pop();
        count1=clusterN.pop();
        circ_edges.clear();
	if (n != 0)
	{
	  while(!embedding[count1-n]->queueEmpty())
            circ_edges.append(eR[embedding[count1-n]->deQueue()]);
	}	
	cl1->set_edge_order(circ_edges);
	forall(cl2,cl1->get_subclusters()) {
	  clusterStack.push(cl2);
	  clusterN.push(count++);
	}
	
    }

	C.compute_faces();

	if (C.is_map() == false)
		cerr << "keine map!" << endl;
	if (Is_Planar_Map(C) == false)
		cerr << "keine planar map!" << endl;

    //delete[] eS; //
    //delete[] eT; //
    delete[] eR;
    delete[] nR;
}

/*****************************************************************************/
/*****************************                ********************************/
/***************************** addVirtualEdge ********************************/
/*****************************                ********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::addVirtualEdge(int storeVirtNode, int storeNode,
                                              int cluster, int* inCluster)
{
    int foundEdge = 0;
    outEdge element;
    int* markHubs = new int[actual_N];

    for (int i = 0; i < actual_N; i++)
        markHubs[i] = 0;

    hubs[cluster]->startAtBottom();

    while (!hubs[cluster]->readLast())
        markHubs[hubs[cluster]->readNext()] = 1;

    updateAdjList[storeNode].reset();

    while (updateAdjList[storeNode].out(element))
        if (!markHubs[element.getNode()] && inCluster[element.getNode()] &&
            !foundEdge)
            {
                foundEdge = 1;
                int otherVirtNode = element.getNode();

                outEdge element1(storeVirtNode, actual_M);
                outEdge element2(otherVirtNode, actual_M);

                updateAdjList[otherVirtNode].in(element1);
                updateAdjList[storeVirtNode].in(element2);
            }

    delete [] markHubs;
}

/*****************************************************************************/
/*****************************                           *********************/
/***************************** change_and_correctNumbers *********************/
/*****************************                           *********************/
/*****************************************************************************/

void C_planarConnectedC_Graph::change_and_correctNumbers(clusterData* info)
{

    linkedList<outEdge>* saveAdjList = new linkedList<outEdge>[info->nodes];

    // ***********************************************************************
    // *** Change the consecutive numbering of the planar embedded cluster ***
    // *** and store the new numbering in the array 'saveAdjList'.         ***
    // ***********************************************************************

    for (int k = 1; k < info->cluster_graph->getN(); k++)
        // Start with k = 1, because node no. 0 is the virtual vertex and
        // doesn't belong to the cluster.
        {
            int number = info->saveNodeNumbers[k];
            linkedList<outEdge>* copyAdjList = new linkedList<outEdge>;

            copyAdjList->append(info->cluster_graph->getAdjList(k));
            copyAdjList->reverse();
            copyAdjList->reset();

            outEdge storeElement;

            while (copyAdjList->out(storeElement))
                if (storeElement.getNode() == 0)
                    {
                        if ((info->oneVirtEdge &&
                             info->saveEdgeNumbers[storeElement.getNumber()] != 
                             info->edges - 1) || !info->oneVirtEdge)
                            {
                                outEdge inElement(info->graphEnd[storeElement.
                                                                getNumber()],
                                                  info->saveEdgeNumbers
                                                  [storeElement.getNumber()]);

                                saveAdjList[number].in(inElement);
                            }
                    }
                else
                    {
                        outEdge inElement(info->saveNodeNumbers[storeElement.
                                                               getNode()],
                                          info->saveEdgeNumbers[storeElement.
                                                               getNumber()]);
                        
                        saveAdjList[number].in(inElement);
                    }
            delete copyAdjList;
        }
    // 'saveAdjList' now contains the planar embedded cluster in the changed
    // numbering.

    // ************************************************************************
    // *** Change the numbering of the remaining graph (without the         ***
    // *** cluster) and store the new numbering in the array 'saveAdjList'. ***
    // ************************************************************************

    // Notice: No non empty list of 'saveAdjList' gets another entry.

    // Determine the number of vertices contained in the representative graph
    // of the actual cluster.

    int nodeCounter = actual_N - info->nodes + info->cluster_graph->getN() - 1;
        // actual_N = number of vertices in the actual graph G (before
        //            the update)
        // info->nodes = number of vertices in the graph G after the update
        // info->cluster_graph->getN() - 1 = number of vertices in the extended
        //                                   subgraph induced by the cluster
        //                                   without the virtual vertex
        // nodeCounter = number of vertices in the representative graph

	int j;
    for (j = nodeCounter; j < actual_N; j++)
        {
            outEdge storeElement;

            updateAdjList[j].reverse();
            updateAdjList[j].reset();

            while (updateAdjList[j].out(storeElement))
                {
                    if (0 <= storeElement.getNode() &&
                        storeElement.getNode() < nodeCounter)
                        // The corresponding node belongs to the cluster.
                        {
                        if ((info->oneVirtEdge &&
                             info->edgeNumbers[storeElement.getNumber()] != 
                             info->edges - 1) || !info->oneVirtEdge)
                            {
                                outEdge inElement(info->clusterEnd
                                                  [storeElement.getNumber()],
                                                  info->edgeNumbers
                                                  [storeElement.getNumber()]);

                                saveAdjList[info->nodeNumbers[j]].
                                    in(inElement);
                            }
                        }

                    if (nodeCounter <= storeElement.getNode() &&
                        storeElement.getNode() < actual_N)
                        // The corresponding node belongs to the remaining
                        // graph.
                        {
                            outEdge inElement(info->nodeNumbers
                                              [storeElement.getNode()],
                                              info->edgeNumbers
                                              [storeElement.getNumber()]);

                            saveAdjList[info->nodeNumbers[j]].in(inElement);
                        }
                }
        }

    // Change the old embedding to the new computed embedding.

    delete [] updateAdjList;

    updateAdjList = new linkedList<outEdge>[info->nodes];
    for (j = 0; j < info->nodes; j++)
        updateAdjList[j].append(&saveAdjList[j]);

    delete [] saveAdjList;
    saveAdjList = NULL;
}

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** changeWheel ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::changeWheel(int hubNumber, int* &rim)
{
    // (1) Determine the nodes on the rim; mark these nodes in the array 'rim'.
    // (2) Perform the following steps for each rim-node:
    //     (a) Reduce 'updateAdjList[rimBNode]' to the needed parts.
    //     (b) Put these parts together in the correct order.
    // (3) Delete the marks of the rim-nodes in the array 'rim'.

    // *********************** part (1) **************************************

    outEdge getElement;

    updateAdjList[hubNumber].reset();

    while (updateAdjList[hubNumber].out(getElement))
        rim[getElement.getNode()] = 1;

    // *********************** part (2) **************************************

    updateAdjList[hubNumber].reset();

    while (updateAdjList[hubNumber].out(getElement))
        {
            int rimBNode = getElement.getNode(); 
            outEdge outElement;
            linkedList<outEdge>* changeAdjList = new linkedList<outEdge>[3];

            updateAdjList[rimBNode].reset();

            // *************** step (a) **************************************

            // Go to the position of the element in the array
            // 'updateAdjList[rimBNode]' that corresponds to HUB.

            updateAdjList[rimBNode].out(outElement);

            while (outElement.getNode() != hubNumber)
                updateAdjList[rimBNode].out(outElement);

            outEdge hubElement(outElement.getNode(), outElement.getNumber());

            // Collect the elements between the position of HUB and the
            // position of R1 in the array 'updateAdjList[rimBNode]'.

            if (!(updateAdjList[rimBNode].out(outElement)))
                {
                    updateAdjList[rimBNode].reset();
                    updateAdjList[rimBNode].out(outElement);
                }

            while (rim[outElement.getNode()] == 0)
                {
                    changeAdjList[0].in(outElement);

                    if (!(updateAdjList[rimBNode].out(outElement)))
                        {
                            updateAdjList[rimBNode].reset();
                            updateAdjList[rimBNode].out(outElement);
                        }
                }

            outEdge neighbour1(outElement.getNode(), outElement.getNumber());

            // Collect the elements between the position of R1 and the
            // position of R2 in the array 'updateAdjList[rimBNode]'.

            if (!(updateAdjList[rimBNode].out(outElement)))
                {
                    updateAdjList[rimBNode].reset();
                    updateAdjList[rimBNode].out(outElement);
                }

            while (rim[outElement.getNode()] == 0)
                {
                    changeAdjList[1].in(outElement);

                    if (!(updateAdjList[rimBNode].out(outElement)))
                        {
                            updateAdjList[rimBNode].reset();
                            updateAdjList[rimBNode].out(outElement);
                        }
                }

            outEdge neighbour2(outElement.getNode(), outElement.getNumber());

            // Collect the elements between the position of R2 and the
            // position of HUB in the array 'updateAdjList[rimBNode]'.

            if (!(updateAdjList[rimBNode].out(outElement)))
                {
                    updateAdjList[rimBNode].reset();
                    updateAdjList[rimBNode].out(outElement);
                }

            while (outElement.getNode() != hubNumber)
                {
                    changeAdjList[2].in(outElement);

                    if (!(updateAdjList[rimBNode].out(outElement)))
                        {
                            updateAdjList[rimBNode].reset();
                            updateAdjList[rimBNode].out(outElement);
                        }
                }

            // *************** step (b) **************************************

            // Change the order in 'updateAdjList[rimBNode]' to the following
            // order:
            //         R2--HUB--R1--changeAdjList[0]--changeAdjList[1]-
            //         -changeAdjList[2]

            updateAdjList[rimBNode].clear();

            for (int k = 2; k >= 0; k--)
                {
                    changeAdjList[k].reverse();
                    updateAdjList[rimBNode].append(&changeAdjList[k]);
                }

            updateAdjList[rimBNode].in(neighbour1);
            updateAdjList[rimBNode].in(hubElement);
            updateAdjList[rimBNode].in(neighbour2);

            delete [] changeAdjList;
        }

    // ********************** part (3) ***************************************

    updateAdjList[hubNumber].reset();

    while (updateAdjList[hubNumber].out(getElement))
        rim[getElement.getNode()] = 0;
}

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** clusterTest ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::clusterTest(treeElement* clusterPtr)
{
    int clusterPlanarTest = 0;
    clusterData* datas = new clusterData;

    datas->nodes = actual_N;
    datas->edges = actual_M;
    datas->number = tree->getItem(clusterPtr);

    int index = datas->number - n;

    // ************* part (1) ************************************************

    // Determine the hubs of the wheel-graphs contained in the direct
    // sub-clusters of the actual cluster. These hubs were generated in
    // the sub-clusters and possibly new numbered.

    int node;

    clusterNodes[index].reset();

    while (clusterNodes[index].out(node))
        if (node >= actual_N && node < (clusterNumber + actual_N))
            // 'node' represents a cluster.
            while (!generatedHubs[node - actual_N]->stackEmpty())
                hubs[index]->push(generatedHubs[node - actual_N]->pop());

    // ************* part (2) ************************************************

    if (clusterPtr == tree->getRoot())
        {
            // The root-cluster is reached. Try to embed it planar and return
            // with the result of the embedding.

            datas->cluster_graph = new extendedPlanGraph;
            extendPQdirected<outEdge, int, int>* P = NULL;

            datas->cluster_graph->readGraph(actual_N, actual_M, updateAdjList,
                                            0, 0, "Root-Cluster");
            clusterPlanarTest = 
                datas->cluster_graph->
                embedGraph((PQdirected<outEdge,int,int>* &)P,1,1,0);

            if (clusterPlanarTest)
                {
                S->push(datas);
                }
            else
                {

                    P = NULL;

                    delete datas;
                    datas = NULL;
                }

            return clusterPlanarTest;
        }

    // ************* part (3) ************************************************

    // Construct the subgraph induced by the cluster and test it for
    // connection.

    // Determine the number of vertices contained in the subgraph.

    int countN = findNodes(index);

    // Determine the numbers of the vertices contained in the subgraph.

    int* nodesInCluster = new int[actual_N];
	int k;
    for (k = 0; k < actual_N; k++)
        nodesInCluster[k] = 0;

    clusterNodes[index].reset();

    while (clusterNodes[index].out(node))
        nodesInCluster[node] = 1;

    // Determine the number of edges contained in the subgraph and whether
    // the cluster contains only one node incident to virtual edges of the
    // cluster or not.
    
    int virtualEdges = 0; // counts the number of virtual edges of the cluster
    int virtNode = -1, virtClusterNode = -1;

    datas->oneVirtEdge = 0;
    int countM = findEdges(nodesInCluster, virtualEdges, index,
                           datas->oneVirtEdge, virtNode, virtClusterNode);

    // If there is only one node in the cluster incident to virtual edges of
    // the cluster, insert an additional virtual edge.

    if (datas->oneVirtEdge == 1)
        {
            addVirtualEdge(virtNode, virtClusterNode, index, nodesInCluster);
            datas->edges = actual_M + 1;
        }

    // Determine the original numbers of the edges in the actual graph G and
    // vice versa. This is possible, when the edges are not included in
    // clusters that were processed during earlier function calls of
    // 'clusterTest'. If there was inserted an additional virtual edge for
    // the cluster, the number of the edge is 'actual_M' and the
    // corresponding edge number at the beginning of the C-planarity test
    // 'additionalEdgesCounter'.

    additionalEdgesCounter += datas->oneVirtEdge;
    actual_M += datas->oneVirtEdge;

    datas->originalToActual = new int[additionalEdgesCounter];
    datas->actualToOriginal = new int[actual_M];

	int j;
    for (j = 0; j < additionalEdgesCounter; j++)
        datas->originalToActual[j] = -1;

    for (j = 0; j < actual_M; j++)
        datas->actualToOriginal[j] = -1;

    if (datas->oneVirtEdge == 1)
        {
            datas->originalToActual[additionalEdgesCounter - 1] = actual_M - 1;
            datas->actualToOriginal[actual_M - 1] = additionalEdgesCounter - 1;
        }

    saveOriginalEdgeNumbers(datas);

       // Determine the subgraph.

       // Construct the adjacency lists of the subgraph; during the
       // construction store the data necessary for the generation of
       // the extended subgraph.

    datas->oldNode = new int[actual_N];
    datas->oldEdge = new int[actual_M];
    datas->saveNodeNumbers = new int[countN + 1];
    datas->saveEdgeNumbers = new int[countM + virtualEdges];
    datas->graphEnd = new int[countM + virtualEdges];
    linkedList<outEdge>* clusterAdjList = new linkedList<outEdge>[countN+1];
    linkedList<outEdge>* virtualEdgesList = new linkedList<outEdge>[countN+1];

    for (j = 0; j < actual_N; j++)
        datas->oldNode[j] = -1;
    for (k = 0; k < actual_M; k++)
        datas->oldEdge[k] = -1;

    S->push(datas);

    findSubGraph(countN, countM, nodesInCluster, clusterAdjList,
                 virtualEdgesList, index);

       // Test the subgraph for connection.

    extendedPlanGraph* graph = new extendedPlanGraph;
    graph->readGraph(countN, countM, clusterAdjList, 0, 0, "Cluster-Graph");

    if (!(clusterPlanarTest = graph->connectionTest()))
        cout << "WARNING (C_planarConnectedC_Graph::clusterTest):"
             << "\n\tthe graph induced by cluster no. " << datas->number + 1
             << " is not connected!" << endl;

    delete graph;

    // ************* part (4) ************************************************

    // Construct the extended subgraph and try to embed it planar.

    extendPQdirected<outEdge,int,int>* P = NULL;

    if (clusterPlanarTest)
        {

            // Construct the extended subgraph.

            findExtendedCluster(countN, clusterAdjList, virtualEdgesList);

            // Start the planar embedding of the extended subgraph.

            countM += virtualEdges;

            datas->cluster_graph = new extendedPlanGraph;

            datas->cluster_graph->readGraph(countN + 1, countM, clusterAdjList,
                                            0, 0, "Cluster-Graph");
            clusterPlanarTest =
                datas->cluster_graph->
                embedGraph((PQdirected<outEdge,int,int>* &)P,1,0,0);

            if (!clusterPlanarTest)
                {
                    P = NULL;
                }

        }

    delete [] clusterAdjList;
    delete [] virtualEdgesList;

    // ************* part (5) ************************************************

    // Construct the representative graph of the cluster according to the
    // PQ-tree returned from the planar embedding.

    if (clusterPlanarTest)
        {
            // Determine the number of vertices and edges contained in the
            // representative graph; these is necessary to generate storage
            // place of correct size.

            int edge = 0, newActN = 0, newActM = 0;

            node = 0;

            countPQTree(P->getRoot(), node, edge);

            // number of vertices after the update
            newActN = actual_N - countN + node;
            
            // number of edges after the update
            newActM = actual_M - countM + edge; 

            // Construct the representative graph.

            int countNode = 0, countEdge = 0, store_node = node;
            linkedList<outEdge>* helpAdjList = new linkedList<outEdge>[newActN];
            datas->storeNumbers = new int[actual_N];
            datas->storeEdges   = new int[actual_M];
            datas->leafBKeys     = new leafKey<outEdge,int,int>*[actual_M];
            datas->nodeNumbers  = new int[newActN];
            datas->edgeNumbers  = new int[newActM];
            datas->clusterEnd   = new int[newActM];

            for (int i = 0; i < actual_N; i++)
                datas->storeNumbers[i] = -1;
            for (int j = 0; j < actual_M; j++)
                {
                    datas->storeEdges[j] = -1;
                    datas->leafBKeys[j] = NULL;
                }
            for (int k = 0; k < newActN; k++)
                datas->nodeNumbers[k] = -1;
            for (int l = 0; l < newActM; l++)
                {
                    datas->edgeNumbers[l] = -1;
                    datas->clusterEnd[l] = -1;
                }

            constructWheelGraph(P->getRoot(), helpAdjList, countNode,
                                countEdge, node, 0, index);

            // ************* part (6) ****************************************

            // Replace the subgraph induced by the cluster by the
            // representative graph and construct a consecutive numbering
            // of graph G.

               // Continue the new numbering of G started during the
               // construction of the representative graph.

            updateGraph(helpAdjList, node, edge, nodesInCluster);

               // Change the numbering in the arrays 'clusterNodes' and
               // 'generatedHubs'.

            correctClusterNodes_and_generatedHubs(index, store_node, newActN,
                                                  datas->storeNumbers);

               // Take over the new numbered graph G to the adjacency lists of
               // array 'updateAdjList'.

            actual_N = newActN;
            actual_M = newActM;

            delete [] updateAdjList;
            updateAdjList = helpAdjList;

        }

    delete [] datas->storeNumbers;
    delete [] datas->storeEdges;
    datas->storeEdges = datas->storeNumbers = NULL;


    delete [] nodesInCluster;
    return clusterPlanarTest;
}

/*****************************************************************************/
/*****************************                  ******************************/
/***************************** constructC_Graph ******************************/
/*****************************                  ******************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::constructC_Graph(const ClusterGraph& C)
{
    int readIn = clusterGraph::constructC_Graph(C);

    if (readIn)
        {
            C_planar = -1;
            C_embedded = -1;
        }

    return readIn;
}

/*****************************************************************************/
/*****************************                     ***************************/
/***************************** constructWheelGraph ***************************/
/*****************************                     ***************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::constructWheelGraph(pqNode<outEdge,int,int>* root,
                                                   linkedList<outEdge>* &hAL,
                                                   int &count_N, int &count_M,
                                                   int &_node, int parentNum,
                                                   int cluster)
{
    if (root->type() == P_NODE)
        {
            // The P-node represents a cutvertex, possibly it can also be the
            // "first" vertex of a 2-connected component (for each child that
            // is a Q-node).

            int mark = 1;
            int saveCount_N = count_N; // Store the new number of the node
                                       // that was assigned during the last
                                       // function call.
            pqNode<outEdge,int,int>* child = root->referenceChild();

            while (mark || child != root->referenceChild())
                {
                    mark = 0;

                    if (child->type() == P_NODE)
                        {
                            // The actual node is only a cutvertex.

                            // new number of the child node

                            count_N++;

                            // connection between the two P-nodes
                    
                            outEdge element_1(saveCount_N, count_M);
                            outEdge element_2(count_N, count_M++);

                            hAL[saveCount_N].in(element_2);
                            hAL[count_N].in(element_1);
                    
                            // continue with the child

                            constructWheelGraph(child, hAL, count_N, count_M,
                                                _node, saveCount_N, cluster);
                        }
                    else if (child->type() == Q_NODE)
                        {
                            // A wheel has to be constructed next, because the
                            // actual node belongs to a 2-connected component.

                            // number for the hub of the wheel

                            count_N++;

                            // connection between the actual node and the hub
                            // of the new wheel-subgraph

                            outEdge element_1(saveCount_N, count_M);
                            outEdge element_2(count_N, count_M++);

                            hAL[saveCount_N].in(element_2);
                            hAL[count_N].in(element_1);

                            // Store the number of the hubs contained in the
                            // repr. graph of this cluster in a stack to
                            // identify the wheel-subgraphs in later
                            // calculation.

                            generatedHubs[cluster]->push(count_N++);
                            
                            // connection between the actual node and the
                            // rim-node that corresponds with the
                            // 'leftEndmost'-child of the Q-node
                            
                            outEdge element_3(saveCount_N, count_M);
                            outEdge element_4(count_N, count_M++);
                    
                            hAL[saveCount_N].in(element_4);
                            hAL[count_N].in(element_3);

                            // continue with the child

                            constructWheelGraph(child, hAL, count_N, count_M,
                                                _node, saveCount_N, cluster);
                        }
                    else if (child->type() == LEAF)
                        {
                            // The actual node is incident to a virtual edge.

                            // Look at the element stored in the leaf and
                            // assign new numbers to the corresponding virtual
                            // edge and the incident node which belongs to the
                            // cluster

                            outEdge* element = child->getKey()->
                                userStructKey();
                            clusterData* D = S->pop();
                            int number = -1;

                            // Store the leafKey of the leaf that corresponds with
                            // the virtual edge; necessary to permutate the
                            // leafs of the PQ-tree.

                            D->leafBKeys[D->saveEdgeNumbers[element->getNumber()]]
                                = child->getKey();

                            // Assign a new number to node
                            // D->graphEnd[element->getNumber()], if this was not
                            // done during earlier function calls. Store for
                            // the old number the new number and vice versa.

                            if (D->storeNumbers[D->graphEnd[element->getNumber()]]
                                == -1)
                                {
                                    D->nodeNumbers[_node] =
                                        D->graphEnd[element->getNumber()];
                                    D->storeNumbers
                                        [D->graphEnd[element->getNumber()]] =
                                        _node;
                                }
                            else
                                number = D->storeNumbers
                                    [D->graphEnd[element->getNumber()]];

                            // Assign a new number to node
                            // D->saveNodeNumbers[element->getNode()], if this
                            // was not done during earlier function calls.
                            // Store for the old number the new number and
                            // vice versa.

                            if (D->storeNumbers[D->saveNodeNumbers
                                               [element->getNode()]] == -1)
                                {
                                    D->storeNumbers
                                        [D->saveNodeNumbers[element->getNode()]] =
                                        saveCount_N;
                                    D->nodeNumbers[saveCount_N] =
                                        D->saveNodeNumbers[element->getNode()];
                                }

                            // Assign a new number to edge
                            // D->saveEdgeNumbers[element->getNumber()]. Store
                            // for the old number the new number and vice
                            // versa.
                    
                            D->edgeNumbers[count_M] =
                                D->saveEdgeNumbers[element->getNumber()];
                            D->storeEdges[D->saveEdgeNumbers
                                         [element->getNumber()]] = count_M;

                            // Store for the new numbered edge the old
                            // incident node, that belongs to the cluster;
                            // necessary to identify virtual edges in later
                            // calculations.
                            
                            D->clusterEnd[count_M] = 
                                D->saveNodeNumbers[element->getNode()];
                    
                            S->push(D);

                            // connection between the remaining graph and the
                            // representative graph

                            if (number == -1)  // connection, if node and edge
                                               // get new numbers
                                {
                                    outEdge element_1(saveCount_N, count_M);
                                    outEdge element_2(_node, count_M++);

                                    hAL[saveCount_N].in(element_2);
                                    hAL[_node++].in(element_1);
                                }
                            else  // connection, if only the edge gets a
                                  // new number
                                {
                                    outEdge element_1(saveCount_N, count_M);
                                    outEdge element_2(number, count_M++);

                                    hAL[saveCount_N].in(element_2);
                                    hAL[number].in(element_1);
                                }
                        }
                    else if (child->type() == IND)
                        cout << "ERROR (C_planarConnectedC_Graph::"
                             << "constructWheelGraph):"
                             << "\n\tP-node posses an indicator as child!"
                             << endl;
                    else
                        cout << "ERROR (C_planarConnectedC_Graph::"
                             << "constructWheelGraph):"
                             << "\n\terror in the PQ-tree!" << endl;

                    child = child->getNextSib(NULL);
                }
        }

    if (root->type() == Q_NODE)
        {
            // The actual node represents a 2-connected component; therefore
            // a wheel-subgraph is constructed.

            // Store the numbers of the wheel-node from which the wheel is
            // reached (parentNum), the hub (hubNum) and the node, that
            // corresponds with the 'leftEndmost'-child of the Q-node
            // (saveCount_N). All these numbers were assigned during the
            // last function call.
            
            // int parentNum = count_N - 2;
            int hubNum = count_N - 1;
            int saveCount_N = count_N;
            pqNode<outEdge,int,int>* child = root->getEndmost(0);
            pqNode<outEdge,int,int>* left = NULL;
            pqNode<outEdge,int,int>* right = NULL;

            while (child != NULL)
                {
                    int mark = 1;

                    if (child->type() == P_NODE)
                        {
                            constructWheelGraph(child, hAL, count_N, count_M,
                                                _node, saveCount_N, cluster);
                        }
                    else if (child->type() == Q_NODE)
                        {
                            // A wheel has to be constructed next, because
                            // the node also belongs to another 2-connected
                            // component.

                            // number for the hub of the new wheel

                            count_N++;

                            // connection between the actual node and the
                            // hub of the new wheel-subgraph

                            outEdge element_1(saveCount_N, count_M);
                            outEdge element_2(count_N, count_M++);

                            hAL[saveCount_N].in(element_2);
                            hAL[count_N].in(element_1);

                            // Store the number of the hubs contained in
                            // the repr. graph of this cluster in a stack
                            // to identify the wheel-subgraphs in later
                            // calculation.

                            generatedHubs[cluster]->push(count_N++);

                            // connection between the actual node and the
                            // rim-node that corresponds with the
                            // 'leftEndmost'-child of the Q-node

                            outEdge element_3(saveCount_N, count_M);
                            outEdge element_4(count_N, count_M++);

                            hAL[saveCount_N].in(element_4);
                            hAL[count_N].in(element_3);

                            // continue with the child
                    
                            constructWheelGraph(child, hAL, count_N, count_M,
                                                _node, saveCount_N, cluster);
                        }
                    else if (child->type() == LEAF)
                        {
                            // The actual node is incident to a virtual edge.

                            // Look at the element stored in the leaf and
                            // assign new numbers to the corresponding virtual
                            // edge and the incident node which belongs to the
                            // cluster

                            outEdge* element = child->getKey()->
                                userStructKey();
                            clusterData* D = S->pop();
                            int number = -1;
                    
                            // Store the leafKey of the leaf that corresponds 
                            // with the virtual edge; necessary to permutate
                            // the leafs of the PQ-tree.

                            D->leafBKeys[D->saveEdgeNumbers[element->getNumber()]]
                                = child->getKey();

                            // Assign a new number to node
                            // D->graphEnd[element->getNumber()], if this was
                            // not done during earlier function calls. Store
                            // for the old number the new number and vice
                            //versa.

                            if (D->storeNumbers[D->graphEnd
                                               [element->getNumber()]] == -1)
                                {
                                    D->nodeNumbers[_node] =
                                        D->graphEnd[element->getNumber()];
                                    D->storeNumbers
                                        [D->graphEnd[element->getNumber()]]
                                        = _node;
                                }
                            else
                                number = D->storeNumbers[D->graphEnd
                                                        [element->getNumber()]];

                            // Assign a new number to node
                            // D->saveNodeNumbers[element->getNode()], if this
                            // was not done during earlier function calls.
                            // Store for the old number the new number and
                            // vice versa.

                            if (D->storeNumbers[D->saveNodeNumbers
                                               [element->getNode()]] == -1)
                                {
                                    D->storeNumbers
                                        [D->saveNodeNumbers[element->getNode()]]
                                        = saveCount_N;
                                    D->nodeNumbers[saveCount_N] =
                                        D->saveNodeNumbers[element->getNode()];
                                }
                    
                            // Assign a new number to edge
                            // D->saveEdgeNumbers[element->getNumber()].
                            // Store for the old number the new number and
                            // vice versa.

                            D->edgeNumbers[count_M] =
                                D->saveEdgeNumbers[element->getNumber()];
                            D->storeEdges[D->saveEdgeNumbers
                                         [element->getNumber()]] = count_M;
                    
                            // Store for the new numbered edge the old
                            // incident node,that belongs to the cluster;
                            // necessary to identify virtual edges in
                            // later calculations.

                            D->clusterEnd[count_M] = 
                                D->saveNodeNumbers[element->getNode()];
                    
                            S->push(D);

                            // connection between the remaining graph and
                            // the repr. graph

                            if (number == -1)  // connection, if node and
                                               // edge get new numbers
                                {
                                    outEdge element_1(saveCount_N, count_M);
                                    outEdge element_2(_node, count_M++);
                            
                                    hAL[saveCount_N].in(element_2);
                                    hAL[_node++].in(element_1);
                                }
                            else  // connection, if only the edge gets a
                                  // new number
                                {
                                    outEdge element_1(saveCount_N, count_M);
                                    outEdge element_2(number, count_M++);

                                    hAL[saveCount_N].in(element_2);
                                    hAL[number].in(element_1);
                                }
                        }
                    else if (child->type() == IND)
                        mark = 0;
                    else
                        cout << "ERROR (C_planarConnectedC_Graph::"
                             << "constructWheelGraph):"
                             << "\n\terror in the PQ-tree!" << endl;

                    if (mark)
                        {
                            // All children of a Q-node (that are not
                            // indicators) have to be connected with
                            // the hub of the wheel and with the node
                            // that corresponds with the right Q-node
                            // sibling (that is not an indicator).

                            // connection between the actual node and the hub
                    
                            outEdge element_5(saveCount_N, count_M);
                            outEdge element_6(hubNum, count_M++);

                            hAL[saveCount_N].in(element_6);
                            hAL[hubNum].in(element_5);

                            // connection between the actual node and the
                            // node that corresponds with the right Q-node
                            // sibling (that is not an indicator). If there
                            // is no such right Q-node sibling, the node is
                            // connected with the "first" node of the wheel,
                            // i.e. the node from which the wheel was reached.

                            pqNode<outEdge,int,int>* helpPtr = child;
                            pqNode<outEdge,int,int>* save1 = left,* save2;

                            while (helpPtr->getNextSib(save1) != NULL && 
                                   helpPtr->getNextSib(save1)->type() == IND)
                                {
                                    save2 = helpPtr->getNextSib(save1);
                                    save1 = helpPtr;
                                    helpPtr = save2;
                                }

                            if (helpPtr->getNextSib(save1) != NULL)
                                {
                                    count_N++;
                                    
                                    outEdge element_7(saveCount_N, count_M);
                                    outEdge element_8(count_N, count_M++);

                                    hAL[saveCount_N].in(element_8);
                                    hAL[count_N].in(element_7);
                                }
                            else
                                {
                                    outEdge element_7(saveCount_N, count_M);
                                    outEdge element_8(parentNum, count_M++);

                                    hAL[saveCount_N].in(element_8);
                                    hAL[parentNum].in(element_7);
                                }
                        }
                    
                    saveCount_N = count_N; // Store the number of the
                                           // Q-node-child that was new
                                           // numbered last.
                    right = child->getNextSib(left);
                    left = child;
                    child = right;
                }
        }
}

/*****************************************************************************/
/*********************                                       *****************/
/********************* correctClusterNodes_and_generatedHubs *****************/
/*********************                                       *****************/
/*****************************************************************************/

void C_planarConnectedC_Graph::correctClusterNodes_and_generatedHubs(int
                               cluster, int storeNode, int newN, int* newBNode)
{
    for (int counter = 0; counter < clusterNumber; counter++)
        {
            // *************************************************************
            // ******* Change the numbering in array 'clusterNodes'. *******
            // *************************************************************

            // Change the numbers stored for each cluster that is not the
            // actual cluster.

            if (counter != cluster)
                {
                    int start = 1, number;
                    element<int>* previous,* storeActual;

                    clusterNodes[counter].reset();
                    previous = storeActual =  clusterNodes[counter].
                        getActual();

                    while (clusterNodes[counter].out(number))
                        {
                            clusterNodes[counter].
                                remove(storeActual, previous);

                            if (number >= actual_N &&
                                number < (clusterNumber + actual_N))
                                // 'number' is the number of a cluster.
                                clusterNodes[counter].
                                    in(number - actual_N + newN);
                            else if (number >= 0 && number < actual_N)
                                // 'number' is the number of a vertex.
                                clusterNodes[counter].in(newBNode[number]);
                            else
                                cout << "ERROR (C_planarConnectedC_Graph::"
                                     << "correctClusterNodes_and_generated"
                                     << "Hubs): " << "\n\tthe found number "
                                     << "is neither a cluster-"
                                     << "number nor a vertex-number!" << endl;
                            
                            if (start)
                                {
                                    start = 0;
                                    previous = clusterNodes[counter].
                                        getFirst();
                                }

                            storeActual = clusterNodes[counter].getActual();
                        }
                }
            else
                // If 'counter' = 'cluster', the nodes of the actual cluster
                // are replaced by the nodes of the corresponding repr. graph.

                {
                    clusterNodes[counter].clear();

                    for (int k = 0; k < storeNode; k++)
                        clusterNodes[counter].in(k);
                }

            // **************************************************************
            // ***** Change the numbering of the array 'generatedHubs'. *****
            // **************************************************************

            pqStack<int>* helpStack = new pqStack<int>;
            
            if (counter != cluster)
                {
                    while (!generatedHubs[counter]->stackEmpty())
                        helpStack->push(newBNode[generatedHubs[counter]
                                               ->pop()]);

                    while (!helpStack->stackEmpty())
                        generatedHubs[counter]->push(helpStack->pop());

                }
            delete helpStack;
        }
}

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** countPQTree ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::countPQTree(pqNode<outEdge,int,int>* root,
                                           int &countNode, int &countEdge)
{
    if (root->type() == P_NODE)
        {
            // The actual node represents a cutvertex, possibly it can also
            // be the "first" vertex of a 2-connected component (for each
            // child that is a Q-node.

            int mark = 1;
            pqNode<outEdge,int,int>* child = root->referenceChild();

            countNode++;   // node that corresponds to the actual PQ-tree node
            while (child && (mark || root->referenceChild() != child))
                {
                    mark = 0;

                    if (child->type() == P_NODE)
                        {
                            // The actual node is only a cutvertex.

                            countEdge++; // edge to the next P-node
                            countPQTree(child, countNode, countEdge);
                        }
                    else if (child->type() == Q_NODE)
                        {
                            // A wheel has to be constructed next, because the
                            // actual node belongs to a 2-connected component.

                            countEdge += 2; // edges to the hub of the wheel
                                            // and to one of the two adjacent
                                            // nodes on the rim

                            countPQTree(child, countNode, countEdge);
                        }
                    else if (child->type() == LEAF)
                        {
                            // The actual node is incident to a virtual edge.

                            countEdge++;   // virtual edge
                        }
                    else
                        cout << "ERROR (C_planarConnectedC_Graph::countPQTree"
                             << "):\n\tP-node posses a child that is neither "
                             << "a P- or Q-node nor a leaf!" << endl;

                    child = child->getNextSib(NULL);

                }
        }

    if (root->type() == Q_NODE)
        {
            // The actual node represents a 2-connected component; therefore
            // a wheel-subgraph is constructed.

            pqNode<outEdge,int,int>* child = root->getEndmost(0);
            pqNode<outEdge,int,int>* left = NULL;
            pqNode<outEdge,int,int>* right = NULL;

            countNode++;   // node that represents the hub of the wheel

            while (child != NULL)
                {
                    if (child->type() == P_NODE)
                        {
                            countEdge += 2; // edges to the hub of the wheel
                                            // and to the rim-node that
                                            // corresponds with the right
                                            // Q-node sibling (that is not
                                            // an indicator) of 'child'
                            countPQTree(child, countNode, countEdge);
                        }
                    else if (child->type() == Q_NODE)
                        {
                            // The next child contained in the wheel is the
                            // "first" node of a new 2-connected component,
                            // so another wheel-subgraph has to be constructed.

                            countNode++;    // node that represents the actual
                                            // node
                            countEdge += 4; // edges to the hub and to the
                                            // rim-node of the already
                                            // existing wheel and edges to
                                            // the hub and to one of the two
                                            // rim-nodes of the next
                                            // wheel-subgraph
                            countPQTree(child, countNode, countEdge);
                        }
                    else if (child->type() == LEAF)
                        {
                            // The actual node is incident to a virtual edge.

                            countNode++;    // node that represents the actual
                                            // node
                            countEdge += 3; // edges to the hub of the wheel
                                            // and to the next node on the rim
                                            // and the virtual edge

                        }
                    else if (child->type() != IND)
                        cout << "ERROR (C_planarConnectedC_Graph::countPQTree"
                             << "):\n\tQ-node posses a child that is neither "
                             << "a P- or Q-node nor a leaf nor an indicator!"
                             << endl;

                    right = child->getNextSib(left);
                    left = child;
                    child = right;
                }
        }
}

/*****************************************************************************/
/*****************************                 *******************************/
/***************************** C_planarCluster *******************************/
/*****************************                 *******************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::C_planarCluster(treeElement* helpPtr)
{
    if (helpPtr && C_planar)
        {
            C_planarCluster(tree->getChild(helpPtr));

            if (C_planar)
                C_planarCluster(tree->getNext(helpPtr));

            if(C_planar && tree->getItem(helpPtr) >= n)
                C_planar = clusterTest(helpPtr);
        }

}

/*****************************************************************************/
/*****************************               *********************************/
/***************************** C_planarEmbed *********************************/
/*****************************               *********************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::C_planarEmbed()
{
    if (!C_planarTest(0))
        {
            return C_embedded = 0;
        }
    else if (n == 0 && m == 0 && clusterNumber == 1)
        {
            return C_embedded = 1;
        }
    else
        {
            // The clustered graph is a C-planar connected clustered graph.

            
    // *************************** part (A) **********************************


            // Perform the steps (1) - (5) for each cluster in stack S:

            embedding = new pqQueue<int>*[clusterNumber];

			int j;
            for (j = 0; j < clusterNumber; j++)
                embedding[j] = new pqQueue<int>;

            while (!S->stackEmpty())
                {
                    clusterData* D = S->pop();

                 // ********** step (1) **************************************


                    // If the root-cluster is reached the complete planar
                    // embedding of the cluster is copied to the adjacency
                    // lists of 'updateAdjList'.

                    if (D->number == tree->getItem(tree->getRoot()))
                        {
                            for (int k = 0; k < actual_N; k++)
                                {
                                    updateAdjList[k].clear();
                                    updateAdjList[k].append(D->cluster_graph->
                                                            getAdjList(k));
                                }

                        }
                    else
                        {

                 // ********** step (2) **************************************


                            // If not the root-cluster is reached, the
                            // interrupted planar embedding of the cluster
                            // is completed according to the (final) order
                            // of the virtual edges of the cluster.

                            D->cluster_graph->completeEmbedding(D->leafBKeys,
                                     D->originalToActual, D->actualToOriginal,
                                     embedding, D->number-n, m);

                 // ********** step (3) **************************************


                            // Replace the representative graph of the
                            // cluster by the planar embedded cluster and
                            // make an update of G.

                            change_and_correctNumbers(D);

                            // number of nodes in the actual graph
                            actual_N = D->nodes;

                            // number of edges in the actual graph
                            actual_M = D->edges - D->oneVirtEdge;

                            // The calculation for this cluster is finished
                            // now, so it is possible to delete the stored
                            // data of this cluster.

                            delete [] D->nodeNumbers;
                            delete [] D->edgeNumbers;
                            delete [] D->saveNodeNumbers;
                            delete [] D->saveEdgeNumbers;
                            delete [] D->graphEnd;
                            delete [] D->clusterEnd;

                            for (int k = 0; k < actual_M + D->oneVirtEdge; k++)
                                D->leafBKeys[k] = NULL;
                            delete [] D->leafBKeys;                          

                            D->nodeNumbers = D->edgeNumbers = NULL;
                            D->saveNodeNumbers = D->saveEdgeNumbers = NULL;
                            D->graphEnd = D->clusterEnd = NULL;
                            D->leafBKeys = NULL;
                        }

                    delete D->cluster_graph;
                    D->cluster_graph = NULL;

                    delete [] D->originalToActual;
                    delete [] D->actualToOriginal;
                    D->originalToActual = D->actualToOriginal = NULL;

                 // ********** step (4) *************************************


                    // If the function reaches the last cluster in stack S,
                    // the function stops here. Otherwise modify the
                    // wheel-subgraphs of the cluster to canonically wheels.

                    // Notice: The actual graph now has the same size and
                    //         numbering as at the time before function
                    //         'clusterTest()' was called for this cluster.

                    if (!S->stackEmpty())
                        {
                            int index = D->number - n;

                            int* rimNodes = new int[actual_N];

                            for (int j = 0; j < actual_N; j++)
                                rimNodes[j] = 0;

                            while (!hubs[index]->stackEmpty())
                                changeWheel(hubs[index]->pop(), rimNodes);

                            delete [] rimNodes;

                 // ********** step (5) **************************************


                            // Determine the final order of the virtual edges
                            // of the next cluster in stack S.

                            findVirtualOrder(embedding);

                        }

                    delete D;
                    D = NULL;
                }

    // *************************** part (B) ***********************************


            // If the underlying graph containted loops, the loops were removed
            // at the beginning of the C-planarity test. Now reinsert the loops
            // and remake the edge numbering consecutively.

            if (m != actual_M)
                {
                    linkedList<outEdge>* helpAdjList =
                        new linkedList<outEdge>[n];

                    for (int i = 0; i < n; i++)
                        {
                            outEdge outElement;

                            updateAdjList[i].reverse();
                            updateAdjList[i].reset();

                            while (updateAdjList[i].out(outElement))
                                {
                                    outEdge inElement(outElement.getNode(),
                                                      rememberEdgeNumbers
                                                      [outElement.getNumber()]);

                                    helpAdjList[i].in(inElement);
                                }
                           
                            updateAdjList[i].clear();
                            helpAdjList[i].append(&saveLoops[i]);
                            actual_M += saveLoops[i].count();
                        }

                    delete [] updateAdjList;

                    updateAdjList = new linkedList<outEdge>[n];
                    for (int j = 0; j < n; j++)
                        updateAdjList[j].append(&helpAdjList[j]);

                    delete [] helpAdjList;
                    helpAdjList = NULL;

                    delete [] rememberEdgeNumbers;
                    rememberEdgeNumbers = NULL;
                }            

    // *************************** part (C) ***********************************


            // Copy the determined C-planar embedding to 'adjacencyList' and
            // delete the unneeded storage space.

            for (int k = 0; k < clusterNumber; k++)
                {
                    delete hubs[k];
                    hubs[k] = NULL;
                }

            if (S)
                {
                    clusterData* checkDataPtr = NULL;

                    while (!S->stackEmpty())
                        if ((checkDataPtr = S->pop()))
                            delete checkDataPtr;
                    
                    delete    S;
                    S = NULL;
                }

            delete [] clusterNodes;
            delete [] saveLoops;
            delete [] hubs;

            clusterNodes = NULL;
            saveLoops = NULL;
            hubs = NULL;

            // copy the C-planar embedding to the array 'adjacencyList'

            delete [] adjacencyList;

            adjacencyList = new linkedList<outEdge>[n];
            for (j = 0; j < n; j++)
                adjacencyList[j].append(&updateAdjList[j]);

            delete [] updateAdjList;
            updateAdjList = NULL;

            return C_embedded = 1;
        }
}

/*****************************************************************************/
/*****************************              **********************************/
/***************************** C_planarTest **********************************/
/*****************************              **********************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::C_planarTest(int onlyPlanar)
{

    // test, if G is the empty clustered graph

    if (n == 0 && m == 0 && clusterNumber == 1)
        {
            return C_planar = 1;
        }

    // test the underlying graph G for connection and planarity

    extendedPlanGraph* graph = new extendedPlanGraph;

    graph->readGraph(n, m, adjacencyList, 0, 0, graphName);

    if (!(C_planar = graph->connectionTest()))
        cout << "ERROR (C_planarConnectedC_Graph::C_planarTest):"
             << "\n\tgraph " << graphName << " is not connected!" << endl;
    else C_planar = graph->isPlanarC();

    delete graph;

    if (!C_planar) // graph G is not connected or not planar
        return 0;
    else
        {
            // C-planarity test is performed on the copied arrays
            // 'clusterNodes' and 'updateAdjList', because several
            // updates have to be done for G

            actual_N = n;
            actual_M = m;
            additionalEdgesCounter = m;

            updateAdjList = new linkedList<outEdge>[n];
            for (int j = 0; j < n; j++)
                updateAdjList[j].append(&adjacencyList[j]);

            clusterNodes = new linkedList<int>[clusterNumber];
			int k;
            for (k = 0; k < clusterNumber; k++)
                clusterNodes[k].append(&c_Nodes[k]);

            S = new pqStack<clusterData*>;

            // remove all loops from 'updateAdjList', if G contains loops;
            // number the remained edges consecutively

            saveLoops = new linkedList<outEdge>[n];

            getLoops(); 

            // generate the arrays 'generatedHubs' and 'hubs' for storing
            // the hubs of each cluster

            generatedHubs = new pqStack<int>*[clusterNumber];
            hubs = new pqStack<int>*[clusterNumber];

            for (k = 0; k < clusterNumber; k++)
                {
                    generatedHubs[k] = new pqStack<int>;
                    hubs[k] = new pqStack<int>;
                }

            // start the calculations for each cluster

            C_planarCluster(tree->getRoot());

            // Delete the generated space, if C is not C-planar or if no
            // embedding is determined.

            if (onlyPlanar || !C_planar)
                {
                    delete [] saveLoops;
                    delete [] updateAdjList;
                    delete [] clusterNodes;
                    
                    if (S)
                        {
                            clusterData* checkDataPtr = NULL;

                            while (!S->stackEmpty())
                                if ((checkDataPtr = S->pop()))
                                    delete checkDataPtr;

                            delete    S;
                            S = NULL;
                        }

                    if (rememberEdgeNumbers)
                        delete [] rememberEdgeNumbers;

                    for (int k = 0; k < clusterNumber; k++)
                        delete hubs[k];

                    delete [] hubs;

                    saveLoops = updateAdjList = NULL;
                    clusterNodes = NULL;
                    rememberEdgeNumbers = NULL;
                    hubs = NULL;
                }

            for (k = 0; k < clusterNumber; k++)
                delete generatedHubs[k];

            delete [] generatedHubs;
            generatedHubs = NULL;

            C_embedded = C_planar;
        }
       
    return C_planar;

}

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** deleteGraph ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::deleteGraph()
{
    if (clusterNodes)
        {
            delete [] clusterNodes;
            clusterNodes = NULL;
        }

    if (saveLoops)
        {
            delete [] saveLoops;
            saveLoops = NULL;
        }

    if (rememberEdgeNumbers)
        {
            delete [] rememberEdgeNumbers;
            rememberEdgeNumbers = NULL;
        }

    if (S)
        {
            while (!S->stackEmpty())
                delete S->pop();

            delete S;
            S = NULL;
        }

    if (updateAdjList && adjacencyList != updateAdjList)
            {
                delete [] updateAdjList;
                updateAdjList = NULL;
            }

    if (generatedHubs)
        {
            for (int i = 0; i < clusterNumber; i++)
                    delete generatedHubs[i];

            delete [] generatedHubs;
            generatedHubs = NULL;
        }

    if (hubs)       
        {
            for (int i = 0; i < clusterNumber; i++)
                    delete hubs[i];

            delete [] hubs;
            hubs = NULL;
        }

    if (embedding)
        {
            for (int i = 0; i < clusterNumber; i++)
                {
                    delete embedding[i];
                    embedding[i] = NULL;
                }

            delete [] embedding;
            embedding = NULL;
        }

    C_planar = C_embedded = 1;
    actual_M = actual_N = 0;

    clusterGraph::deleteGraph();
}

/*****************************************************************************/
/*****************************           *************************************/
/***************************** findEdges *************************************/
/*****************************           *************************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::findEdges(int* inBNodes, int &v_Edges,
                                        int cluster, int& onlyOne,
                                        int& storeVirtNode, int& storeNode)
{
    int count = 0, number;

    clusterNodes[cluster].reset();

    while (clusterNodes[cluster].out(number))
        {
            outEdge getElement;

            updateAdjList[number].reset();

            while (updateAdjList[number].out(getElement))
                {
                    if (inBNodes[getElement.getNode()])
                        count++;
                    else
                        {
                            v_Edges++;

                            if (!onlyOne)
                                {
                                    storeNode = number;
                                    storeVirtNode = getElement.getNode();
                                    onlyOne++;
                                }
                            else if (number != storeNode)
                                onlyOne++;
                        }

                }
        }

    if (onlyOne == 1)
        v_Edges++;
    else
        onlyOne = 0;


    return count = count / 2;

}

/*****************************************************************************/
/*****************************                     ***************************/
/***************************** findExtendedCluster ***************************/
/*****************************                     ***************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::findExtendedCluster(int nodes,
                                                   linkedList<outEdge>* &cAL,
                                                   linkedList<outEdge>* vEL)
{

    // Change the 0. adjacency to the (nodes). adjacency list, because the
    // virtual vertex gets the number '0'.

    cAL[nodes].append(&cAL[0]);
    cAL[0].clear();

    // Change in array 'cAL' the number '0' to 'nodes', which is the highest
    // vertex number in the extended subgraph.

    outEdge checkElement;

    cAL[nodes].reset();

    while (cAL[nodes].out(checkElement))
        {
            int start = 1;
            outEdge testElement;
            element<outEdge>* previous,* storeActual;

            cAL[checkElement.getNode()].reset();
            previous = storeActual = cAL[checkElement.getNode()].getActual();

            while (cAL[checkElement.getNode()].out(testElement))
                {
                    if (testElement.getNode() == 0)
                        {
                            outEdge newElement(nodes, testElement.getNumber());
                            cAL[checkElement.getNode()].remove(storeActual,
                                                            previous);
                            cAL[checkElement.getNode()].in(newElement);

                            if (start)
                                {
                                    start = 0;
                                    previous = cAL[checkElement.getNode()].
                                        getFirst();
                                }                           
                        }
                    else
                        {
                            start = 0;
                            previous = storeActual;
                        }
                    storeActual = cAL[checkElement.getNode()].getActual();
                }
        }

    // Correct the entries for the numbers '0' and 'nodes' in the arrays
    // 'oldNode' and 'saveNodeNumbers'.

    clusterData* D = S->pop();
    D->oldNode[D->saveNodeNumbers[0]] = nodes;
    D->saveNodeNumbers[nodes] = D->saveNodeNumbers[0];
    D->saveNodeNumbers[0] = -1;
    S->push(D);

    // Add the virtual elements to the adjacency lists of the (extended)
    // subgraph.

    for (int j = 0; j <= nodes; j++)
        cAL[j].append(&vEL[j]);

}

/*****************************************************************************/
/*****************************           *************************************/
/***************************** findNodes *************************************/
/*****************************           *************************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::findNodes(int cluster)
{
    int number;
    element<int>* previous,* storeActual;

    clusterNodes[cluster].reset();
    previous = storeActual = clusterNodes[cluster].getActual();

    while (clusterNodes[cluster].out(number))
        {
            if (number >= actual_N && number < (clusterNumber + actual_N))
                // 'number' is the number of a cluster
                {

                    if (previous == storeActual)
                        // First element in 'clusterNodes[cluster]' is
                        // a cluster
                        previous = clusterNodes
                            [cluster].copyToTop(&clusterNodes
                                                [number - actual_N]); 
                    else
                        clusterNodes[cluster].append(&clusterNodes
                                                     [number - actual_N]);

                    clusterNodes[cluster].remove(storeActual, previous);
                    storeActual = clusterNodes[cluster].getActual();
                    clusterNodes[number - actual_N].clear();
                }
            else
                // 'number' is the number of a node contained in the cluster
                {
                    previous = storeActual;
                    storeActual = clusterNodes[cluster].getActual();
                }
        }

    return clusterNodes[cluster].count();
}

/*****************************************************************************/
/*****************************              **********************************/
/***************************** findSubGraph **********************************/
/*****************************              **********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::findSubGraph(int nodes, int edges, int* inBNodes,
                                            linkedList<outEdge>* &cAL,
                                            linkedList<outEdge>* &vEL,
                                            int cluster)
{

    int nodeCount = 0, edgeCount = 0, number, storeN_Count;
    outEdge checkElement;
    clusterData* D = S->pop();

    clusterNodes[cluster].reset();

    // Visit all vertices of the cluster.

    while(clusterNodes[cluster].out(number))
        {
            if (D->oldNode[number] == -1)
                // BNode 'number' was not processed before.
                {
                    storeN_Count = D->oldNode[number] = nodeCount;
                    D->saveNodeNumbers[nodeCount++] = number;
                }
            else
                storeN_Count = D->oldNode[number];

            updateAdjList[number].reset();

            while (updateAdjList[number].out(checkElement))
                // Change the numbers of the adjacent nodes and incident edges
                // of node 'number'.
                {
                    if (inBNodes[checkElement.getNode()])
                        // Element of the subgraph was found.
                        {
                            if (D->oldNode[checkElement.getNode()] == -1)
                                // The node and the edge numbers were not
                                // changed before.
                                {
                                    outEdge firstElement(storeN_Count,
                                                         edgeCount);
                                    outEdge secondElement(nodeCount,
                                                          edgeCount);

                                    D->oldNode[checkElement.getNode()] =
                                        nodeCount;
                                    D->oldEdge[checkElement.getNumber()] =
                                        edgeCount;
                                    D->saveNodeNumbers[nodeCount] =
                                        checkElement.getNode();
                                    D->saveEdgeNumbers[edgeCount] =
                                        checkElement.getNumber();
                                    D->graphEnd[edgeCount] = -1;
                                    cAL[storeN_Count].in(secondElement);
                                    cAL[nodeCount].in(firstElement);
                                    nodeCount++;
                                    edgeCount++;
                                }

                            if (D->oldEdge[checkElement.getNumber()] == -1)
                                // Only the edge number was not changed before.
                                {
                                    outEdge firstElement(storeN_Count,
                                                         edgeCount);
                                    outEdge secondElement(D->oldNode
                                                          [checkElement.
                                                          getNode()], edgeCount);

                                    D->graphEnd[edgeCount] = -1;
                                    D->saveEdgeNumbers[edgeCount] =
                                        checkElement.getNumber();
                                    D->oldEdge[checkElement.getNumber()] =
                                        edgeCount;
                                    cAL[storeN_Count].in(secondElement);
                                    cAL[D->oldNode[checkElement.getNode()]].in
                                        (firstElement);
                                    edgeCount++;
                                }
                        }
                    else
                        // Virtual element of the extended subgraph was
                        // found; these elements are stored in the array
                        // 'vEL' and added to the adjacency lists of array
                        // 'cAL' in function 'findExtendedCluster'.
                        {
                            outEdge virtualElement(0, edges);

                            D->graphEnd[edges] = checkElement.getNode();
                            D->saveEdgeNumbers[edges] = checkElement.getNumber();
                            D->oldEdge[checkElement.getNumber()] = edges;

                            if (storeN_Count == 0)
                                {
                                    outEdge newElement(nodes, edges);

                                    vEL[0].in(newElement);
                                    vEL[nodes].in(virtualElement);
                                }
                            else
                                {
                                    outEdge newElement(storeN_Count, edges);

                                    vEL[0].in(newElement);
                                    vEL[storeN_Count].in(virtualElement);
                                }
                            edges++;
                        }
                }
            updateAdjList[number].clear();
        }

    // Initialize the entry; used in function 'findExtendedCluster'.

    D->saveNodeNumbers[nodes] = -1;

    S->push(D);

}

/*****************************************************************************/
/*****************************                  ******************************/
/***************************** findVirtualOrder ******************************/
/*****************************                  ******************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::findVirtualOrder(pqQueue<int>** &embedVirtual)
{

  // *************************************************************************
  // ** Determine the number of vertices in the representative graph of the **
  // ** top cluster of stack S.                                             **
  // *************************************************************************

    clusterData* info = S->pop();
    int number = actual_N - info->nodes + info->cluster_graph->getN() - 1,
        cluster = info->number - n;

    S->push(info);

  // **************************************************************************
  // ** Run through the representative graph of the cluster, similar to DFS. **
  // ** Start at node no. 0; this node is always incident to a virtual edge. **
  // **************************************************************************

    int virtualNode = 0;
    int* marked = new int[number];
    pqStack<int> reached;
    outEdge element;

    for (int k = 0; k < number; k++)
        {
            marked[k] = -1;
            updateAdjList[k].reset();
        }

    // Determine the first virtual edge in the adjacency list of node no. 0.

    updateAdjList[virtualNode].out(element);

    while (info->clusterEnd[element.getNumber()] == -1)
        updateAdjList[virtualNode].out(element);

    // Store the first virtual edge of node '0' in queue 'embedVirtual' and
    // 'embedding' respectively.

    embedVirtual[cluster]->enQueue(info->actualToOriginal
                                   [info->edgeNumbers[element.getNumber()]]);

    // Store the node with the virtual edge in stack 'reached' and mark it with
    // the number of the edge from which the node is reached (here it is the
    // virtual edge).

    marked[virtualNode] = element.getNumber();
    reached.push(virtualNode);

    // Work off the stack 'reached'.

    while (!reached.stackEmpty())
        {
            int tail = reached.pop();
            reached.push(tail);

            if (updateAdjList[tail].out(element))
            // The end of the actual adjacency list is not reached yet.
                {
                    int edge = element.getNumber(), head = element.getNode();

                    if (marked[tail] != edge)
                    // The scanning of the actual adjacency list is incomplete.
                        {
                            if (info->clusterEnd[edge] == -1)
                            // The found edge is not a virtual edge.
                                {
                                    if (marked[head] == -1)
                                    // BNode 'head' was not reached before.
                                        {
                                            // Determine in the adjacency
                                            // list of 'head' the edge where
                                            // 'head' is reached/ from.

                                            outEdge outElement;

                                            updateAdjList[head].
                                                out(outElement);

                                            while (outElement.getNumber() != edge)
                                                updateAdjList[head].
                                                    out(outElement);

                                            // Work on in the next step
                                            // with 'head'.

                                            marked[head] = edge;
                                            reached.push(head);
                                        }
                                }
                            else
                            // The found edge is a virtual edge.
                                {
                                    // Insert the edge in queue
                                    // 'embedVirtual[cluster]' and
                                    // 'embedding[cluster]' respectively.

                                    embedVirtual[cluster]->
                                        enQueue(info->
                                                actualToOriginal[info->
                                                                edgeNumbers
                                                                [element.
                                                                getNumber()]]);
                                }
                        }
                    else
                    // The actual adjacency list is worked off completely,
                    // so remove the corresponding node from the stack
                    // 'removed'.

                        reached.pop();
                }
            else
            // The end of the actual adjacency list is reached, but not all
            // elements of the list are scanned, so reset the list again.

                updateAdjList[tail].reset();
        }

    delete [] marked;
    delete [] info->oldNode;
    delete [] info->oldEdge;

    info->oldNode = info->oldEdge = 0;
}

/*****************************************************************************/
/*****************************               *********************************/
/***************************** getC_embedded *********************************/
/*****************************               *********************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::getC_embedded() const
{
    return C_embedded;
}

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** getC_planar ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

int C_planarConnectedC_Graph::getC_planar() const
{
    return C_planar;
}

/*****************************************************************************/
/*****************************              **********************************/
/***************************** getEmbedding **********************************/
/*****************************              **********************************/
/*****************************************************************************/

pqQueue<int>** C_planarConnectedC_Graph::getEmbedding() const
{
    return embedding;
}

/*****************************************************************************/
/*****************************          **************************************/
/***************************** getLoops **************************************/
/*****************************          **************************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::getLoops()
{
    // Search the loops in the array 'updateAdjList', remove and store them in
    // the array 'saveLoops'.

    for (int i = 0; i < n; i++)
        {
            saveLoops[i].clear();
            updateAdjList[i].reset();

            int start = 1;
            outEdge testElement;
            element<outEdge>* previous,* storeActual;

            previous = storeActual = updateAdjList[i].getActual();

            while (updateAdjList[i].out(testElement))
                {
                    if (testElement.getNode() == i)
                        {
                            saveLoops[i].in(testElement);
                            updateAdjList[i].remove(storeActual, previous);
                            actual_M--;

                            if (start)
                                previous = updateAdjList[i].getFirst();
                        }
                    else
                        {
                            start = 0;
                            previous = storeActual;
                        }

                    storeActual = updateAdjList[i].getActual();
                }
        }

    // If the graph contains loops, determine a new and consecutively
    // edge numbering.

    if (m != actual_M)
        {
            int counter = 0;
            rememberEdgeNumbers = new int[actual_M];
            int* saveEdgeNumber = new int[m];
            linkedList<outEdge>* helpAdjList = new linkedList<outEdge>[n];

			int j;
            for (j = 0; j < m; j++)
                saveEdgeNumber[j] = -1;

            for (int i = 0; i < n; i++)
                {
                    outEdge outElement;

                    updateAdjList[i].reset();

                    while (updateAdjList[i].out(outElement))
                        {
                            if (saveEdgeNumber[outElement.getNumber()] == -1)
                                {
                                    outEdge inElement(outElement.getNode(),
                                                      counter);

                                    rememberEdgeNumbers[counter] =
                                        outElement.getNumber();
                                    saveEdgeNumber[outElement.getNumber()] =
                                        counter++;
                                    helpAdjList[i].in(inElement);
                                }
                            else
                                {
                                    outEdge inElement(outElement.getNode(),
                                                      saveEdgeNumber
                                                      [outElement.getNumber()]);

                                    helpAdjList[i].in(inElement);
                                }
                        }
                }

            delete [] saveEdgeNumber;
            delete [] updateAdjList;

            updateAdjList = new linkedList<outEdge>[n];
            for (j = 0; j < n; j++)
                updateAdjList[j].append(&helpAdjList[j]);
            
            delete [] helpAdjList;
            helpAdjList = NULL;
        }
}

/*****************************************************************************/
/****************************                         ************************/
/**************************** saveOriginalEdgeNumbers ************************/
/****************************                         ************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::saveOriginalEdgeNumbers(clusterData* & info)
{
    int edgesBound;

    if (info->oneVirtEdge)
        edgesBound = actual_M - 1;
    else
        edgesBound = actual_M;


    if (S->stackEmpty() && rememberEdgeNumbers)
        {
            for (int count = 0; count < edgesBound; count++)
                {
                    info->actualToOriginal[count] = rememberEdgeNumbers[count];
                    info->originalToActual[rememberEdgeNumbers[count]] = count;
                }
        }
    else if (S->stackEmpty())
        for (int count = 0; count < edgesBound; count++)
            info->actualToOriginal[count] = info->originalToActual[count]
                = count;
    else
        {
            clusterData* beforeInfo = S->pop();
            
            for (int count = 0; count < edgesBound; count++)
                if (beforeInfo->edgeNumbers[count] != -1)
                    if (beforeInfo->actualToOriginal[beforeInfo->
                                                    edgeNumbers[count]] != -1)
                    // the edge 'count' is not included in clusters which were
                    // processed in previous function calls of 'clusterTest',
                    // therefore a corresponding number in the original graph
                    // exists
                        {
                            info->actualToOriginal[count] =
                                beforeInfo->actualToOriginal[beforeInfo->
                                                            edgeNumbers
                                                            [count]];
                            info->originalToActual[info->
                                                  actualToOriginal
                                                  [count]] = count;
                        }
                
            S->push(beforeInfo);
        }
}

/*****************************************************************************/
/*****************************             ***********************************/
/***************************** updateGraph ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

void C_planarConnectedC_Graph::updateGraph(linkedList<outEdge>* &hAL,
                                           int countNode, int countEdge,
                                           int* inCluster)
{
    int storeNode = 0;
    clusterData * D = S->pop();

    for (int number = 0; number < actual_N; number++)
        {
            outEdge checkElement;

            if (inCluster[number] == 0)
            // It just has to be looked at the adjacency lists of nodes, which
            // dont belong to the cluster. The other nodes are contained in
            // the repr. graph, so they were handled in function
            // 'constructWheelGraph'.
                {
                    if (D->storeNumbers[number] == -1)
                    // BNode 'number' was not new numbered before.
                        {
                            storeNode = D->storeNumbers[number] = countNode;
                            D->nodeNumbers[countNode++] = number;
                        }
                    else
                    // BNode 'number' was already new numbered, so get the
                    // new number from the array 'storeNumbers'.
                        storeNode = D->storeNumbers[number];
                }

            // Notice: All adjacency lists (in 'updateAdjList') of
            //         cluster-nodes are empty. And each virtual element,
            //         that means that the corresponding edge is a virtual
            //         edge, was completely new numbered in function
            //         'constructWheelGraph'.

            updateAdjList[number].reset();

            while (updateAdjList[number].out(checkElement))
                {
                    if (D->storeNumbers[checkElement.getNode()] == -1 && 
                        D->storeEdges[checkElement.getNumber()] == -1)
                        // BNode and edge have to new numbered.
                        {
                            outEdge element_1(storeNode, countEdge);
                            outEdge element_2(countNode, countEdge);

                            D->storeNumbers[checkElement.getNode()] = countNode;
                            D->storeEdges[checkElement.getNumber()] = countEdge;
                            D->nodeNumbers[countNode] = checkElement.getNode();
                            D->edgeNumbers[countEdge] = checkElement.getNumber();
                            hAL[storeNode].in(element_2);
                            hAL[countNode].in(element_1);
                            countNode++;
                            countEdge++;
                        }

                    if (D->storeEdges[checkElement.getNumber()] == -1)
                        // Only the edge has to be new numbered.
                        {
                            outEdge element_1(storeNode, countEdge);
                            outEdge element_2(D->storeNumbers
                                              [checkElement.getNode()],countEdge);

                            D->storeEdges[checkElement.getNumber()] = countEdge;
                            D->edgeNumbers[countEdge] = checkElement.getNumber();
                            hAL[storeNode].in(element_2);
                            hAL[D->storeNumbers[checkElement.getNode()]].
                                in(element_1);
                            countEdge++;
                        }
                }
        }

    S->push(D);
}





