/*******************************************************************************
+
+  AGD 1.2
+
+
+  extendedPlanGraph.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



/*********************************************************************

  FILENAME  : extendedPlanGraph.cc
  VERSION   : 1.0.0
  AUTHOR    : Christoph Liebel
  LANGUAGE  : C++

*********************************************************************/


#include <LEDA/graph/graph.h>
#include <AGD/impl/cplanarity/extendedPlanGraph.h>


/*****************************************************************************/
/*****************************                   *****************************/
/***************************** extendedPlanGraph *****************************/
/*****************************                   *****************************/
/*****************************************************************************/

extendedPlanGraph::extendedPlanGraph()
  : planGraph()
{}

/*****************************************************************************/
/*****************************                    ****************************/
/***************************** ~extendedPlanGraph ****************************/
/*****************************                    ****************************/
/*****************************************************************************/

extendedPlanGraph::~extendedPlanGraph()
{
    if (adjacencyList)
        {
            if (adjacencyList == embeddedAdjacencyList)
                embeddedAdjacencyList = 0;
            else
                {
                    delete [] embeddedAdjacencyList;
                    embeddedAdjacencyList = 0;
                }

            delete [] adjacencyList;
            adjacencyList = 0;
        }

    if (direction)
        {
            delete [] direction;
            direction = 0;
        }

    if (weight)
        {
            delete [] weight;
            weight = 0;
        }

    if (D)
        {
            delete [] D;
            D = 0;
        }

    if (PQ)
        {
            delete PQ;
            PQ = 0;
        }

    if (stNode)
        {
            delete[] stNode;
            stNode = 0;
        }

    if (marked)
        {
            delete[] marked;
            marked = 0;
        }

    n = m = mSingle = 0;

}

/*****************************************************************************/
/*****************************                   *****************************/
/***************************** completeEmbedding *****************************/
/*****************************                   *****************************/
/*****************************************************************************/

void extendedPlanGraph::completeEmbedding(leafKey<outEdge,int,int>** virtualElements,
                                          int* oTA, int* aTO,
                                          pqQueue<int>** embedVirtual,
                                          int cluster, int numberOfEdges)
{
    /* (1) Reduce the PQ-tree in pairs to all neighboured elements in the
           pqQueue "embedVirtual" (in the last-in-first-out order of the pqQueue).
           The corresponding leafKeys of the PQ-tree leafs have to be determined
           for that.
           
       (2) Execute the reduction- and addition-step for the virtual node
           (that is the node with number 0).

       (3) Work off all indicator pqStacks 'toBeReversed' and 'notToBeReversed'.

       (4) Complete the planar embedding of the graph and delete the
           unnecessary storage place. */

// *************************** part (1) **************************************

    outEdge* firstElement = 0;

    if (embedVirtual[cluster]->queueSize())
        {
            int ok = 1;
            int size = embedVirtual[cluster]->queueSize();

            for (int counter = 1; counter < size; counter++)
                {
                    leafKey<outEdge,int,int>** toReduceElements =
                        new leafKey<outEdge,int,int>*[2];
                    int embElement1 = oTA[embedVirtual[cluster]
                                                            ->deQueue()],
                        embElement2 = oTA[embedVirtual[cluster]
                                                            ->front()];

                    // Delete the edge corresponding to 'embElement1' if the edge
                    // is an additional edge

                    if (aTO[embElement1] < numberOfEdges)
                        embedVirtual[cluster]->enQueue(aTO[embElement1]);

                    toReduceElements[0] = virtualElements[embElement1];
                    toReduceElements[1] = virtualElements[embElement2];

                    if (counter == 1)
                        firstElement = virtualElements[embElement1]->
                            userStructKey();
    
                    if (ok)
                        {
                            ok = PQ->Reduction(2, toReduceElements);
                            PQ->emptyAllPertinentNodes();
                        }

                    delete [] toReduceElements;
                }

            // Places the first element at the end of the pqQueue
            // "embedVirtual[cluster]". Delete the corresponding edge, if the
            // edge is an additional edge.

            int element = embedVirtual[cluster]->deQueue();

            if (element < numberOfEdges)
                embedVirtual[cluster]->enQueue(element);
        }

// *************************** part (2) **************************************

    nodeData<outEdge,int,int>* d = &D[0];

    PQ->Reduction(d->lowerCount, d->lowerNumbered);
    PQ->addition(0, d->higherCount, d->higherNumbered, d->upwardAdjacencyList,
                 d->toBeReversed, d->notToBeReversed);
    PQ->emptyAllPertinentNodes();

    pqStack<int>* help;

    // Check, whether the virtual edges are now in the same order as they
    // are in the pqQueue "embedVirtual" or whether they are in the reversed
    // order. If they are in reversed order, turn over the adjacency list
    // and the indicator pqStacks of the virtual node.
    // Notice: The order of the virtual edges is the same order as in the
    //         pqQueue, if it is the reversed order of the edges in the
    //         adjacency list of the virtual node.

    if (embedVirtual[cluster]->queueSize())
        {
            outEdge checkElement;

            d->upwardAdjacencyList->reset();
            d->upwardAdjacencyList->out(checkElement);

            if ((*firstElement) == checkElement)
                {
                    d->upwardAdjacencyList->reverse();

                    help = d->toBeReversed;
                    d->toBeReversed = d->notToBeReversed;
                    d->notToBeReversed = help;
                }
        }

//    delete PQ;
//    PQ = NULL;

    for (int k = 0; k < d->lowerCount; k++)
        delete d->lowerNumbered[k];

    d->lowerCount = d->higherCount = 0;

// *************************** part (3) **************************************

    int reverse;

    for (int st = stCount - 1; st >= 0; st--)
        {
            d = &D[stNode[st]];

            while (!d->toBeReversed->stackEmpty())
                {
                    reverse = d->toBeReversed->pop();
                    D[reverse].upwardAdjacencyList->reverse();

                    help = D[reverse].toBeReversed;
                    D[reverse].toBeReversed = D[reverse].notToBeReversed;
                    D[reverse].notToBeReversed = help;
                }

            d->notToBeReversed->cleanupStack();
        }

// *************************** part (4) **************************************

    entireEmbed(0);
    embedded = 1;

    if (adjacencyList != embeddedAdjacencyList) {
		delete [] adjacencyList;
		adjacencyList = embeddedAdjacencyList;
	}
}

/*****************************************************************************/
/*****************************                ********************************/
/***************************** connectionTest ********************************/
/*****************************                ********************************/
/*****************************************************************************/

int extendedPlanGraph::connectionTest()
{
	if (n <= 1) return 1;

    if (n == 0 && m == 0)
        {
            return 1;
        }
    else if (n == 0 && m != 0)
        {
            cout << "ERROR (extendedPlanGraph::connectionTest): \n\t"
                 << "It is not a graph!" << endl;
            return 0;
        }
    else if (n != 0 && m == 0)
        {
            cout << "WARNING (extendedPlanGraph::connectionTest): \n\t"
                 << "The graph has nodes but no edges; the graph is "
                 << "not connected!" << endl;
            return 0;
        }
            
  int countReached = 0;
  int* mark = new int[n];
  pqStack<int> reached;

  for (int j = 0; j < n; j++)
    {
      mark[j] = 0;
      adjacencyList[j].reset();
    }

  mark[0] = 1;
  countReached++;
  reached.push(0);

  int head, tail;
  outEdge helpElement;

  while (!reached.stackEmpty())
    {
      tail = reached.pop();
      reached.push(tail);
      if (adjacencyList[tail].out(helpElement))
        {
          head = helpElement.getNode();
          if (mark[head] == 0)
            {
              mark[head] = 1;
              countReached++;
              reached.push(head);
            }
        }
      else
        reached.pop();
    }

  delete [] mark;

  if (countReached == n)
    return 1;
  else
    return 0;

}

/*****************************************************************************/
/*****************************            ************************************/
/***************************** getAdjList ************************************/
/*****************************            ************************************/
/*****************************************************************************/

basicList<outEdge>* extendedPlanGraph::getAdjList() const
{
    return adjacencyList;
}


/*****************************************************************************/
/*****************************                 *******************************/
/***************************** getAdjList(int) *******************************/
/*****************************                 *******************************/
/*****************************************************************************/

basicList<outEdge>* extendedPlanGraph::getAdjList(int number) const
{
    return &adjacencyList[number];
}

/*****************************************************************************/
/*****************************      ******************************************/
/***************************** getM ******************************************/
/*****************************      ******************************************/
/*****************************************************************************/

int extendedPlanGraph::getM() const
{
    return m;
}

/*****************************************************************************/
/*****************************      ******************************************/
/***************************** getN ******************************************/
/*****************************      ******************************************/
/*****************************************************************************/

int extendedPlanGraph::getN() const
{
    return n;
}

/*****************************************************************************/
/*****************************           *************************************/
/***************************** readGraph *************************************/
/*****************************           *************************************/
/*****************************************************************************/

void extendedPlanGraph::readGraph(int nodes, int edges,
                                  basicList<outEdge>* helpAdjList,
                                  int* helpDirect, double* helpWeight,
                                  const char*  helpName)
{
  n = nodes;
  m = edges;
  strcpy(graphName, helpName);

  if (n >= 0)
      {
          adjacencyList = new basicList<outEdge>[n];
          for (int j = 0; j < n; j++)
              adjacencyList[j].append(&helpAdjList[j]);

          // If m <= 2 the graph is already planar and embedded.

          embedded = m < 3;
      }
  else
      cout << "ERROR (extendedPlanGraph::readGraph): \n\tNumber of nodes is "
          << "negative!" << endl;

  if (helpDirect && m)
    {
      direction = new int[m];
      for (int i = 0; i < m; i++)
        direction[i] = helpDirect[i];
    }

  if (helpWeight && m)
    {
      weight = new double[m];
      for (int i = 0; i < m; i++)
        weight[i] = helpWeight[i];
    }

  countMulti();

}



