/*******************************************************************************
+
+  AGD 1.2
+
+
+  IOPoints.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



#include "IOPoints.h"


list_item IOPoints::search_real_forward(node v, list_item it) const
{
	const list<InOutPoint> &L = _out[v];

	while (it && marked(L[it].e))
		it = L.succ(it);

	return it;
}

list_item IOPoints::search_real_backward(node v, list_item it) const
{
	const list<InOutPoint> &L = _out[v];

	while (it && marked(L[it].e))
		it = L.pred(it);

	return it;
}

void IOPoints::restore_deg1_nodes(PlaneGraphCopy &PG)
{
	list<node> deg1s;

	PG.restore_deg1_nodes(deg1s);

	node v;
	forall(v,deg1s) {
		edge e = PG.first_adj_edge(v);
		mark[e] = mark[PG.reversal(e)] = true;
	}
}

edge IOPoints::switch_begin_in(node v)
{
	list<InOutPoint> &Lin  = _in [v];
	list<InOutPoint> &Lout = _out[v];
	list_item it;
	edge e = nil;

	while ((it = Lin.first()) != nil && marked(e = Lin[it].e))
		the_point_of[e] =&(Lout[Lout.push_front(Lin.pop_front())]);

	return it ? e : nil;
}

edge IOPoints::switch_end_in(node v)
{
	list<InOutPoint> &Lin  = _in [v];
	list<InOutPoint> &Lout = _out[v];
	list_item it;
	edge e = nil;

	while ((it = Lin.last()) != nil && marked(e = Lin[it].e))
		the_point_of[e] = &(Lout[Lout.push_back(Lin.pop_back())]);

	return it ? e : nil;
}

void IOPoints::switch_begin_out(node v)
{
	list<InOutPoint> &Lin  = _in [v];
	list<InOutPoint> &Lout = _out[v];

	edge e = Lout[Lout.first()].e;
	the_point_of[e] = &(Lin[Lin.push_front(Lout.pop_front())]);
}

void IOPoints::switch_end_out(node v)
{
	list<InOutPoint> &Lin  = _in [v];
	list<InOutPoint> &Lout = _out[v];

	edge e = Lout[Lout.last()].e;
	the_point_of[e] = &(Lin[Lin.push_back(Lout.pop_back())]);
}

void IOPoints::num_deg1(node v, int &xl, int &xr,
	bool double_count) const
{
	const list<InOutPoint> &L = _out[v];
	list_item it;

	xl = xr = 0;
	for (it = L.first(); it && marked(L[it].e); it = L.succ(it))
		xl++;

	if (double_count || it) // avoid double counting if all are marked
		for (it = L.last(); it && marked(L[it].e); it = L.pred(it))
			xr++;
}

InOutPoint IOPoints::middle_neighbor(node z1) const
{
	const list<InOutPoint> &L = _in[z1];

	list_item it, it_found = nil;
	int       i,  pos = (L.length()-1)/2;

	for (it = L.succ(L.first()), i = 1; i <= pos || !it_found; it = L.succ(it), i++)
		if (!marked(L[it].e))
			it_found = it;

	return L[it_found];
}
