/*******************************************************************************
+
+  AGD 1.2
+
+
+  FlowCompaction.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 709 "Lweb/compaction.lw"
#include <AGD/FlowCompaction.h>
#include <AGD/impl/CompactionPair.h>

// for debug:
//#include <LEDA/graphics/graphwin.h>
//#include <cassert>

using namespace leda;


AGDModule *FlowCompaction::clone() const { 
  FlowCompaction *C = new FlowCompaction(); 
  return C; 
}

int FlowCompaction::call(const OrthogonalRepresentation &H, GridLayout& L) {
  
  string msg; if (!H.check(msg)) error_handler(1, msg + " (H)");

  OrthogonalRepresentation H_prime(H, true);
  
  if (!H_prime.check(msg)) error_handler(1, msg + " (H' after construction)");

  H_prime.dissect(); 

  if (!H_prime.check(msg)) error_handler(1, msg + " (H' after dissect)");

  CompactionPair C(H_prime);
  
  C.compute_coords_properly(CompactionPair::vertical);
  C.compute_coords_properly(CompactionPair::horizontal);

  if (debug_) {
   /*  W_debug.set_graph(H_prime); */
/*     node v; forall_nodes(v, H_prime) { */
/*       W_debug.set_position(v, point(100*C.x_coord(v),  */
/*                                100*C.y_coord(v))); */
/*     } */
/*     W_debug.open(); */
/*     W_debug.fill_window(); */

  }

  node v; forall_nodes(v, H) {
    L.x()[v] = (int) C.x_coord(H_prime.copy(v));
    L.y()[v] = (int) C.y_coord(H_prime.copy(v));
  }

  return L.total_manhattan_edge_length();
  
}



double FlowCompaction::call(GridLayout& L) {
  
  double l_old, l_new; 
  node v;

  do {
    
    l_old = L.total_manhattan_edge_length();

    CompactionPair X(L);
    X.compute_coords_properly(CompactionPair::vertical);
    forall_nodes(v, L.get_graph()) L.x()[v] = (int) X.x_coord(v);
       
    CompactionPair Y(L);
    Y.compute_coords_properly(CompactionPair::horizontal);
    forall_nodes(v, L.get_graph()) L.y()[v] = (int) Y.y_coord(v);

    l_new = L.total_manhattan_edge_length();
    
    //assert(l_old - l_new >= 0.0);
	if (l_old - l_new < 0.0)
		error_handler(1,"FlowCompaction::call: l_old - l_new < 0.0");

  } while (l_old - l_new > 0.0);
  
  return l_new;  
}


