/*******************************************************************************
+
+  AGD 1.2
+
+
+  OrthogonalLayout.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 830 "Lweb/OrthoLayout.lw"
#include <AGD/OrthogonalLayout.h>

using namespace leda;

OrthogonalLayout::OrthogonalLayout() : Ovid() { 

  _allowBends = true;
  _containsCages = true;
  _blowUp = false;

  add_post_rule(key::orthogonal);

}

AGDModule* OrthogonalLayout::clone() const { 

  OrthogonalLayout *O = new OrthogonalLayout;
    
  O->variant(variant());
  O->output_level(output_level());
  O->traditional(traditional());
  O->progressive(progressive());
  O->verify(verify());

  O->set_compactor(compactor.get());

  return O; 
 
}


//#line 863 "Lweb/OrthoLayout.lw"
void OrthogonalLayout::preprocessing(OrthogonalRepresentation &H) {

  if (!Is_Connected(H)) { 
//#line 582 "Lweb/OrthoLayout.lw"
if (_outputLevel > 3) INF("making the input graph temporarily connected...");
edge e_ref = nil;
if (H.number_of_edges() > 0) e_ref = H.first_face_edge(H.get_external_face());
Make_Connected(H);
list<edge> R; H.make_map(R); H.compute_faces();
if (e_ref) H.set_external_face(H.face_of(e_ref));
else H.set_external_face(H.first_face());

//#line 865 "Lweb/OrthoLayout.lw"
                                                             } 
  
//#line 596 "Lweb/OrthoLayout.lw"
  
INF("transforming highdeg vertices into cages... ");

node v; forall_nodes(v, H) if (outdeg(v) > 4) H.expand(v);
  
DONE();

//#line 866 "Lweb/OrthoLayout.lw"
                                               ;

}

void OrthogonalLayout::postprocessing(OrthogonalRepresentation &H,
                                      GridLayout &L) {

  INF("putting highdeg vertices in their cages... "); 

  face f; list<face> all_faces = H.all_faces(); // workaround
  forall(f, all_faces) if (H.is_cage(f)) {
  
    int xmin = MAXINT, ymin = MAXINT, xmax = -MAXINT, ymax = -MAXINT;
  
    edge e; forall_face_edges(e, f) { 
      node c = source(e); 
      if (L.x()[c] < xmin) xmin = L.x()[c]; 
      if (L.x()[c] > xmax) xmax = L.x()[c]; 
      if (L.y()[c] < ymin) ymin = L.y()[c]; 
      if (L.y()[c] > ymax) ymax = L.y()[c]; 
    }
  
    node v = H.collapse(f);

    L.x()[v] = xmin; L.y()[v] = ymin;
    L.width(v, xmax - xmin);
    L.height(v, ymax - ymin);
    
  }

  DONE();
  
}

