/*******************************************************************************
+
+  AGD 1.2
+
+
+  Ovid.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 466 "Lweb/Ovid.lw"
#include <AGD/Ovid.h>
#include <AGD/FlowCompaction.h>

#include <LEDA/graph/face_array.h>
#include <LEDA/graph/graph_alg.h>


using namespace leda;

Ovid::Ovid() {
  
  _allowBends = true;
  _containsCages = false;
  _blowUp = false;
  
  k_ = 2;

  _allowNonPosDeg = _allowNegDeg =
  _forbidSmallZeroDeg = 
  _verify = true;

  _variant = false;
  _verify = true; 

  _traditional = false; _progressive = false; _compactLayout = true;
  debug_ = false;

  _outputLevel = 0;
  n_internal_ = m_internal_ = l_internal_ = 0;
  t_compact_ = 0.0f;

  add_precondition(key::planar); 
  add_post_rule(key::no_crossings); 

  PreCond PRE;  
  PRE[1] << key::planar << key::connected; // xxx: connected?

  compactor.init(PRE, PostCond(), FlowCompaction());

}


void Ovid::do_call(OrthoRepGridConcept& C) {
  edge_map<int> max_b(C.get_graph().original(), 1000);
  if (!do_call(C, max_b)) error_handler(1, "Ovid::do_call. No feasible "
                                        "flow.");
}

bool Ovid::do_call(OrthoRepGridConcept& C, 
                   const edge_map<int>& max_b) {

  OrthogonalRepresentation &H = C.get_representation();

  if (H.number_of_nodes() <= 1) return true;  
  if (_outputLevel > 0) LINE();

  H.output_level(_outputLevel);
  
  preprocessing(H);
  n_internal_ = H.number_of_nodes();
  m_internal_ = H.number_of_edges();

  edge_map<int> curr_max_b(H, 1000);
  edge e; forall_edges(e, H) if (H.original(e))
    curr_max_b[e] = max_b[H.original(e)];

  if (!compute_optimal_shape(H, curr_max_b)) return false; 

  if (_outputLevel > 3) cout << H << endl;
  string msg; if (_verify && !H.check(msg)) error_handler(0, msg);

  H.normalize(); H.orientate();
  
  if (_blowUp) { 
//#line 878 "Lweb/Ovid.lw"
INF("blowing up the cages... ");

face f; forall_faces(f, H) if (H.is_cage(f)) { 
  int *dim = new int[4]; dim[0] = dim[1] = dim[2] = dim[3] = 0; 
  edge e; forall_face_edges(e, f) ++dim[(int)H.dir(e)];
  if (leda::max(dim[0], dim[2]) == 1) { 
//#line 897 "Lweb/Ovid.lw"
forall_face_edges(e, f) if (H.hor(e)) break; 
H.split(e);


//#line 883 "Lweb/Ovid.lw"
                                                                          }; 
  if (leda::max(dim[1], dim[3]) == 1) { 
//#line 892 "Lweb/Ovid.lw"
forall_face_edges(e, f) if (H.ver(e)) break; 
H.split(e);

//#line 884 "Lweb/Ovid.lw"
                                                                        }; 
  delete[] dim; 
}

DONE();

//#line 537 "Lweb/Ovid.lw"
                                                  }

  if (_outputLevel > 3)  cout << "before compaction:\n" << H << endl;
  
  if (_verify && !H.check(msg)) error_handler(1, msg);

  GridLayout &L = C.get_layout(); // obacht: spaeter wieder gridlayout
  // fr H machen, dann nach der kompaktierung umsetzen (s. unten)
  // oder: C.get_layout() nehmen, dann aber grosse knoten mitgeben...
  //GridLayout L(H);
//  list<Label> Lambda = C.get_labels();

  float T = used_time(); 

  compactor.get().debug(debug_); // pass on the debugging property
  //l_internal_ = compactor.get().call(H, L, Lambda);
  l_internal_ = compactor.get().call(H, L);
  t_compact_ = used_time(T);

  if (_compactLayout) compactor.get().call(L);

  // hier von L auf C.get_layout() uebertragen
  // obacht: fehlt noch.
  // pruefen: war C.get_layout() wirklich fuer G und nicht fuer H?
  
  postprocessing(H, L);

  return true;
}




//#line 577 "Lweb/Ovid.lw"
bool Ovid::compute_optimal_shape(OrthogonalRepresentation& H, 
                                 const edge_map<int>& max_b) {

  
//#line 595 "Lweb/Ovid.lw"
INF("constructing the network... ");
graph N;

//#line 612 "Lweb/Ovid.lw"
face f_0 = H.get_external_face();       

node_map<node> V2U(H), U2V(N);
node_map<n_type> node_type(N);
node v, i, j; face f; edge e, a;

forall_edges(e, H) H.set_angle(e, 0);

forall_nodes(v, H) {
  i = N.new_node();
  V2U[v] = i; U2V[i] = v;
  node_type[i] = ((outdeg(v) > 4) ? high : low);
}

face_array<node> F2U(H); node_map<face> U2F(N);

forall_faces(f, H) {
  j = N.new_node();
  F2U[f] = j; U2F[j] = f;
  node_type[j] = ((f == f_0) ? outer : inner);
}

//#line 597 "Lweb/Ovid.lw"
                             ;

//#line 642 "Lweb/Ovid.lw"
node_array<int> b(N, 0);

if (_traditional) {
  
  forall_nodes(i, N) switch (node_type[i]) {
  case high: 
  case low:   b[i] = pow2(k_) - outdeg(U2V[i]); break;
  case inner: b[i] = pow2(k_) - H.adj_edges(U2F[i]).size(); break;
  case outer: b[i] = pow2(k_) - H.adj_edges(U2F[i]).size() - pow2(k_+1); break;
  default: error_handler(1, "unknown node type in network");
  }
  
} else {
  
  forall_nodes(i, N) switch (node_type[i]) {
  case high:
  case low:  b[i] = 2*(outdeg(U2V[i]) - 2); break;
  case inner: b[i] = -4; break;
  case outer: b[i] = +4; break;
  default: error_handler(1, "unknown node type in network");
  }
}

if (_verify) {
  int checksum = 0; forall_nodes(i, N) checksum += b[i];
  if (checksum != 0) error_handler(1, "supply does not equal demand");
}

//#line 598 "Lweb/Ovid.lw"
                         ;

//#line 678 "Lweb/Ovid.lw"
edge_map<a_type> arc_type(N);
edge_map<edge>   e_corr(N);

edge_map<bool> allow_concave_bend(H, true);
edge e_G; 
forall(e_G, _noConcaveBends) forall(e, H.chain(e_G)) 
  allow_concave_bend[e] = false;
forall_faces(f, H) if (H.is_cage(f)) forall_face_edges(e, f)
  allow_concave_bend[e] = false;
     
if (_allowBends) { 
//#line 696 "Lweb/Ovid.lw"
forall_faces(f, H) forall_face_edges(e, f) {
  if (allow_concave_bend[e]) {
    face g = H.face_of(H.reversal(e));            
    a = N.new_edge(F2U[f], F2U[g]);
    arc_type[a] = (H.is_cage(f) ? cage_bend : bend); 
    e_corr[a] = e;
  }
}

//#line 688 "Lweb/Ovid.lw"
                                        }

//#line 709 "Lweb/Ovid.lw"
forall_edges(e, H) {
  v = target(e); f = H.face_of(e);
  a = N.new_edge(V2U[v], F2U[f]);
  arc_type[a] = (H.is_cage(f) ? cage_angle : angle); 
  e_corr[a] = e;
}

//#line 690 "Lweb/Ovid.lw"
if (_allowNonPosDeg) { 
//#line 720 "Lweb/Ovid.lw"
forall_edges(e, H) {
  v = target(e); f = H.face_of(e);
  if (_traditional) {
    if (outdeg(v) > 4 || !_forbidSmallZeroDeg) {
      a = N.new_edge(F2U[f], V2U[v]);
      arc_type[a] = neg_angle; e_corr[a] = e;
    }
  } else {
    a = N.new_edge(F2U[f], V2U[v]);
    arc_type[a] = neg_angle; e_corr[a] = e;
  }
}

//#line 690 "Lweb/Ovid.lw"
                                                      }

//#line 599 "Lweb/Ovid.lw"
                            ;

//#line 739 "Lweb/Ovid.lw"
edge_array<int> u(N), l(N, 0), c(N);

const int u_max = pow2(k_) * H.number_of_nodes(),
  u_neg_angle = (_allowNegDeg ? pow2(k_) : 1);

if (_traditional) {
  
  forall_edges(a, N) switch(arc_type[a]) {
  case angle: u[a] = pow2(k_) - 1; c[a] = 0; break;
  case cage_angle: u[a] = 1; c[a] = 0; if (_variant) l[a] = 1; break;
  case neg_angle: u[a] = u_neg_angle; c[a] = 1; break;
  case bend: u[a] = leda::min(max_b[e_corr[a]], u_max); c[a] = 1; break;
  case cage_bend: u[a] = 0; c[a] = 100; break;
  default: error_handler(1, "Ovid: bad type");
  }
  
} else {

  int angle_cost, neg_angle_cost, bend_cost;
  if (_progressive) {  
    angle_cost = neg_angle_cost = 1; 
    bend_cost = 3 * H.number_of_nodes();
  } else {
    angle_cost = neg_angle_cost = 0; bend_cost = 1; 
  }
  
  
  forall_edges (a, N) switch(arc_type[a]) {
  case angle: u[a] = 1; c[a] = angle_cost; break;
  case neg_angle: u[a] = 2; c[a] = neg_angle_cost; break;
  case bend: u[a] = 1000; c[a] = bend_cost; break;
  case cage_bend: u[a] = 0; c[a] = 0; break;
  case cage_angle: u[a] = 1; c[a] = 0; if (_variant) l[a] = 1; break;
  default: error_handler(1, "Ovid::compute_optimal_shape. Unknown angle type");
  
  }
}

//#line 600 "Lweb/Ovid.lw"
                                               ;
DONE();

//#line 580 "Lweb/Ovid.lw"
                           ;
  
//#line 783 "Lweb/Ovid.lw"
INF("computing the minimal cost flow... ");

edge_array<int> chi(N, 0); 
bool feasible = MIN_COST_FLOW(N, l, u, c, b, chi);

int cost = 0;
forall_edges(a, N)  cost += chi[a] * c[a];

if (_outputLevel > 0) cout << string("c(x) = %d ", cost);

DONE();


//#line 581 "Lweb/Ovid.lw"
                                   ;
  
//#line 803 "Lweb/Ovid.lw"
INF("deriving the orthogonal representation... ");

if (_traditional) {

  int factor = pow2(2 - k_);
  
  forall_edges(a, N) {
    e = e_corr[a];
    switch (arc_type[a]) {
    case neg_angle: 
      H.set_angle(e, H.get_angle(e) - chi[a] * 90 * factor); 
      break;
    case cage_angle:
    case angle:     
      H.set_angle(e, H.get_angle(e) + (chi[a] + 1) * 90 * factor);
      break;
    case bend:      
    case cage_bend: 
      { 
//#line 860 "Lweb/Ovid.lw"
if (_traditional) {
  if (chi[a] > 0) {
    H.set_string(e, '0', chi[a]);
    H.set_string(H.reversal(e), '1', chi[a]);
  }
} else {
  if (chi[a] > 0) {
    H.set_string(e, '1', chi[a]);
    H.set_string(H.reversal(e), '0', chi[a]);
  }
}
  
//#line 821 "Lweb/Ovid.lw"
                               }; 
      break;
    default: error_handler(1, "fatal error 3");
    }
  }
} else {
 
  forall_edges(e, H) H.set_angle(e, 180);

  forall_edges(a, N) { 
    e = e_corr[a];
    switch (arc_type[a]) {
    case cage_angle:
    case angle: H.set_angle(e, H.get_angle(e) - chi[a] * 90); break;
    case neg_angle: H.set_angle(e, H.get_angle(e) + chi[a] * 90); break;
    case cage_bend:
    case bend: { 
//#line 860 "Lweb/Ovid.lw"
if (_traditional) {
  if (chi[a] > 0) {
    H.set_string(e, '0', chi[a]);
    H.set_string(H.reversal(e), '1', chi[a]);
  }
} else {
  if (chi[a] > 0) {
    H.set_string(e, '1', chi[a]);
    H.set_string(H.reversal(e), '0', chi[a]);
  }
}
  
//#line 837 "Lweb/Ovid.lw"
                                        }; break;
    default: error_handler(1, "Ovid:compute_optimal_shape. "
                           "Unknown angle type");
    }
  } 

}



DONE();

//#line 582 "Lweb/Ovid.lw"
                               ;

  return feasible;

}

