/*******************************************************************************
+
+  AGD 1.2
+
+
+  basicGraph.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



/************************************************************************

         Filename   :   basicGraph.cc
         Version    :   1.0.1
         Author     :   Christoph Buchheim
         Language   :   C++

************************************************************************/

#if defined(__BORLANDC__)
// turn off some warnings
#pragma option -w-inl -w-ccc -w-aus -w-eff -w-par -w-rch -w-pia
#endif


#ifdef NOAGD

#include "basicGraph.h"
#include "pqQueue.h"

#else

#include <AGD/impl/planarity/basicGraph.h>
#include <AGD/impl/PQTree/pqQueue.h>

#endif

#include <LEDA/system/basic.h>
#include <LEDA/internal/std/string.h>


basicGraph::basicGraph():n(0),m(0),mSingle(0),adjacencyList(0),
                         direction(0),weight(0) {
  strcpy(graphName,"not initialized");
}

basicGraph::~basicGraph() {
  deleteGraph();
}

int basicGraph::countMulti() {
  int multiCount=0;
  int loopCount=0;
  int* neighbour=new int[n];
  int nodeNumber;
  for(nodeNumber=0;nodeNumber<n;nodeNumber++)
    neighbour[nodeNumber]=-1;
  outEdge nextEdge;
  for(nodeNumber=0;nodeNumber<n;nodeNumber++) {
    adjacencyList[nodeNumber].reset();
    while(adjacencyList[nodeNumber].out(nextEdge)) {
      if(nextEdge.getNode()==nodeNumber) loopCount++;
      else if(neighbour[nextEdge.getNode()]==nodeNumber) multiCount++;
      else neighbour[nextEdge.getNode()]=nodeNumber;
    }
  }
  mSingle=m-loopCount-multiCount/2;
  delete[] neighbour;
  return loopCount;
}

void basicGraph::deleteGraph() {
  if(adjacencyList) {
    delete[] adjacencyList;
    if(direction) delete[] direction;
    if(weight) delete[] weight;
    adjacencyList=0;
    direction=0;
    weight=0;
    n=m=mSingle=0;
    strcpy(graphName,"deleted");
  }
}

#ifdef NOAGD
int basicGraph::readGraph(char* filename) {
  deleteGraph();
  ifstream graphIn(filename,ios::in);
  if(!graphIn) {
    cerr<<"Cannot build graph: File not found."<<endl;
    return 0;
  }
  pqQueue<int> directionQueue;
  pqQueue<double> weightQueue;
  stringstream lineIn;
  char line[128],word[128];
  long checkSum=0,check;
  int target,source,edgeNumber=0;
  double wgh;
  int directed=0,weighted=0;
  int beginF=0,graphF=0,endF=0,checksumF=0;
  imap_ord('1');
  while(!endF&&graphIn.getline(line,128)) {
    lineIn<<line;
    strcat(line,"\n");
    if(line[0]=='*') {
      lineIn>>word;
      if(!strcmp(word,"*BEGIN")) {
        
#ifdef NOAGD
if(beginF) cerr<<"Warning: *BEGIN already found."<<endl;
else {
  beginF=1;
  checkSum=new_checksum(line,checkSum);
  lineIn.getline(word,128);
  strcpy(graphName,word+1);
}
#endif

      }
      else if(!strcmp(word,"*END")) {
        
#ifdef NOAGD
endF=1;
if(!beginF) cerr<<"Warning: *END without *BEGIN."<<endl;
if(!graphF) cerr<<"Warning: Graph is empty."<<endl;
if(!checksumF) cerr<<"Warning: No checksum found."<<endl;
lineIn.getline(word,128);
if(strcmp(graphName,word+1)&&beginF)
  cerr<<"Warning: *END with wrong name of graph."<<endl;
#endif

      }
      else if(!strcmp(word,"*GRAPH")) {
        
#ifdef NOAGD
if(graphF) cerr<<"Warning: *GRAPH already found."<<endl;
else {
  graphF=1;
  if(!beginF) cerr<<"Warning: *GRAPH without *BEGIN."<<endl;
  checkSum=new_checksum(line,checkSum);
  lineIn>>n>>m;
  if(n) {
    adjacencyList=new basicList<outEdge>[n];
    if(!adjacencyList) {
      cerr<<"Cannot build graph: Could not allocate memory."<<endl;
      return 0;
    }
  }
  while(lineIn>>word) {
    if(!strcmp(word,"DIRECTED")) directed=1;
    else if(!strcmp(word,"WEIGHTED")) weighted=1;
    else if(strcmp(word,"UNDIRECTED")&&strcmp(word,"UNWEIGHTED")) {
      cerr<<"Warning: Do no understand \""<<word;
      cerr<<"\" in *GRAPH line"<<endl;
    }
  }
}
#endif

      }
      else if(!strcmp(word,"*CHECKSUM")) {
        
#ifdef NOAGD
checksumF=1;
lineIn>>check;
if(check==-1) cerr<<"Warning: Checksum ignored."<<endl;
else if(checkSum!=check) cerr<<"Warning: Checksum wrong."<<endl;
while(lineIn>>word) {
  cerr<<"Warning: Do not understand \""<<word;
  cerr<<"\" in *CHECKSUM line."<<endl;
}
#endif

      }
      else {
        if(strcmp(word,"*"))
          cerr<<"Warning: Do not understand line: "<<line;
        while(lineIn>>word);
      } 
    }
    else {
      
#ifdef NOAGD
if(!graphF) cerr<<"Warning: *GRAPH not found yet."<<endl;
else {
  checkSum=new_checksum(line,checkSum);
  lineIn>>source>>target;
  if(target<1||target>n||source<1||source>n) {
    cerr<<"Warning: Vertex of edge ";
    cerr<<target<<"-"<<source<<" doesn't exist."<<endl;
  }
  else {
    target--; source--;
    adjacencyList[target].in(outEdge(source,edgeNumber));
    if(target!=source) adjacencyList[source].in(outEdge(target,edgeNumber));
    if(directed) directionQueue.enQueue(target);
    if(weighted) {
      lineIn>>wgh; 
      weightQueue.enQueue(wgh);
    }
    edgeNumber++;
  }
  while(lineIn>>word) {
    cerr<<"Warning: Do not understand \""<<word;
    cerr<<"\" in edge line."<<endl;
  }
}
#endif

    }
    lineIn.clear();
  }
  if(!graphIn) cerr<<"Warning: Error reading from file."<<endl;
  if(!beginF) {
    cerr<<"Warning: *BEGIN not found."<<endl;
    strcpy(graphName,"");
  }
  if(!endF) cerr<<"Warning: *END not found."<<endl;
  if(!graphF) cerr<<"Warning: *GRAPH not found."<<endl;
  if(!checksumF) cerr<<"Warning: *CHECKSUM not found."<<endl;
  if(edgeNumber!=m) {
    cerr<<"Warning: Wrong number of edges."<<endl;
    m=edgeNumber;
  }
  if(directed) {
    direction=new int[m];
    for(int edgeCount=0;edgeCount<m;edgeCount++)
      direction[edgeCount]=directionQueue.deQueue();
  }
  if(weighted) {
    weight=new double[m];
    for(int edgeCount=0;edgeCount<m;edgeCount++)
      weight[edgeCount]=weightQueue.deQueue();
  }
  countMulti();
  return 1;
}
#endif

#ifdef NOAGD
void basicGraph::writeGraph(int format,char* filename) {
  ostream* fout;
  fstream fileOut;
  if(filename) {
    fileOut.open(filename,ios::out|ios::trunc);
    fout=&fileOut;
  }
  else fout=&cout;
  stringstream lineOut;
  int* marked=new int[m];
  outEdge neighbour;
  long checkSum=0;
  char line[128];
  for(int number=0;number<m;number++) marked[number]=0;
  switch(format) {
    case 0: (*fout)<<n<<" "<<m<<endl; 
            break;
    case 1: (*fout)<<"Name of graph: "<<graphName<<endl;
            break;
    case 2: lineOut<<"*BEGIN "<<graphName<<endl
                   <<"*GRAPH "<<n<<" "<<m;
            if(direction) lineOut<<" DIRECTED";
            else lineOut<<" UNDIRECTED";
            if(weight) lineOut<<" WEIGHTED";
            else lineOut<<" UNWEIGHTED";
            imap_ord('1');
            while(lineOut.getline(line,128)) {
              (*fout)<<line<<endl;
              strcat(line,"\n");
              checkSum=new_checksum(line,checkSum);
            }
            lineOut.clear();
  }
  for(int nodeNumber=0;nodeNumber<n;nodeNumber++) {
    if(format==1) (*fout)<<nodeNumber+1<<"-";
    adjacencyList[nodeNumber].reset();
    while(adjacencyList[nodeNumber].out(neighbour)) 
    switch(format) {
      case 0: if(direction&&direction[neighbour.getNumber()]!=-1
                 &&direction[neighbour.getNumber()]!=neighbour.getNode())
                (*fout)<<"("<<neighbour.getNode()+1<<") ";
              else (*fout)<<neighbour.getNode()+1<<" ";
              (*fout)<<neighbour.getNumber()+1<<" ";
              break;
      case 1: if(direction&&direction[neighbour.getNumber()]!=-1
                 &&direction[neighbour.getNumber()]!=neighbour.getNode())
                (*fout)<<"("<<neighbour.getNode()+1<<") ";
              else (*fout)<<neighbour.getNode()+1<<" ";
              break;
      case 2: if(!direction&&!marked[neighbour.getNumber()]||direction
                 &&(direction[neighbour.getNumber()]==-1
                    ||direction[neighbour.getNumber()]==neighbour.getNode())) {
                marked[neighbour.getNumber()]=1;
                lineOut<<nodeNumber+1<<" "
                       <<neighbour.getNode()+1;
                if(weight) lineOut<<" "<<weight[neighbour.getNumber()];
                lineOut<<endl;
                lineOut.getline(line,128);
                (*fout)<<line<<endl;
                strcat(line,"\n");
                checkSum=new_checksum(line,checkSum);
              }
    }
    if(format==0) (*fout)<<"-1"<<endl;
    if(format==1) (*fout)<<endl;
  }
  if(format==2) {
    (*fout)<<"*CHECKSUM "<<checkSum<<endl;
    (*fout)<<"*END "<<graphName<<endl;
  }
  delete[] marked;
}
#endif


