#include <LEDA/d3_gl_scene.h>

LEDA_BEGIN_NAMESPACE 


d3_gl_scene::~d3_gl_scene()
{		
	tex_face_list.clear();
	tex_info_list.clear();

	clear_lists();
	
	delete Hp;
	delete pos1;	
	
	while(!seq_mat.empty())
		delete[] seq_mat.pop();

	while(!sequences.empty()){
		transformation_list& transforms = sequences[sequences.first()];
		while(!transforms.empty())
			delete[] transforms.pop().second();
		sequences.erase(sequences.first());
	}		
}

void d3_gl_scene::init()
{
  draw_nodes  = true;
  draw_edges  = true;
  draw_faces  = true; 
  visible     = true;  
  draw_obj_border = false;
  texturing  = false;
  node_width = 2.0f;  
  edge_width = 2.0f;       
  
  clear_matrix();       
  
  sequences.clear();
  seq_mat.clear();    
  
  d3_gl_win = NULL;   
  
  border_line_width = 5.0f;
  border_line_color.set_color(green);
  
  set_color(red,Hp->all_nodes(),false);	
  set_color(black,Hp->all_edges(),false);	
  set_color(yellow,Hp->all_faces(),false);
}

void d3_gl_scene::clear_lists()
{
  col_node_list.clear();
  col_edge_list.clear();
  col_face_list.clear();
  tex_face_list.clear();
}

void d3_gl_scene::set_graph(graph& G)
{
  Hp = &G;
  clear_lists();
}

void d3_gl_scene::set_node_array(const node_array<vector>& pos)
{
  clear_lists();
	
  if(pos1 != NULL) delete pos1;
  pos1 = new node_array<vector>(pos);  

  vector cg(0,0,0);
  node v;

  forall_nodes(v,*Hp) cg += (*pos1)[v];
  g_center = cg/Hp->number_of_nodes(); 
  
  comp_coord_extrema(xmin,xmax,ymin,ymax,zmin,zmax);  
  set_border();  
}

void d3_gl_scene::set_node_array(const node_array<rat_vector>& pos)
{
  col_node_list.clear();
  col_edge_list.clear();
  col_face_list.clear();
  tex_face_list.clear();

  if(pos1 != NULL) delete pos1;
  pos1 = new node_array<vector>(*Hp,vector(3));
  
  vector cg(0,0,0);
  node v;

  forall_nodes(v,*Hp) 
  { (*pos1)[v] = pos[v].to_vector();
    cg += (*pos1)[v];
   }
  
  g_center = cg/Hp->number_of_nodes();  
  comp_coord_extrema(xmin,xmax,ymin,ymax,zmin,zmax);
  set_border();  
}

void d3_gl_scene::set_border()
{
	border[0] = vector(xmin-1.0, ymin-1.0, zmin-1.0);
	border[1] = vector(xmin-1.0, ymin-1.0, zmax+1.0);
	border[2] = vector(xmin-1.0, ymax+1.0, zmax+1.0);
	border[3] = vector(xmin-1.0, ymax+1.0, zmin-1.0);
	border[4] = vector(xmax+1.0, ymin-1.0, zmin-1.0);
	border[5] = vector(xmax+1.0, ymin-1.0, zmax+1.0);
	border[6] = vector(xmax+1.0, ymax+1.0, zmax+1.0);
	border[7] = vector(xmax+1.0, ymax+1.0, zmin-1.0);
}

void d3_gl_scene::comp_coord_extrema(double& x0, double& x1, double& y0 ,
									 double& y1, double& z0, double& z1)
{
	x0 = 0.0;	x1 = 0.0;
	y0 = 0.0;	y1 = 0.0;
	z0 = 0.0;	z1 = 0.0;  

	vector vec;
	node v;

	forall_nodes(v,*Hp){
		vec = (*pos1)[v];
		
		if(vec[0]>x1)
			x1 = vec[0];
		if(vec[0]<x0)
		    x0 = vec[0];

		if(vec[1]>y1)
			y1 = vec[1];
		if(vec[1]<y0)
			y0 = vec[1];

		if(vec[2]>z1)
			z1 = vec[2];
		if(vec[2]<z0)
			z0 = vec[2];
	}
}

void d3_gl_scene::comp_optimal_coord(double& x0, double& x1, double& y0 ,
									 double& y1, double& z0, double& z1)
{
	comp_coord_extrema(x0,x1,y0,y1,z0,z1);
	double help;

	help = x0;
	x0 = (x0 - 2.0f) - ((x1-x0)*0.5f);
	x1 = (x1 + 2.0f) + ((x1-help)*0.5f);
	help = y0;
	y0 = (y0 - 2.0f) - ((y1-y0)*0.5f);
	y1 = (y1 + 2.0f) + ((y1-help)*0.5f);
	help = z0;
	z0 = (z0 - 2.0f) - ((z1-z0)*0.5f);
	z1 = (z1 + 2.0f) + ((z1-help)*0.5f);		
}


void d3_gl_scene::print_ppl(point_pointer_list& ppl)
{
	point_pointer pp;
	forall(pp,ppl)
		cout<<pp<<" ";
}

bool d3_gl_scene::cmp_two_ppl(const point_pointer_list& ppl1,
							  const point_pointer_list& ppl2, bool print)
{
	if(ppl1.length() != ppl2.length()) return false;

	list_item it1 = ppl1.first();
	list_item it2 = ppl2.first();

	while(it1 != nil && it2 != nil){
		if(ppl1[it1] != ppl2[it2]){
			if(print)
				cout<<"ERROR : "<<ppl1[it1]<<"  "<<ppl2[it2];
			return false;
		}
		it1 = ppl1.succ(it1);
		it2 = ppl2.succ(it2);
	}

	return true;
}

void d3_gl_scene::nodes_to_points(const list<node>& n_list, point_pointer_list& p_list)
{
	node n;
	forall(n,n_list){
		vector& v = (*pos1)[n];
		p_list.append(&v);
	}
}

void d3_gl_scene::edges_to_points(const list<edge>& e_list, pp_tuple_list& L)
{	
	edge e;
	forall(e,e_list){
		node n1 = Hp->source(e);
		node n2 = Hp->target(e);
		vector& v1 = (*pos1)[n1];
		vector& v2 = (*pos1)[n2];
		L.append(pp_tuple(&v1,&v2));		
	}	
}

void d3_gl_scene::faces_to_points(const list<face>& f_list, ppl_list& L)
{
	face f;
	forall(f,f_list){
		point_pointer_list temp_list;
		edge first = Hp->first_face_edge(f);
		edge e = first;

		do{
			vector& v1 = (*pos1)[source(e)];			
			temp_list.append(&v1);			
			e = Hp->face_cycle_succ(e);
		}while(e!=first);

		L.append(temp_list);
	}	
}


void d3_gl_scene::set_color(const d3_gl_col& c,
			   const list<edge>& new_list, const bool& duplicate_test)
{
	if(Hp->number_of_edges() == 0) return;

	pp_tuple_list L;
	edges_to_points(new_list,L);
	pp_tuple new_tuple;
		
	bool appended = false;	
	
	list_item it = col_edge_list.first();

	while(it != nil){
		col_edge_tuple& c_tuple = col_edge_list[it];
		pp_tuple_list& PTS = c_tuple.second();				

		if(c == c_tuple.first()){
			
			forall(new_tuple,L){
				if(duplicate_test){

					list_item it2 = PTS.first();
					while(it2 != nil){
						pp_tuple& old_tuple = PTS[it2];
						if(old_tuple == new_tuple || // if(edge OR inverse edge) do not insert
							(old_tuple.first() == new_tuple.second() && old_tuple.second() == new_tuple.first()))
						{ break; }

						it2 = PTS.succ(it2);
					}
					if(it2 == nil)
						PTS.append(new_tuple);									

				}
				else PTS.append(new_tuple);
			}
			appended = true;
			
		}
		else{
			forall(new_tuple,L){
				list_item it2 = PTS.first();
				while(it2 != nil){
					pp_tuple& old_tuple = PTS[it2];
					if(old_tuple == new_tuple || // remove edge OR inverse edge
						(old_tuple.first() == new_tuple.second() && old_tuple.second() == new_tuple.first()))
					{
						list_item temp = PTS.succ(it2);
						PTS.del_item(temp);
						it2 = temp;
					}

					it2 = PTS.succ(it2);
				}
			}			
		}

		if(PTS.empty()){			
			list_item it2 = col_edge_list.succ(it);
			col_edge_list.del_item(it);
			it = it2;
		}
		else
			it = col_edge_list.succ(it);
	}	

	if(!appended){ 
		col_edge_list.append(col_edge_tuple(c,L));
	}
}

void d3_gl_scene::set_color(const d3_gl_col& c,
			   const list<node>& new_list, const bool& duplicate_test)
{
	if(Hp->number_of_nodes() == 0) return;
	
	point_pointer_list pp_list;
	nodes_to_points(new_list,pp_list);
		
	bool appended = false;		

	list_item it = col_node_list.first();

	while(it != nil){
		col_node_tuple& c_tuple = col_node_list[it];
		point_pointer_list& PTS = c_tuple.second();
		point_pointer pp;

		if(c == c_tuple.first()){							
			forall(pp,pp_list){
				if(duplicate_test){
					if(!PTS.rank(pp))
						PTS.append(pp);				
				}
				else PTS.append(pp);
			}
			appended = true;													
		}
		else{ 
			forall(pp,pp_list){
				PTS.remove(pp);												
			}		
		}

		if(PTS.empty()){			
			list_item it2 = col_node_list.succ(it);
			col_node_list.del_item(it);
			it = it2;
		}
		else
			it = col_node_list.succ(it);
	}	

	if(!appended){ 
		col_node_list.append(col_node_tuple(c,pp_list));
	}		
}

void d3_gl_scene::set_color(const d3_gl_col& c,
			   const list<face>& new_list, const bool& duplicate_test)
{
	if(Hp->number_of_faces() == 0) return;

	ppl_list L;
	faces_to_points(new_list,L);
		
	bool appended = false;		

	list_item it = col_face_list.first();

	while(it != nil){
		col_face_tuple& c_tuple = col_face_list[it];
		ppl_list& PTS = c_tuple.second();
		point_pointer_list elem;

		if(c == c_tuple.first()){							
			forall(elem,L){
				if(duplicate_test){
					if(!PTS.rank(elem))
						PTS.append(elem);				
				}
				else PTS.append(elem);
			}
			appended = true;													
		}
		else{ 
			forall(elem,L){
				PTS.remove(elem);												
			}		
		}

		if(PTS.empty()){			
			list_item it2 = col_face_list.succ(it);
			col_face_list.del_item(it);
			it = it2;
		}
		else
			it = col_face_list.succ(it);
	}	

	if(!appended){ 
		col_face_list.append(col_face_tuple(c,L));
	}	
}

void d3_gl_scene::set_color(const color& c, const list<edge>& new_list,
							const bool& duplicate_test)
{	set_color(d3_gl_col(c),new_list,duplicate_test);   }

void d3_gl_scene::set_color(const color& c, const list<node>& new_list,
							const bool& duplicate_test)
{	set_color(d3_gl_col(c),new_list,duplicate_test);   }

void d3_gl_scene::set_color(const color& c, const list<face>& new_list,
							const bool& duplicate_test)
{   set_color(d3_gl_col(c),new_list,duplicate_test);   }


void d3_gl_scene::change_color(const d3_gl_col& c, const edge& e)
{
	if(Hp->number_of_edges() == 0) return;

	pp_tuple_list L;
	edges_to_points(list<edge>(e),L);
	pp_tuple find = L[L.first()];

	list_item it = col_edge_list.first();

	while(it != nil){
		col_edge_tuple& c_tuple = col_edge_list[it];
		pp_tuple_list& PTS = c_tuple.second();

		list_item it2 = PTS.first();
		while(it2 != nil){
			pp_tuple& elem = PTS[it2];
			if(elem == find || // edge OR inverse edge
				(elem.first() == find.second() && elem.second() == find.first()))
			{ 
				c_tuple.first() = c;
				break;
			}
			it2 = PTS.succ(it2);
		}		

		it = col_edge_list.succ(it);
	}	
}

void d3_gl_scene::change_color(const d3_gl_col& c, const node& n)
{
	if(Hp->number_of_nodes() == 0) return;	

	point_pointer_list L;
	nodes_to_points(list<node>(n),L);

	list_item it = col_node_list.first();

	while(it != nil){
		col_node_tuple& c_tuple = col_node_list[it];
		point_pointer_list& PTS = c_tuple.second();

		if(PTS.rank(L.head()) > 0)
			c_tuple.first() = c;
		it = col_node_list.succ(it);
	}	
}

void d3_gl_scene::change_color(const d3_gl_col& c, const face& f)
{
	if(Hp->number_of_faces() == 0) return;

	ppl_list L;
	faces_to_points(list<face>(f),L);

	list_item it = col_face_list.first();

	while(it != nil){
		col_face_tuple& c_tuple = col_face_list[it];
		ppl_list& PTS = c_tuple.second();

		list_item it2 = PTS.first();		

		while(it2 != nil){			
			
			if(cmp_two_ppl(L[L.get_item(0)],PTS[it2])){
				c_tuple.first() = c;				
			}			

			it2 = PTS.succ(it2);
		}		

		it = col_face_list.succ(it);
	}	
}

void d3_gl_scene::change_color(const color& c, const edge& e)
{	change_color(d3_gl_col(c),e);	}

void d3_gl_scene::change_color(const color& c, const node& n)
{	change_color(d3_gl_col(c),n);	}

void d3_gl_scene::change_color(const color& c, const face& f)
{	change_color(d3_gl_col(c),f);	}


bool d3_gl_scene::get_color(const edge& e, d3_gl_col& c)
{ 
	if(Hp->number_of_edges() == 0) return false;

	pp_tuple_list L;
	edges_to_points(list<edge>(e),L);
	pp_tuple& orig = L[L.first()];

	col_edge_tuple c_tuple;
	pp_tuple elem;

	forall(c_tuple, col_edge_list){
		forall(elem, c_tuple.second()){
			if(elem == orig || (elem.first() == orig.second() && elem.second() == orig.first()))
			{ 
				c = c_tuple.first();
				return true;
			}
		}
	} 

	return false;	
}

bool d3_gl_scene::get_color(const node& n, d3_gl_col& c)
{ 
	if(Hp->number_of_nodes() == 0) return false;

	point_pointer_list L;
	nodes_to_points(list<node>(n),L);
	point_pointer& orig = L[L.get_item(0)];	

	col_node_tuple c_tuple;
	point_pointer  elem;

	forall(c_tuple, col_node_list){
		forall(elem, c_tuple.second()){
			if(elem == orig){
				c = c_tuple.first();
				return true;				
			}
		}
	} 

	return false;
}

bool d3_gl_scene::get_color(const face& f, d3_gl_col& c)
{
	if(Hp->number_of_faces() == 0) return false;
	
	ppl_list L;
	faces_to_points(list<face>(f),L);

	col_face_tuple c_tuple;
	point_pointer_list elem2;

	forall(c_tuple, col_face_list){
		forall(elem2, c_tuple.second()){
			if(cmp_two_ppl(elem2,L.head())){
				c = c_tuple.first();
				return true;
			}
		}
	} 

	return false;
}


int d3_gl_scene::get_texture(const face& f)
{
	if(Hp->number_of_faces() == 0){		
		return -1;
	}

	tex_face_tuple t_tuple;	
	ppl_list L;
	faces_to_points(list<face>(f),L);	

	list_item it = tex_face_list.first();

	while(it != nil){
		tex_face_tuple& t_tuple = tex_face_list[it];
		ppl_list& L2 = t_tuple.second();

		list_item it2 = L2.first();		

		while(it2 != nil){
			if(cmp_two_ppl(L[L.get_item(0)],L2[it2]))
				return t_tuple.first();			

			it2 = L2.succ(it2);
		}		

		it = tex_face_list.succ(it);
	}	
	 
	return -1;
}


int d3_gl_scene::number_of_textures()
{
	if(tex_info_list.length() != tex_face_list.length())
		return -1;
	else
		return tex_info_list.length();
}

//--------------------------------------------------------------------------------

int d3_gl_scene::reg_tex(char* t_data, const int& w, const int& h,
						  const GLint& filter)
{	
	if(d3_gl_win == NULL) return -1;
	else if(d3_gl_win->gl_win == NULL) return -1;			

	GLuint t_num;
	d3_gl_win->reg_tex(&t_num,t_data,w,h,filter);

	tex_info_list.append(tex_info(t_data,w,h,filter));
	tex_face_list.push(tex_face_tuple(t_num,ppl_list())); // empty list		

	return t_num;
}

int d3_gl_scene::reg_tex(char* t_data, const int& w, const int& h,
						  const GLint& filter, const list<face>& faces)
{
	GLuint t_num = reg_tex(t_data,w,h,filter);
	
	tex_face_tuple& tex_tuple = tex_face_list[tex_face_list.first()];
	ppl_list& L = (ppl_list&) tex_tuple.second();

	faces_to_points(faces,L);
	
	return t_num;
}

void d3_gl_scene::unreg_tex(const GLuint& t_num)
{
	list_item it  = tex_face_list.first();
	list_item it2 = tex_info_list.first();	

	while(it != nil)
	{
		if(tex_face_list[it].first() == t_num){
			tex_face_list.del(it);			
			tex_info_list.del(it2);			
			
			return;
		}		
		it  = tex_face_list.succ(it);
		it2 = tex_info_list.succ(it2);		
	}	
}

void d3_gl_scene::re_reg_all_textures()
{
	list_item it  = tex_face_list.first();
	list_item it2 = tex_info_list.first();		

	while(it != nil)
	{
		tex_face_tuple& tex_tuple = tex_face_list[it];
		tex_info& t_info = tex_info_list[it2]; 
		char* data = t_info.first();
		int w = t_info.second();
		int h = t_info.third();
		GLuint filter = t_info.fourth();
		
		it  = tex_face_list.succ(it);
		it2 = tex_info_list.succ(it2);		
		
		unreg_tex(tex_tuple.first());
		reg_tex(data,w,h,filter);

		tex_face_tuple& T = tex_face_list[tex_face_list.first()];
		T.second() = tex_tuple.second();	    
	}
}

//-------------------------------------------------------------------------------------

void d3_gl_scene::set_texture(const GLuint& tex_num , const list<face>& faces)
{				
	bool is_set = false;
	ppl_list L;
	faces_to_points(faces,L);
	point_pointer_list elem;	
	
	list_item it = tex_face_list.first();

	while(it != nil){
		tex_face_tuple& tex_tuple = tex_face_list[it];
		ppl_list& PTS = tex_tuple.second();
		
		forall(elem,L){
			if(tex_num == tex_tuple.first()){				
				if(PTS.rank(elem) == 0){
					PTS.append(elem);						
				}
				is_set = true; // do not "break" at this point
			}
			else{
				PTS.remove(elem);
			}
		}		

		it = tex_face_list.succ(it);
	}
	
	if(!is_set){
		tex_face_list.append(tex_face_tuple(tex_num,L));
	}	
}


void d3_gl_scene::change_texture(const GLuint& tex_num, const face& f)
{
	if(Hp->number_of_faces() == 0) return;

	ppl_list L;
	faces_to_points(list<face>(f),L);

	list_item it = tex_face_list.first();

	while(it != nil){
		tex_face_tuple& tex_tuple = tex_face_list[it];
		ppl_list& PTS = tex_tuple.second();

		list_item it2 = PTS.first();		

		while(it2 != nil){			
			
			if(cmp_two_ppl(L[L.get_item(0)],PTS[it2])){
				tex_tuple.first() = tex_num;				
			}			

			it2 = PTS.succ(it2);
		}		

		it = tex_face_list.succ(it);
	}		
}

void d3_gl_scene::set_tex_filter(const GLuint& tex_num, const GLint& filter)
{
	list_item tuple_it  = tex_face_list.first();
	list_item info_it   = tex_info_list.first();

	while(tuple_it != nil){
		tex_face_tuple& T = tex_face_list[tuple_it];

		if(tex_num == T.first()){
			if(filter == GL_NEAREST || filter == GL_LINEAR){
				tex_info_list[info_it].fourth() = filter;

				if(d3_gl_win == NULL) return;
				if(d3_gl_win->gl_win == NULL) return;

				d3_gl_win->gl_win->gl_start();

				glBindTexture(GL_TEXTURE_2D,tex_num);
				glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);  
				glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);

				d3_gl_win->gl_win->gl_finish();

				return;
			}			
		}

		tuple_it  = tex_face_list.succ(tuple_it);
		info_it   = tex_info_list.succ(info_it);
	}				
}

GLint d3_gl_scene::get_tex_filter(const GLuint& tex_num)
{
	list_item tuple_it  = tex_face_list.first();
	list_item info_it   = tex_info_list.first();

	while(tuple_it != nil){
		tex_face_tuple& T = tex_face_list[tuple_it];

		if(tex_num == T.first()){
			return tex_info_list[info_it].fourth();
		}

		tuple_it  = tex_face_list.succ(tuple_it);
		info_it   = tex_info_list.succ(info_it);
	}

	return -1;		
}

char* d3_gl_scene::get_tex_data(const GLuint& tex_num)
{
	tex_face_tuple tex_tuple;
	int index = 0;

	forall(tex_tuple,tex_face_list)
	{
		if(tex_num == tex_tuple.first()){
			list_item it = tex_info_list.get_item(index);
			return tex_info_list[it].first();
		}
		++index;
	}
	return NULL;
}

void d3_gl_scene::draw() 
{ 
  glPushMatrix(); 
  
  double actual_mat[16];
  for(int i=0; i<16;i++)
		  actual_mat[i] = trans_mat[i]; 
  
  if(seq_mat.length()){	  
	  clear_matrix();	  

	  list_item mat  = seq_mat.first();
	  list_item kind = sequences.first();	  

	  while(mat!=nil){		  
		  glMultMatrixd(seq_mat[mat]);
		  transformation_list& transforms = sequences[kind];

		  list_item it = transforms.first();
		  while(it != nil){
			  transformation& T = transforms[it];
			  GLdouble*& d = T.second();

			  if(T.first() == D3_GL_TRANS){
				  glTranslated(d[0],d[1],d[2]);				  
				  d[0] += d[3];
				  d[1] += d[4];
				  d[2] += d[5];
			  }
			  else if(T.first() == D3_GL_SCALE){
				  glScaled(d[0],d[1],d[2]);				  
				  d[0] += d[3];
				  d[1] += d[4];
				  d[2] += d[5];
			  }
			  else if(T.first() == D3_GL_ROT){
				  glRotated(d[0],d[2],d[3],d[4]);				  
				  d[0] += d[1];
			  }  

			  it = transforms.succ(it);
		  }          		  

		  mat  = seq_mat.succ(mat);
		  kind = sequences.succ(kind);
	  }	  
	  
	  glMultMatrixd(actual_mat);
  }          
  
  if(draw_obj_border){
	  glLineWidth(border_line_width);	  
	  border_line_color();
	  
	  glBegin(GL_LINE_LOOP); // left
		gl_vertex(border[0]); 
		gl_vertex(border[1]); 
		gl_vertex(border[2]); 
		gl_vertex(border[3]); 
	  glEnd();

	  glBegin(GL_LINE_LOOP); // right
		gl_vertex(border[4]); 
		gl_vertex(border[5]); 
		gl_vertex(border[6]); 
		gl_vertex(border[7]); 
	  glEnd();

	  glBegin(GL_LINES);     // the rest
		gl_vertex(border[0]); gl_vertex(border[4]);
		gl_vertex(border[1]); gl_vertex(border[5]);
		gl_vertex(border[2]); gl_vertex(border[6]);
		gl_vertex(border[3]); gl_vertex(border[7]);		
	  glEnd();	  	       
  }  
  
  if(draw_faces)
  {
    list_item it1 = col_face_list.first();
	while(it1 != nil){
		col_face_tuple& T = col_face_list[it1];
		d3_gl_col& temp_col = T.first();
		temp_col();		

		ppl_list& L = T.second();
		list_item it2 = L.first();
		while(it2 != nil){
			point_pointer_list& ppl = L[it2];			
			point_pointer pp;
			
			glBegin(GL_POLYGON);

			forall(pp,ppl)
				gl_vertex(*pp);			    										

			glEnd();			

			it2 = L.succ(it2);
		}

		it1 = col_face_list.succ(it1);
	}     
  }  
  
  if(texturing)
  {	 	  
      glEnable(GL_TEXTURE_2D);
	  
	  list_item it1 = tex_face_list.first();

	  while(it1 != nil){

		  tex_face_tuple& T = tex_face_list[it1];
		  glBindTexture(GL_TEXTURE_2D, T.first());
		  ppl_list& L = T.second();			  

		  list_item it2 = L.first();
		  while(it2 != nil){
			  point_pointer_list& ppl = L[it2];

			  glBegin(GL_POLYGON);			  

			  if(ppl.length() > 2){ 				

				list_item it3 = ppl.first();				
				glTexCoord2f(1.0f,1.0f); gl_vertex(*ppl[it3]);								
				
				it3 = ppl.succ(it3);				
				glTexCoord2f(0.0f,1.0f); gl_vertex(*ppl[it3]);						

				it3 = ppl.succ(it3);				
				glTexCoord2f(0.0f,0.0f); gl_vertex(*ppl[it3]);								

				it3 = ppl.succ(it3);
				if(it3 != nil){					
					glTexCoord2f(1.0f,0.0f); gl_vertex(*ppl[it3]);					
				}				
			  }			  

			  glEnd();

			  it2 = L.succ(it2);
		  }			  
		  it1 = tex_face_list.succ(it1);		  
	  }

	  glDisable(GL_TEXTURE_2D);	  	  
  }    
  
  if(draw_edges)
  {	
    glLineWidth(edge_width);
	glBegin(GL_LINES);

	list_item it = col_edge_list.first();
	while(it != nil){
		col_edge_tuple& T = col_edge_list[it];
		d3_gl_col& temp_col = T.first();
		temp_col();	

		pp_tuple_list& L = T.second();
		list_item it2 = L.first();
		while(it2 != nil){
			pp_tuple& ppt = L[it2];
			gl_vertex(*ppt.first());
			gl_vertex(*ppt.second());
			it2 = L.succ(it2);
		}

		it = col_edge_list.succ(it);
	}    	

	glEnd();	
   }      

  
  if (draw_nodes)
  {
	glPointSize(node_width);
	glBegin(GL_POINTS);

	list_item it = col_node_list.first();
	while(it != nil){
		col_node_tuple& T = col_node_list[it];
		d3_gl_col& temp_col = T.first();
		temp_col();	

		point_pointer_list& L = T.second();
		point_pointer pp;		
		forall(pp,L){
			gl_vertex(*pp);			
		}		

		it = col_edge_list.succ(it);
	} 		

	glEnd();	
  }     
  
  glPopMatrix();  
  
  set_matrix(actual_mat);
}


// ----------------------------------------------
// affine transformations

void d3_gl_scene::mult_matrix(const double* m)
{
	double result[] = { 0.0, 0.0, 0.0, 0.0,
		                0.0, 0.0, 0.0, 0.0,
					    0.0, 0.0, 0.0, 0.0,
					    0.0, 0.0, 0.0, 0.0 };

	int k;

	for(k=0; k<4; k++)
		for(int i=0; i<4; i++)
			for(int j=0; j<4; j++)
				result[i*4+k] += trans_mat[j*4+k]*m[j+i*4];		

	for(k=0; k<16; k++)
		trans_mat[k] = result[k];
}

void d3_gl_scene::clear_matrix()
{
	for(int i=0; i<16 ;i++){
	  if(i % 5 == 0)
		trans_mat[i] = 1.0;
	  else
		trans_mat[i] = 0.0;
	}
}

void d3_gl_scene::print_matrix()
{	
	for(int j=0; j<4 ; j++){
	  for(int k=0; k<4; k++)
		  cout<<trans_mat[k*4+j]<<"  ";
	  cout<<"\n";
	}
	cout<<"\n";
}

void d3_gl_scene::set_matrix(const double* m)
{
	for(int i=0; i<16; i++)
		trans_mat[i] = m[i];
}

double* d3_gl_scene::get_matrix()
{ return trans_mat; }



void d3_gl_scene::translate(const double& x, const double& y, const double& z,
							const bool& first, const bool& update)
{
	GLdouble T[] = { 1.0, 0.0, 0.0, 0.0,
			   	     0.0, 1.0, 0.0, 0.0,
				     0.0, 0.0, 1.0, 0.0,
				     x  , y  , z  , 1.0 };
	if(first){
		GLdouble temp[16];
		for(int i=0; i<16; i++)
			temp[i] = trans_mat[i];

		set_matrix(T);
		mult_matrix(temp);
	}
	else
		mult_matrix(T);		
}


void d3_gl_scene::scale(const double& xf, const double& yf, const double& zf,
						const bool& keep_pos, const bool& first)
{
	GLdouble S[] = { xf , 0.0, 0.0, 0.0,
				     0.0, yf , 0.0, 0.0,
				     0.0, 0.0, zf , 0.0,
				     0.0, 0.0, 0.0, 1.0 };

	if(keep_pos){
		if(first){
			GLdouble temp[16];
			for(int i=0; i<16; i++)
				temp[i] = trans_mat[i];

			clear_matrix();
			translate(g_center[0],g_center[1],g_center[2]);
			mult_matrix(S);
			translate(-g_center[0],-g_center[1],-g_center[2]);
			mult_matrix(temp);
		}
		else
			translate(g_center[0],g_center[1],g_center[2]);
			mult_matrix(S);
			translate(-g_center[0],-g_center[1],-g_center[2]);
	}
	else{
		if(first){
			GLdouble temp[16];
			for(int i=0; i<16; i++)
				temp[i] = trans_mat[i];

			set_matrix(S);
			mult_matrix(temp);
		}
		else
			mult_matrix(S);
	}
}

void d3_gl_scene::rotate(const double& alpha, double x, double y,double z, const bool& first)
{
	double l = ::sqrt(x*x + y*y + z*z);
	x = x/l;
	y = y/l;
	z = z/l;

	double cosa = cos(vgl::deg_to_rad(alpha));
	double sina = sin(vgl::deg_to_rad(alpha));	

	GLdouble R[] =
	{ x*x+cosa-cosa*x*x  , y*x-cosa*y*x+sina*z, z*x-cosa*z*x-sina*y, 0.0,
	  x*y-cosa*x*y-sina*z, y*y+cosa-cosa*y*y  , z*y-cosa*z*y+sina*x, 0.0,
	  x*z-cosa*x*z+sina*y, y*z-cosa*y*z-sina*x, z*z+cosa-cosa*z*z  , 0.0,
	  0.0                , 0.0                , 0.0                , 1.0 };

	if(first){
		GLdouble temp[16];
		for(int i=0; i<16; i++)
			temp[i] = trans_mat[i];

		set_matrix(R);
		mult_matrix(temp);
	}
	else
		mult_matrix(R);
}


void d3_gl_scene::get_modelview_matrix(GLdouble* matrix)
{
  glPushMatrix(); 
  
  double actual_mat[16];
  for(int i=0; i<16;i++)
		  actual_mat[i] = trans_mat[i];

  if(seq_mat.length()){	  
	  clear_matrix();	  

	  list_item mat  = seq_mat.first();
	  list_item kind = sequences.first();	  

	  while(mat!=nil){		  
		  glMultMatrixd(seq_mat[mat]);
		  transformation_list& transforms = sequences[kind];

		  list_item it = transforms.first();
		  while(it != nil){
			  transformation& T = transforms[it];
			  GLdouble*& d = T.second();

			  if(T.first() == D3_GL_TRANS){
				  glTranslated(d[0],d[1],d[2]);				  
				  d[0] += d[3];
				  d[1] += d[4];
				  d[2] += d[5];
			  }
			  else if(T.first() == D3_GL_SCALE){
				  glScaled(d[0],d[1],d[2]);				  
				  d[0] += d[3];
				  d[1] += d[4];
				  d[2] += d[5];
			  }
			  else if(T.first() == D3_GL_ROT){
				  glRotated(d[0],d[2],d[3],d[4]);				  
				  d[0] += d[1];
			  }  

			  it = transforms.succ(it);
		  }          		  

		  mat  = seq_mat.succ(mat);
		  kind = sequences.succ(kind);
	  }		  

	  mult_matrix(actual_mat);
  }    

  glMultMatrixd(trans_mat);
  glGetDoublev(GL_MODELVIEW_MATRIX,matrix);
  glPopMatrix();

  set_matrix(actual_mat);
}


int d3_gl_scene::begin_sequence()
{
	// add empty list 
	sequences.append(transformation_list());

	// store actual matrix
	GLdouble* trans_mat2 = new double[16];
	for(int i=0; i<16; i++)
		trans_mat2[i] = trans_mat[i];
	seq_mat.append(trans_mat2);

	// reset actual matrix
	clear_matrix();

	return (seq_mat.length()-1);
}

// has currently no function ; but it helps to structure the code
int d3_gl_scene::end_sequence()
{   return (seq_mat.length()-1); }


void d3_gl_scene::remove_sequence(const int& index)
{
	if(index < 0 || index > (sequences.length()-1)){
		LEDA_EXCEPTION(1,"d3_gl_scene:: index out of range");
		return;
	}
	
	double* stored_mat = seq_mat[seq_mat.get_item(index)];
	double  actual_mat[16];

	for(int i=0; i<16; i++)
		actual_mat[i] = trans_mat[i];

	set_matrix(stored_mat);	
	
	transformation t;
	transformation_list& transforms = sequences[sequences.get_item(index)];	  	

	forall(t,transforms){
      GLdouble*& d = t.second();
	  if(t.first() == D3_GL_TRANS){		  
		  translate(d[0],d[1],d[2],false,true);
		  d[0] += d[3];
		  d[1] += d[4];
		  d[2] += d[5];
	  }
	  else if(t.first() == D3_GL_SCALE){		  
		  scale(d[0],d[1],d[2],false,false);
		  d[0] += d[3];
		  d[1] += d[4];
		  d[2] += d[5];
	  }
	  else if(t.first() == D3_GL_ROT){		  
		  rotate(d[0],d[2],d[3],d[4],false);
		  d[0] += d[1];
	  }
	  delete[] d;
	}		  	  
	  
	transforms.clear();
	
	mult_matrix(actual_mat);

	delete[] seq_mat.del(seq_mat.get_item(index));
	sequences.del(sequences.get_item(index));
}


int d3_gl_scene::seq_translate(const double* params)
{
	GLdouble* trans_kind = new double[6];	
	for(int i=0; i<6; i++)
		trans_kind[i] = params[i];

	list_item it = sequences.last();
	sequences[it].append(transformation(D3_GL_TRANS,trans_kind));
	return (sequences[it].length()-1);
}


int d3_gl_scene::seq_scale(const double* params)
{
	GLdouble* trans_kind = new double[6];	
	for(int i=0; i<6; i++)
	 	trans_kind[i] = params[i];

	list_item it = sequences.last();
	sequences[it].append(transformation(D3_GL_SCALE,trans_kind));
	return (sequences[it].length()-1);
}


int d3_gl_scene::seq_rotate(const double* params)
{
    GLdouble* trans_kind = new double[5];	
	for(int i=0; i<5; i++)
		trans_kind[i] = params[i];

	list_item it = sequences.last();
	sequences[it].append(transformation(D3_GL_ROT,trans_kind));
	return (sequences[it].length()-1);
}

int d3_gl_scene::seq_translate(const double& x , const double& y , const double& z,
	 						   const double& x2, const double& y2, const double& z2)
{
	double params[6];
	params[0] = x;  params[1] = y;  params[2] = z;
	params[3] = x2; params[4] = y2; params[5] = z2;

	return seq_translate(params);
}


int d3_gl_scene::seq_scale(const double& xf , const double& yf , const double& zf,
						   const double& xf2, const double& yf2, const double& zf2)
{
	double params[6];
	params[0] = xf;  params[1] = yf;  params[2] = zf;
	params[3] = xf2; params[4] = yf2; params[5] = zf2;

	return seq_scale(params);

}


int d3_gl_scene::seq_rotate(const double& start_angle, const double& alpha, const double& x,
						    const double& y, const double& z)
{
	double params[5];
	params[0] = start_angle;  params[1] = alpha;
	params[2] = x; params[3] = y; params[4] = z;

	return seq_rotate(params);
}


void d3_gl_scene::set_seq_trans(const int& index1, const int& index2, const double* params)
{
	if(index1 < 0 || index1 > (sequences.length()-1))
		LEDA_EXCEPTION(1,"d3_gl_scene: index out of range");

	transformation_list& trans_list = sequences[sequences.get_item(index1)];

	if(index2 < 0 || index2 > (trans_list.length()-1))
		LEDA_EXCEPTION(1,"d3_gl_scene: index out of range");

	transformation& T = trans_list[trans_list.get_item(index2)];

	int l;
	if(T.first() == D3_GL_TRANS || T.first() == D3_GL_SCALE)
		 l = 6;
	else l = 5;

	GLdouble*& d = T.second();

	for(int i=0; i<l; i++)
		d[i] = params[i];
}


void d3_gl_scene::get_seq_trans(const int& index1, const int& index2, double* params)
{
	if(index1 < 0 || index1 > (sequences.length()-1))
		LEDA_EXCEPTION(1,"d3_gl_scene: index out of range");

	transformation_list& trans_list = sequences[sequences.get_item(index1)];

	if(index2 < 0 || index2 > (trans_list.length()-1))
		LEDA_EXCEPTION(1,"d3_gl_scene: index out of range");

	transformation& T = trans_list[trans_list.get_item(index2)];

	int l;
	if(T.first() == D3_GL_TRANS || T.first() == D3_GL_SCALE)
		 l = 6;
	else l = 5;

	GLdouble*& d = T.second();

	for(int i=0; i<l; i++)
		params[i] = d[i];	
}


double* d3_gl_scene::get_seq_trans_a(const int& index1, const int& index2)
{
	if(index1 < 0 || index1 > (sequences.length()-1))
		return NULL;

	transformation_list& trans_list = sequences[sequences.get_item(index1)];

	if(index2 < 0 || index2 > (trans_list.length()-1))
		LEDA_EXCEPTION(1,"d3_gl_scene: index out of range");

	transformation& T = trans_list[trans_list.get_item(index2)];

	return T.second();
}



void d3_gl_scene::direct_translate(const double& x, const double& y, const double& z)
{
	if (Hp == NULL) return; 

	vector vec(x,y,z);
	node v;
	forall_nodes(v,*Hp) 
		(*pos1)[v] = (*pos1)[v] + vec;

	for(int i=0; i<8; i++)
		border[i] += vec;
	
	g_center  += vec;

	xmin += x; xmax += x;
    ymin += y; ymax += y;
	zmin += z; zmax += z;	
}

void d3_gl_scene::direct_scale(const double& xf, const double& yf,
							   const double& zf, const bool& keep_pos)
{ 
    if(Hp == NULL) return;
	
	vector temp = g_center;
	if(keep_pos) direct_translate(-temp[0],-temp[1],-temp[2]);
   
	node v;
	vector vec;
	forall_nodes(v,*Hp){  		 
		vec = (*pos1)[v];
		vec[0] *= xf;
		vec[1] *= yf;
		vec[2] *= zf;
		(*pos1)[v] = vec;		
	}

	xmin *= xf; xmax *= xf;
	ymin *= yf; ymax *= yf;
	zmin *= zf; zmax *= zf; 
	
	g_center[0] *= xf;
	g_center[1] *= yf;
	g_center[2] *= zf;

	for(int i=0; i<8; i++){
		border[i][0] *= xf;
		border[i][1] *= yf;
		border[i][2] *= zf;
	}

	if(keep_pos){
		direct_translate(temp[0],temp[1],temp[2]);
		set_border();
	}
}

void d3_gl_scene::help_rotate(double& cosa, double& sina, double& x, double& y,
							  double& z, vector& v)
{
	double x2,y2,z2;
	x2 = v[0]*(x*x+cosa-cosa*x*x)   + v[1]*(x*y-cosa*x*y-sina*z ) + v[2]*(x*z-cosa*x*z+sina*y);
	y2 = v[0]*(y*x-cosa*y*x+sina*z) + v[1]*(y*y+cosa-cosa*y*y )   + v[2]*(y*z-cosa*y*z-sina*x);
	z2 = v[0]*(z*x-cosa*z*x-sina*y) + v[1]*(z*y-cosa*z*y+sina*x ) + v[2]*(z*z+cosa-cosa*z*z);
	v[0] = x2;
	v[1] = y2;
	v[2] = z2;

	if(x2<xmin) xmin = x2;
	if(x2>xmax) xmax = x2;
	if(y2<ymin) ymin = y2;
	if(y2>ymax) ymax = y2;
	if(z2<zmin) zmin = z2;
	if(z2>zmax) zmax = z2;
}

void d3_gl_scene::direct_rotate(const double& alpha, double& x, double& y, double& z)					
{
	double l = ::sqrt(x*x + y*y + z*z);
	x = x/l;
	y = y/l;
	z = z/l;

	xmin = 0.0; xmax = 0.0;
	ymin = 0.0; ymax = 0.0;
	zmin = 0.0; zmax = 0.0;

	double cosa = cos(vgl::deg_to_rad(alpha));
	double sina = sin(vgl::deg_to_rad(alpha));	

	node n;		
	forall_nodes(n,*Hp){
		vector& v = (*pos1)[n];		
		help_rotate(cosa,sina,x,y,z,v);		
	}
    
	help_rotate(cosa,sina,x,y,z,g_center);	

	for(int i=0; i<8; i++){		
		help_rotate(cosa,sina,x,y,z,border[i]);		
	}

	vector a(xmin,ymin,zmin);
	vector b(xmax,ymax,zmax);    
	help_rotate(cosa,sina,x,y,z,a);
	help_rotate(cosa,sina,x,y,z,b);	
	xmin = a[0]; xmax = b[0];
	ymin = a[1]; ymax = b[1];
	zmin = a[2]; zmax = b[2];
}


void d3_gl_scene::set_position(const node& n, double x, double y, double z)
{
  (*pos1)[n] = vector(x,y,z); 
  // first_rot(total_1,total_rot2,total_rot3,(*pos1)[v]);
  if(x<ymin) xmin = x;
  if(x>ymax) xmax = x;
  if(y<ymin) ymin = y;
  if(y>ymax) ymax = y;
  if(z<zmin) zmin = z;
  if(z>zmax) zmax = z;
  set_border();

  vector cg(0,0,0);
  node n2;

  forall_nodes(n2,*Hp) cg += (*pos1)[n2];
  g_center = cg/Hp->number_of_nodes();
}


void d3_gl_scene::print_node_list()
{
	col_node_tuple c_tuple;
	cout<<"\n* nodes *";
	forall(c_tuple,col_node_list){
		cout<<"\n"<<c_tuple.first()<<"  "<<c_tuple.second().length();
	}
}

void d3_gl_scene::print_edge_list()
{
	col_edge_tuple c_tuple;
	cout<<"\n* edges *";
	forall(c_tuple,col_edge_list){
		cout<<"\n"<<c_tuple.first()<<"  "<<c_tuple.second().length();
	}
}

void d3_gl_scene::print_face_list()
{
	col_face_tuple c_tuple;
	cout<<"\n* faces *";
	forall(c_tuple,col_face_list){
		cout<<"\n"<<c_tuple.first()<<"  "<<c_tuple.second().length();
	}
}



int compare(const d3_gl_scene& scene1, const d3_gl_scene& scene2)
{ if(scene1.get_node_array() == scene2.get_node_array() )
	 return 0;
  else
	 return -1;
}


int compare(const SC_PTR& scene1, const SC_PTR& scene2)
{ 	if(scene1 == scene2)
		return 0;
	else
		return -1;
} 

int compare(const point_pointer& p1, const point_pointer& p2)
{
	if(p1 == p2)
		return 0;
	else
		return -1;
}

int compare(const point_pointer_list& L1, const point_pointer_list& L2)
{
	if(L1.length() != L2.length()) return -1;

	list_item it1 = L1.first();
	list_item it2 = L2.first();

	while(it1 != nil && it2 != nil){
		if(L1[it1] != L2[it2]) return -1;
		it1 = L1.succ(it1);
		it2 = L2.succ(it2);
	}
	
	return 0;
}


LEDA_END_NAMESPACE
