// compression module includes
#include <LEDA/coding/MTF2.h>

LEDA_BEGIN_NAMESPACE

/// MTF2Coder //////////////////////////////////////////////////////////////////////////////

MTF2Coder::MTF2Coder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

MTF2Coder::MTF2Coder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
MTF2Coder::init_table()
{
	byte i = 0;
	do {
		CyclicSucc[i] = i+1;
	} while (++i);
	Tail = 0xFF;
}

void 
MTF2Coder::init_encoding()
{
	base::init_encoding();
	
	init_table();
	PrevRankWasNotZero = false;
}

void 
MTF2Coder::encode_stream()
{
	byte head = CyclicSucc[Tail];

	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;

		byte rank = 0;
		if (symbol == head) { PrevRankWasNotZero = false; }
		else {
			byte pred_cur_sym;
			byte cur_sym = head;
			do {
				++rank;
				cur_sym = CyclicSucc[pred_cur_sym = cur_sym];
			} while (cur_sym != symbol);

			// update list ...
			if (rank == 1) {
				if (PrevRankWasNotZero) {
					// swap #0 and #1
					CyclicSucc[head] = CyclicSucc[cur_sym];
					CyclicSucc[cur_sym] = head;
					CyclicSucc[Tail] = cur_sym;
					head = cur_sym;
				}
			}
			else {
				if (Tail == cur_sym) Tail = pred_cur_sym;

				// remove cur_sym from list
				CyclicSucc[pred_cur_sym] = CyclicSucc[cur_sym];
				// reinsert after head
				CyclicSucc[cur_sym] = CyclicSucc[head]; // old #1 becomes #2
				CyclicSucc[head] = cur_sym;             // cur_sym becoms new #1
			}

			PrevRankWasNotZero = true;
		}

		write_byte(rank);
	}

	if (c == EOF) {
		encoding_finished();
	}
}

void 
MTF2Coder::init_decoding()
{
	base::init_decoding();

	init_table();
	PrevRankWasNotZero = false;
}

void 
MTF2Coder::decode_stream()
{
	byte head = CyclicSucc[Tail];

	int c;
	while ((c = read_byte()) != EOF) {
		byte rank = (byte) c;

		byte cur_sym;
		if (rank == 0) { cur_sym = head; PrevRankWasNotZero = false; }
		else {
			byte pred_cur_sym = 0; // just to avoid a warning
			cur_sym = head;
			for (byte r = 0; r < rank; ++r) {
				cur_sym = CyclicSucc[pred_cur_sym = cur_sym];
			}

			// update list ...
			if (rank == 1) {
				if (PrevRankWasNotZero) {
					// swap #0 and #1
					CyclicSucc[head] = CyclicSucc[cur_sym];
					CyclicSucc[cur_sym] = head;
					CyclicSucc[Tail] = cur_sym;
					head = cur_sym;
				}
			}
			else {
				if (Tail == cur_sym) Tail = pred_cur_sym;

				// remove cur_sym from list
				CyclicSucc[pred_cur_sym] = CyclicSucc[cur_sym];
				// reinsert after head
				CyclicSucc[cur_sym] = CyclicSucc[head]; // old #1 becomes #2
				CyclicSucc[head] = cur_sym;             // cur_sym becoms new #1
			}

			PrevRankWasNotZero = true;
		}

		if (write_byte(cur_sym) < 0) return;
	}

	decoding_finished();
}

LEDA_END_NAMESPACE
