// compression module includes
#include <LEDA/coding/RLE0.h>

LEDA_BEGIN_NAMESPACE

/// RLE0Coder //////////////////////////////////////////////////////////////////////////////

RLE0Coder::RLE0Coder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

RLE0Coder::RLE0Coder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
RLE0Coder::init_encoding()
{
	base::init_encoding();
	
	LengthZeroRun = 0;
}

void 
RLE0Coder::encode_stream()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;

		if (symbol == 0) ++LengthZeroRun;
		else {
			if (LengthZeroRun > 0) encode_zero_run();
			if (symbol < 0xFE) write_byte(symbol+1);
			else             { write_byte(0xFF); write_byte(symbol - 0xFE);	}
		}
	}

	if (c == EOF) {
		if (LengthZeroRun > 0) encode_zero_run();
		encoding_finished();
	}
}

#define BIT31 0x80000000

void
RLE0Coder::encode_zero_run()
{
	uint32 val = LengthZeroRun+1;
	LengthZeroRun = 0;

	int bits = 31;
	while ((val & BIT31) == 0) { val <<= 1; --bits; }
	val <<= 1; --bits;

	while (bits >= 0) {
		write_byte( (val & BIT31) == 0 ? 0 : 1 );
		val <<= 1; --bits;
	}


/*
	while (val > 1) {
		write_byte( val & 1 );
		val >>= 1;
	}
	*/
}

void 
RLE0Coder::init_decoding()
{
	base::init_decoding();
	
	LengthZeroRun = 0;
	ReadAhead = read_byte();
}

void 
RLE0Coder::decode_stream()
{
	while (LengthZeroRun > 0) {
		--LengthZeroRun;
		if ( write_byte(0x00) < 0) return;
	}

	int c = ReadAhead;
	while (c != EOF) {
		if (c <= 1) {
			uint32 val = 1;
			do {
				val <<= 1; val |= uint32(c);
				c = read_byte();
			} while (c == 0 || c == 1);
			LengthZeroRun = val - 1;

			ReadAhead = c;

			while (LengthZeroRun > 0) {
				--LengthZeroRun;
				if ( write_byte(0x00) < 0) return;
			}
		}
		else {
			byte symbol = (byte) c;

			if (symbol != 0xFF) --symbol;
			else                symbol = 0xFE + (byte) read_byte();

			c = read_byte();
			if ( write_byte(symbol) < 0 ) { ReadAhead = c; return; }
		}
	}

	decoding_finished();
}

LEDA_END_NAMESPACE
