// crypt module includes
#include <LEDA/coding/crypt_auto_decoder.h>

LEDA_BEGIN_NAMESPACE

/// CryptAutoDecoder ////////////////////////////////////////////////////////////////////////

CryptAutoDecoder::CryptAutoDecoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

CryptAutoDecoder::CryptAutoDecoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
CryptAutoDecoder::init_encoding()
{
	convert_keys_to_memory_stream();
	base::init_encoding();
}

void 
CryptAutoDecoder::init_decoding()
{
	convert_keys_to_memory_stream();
	base::init_decoding();
}

void
CryptAutoDecoder::do_reset(reset_type t)
{
	base::do_reset(t);
	if (t == r_hard) {
		clear_keys();
		KeyMemory.clear();
	}
}

void 
CryptAutoDecoder::add_key(const CryptKey& key)
{
	if (can_change_parameters())
		Keys.append(key);
	else error("CryptAutoDecoder::add_key: wrong mode");
}

void 
CryptAutoDecoder::add_keys_in_stream(streambuf* key_streambuf)
{
	istream key_stream(key_streambuf);
	CryptKey key;
	while (key_stream) {
		key_stream >> key;
		if (key_stream) add_key(key);
	}
}

void 
CryptAutoDecoder::add_keys_in_file(const char* file_name)
{
	if (file_name == 0) return;

	filebuf key_stream;
	if (! key_stream.open(file_name, ios::openmode(ios::in | binary))) {
		error("CryptAutoDecoder::add_keys_in_file: open failed");
	}

	add_keys_in_stream(&key_stream);
}

void 
CryptAutoDecoder::convert_keys_to_memory_stream()
{
	set_additional_stream(0, false);

	streamsize sz = 0;
	CryptKey key;
	forall(key, Keys) sz += key.get_size_in_stream();
	KeyMemory.set_size(uint32(sz));
	if (sz == 0) return;

	memory_streambuf* key_buf = new memory_streambuf(KeyMemory, sz, false, true); // wipe
	ostream key_stream(key_buf);
	forall(key, Keys) key_stream << key;
	set_additional_stream(key_buf, true);
}

LEDA_END_NAMESPACE
