#include <LEDA/core/list.h>
#include <LEDA/core/string.h>
#include <LEDA/system/file.h>

#include <LEDA/coding/crypt.h>


using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;
#endif

/*
#ifdef LEDA_COMPRESSION_WA1
#define openmode open_mode
#endif
*/

#define open_in  coder_base::binary_in()
#define open_out coder_base::binary_out()

void copy_streambuffer(streambuf* src, streambuf* tgt)
{
  // copy src to tgt (till src signals EOF)

  char buffer[1024];

  int bytes_read;
  do { bytes_read = src->sgetn(buffer, 1024);
       tgt->sputn(buffer, bytes_read);
  } while (bytes_read == 1024);
}


ostream* setup_ostream(string archive_file)
{
//typedef encoding_ostream<PPMIICoder>   CompressOStream;
  typedef encoding_ostream<DeflateCoder> CompressOStream;

  filebuf* p_raw_filebuf = new filebuf;
  if (!p_raw_filebuf->open(archive_file, open_out)) 
  { cerr << "could not create archive " << archive_file << endl; 
    return 0;
  }

  ostream arc_raw(p_raw_filebuf); 
  // data written to this stream is not encoded
  // every archive starts with "MZ." where "." encodes the format
  // so far we have only two formats: 0x00: compressed, 
  //                                  0x01: compressed & encrypted
  arc_raw << "MZ";
  arc_raw.put((char) 0x00); 
  arc_raw.flush();

  CompressOStream* p_arc = new CompressOStream(p_raw_filebuf, true);
  cout << "  using coder: " << p_arc->get_coder()->get_description() << endl;
  return p_arc;
}


int encode_files(const list<string>& files_to_process, 
                        string archive_file)
{
  //cout << "Packing " << archive_file << " ... " << endl;

  // generate the archive-ostream "arc"
  ostream* p_arc = setup_ostream(archive_file);
  if (! p_arc) return 1;
  ostream& arc = *p_arc;

  // write table_of_contents
  arc << "#" << files_to_process.size() << '\n';
  string filename;
  forall(filename, files_to_process) 
  { if (!is_file(filename)) 
    { cerr << filename << " not found" << endl; 
      return 1; 
     }
    arc << filename << endl;
    arc << size_of_file(filename) << endl;
  }

  // put TOC end marker
  arc.put('#');

  // encode the files
  forall(filename, files_to_process) 
  { cout << "  encoding: " << filename << " ... "; cout.flush();

/*
    // open input file
    filebuf in;
    in.open(filename, open_in);
    if (! in.is_open()) 
    { cerr << filename << "not opened" << endl; 
      return 1; 
     }

    // encode file into archive
    copy_streambuffer(&in, arc.rdbuf());

    // write EOF-marker (not really necessary)
    arc.put('#'); 
    arc.put('L'); 
    arc.put('#');

    if (arc.fail()) 
    { cout << "failure" << endl; 
      return 1; 
     }

*/

    //ostream out(arc.rdbuf());
    ifstream in(filename);

    while (in) 
    { char c;
      in.get(c);
      arc.put(c);
    }
   
    if (arc.fail()) cout << "failure" << endl; 

    cout << "ok" << endl;
  }

  delete p_arc;
  return 0;
}




int main(int argc, char* argv[])
{
  if (argc < 3) 
  { cerr << "usage: leda_zip archive files" << endl;
    return 1; 
   }

  string archive = argv[1];
  list<string> files;
  for (int i=2; i<argc; ++i) files.append(argv[i]);

  if (is_file(archive))
  { cerr << archive << ": file exists" << endl;
    return 1;
   }

  return encode_files(files, archive);
}

