/*******************************************************************************
+
+  LEDA 7.0
+
+
+  alignment.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

#include<LEDA/string/alignment.h>

LEDA_BEGIN_NAMESPACE

void alignment::setStringA(const string &x) {
	inputA = x;
	m = x.length();
	min_m = 0;
	max_m = m;
	return;
}
string alignment::getStringA() {
	return inputA;
}
void alignment::setStringB(const string &x){
	inputB = x;
	n = x.length();
	min_n = 0;
	max_n = m;
	return;
}
string alignment::getStringB() {
	return inputB;
}
string alignment::getAlA() {
	return alignmentA;
}
string alignment::getAlB() {
	return alignmentB;
}
string alignment::getAlInfo() {
	return aligner;
}

bool alignment::isInAlphabet(char c) {
	return scores.hasScore(int(c), int(c));
}

bool alignment::checkStrings() {
	for (int i = 0; i < m; i++)
		if(!scores.hasScore(inputA[i], inputA[i]))
			return false;
	for (int i = 0; i < n; i++)
		if(!scores.hasScore(inputB[i], inputB[i]))
			return false;
	return true;
}

void alignment::setGapPenalty(int x) {
	gap_penalty = x;
	return;
}
void alignment::setGapIncPenalty(int x) {
	gap_inc_penalty = x;
	return;
}
int alignment::getGapPenalty() {
	return gap_penalty;
}

int alignment::getGapIncPenalty() {
	return gap_inc_penalty;
}

void alignment::readFirstString() {
	cout << "Alignment, First String: " << endl;
	cin >> inputA;
	m = inputA.length();
	min_m = 0;
	max_m = m;
}

void alignment::readSecondString() {
	cout << "Alignment, Second String: " << endl;
	cin >> inputB;
	n = inputB.length();

	min_n = 0;
	max_n = n;
}




// longest common substring
// x_ind and y_ind have to be 0
// Hirschberg
unsigned int alignment::longest_common_substring()
{
	char* x = &(inputA[0]);
	char* y = &(inputB[0]);

	map2<unsigned int, unsigned int, unsigned int> lcs(0);

	unsigned int length = 0;
	
	for (int i = 0; i < m; i++)
		for (int j = 0; j < n; j++)
			if(x[i] != y[j])
				lcs(i+1, j+1) = 0;
			else
			{
				lcs(i+1, j+1) = lcs(i,j) +1;
				if (lcs(i+1,j+1) > length)
				{
					length = lcs(i+1, j+1);
					x_ind = i;
					y_ind = j;
				}
			}
	return length;
}
// longest common subsequence
// Hirschberg

unsigned int alignment::longest_common_subsequence()
{
	char* x = &(inputA[0]);
	char* y = &(inputB[0]);


	map2<unsigned int, unsigned int, unsigned int> lcs(0);

	unsigned int i;
	for (i = 0; i < (unsigned int) m; i++)
		for (int j = 0; j < n; j++)
			if(x[i] == y[j])
				lcs(i+1, j+1) = lcs(i,j) +1;
			else
				lcs(i+1, j+1) = max(lcs(i+1,j), lcs(i, j+1));
	
	i = m;
	unsigned int j = n; 
	unsigned int val, valdiag, valup, valleft;
	while (i > 0 && j > 0)
	{ 
		val = lcs(i,j);
		valdiag = lcs(i-1,j-1);
		valup = lcs(i,j-1);
		valleft = lcs(i-1,j);
		
		
		if (val == valup)
		{
			alignmentA = "-" + alignmentA;
			aligner = " " + aligner;
			alignmentB = string(y[j-1]) + alignmentB;
			j--;
		}
		else 
		{
			if (val == valleft)
			{
				alignmentA = string(x[i-1]) + alignmentA;
				aligner = " " + aligner;
				alignmentB = "-" + alignmentB;
				i--;
			}
			else
				if (val == valdiag + 1)
					{
						alignmentA = string(x[i-1]) + alignmentA;
						aligner = "|" + aligner;
						alignmentB = string(y[j-1]) + alignmentB;
						i--;
						j--;
					}
		}
	}
	while (i > 0)
	{
		alignmentA = string(x[i-1]) + alignmentA;
		aligner = " " + aligner;
		alignmentB = "-" + alignmentB;
		i--;
	}
	while (j > 0)
	{
		alignmentA = "-" + alignmentA;
		aligner = " " + aligner;
		alignmentB = string(y[j-1]) + alignmentB;
		j--;
	}
	return lcs(m,n);
}


void alignment::compute_alignment(int i, int j)
{
	int prev_i = i-1;
	int prev_j = j-1;

	if ((i > 0) && (j > 0) && ((*d)(i,j) == ((*d)(prev_i,prev_j) + scores(int(inputA[prev_i]), int(inputB[prev_j]))))) {
		compute_alignment(prev_i, prev_j);
		alignmentA = alignmentA + inputA[prev_i];
		if (inputA[prev_i] == inputB[prev_j])
			aligner = aligner + "|";
		else
			aligner = aligner + "X";
		alignmentB = alignmentB + inputB[prev_j];
		return;
	}

	if ((i > 0) && ((*d)(i,j) == ((*d)(prev_i, j) - gap_penalty))) {
		compute_alignment(prev_i, j);
		alignmentA = alignmentA + inputA[prev_i];
		aligner = aligner + " ";
		alignmentB = alignmentB + "-";
		return;
	}

	if ((j > 0) && ((*d)(i, j) == ((*d)(i, prev_j) - gap_penalty))) {
		compute_alignment(i, prev_j);
		alignmentA = alignmentA + "-";
		aligner = aligner + " ";
		alignmentB = alignmentB + inputB[prev_j];
		return;
	}
	min_m = i;
	min_n = j;
}

void alignment::compute_aff_alignment(int i, int j)
{
	if ((i > 0) && (j > 0) && (((*td)(i,j)).fourth() == 0)) {
		compute_aff_alignment(i-1, j-1);
		alignmentA = alignmentA + inputA[i-1];
		if (inputA[i-1] == inputB[j-1])
			aligner = aligner + "|";
		else
			aligner = aligner + "X";
		alignmentB = alignmentB + inputB[j-1];
		return;
	}
		
	if ((i > 0) && (((*td)(i,j)).fourth() == 1)) { // here starts a gap in inputA
		int k = i;
		while (i > 0 && ((*td)(i - 1,j)).second() == (((*td)(i, j)).second() + gap_inc_penalty)) {
			i--;
		}

		if (i > 0)
			compute_aff_alignment(i - 1, j);

		while ( i <= k) {
			alignmentA = alignmentA + inputA[i-1];
			aligner = aligner + " ";
			alignmentB = alignmentB + "-";
			i++;
		}
		return;
	}

	if ((j > 0) && (((*td)(i,j)).fourth() == 2)) { // here starts a gap in inputB
		int k = i;
		while (j > 0 && ((*td)(i,j-1)).third() == (((*td)(i, j)).third() + gap_inc_penalty)) {
			j--;
		}

		if (j > 0)
			compute_aff_alignment(i, j-1);

		while ( j <= k) {
			alignmentA = alignmentA + "-";
			aligner = aligner + " ";
			alignmentB = alignmentB + inputB[j-1];
			j++;
		}
		return;
	}

	if (((*td)(i,j)).fourth() == 3){
		min_m = i;
		min_n = j;
	}
}

void alignment::local_sw() {
	d = new array2<int>(m+1, n+1);
	min_n = min_m = 0;
	max_n = n;
	max_m = m;
	al_score = 0;

	for (int i = 0; i <= m; i++)
		(*d)(i, 0) = 0;
	for (int j = 0; j <= n; j++)
		(*d)(0, j) = 0;

	for (int i = 0; i < m; i++) {
		int next_i = i+1;
		for (int j = 0; j < n; j++)
		{
			int next_j = j+1;
			(*d)(next_i, next_j) = max(max(int(max((*d)(i, next_j) - gap_penalty, (*d)(next_i, j) - gap_penalty)), (*d)(i, j) + scores(int(inputA[i]), int(inputB[j]))),int(0));
		}
	}
	
	// from here we start the Backtracing

	al_score = (*d)(max_m, max_n);
	alignmentA = "";
	alignmentB = "";
	aligner = "";
	if (al_score > 0)
		compute_alignment(max_m, max_n);
	else
		al_score = 0;

	delete d;

}

void alignment::global_nw() {
	d = new array2<int>(m+1, n+1);
	min_n = min_m = 0;
	max_n = n;
	max_m = m;
	al_score = NEG_INFTY;
	
	(*d)(0, 0) = 0;

	for (int j = 1; j <= n; j++)
		(*d)(0, j) = - j * gap_penalty;

	for (int i = 0; i < m; i++) {
		int next_i = i+1;
		(*d)(next_i, 0) = (*d)(i, 0) - gap_penalty;
		for (int j = 0; j < n; j++)
		{
			int next_j = j+1;
			(*d)(next_i, next_j) = max(int(max((*d)(i, next_j) - gap_penalty, (*d)(next_i, j) - gap_penalty)), (*d)(i, j) + scores(int(inputA[i]), int(inputB[j])));
		}
	}

	// from here we start the Backtracing

	al_score = (*d)(m, n);
	alignmentA = "";
	alignmentB = "";
	aligner = "";

	compute_alignment(max_m, max_n);

	delete d;

}

void alignment::global_aff_nw() {
	td = new array2<fft >(m+1, n+1);
	min_n = 0;
	min_m = 0;
	max_n = n;
	max_m = m;
	al_score = NEG_INFTY;
	
	((*td)(0, 0)).first() = 0;
	((*td)(0, 0)).second() = NEG_INFTY;
	((*td)(0, 0)).third() = NEG_INFTY;
	((*td)(0, 0)).fourth() = 3;

	for (int j = 0; j < n; j++) {
		int next_j = j+1;

		((*td)(0, next_j)).first() = -(gap_penalty + j*gap_inc_penalty);
		((*td)(0, next_j)).second() = NEG_INFTY;
		((*td)(0, next_j)).third() = -(gap_penalty + j*gap_inc_penalty);
		((*td)(0, next_j)).fourth() = 2;
	}

	for (int i = 0; i < m; i++) {
		int next_i = i+1;

		((*td)(next_i, 0)).first() = -(gap_penalty + i*gap_inc_penalty);
		((*td)(next_i, 0)).second() = -(gap_penalty + i*gap_inc_penalty);
		((*td)(next_i, 0)).third() = NEG_INFTY;
		((*td)(next_i, 0)).fourth() = 1;

		// elaboration
		for (int j = 0; j < n; j++) {
			int next_j = j+1;
			((*td)(next_i, next_j)).first() = ((*td)(i, j)).first() + scores(int(inputA[i]), int(inputB[j]));
			((*td)(next_i, next_j)).second() = max (((*td)(i, next_j)).first() - gap_penalty, ((*td)(i, next_j)).second() - gap_inc_penalty);
			((*td)(next_i, next_j)).third() = max (((*td)(next_i, j)).first() - gap_penalty, ((*td)(next_i, j)).third() - gap_inc_penalty);
			((*td)(next_i, next_j)).fourth() = 0;

			if ((((*td)(next_i, next_j)).second() >= ((*td)(next_i, next_j)).first()) && (((*td)(next_i, next_j)).second() >= ((*td)(next_i, next_j)).third())) {
				((*td)(next_i, next_j)).first() = ((*td)(next_i, next_j)).second();
				((*td)(next_i, next_j)).fourth() = 1;
			}
			else if ((((*td)(next_i, next_j)).third() >= ((*td)(next_i, next_j)).first()) && (((*td)(next_i, next_j)).third() > ((*td)(next_i, next_j)).second())) {
				((*td)(next_i, next_j)).first() = ((*td)(next_i, next_j)).third();
				((*td)(next_i, next_j)).fourth() = 2;
			}
		}
				

	}
	

	// from here we start the Backtracing

	al_score = ((*td)(m, n)).first();
	alignmentA = "";
	alignmentB = "";
	aligner = "";

	compute_aff_alignment(m, n);

	delete td;

}

void alignment::local_aff_sw() {
	td = new array2<fft >(m+1, n+1);
	min_n = 0;
	min_m = 0;
	max_n = n;
	max_m = m;
	al_score = 0;
	
	((*td)(0, 0)).first() = 0;
	((*td)(0, 0)).second() = NEG_INFTY;
	((*td)(0, 0)).third() = NEG_INFTY;
	((*td)(0, 0)).fourth() = 3;

	for (int j = 0; j < n; j++) {
		int next_j = j+1;

		((*td)(0, next_j)).first() = 0;
		((*td)(0, next_j)).second() = NEG_INFTY;
		((*td)(0, next_j)).third() = -(gap_penalty + j*gap_inc_penalty);
		((*td)(0, next_j)).fourth() = 3;
	}

	for (int i = 0; i < m; i++) {
		int next_i = i+1;

		((*td)(next_i, 0)).first() = 0;
		((*td)(next_i, 0)).second() = -(gap_penalty + i*gap_inc_penalty);
		((*td)(next_i, 0)).third() = NEG_INFTY;
		((*td)(next_i, 0)).fourth() = 3;


		for (int j = 0; j < n; j++) {
			int next_j = j+1;

					// elaboration
			((*td)(next_i, next_j)).first() = ((*td)(i, j)).first() + scores(int(inputA[i]), int(inputB[j]));
			((*td)(next_i, next_j)).second() = max (((*td)(i, next_j)).first() - gap_penalty, ((*td)(i, next_j)).second() - gap_inc_penalty);
			((*td)(next_i, next_j)).third() = max (((*td)(next_i, j)).first() - gap_penalty, ((*td)(next_i, j)).third() - gap_inc_penalty);
			((*td)(next_i, next_j)).fourth() = 0;

			if ((((*td)(next_i, next_j)).second() >= ((*td)(next_i, next_j)).first()) && (((*td)(next_i, next_j)).second() >= ((*td)(next_i, next_j)).third())) {
				((*td)(next_i, next_j)).first() = ((*td)(next_i, next_j)).second();
				((*td)(next_i, next_j)).fourth() = 1;
			}
			else if ((((*td)(next_i, next_j)).third() >= ((*td)(next_i, next_j)).first()) && (((*td)(next_i, next_j)).third() > ((*td)(next_i, next_j)).second())) {
				((*td)(next_i, next_j)).first() = ((*td)(next_i, next_j)).third();
				((*td)(next_i, next_j)).fourth() = 2;
			}

			if (0 > ((*td)(next_i, next_j)).first()) {
				((*td)(next_i, next_j)).first() = 0;
				((*td)(next_i, next_j)).fourth() = 3;
			}

			if (((*td)(next_i, next_j)).first() > al_score) {
				al_score = ((*td)(next_i, next_j)).first();
				max_m = next_i;
				max_n = next_j;
			}
		}	
	}
	

	// cleaning

//	al_score = ((*td)(m, n)).first();
	alignmentA = "";
	alignmentB = "";
	aligner = "";

	// from here we start the Backtracing
	if (al_score > 0)
		compute_aff_alignment(max_m, max_n);
	else
		al_score = 0;

	delete td;

}
LEDA_END_NAMESPACE


