/*******************************************************************************
+
+  LEDA 7.0
+
+
+  distance.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

#include<LEDA/string/distance.h>

LEDA_BEGIN_NAMESPACE

void distance::readFirstString() {
	cout << "String Matching, First String: " << endl;
	cin >> x;
	m = x.length();
}

void distance::readSecondString() {
	cout << "String Matching, Second String: " << endl;
	cin >> y;
	n = y.length();
}

void distance::setFirstString(const string &str) {
	x = str;
	m = str.length();
}
	
void distance::setAlgorithm(alg A) {
	current = A;
}

void distance::setAlgorithm(int A) {
	if (A < 1 || A > 4)
		LEDA_EXCEPTION(1, "distance::setAlgorithm: index out of range!");
	current = alg(A);
}

int distance::run() {
	switch(current) {
		case 1: return hamming();
		case 2: return levenshtein();
		case 3: return damerau_levenshtein();
		case 4: return int(general_levenshtein());
		break;
	}
	return -1;
}


void distance::setSecondString(const string &str){
	y = str;
	n = str.length();
}
string distance::getFirstString(){
	return x;
}
string distance::getSecondString(){
	return y;
}



// hamming distance
int distance::hamming() {

	if (n != m) 
		return -1;

	int weight(0);
	int i;
	for(i = 0 ; i < n ; i++) {
		if( x[i] != y[i] ) {
			weight += 1;
		}
	}
	return weight;
}



// levenshtein_distance

int distance::levenshtein()
{   
	array2<int> d(m+1, n+1);
	int i, j, cost(0);
	for (i = 0; i < m+1; i++)
		d(i, 0) = i;
	for (j = 1; j < n+1; j++)
		d(0, j) = j;
	for (i = 0; i < m; i++)
	   for (j = 0; j < n; j++)
	   { if (x[i] == y[j]) 
               cost = 0;
	     else 
               cost = 1;
	     d(i+1, j+1) = min(min(d(i,j+1) + 1, d(i+1,j) + 1), d(i,j) + cost);
	    }
   return d(m, n);
}

// damerau levenshtein distance

int distance::damerau_levenshtein()
{
	array2<int> d(m+1, n+1);
	int i, j, cost(0);
	for (i = 0; i < m+1; i++)
		d(i, 0) = i;
	for (j = 1; j < n+1; j++)
		d(0, j) = j;
	for (i = 0; i < m; i++)
		for (j = 0; j < n; j++)
		{
			if (x[i] == y[j])
				cost = 0;
			else cost = 1;
			d(i+1, j+1) = min(min(d(i, j+1) + 1, d(i+1, j) + 1), d(i, j) + cost);

			if(i > 0 && j > 0 && x[i] == y[j-1] && x[i-1] == y[j])
				d(i+1, j+1) = min(d(i+1, j+1), d(i-1, j-1) + cost);   // transposition)
		}
    return d(m, n);
}

// general levenshtein difference


unsigned int distance::general_levenshtein()
{
  unsigned int* p = new unsigned int[n+1];
  unsigned int* q = new unsigned int[n+1];
  unsigned int* r;

  p[0] = 0;
  for(int j = 1; j <= n; ++j )
    p[j] = p[j-1] + cost_ins;

  for(int i = 1; i <= m; ++i )
    {
      q[0] = p[0] + cost_del;
      for(int j = 1; j <= n; ++j )
        {
          unsigned int d_del = p[j] + cost_del;
          unsigned int d_ins = q[j-1] + cost_ins;
          unsigned int d_sub = p[j-1] + ( x[i-1] == y[j-1] ? 0 : cost_sub );
          q[j] = min( min( d_del, d_ins ), d_sub );
      }
      r = p;
      p = q;
      q = r;
    }

  unsigned int tmp = p[n];
  delete[] p;
  delete[] q;

  return tmp;
}


LEDA_END_NAMESPACE
