/*******************************************************************************
+
+  LEDA 7.0
+
+
+  score_matrix.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

#include<LEDA/string/score_matrix.h>

LEDA_BEGIN_NAMESPACE

void score_matrix::setScore(char A, char B, int sc) {
	if (!score.defined(int(A), int(B)))
		LEDA_EXCEPTION(1,"Letters not defined!");
	else
		score(int(A), int(B)) = sc;
}

int score_matrix::getScore(char A, char B) const {
	if (!score.defined(int(A), int(B)))
	{
		LEDA_EXCEPTION(1,"Letters not defined!");
		return 0;
	}

	else
		return score(int(A), int(B));
}

bool score_matrix::hasScore(char a, char b) const {
	return score.defined(a,b);
}

void score_matrix::setScoreMatch(int sc) {
	int length = (*alpha).getLength();
	for (int i = 0; i < length; i++)
		score(int((*alpha).getLetter(i)), int((*alpha).getLetter(i)) ) = sc;
}

void score_matrix::setScoreMismatch(int sc) {
	int length = (*alpha).getLength();
	for (int i = 0; i < length; i++)
		for (int j = 0; j < length; j++)
			if (i==j)
				continue;
			else
				score(int((*alpha).getLetter(i)), int((*alpha).getLetter(j)) ) = sc;
}

void score_matrix::setGapPenalty(int pen) {
	gap_penalty = pen;
}

void score_matrix::setGapIncPenalty(int incpen) {
	gap_inc_penalty = incpen;
}

void score_matrix::setGapPenalty(int pen, int incpen) {
	gap_inc_penalty = incpen;
	gap_penalty = pen;
}

int score_matrix::getGapPenalty()const {
	return gap_penalty;
}

int score_matrix::getGapIncPenalty()const {
	return gap_inc_penalty;
}

int& score_matrix::operator()(int i, int j)
{ 
  if (score.defined(i, j))
	  return score(i, j);
  else
	  LEDA_EXCEPTION(1,"score_matrix: col index out of range");
  return score(i, j);
}

LEDA_END_NAMESPACE
