#ifdef _MSC_IDE

#include <LEDA/core/d_array.h>
#include <LEDA/core/set.h>
#include <LEDA/system/file.h>

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;
#endif

using namespace leda;

#endif // ifdef _MSC_IDE

#include "HtmlReport.h"

static const string jpg_dir = "report/";

// replace \ -> \\, " -> \", RegExp: ^ -> ", $ -> \\n"
static const string header =
"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\"\n"
"\"http://www.w3.org/TR/html4/loose.dtd\">\n"
"<html>\n"
"	<head>\n"
"		<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">\n"
"		<title>Algorithmic Solutions Software GmbH: LEDA Test Report</title>\n"
"		<style type=\"text/css\">\n"
"			<!--\n"
".Stil1 {	font-family: Arial, Helvetica, sans-serif;\n"
"	font-size: 12px;\n"
"}\n"
".Stil2 {font-size: 14px}\n"
"a:link {\n"
"	color: #333333;\n"
"	text-decoration: none;\n"
"}\n"
"a:visited {\n"
"	text-decoration: none;\n"
"	color: #000000;\n"
"}\n"
"a:hover {\n"
"	text-decoration: none;\n"
"	font-weight: bold;\n"
"	color: #999999;\n"
"}\n"
"a:active {\n"
"	text-decoration: none;\n"
"	color: #000000;\n"
"}\n"
"--></style>\n"
"	</head>\n"
"	<body>\n"
"		<table width=\"100%\" border=\"0\">\n"
"			<tr>\n"
"				<td width=\"67%\"><div align=\"left\"><span class=\"Stil1 titel Stil2\"><strong>LEDA 5.0 Test Report </strong>\n"
"						</span></div>\n"
"				</td>\n"
"				<td width=\"33%\" valign=\"bottom\"><div align=\"right\"><img src=\"" + jpg_dir + "elementrot.jpg\" width=\"243\" height=\"30\"></div>\n"
"				</td>\n"
"			</tr>\n"
"		</table>\n"
"		<table width=\"80%\" border=\"0\">\n"
"			<tr class=\"text\">\n"
"				<td><table width=\"100\" border=\"0\">\n"
"						<tr>\n"
"							<td width=\"50\"><div align=\"center\"><span class=\"Stil1\">MAKE</span></div>\n"
"							</td>\n"
"							<td width=\"50\"><div align=\"center\"><span class=\"Stil1\">RUN</span></div>\n"
"							</td>\n"
"						</tr>\n"
"					</table>\n"
"				</td>\n"
"				<td><span class=\"Stil1\">EXECUTABLE</span></td>\n"
"				<td><div align=\"center\"><span class=\"Stil1\">SRC</span></div></td>\n"
"				<td><span class=\"Stil1\">OUTPUT</span></td>\n"
"			</tr>\n"
"			<tr class=\"text\">\n"
"				<td colspan=\"4\" bgcolor=\"#e5e5e5\"></td>\n"
"			</tr>\n"
;
// stopper: \ "

// replace \ -> \\, " -> \", RegExp: ^ -> ", $ -> \\n"
static const string footer =
"</table>\n"
"<table width=\"100%\" border=\"0\" bgcolor=\"#990000\">\n"
"  <tr>\n"
"    <td><div align=\"right\"><font color=\"#FFFFFF\" size=\"1\" face=\"Arial, Helvetica, sans-serif\">Copyright &copy; 1998-2004\n"
"          Algorithmic Solutions Software GmbH. All rights reserved. </font></div></td>\n"
"  </tr>\n"
"</table>\n"
"<p>&nbsp;</p>\n"
"</body>\n"
"</html>\n"
;
// stopper: \ "

// replace \ -> \\, " -> \", RegExp: ^ -> ", $ -> \\n"
static const string entry_template =
"			<tr class=\"text\">\n"
"				<td width=\"20%\"><table width=\"100\" border=\"0\">\n"
"						<tr>\n"
"							<td width=\"50\" bgcolor=\"#COL_MAK\">&nbsp;</td>\n"
"							<td width=\"50\" bgcolor=\"#COL_RUN\">CMT_RUN</td>\n"
"						</tr>\n"
"					</table>\n"
"				</td>\n"
"				<td><span class=\"Stil1\">EXE\n"
"					</span></td>\n"
"				<td><div align=\"center\"><span class=\"Stil1\">SRC</span></div>\n"
"				</td>\n"
"				<td width=\"30%\"><div align=\"left\"><span class=\"Stil1\">OUT</span></div>\n"
"				</td>\n"
"			</tr>\n"
;
// stopper: \ "

const string cmt_run_std = "&nbsp;";
const string cmt_run_GUI = "<div align=\"center\"><span class=\"Stil1\">GUI</span>";

#define TXT_LINK(TXT, PATH) \
(string("<a href=\"") + PATH +string("\" target=\"_blank\">") + TXT + string("</a>"))

#define BIG_ICON_LINK(ICON, PATH) \
(string("<a href=\"") + PATH + string("\" target=\"_blank\"><img src=\"") + jpg_dir + string("icon-") + ICON + string(".jpg\" width=\"51\" height=\"27\" align=\"absbottom\" border=\"0\"></a>"))

#define SMALL_ICON_LINK(ICON, PATH) \
(string("<a href=\"") + PATH + string("\" target=\"_blank\"><img src=\"") + jpg_dir + string("icon-") + ICON + string(".jpg\" width=\"26\" height=\"27\" align=\"absbottom\" border=\"0\"></a>"))

enum Color { C_Red, C_Yellow, C_Green, C_Grey };
static const char* colors[] =
{ "FF3300", "FFFF66", "66FF66", "E5E5E5" };

/// HtmlReport /////////////////////////////////////////////////////////////////////////////

HtmlReport::HtmlReport(string output_path)
 : ReportPath(output_path)
{
	ReportStream = new ofstream(ReportPath);
}

HtmlReport::~HtmlReport()
{
	if (ReportStream != &cout && ReportStream != &cerr) {
		delete ReportStream;
	}
}

void 
HtmlReport::begin_document()
{
	doc() << header << endl;
}

void 
HtmlReport::end_document()
{
	doc() << footer << endl;
}

void 
HtmlReport::add_target(const TestTarget& tgt)
{
	// determine colors
	Color col_mak = C_Grey;
	if (tgt.make_error())        col_mak = C_Red;
	else if (tgt.make_warning()) col_mak = C_Yellow;
	else if (tgt.make_ok())      col_mak = C_Green;
	Color col_run = C_Grey;
	if (tgt.run_error())         col_run = C_Red;
	else if (tgt.run_ok())       col_run = C_Green;

	// determine comments
	string cmt_run = cmt_run_std;
	if (tgt.uses_GUI()) {
		cmt_run = cmt_run_GUI;
	}

	// determine exe
	string exe = tgt.get_exe_name();
	string exe_path = tgt.get_exe_path();
	if (is_file(exe_path)) exe = TXT_LINK(exe, exe_path);

	// determine source
	string source = SMALL_ICON_LINK("source", tgt.get_source_path());

	// determine outputs
	string outputs;
	if (is_file(tgt.get_make_out_path())) {
		outputs += BIG_ICON_LINK("make-out", tgt.get_make_out_path());
	}
	if (is_file(tgt.get_make_err_path())) {
		outputs += BIG_ICON_LINK("make-err", tgt.get_make_err_path());
	}
	if (!tgt.run_skip()) {
		if (is_file(tgt.get_run_out_path())) {
			outputs += BIG_ICON_LINK("run-out", tgt.get_run_out_path());
		}
		if (is_file(tgt.get_run_err_path())) {
			outputs += BIG_ICON_LINK("run-err", tgt.get_run_err_path());
		}
	}

	doc() << "<!-- " << tgt.get_source_path() << " -->" << endl;
	string entry = entry_template;
	entry = entry.replace("COL_MAK", colors[col_mak]);
	entry = entry.replace("COL_RUN", colors[col_run]);
	entry = entry.replace("CMT_RUN", cmt_run);
	entry = entry.replace("EXE", exe);
	entry = entry.replace("SRC", source);
	entry = entry.replace("OUT", outputs);
	doc() << entry << endl;
}
