#include <windows.h>
#include <shlobj.h>
#include <stdio.h>


HRESULT CreateLink(const char* obj_path,  const char* wdir_path, 
                                          const char* link_path)

{ 
  const char* desc = "LEDA Shortcut";

  HRESULT hres; 
  void*   ptr; 

  // Get a pointer to the IShellLink interface. 
  hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, 
                                                 IID_IShellLink, &ptr); 

  IShellLink* psl = (IShellLink*)ptr;

  if (SUCCEEDED(hres)) 
  { 
    // Set path, working directory, and description 
    psl->SetPath(obj_path); 
    psl->SetWorkingDirectory(wdir_path); 
    psl->SetDescription(desc); 

    // Query IShellLink for the IPersistFile interface for saving the 
    // shortcut in persistent storage. 
    hres = psl->QueryInterface(IID_IPersistFile,&ptr); 

    IPersistFile* ppf = (IPersistFile*)ptr;

    if (SUCCEEDED(hres))
    { WORD wsz[MAX_PATH]; 

      // Ensure that the string is ANSI. 
      MultiByteToWideChar(CP_ACP, 0, link_path, -1, wsz, MAX_PATH); 

      // Save the link by calling IPersistFile::Save. 
      hres = ppf->Save(wsz, TRUE); 
      ppf->Release(); 
     } 

    psl->Release(); 
  } 

  return hres; 
}



void clean_dir(const char* dirname)
{
  DWORD att = GetFileAttributes(dirname);
  if (att == 0xFFFFFFFF || !(att & FILE_ATTRIBUTE_DIRECTORY)) return;
  SetCurrentDirectory(dirname);

  WIN32_FIND_DATA fd;
  HANDLE ha = FindFirstFile("*",&fd);
  if (ha && ha != (HANDLE)0xffffffff)
  { do { DeleteFile(fd.cFileName);
    } while (FindNextFile(ha,&fd));
    FindClose(ha);
   }
}



void read_reg(char* buf, unsigned long buf_sz)
{
  HKEY hKey;
  RegOpenKeyEx(HKEY_CURRENT_USER,
                     "Software",       NULL, KEY_QUERY_VALUE, &hKey);
  RegOpenKeyEx(hKey, "Microsoft",      NULL, KEY_QUERY_VALUE, &hKey);
  RegOpenKeyEx(hKey, "Windows",        NULL, KEY_QUERY_VALUE, &hKey);
  RegOpenKeyEx(hKey, "CurrentVersion", NULL, KEY_QUERY_VALUE, &hKey);
  RegOpenKeyEx(hKey, "Explorer",       NULL, KEY_QUERY_VALUE, &hKey);
  RegOpenKeyEx(hKey, "Shell Folders",  NULL, KEY_QUERY_VALUE, &hKey);

  RegQueryValueEx(hKey,"Programs", NULL,NULL, (unsigned char*)buf,&buf_sz);
}



const int argc = 3;
char* argv[] = { "xlman", "graphwin", "logo" };


main()
{
  const char* leda_path = "D:\\LEDA\\BETA\\win32\\bcc";

  char prog_path[256];

  read_reg(prog_path,256);

  sprintf(prog_path+strlen(prog_path),"\\LEDA");

  clean_dir(prog_path);

  CreateDirectory(prog_path,NULL);

  char buf1[256];
  char buf2[256];

  CoInitialize(NULL);
  HRESULT hres;

  for(int i=0; i<argc; i++)
  { sprintf(buf1,"%s\\%s.exe",leda_path,argv[i]);
    sprintf(buf2,"%s\\%s.lnk",prog_path,argv[i]);
    HRESULT hres = CreateLink(buf1,leda_path,buf2);
    if (!SUCCEEDED(hres)) printf("hres = %d\n",hres);
   }

  return 0;
}

