#define LEDA_CHECKING_OFF

#include <LEDA/graph.h>
#include <LEDA/s_graph.h>
#include <LEDA/dimacs_graph.h>

#include <LEDA/templates/max_flow1.t>

#include <LEDA/assert.h>


int main(int argc, char** argv)
{
  graph G(3,2);
  node s=0;
  node t=0;

  edge_array<int>  cap;
  assert(cap.use_edge_data(G));

  edge_array<int>  flow;
  assert(flow.use_edge_data(G));


  cout << endl;
  cout << "reading DIMACS maxflow problem from std input ... ";
  cout << flush;

  read_dimacs_mf(cin,G,s,t,cap);
  cout << endl;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;


  float h = 1.0;
  if (argc >= 2)  h = atof(argv[1]);


  float T = used_time();
  cout << "MAX_FLOW    " << flush;

  max_flow<int,graph>  mf(h);

  int f = mf(G,s,t,cap,flow);

  cout << endl;
  cout << "     ";
  cout << string("time: %6.2f sec  f = %d",used_time(T),f) << endl;
  cout << "     ";
  cout << string("pushes:%d  relabels:%d  updates:%d  insp:%d  gap nodes: %d",
                  mf.pushes(), mf.relabels(), mf.updates(), 
                  mf.edge_inspections(), mf.gap_nodes()) << endl;
  cout << endl;

  mf.check();

  return 0;
}




