#define LEDA_CHECKING_OFF

#include <LEDA/graph.h>
#include <LEDA/s_graph.h>
#include <LEDA/dimacs_graph.h>

#include <LEDA/templates/max_flow1.t>

#include <LEDA/assert.h>


int main(int argc, char** argv)
{
  typedef s_graph<opposite_graph,3,2> static_graph;

  static_graph G;
  static_graph::node s=0;
  static_graph::node t=0;

  edge_array<double,static_graph>  cap;
  assert(cap.use_edge_data(G));

  edge_array<double,static_graph>  flow;
  assert(flow.use_edge_data(G));


  cout << endl;
  cout << "reading DIMACS maxflow problem from std input ... ";
  cout << flush;

  read_dimacs_mf(cin,G,s,t,cap);
  cout << endl;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;



  cout << "MAX_FLOW    " << flush;

  max_flow<double,static_graph>  mf;

  if (argc >= 2)  
  { float h = atof(argv[1]);
    mf.set_heuristic(h);
   }

  float T = used_time();

  double f = mf(G,s,t,cap,flow);

  cout << endl;
  cout << "     ";
  cout << string("time: %6.2f sec  f = %.2f",used_time(T),f) << endl;
  cout << "     ";
  cout << string("pushes:%d  relabels:%d  updates:%d  insp:%d  gap nodes: %d",
                  mf.pushes(), mf.relabels(), mf.updates(), 
                  mf.inspections(), mf.gap_nodes()) << endl;
  cout << endl;

  mf.check();

  return 0;
}




