#include <LEDA/graph.h>
#include <LEDA/templates/dijkstra.t>
#include <LEDA/dimacs.h>
#include <LEDA/stream.h>

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::ifstream;
using std::cout;
using std::cin;
using std::flush;
using std::endl;
#endif


int main(int argc, char** argv)
{
  int node_slots = 1;
  int edge_slots = 1;

  GRAPH<int,int> G(node_slots,edge_slots);

  edge_array<int>& cost = G.edge_data();

  node_array<int>& dist = G.node_data();

  node_array<edge> pred;
  pred.use_node_data(G,0);

  node s;

  cout << endl;

  if (argc < 2) 
    { cout << "reading DIMACS shortest paths problem from std input ... ";
      cout << flush;
      Read_Dimacs_SP(cin,G,s,cost);
      cout << endl;
     }
  else
    { cout << "reading DIMACS shortest paths problem from " << argv[1] << " ... ";
      cout <<flush;
      ifstream in(argv[1]);
      Read_Dimacs_SP(in,G,s,cost);
     }

  cout << endl;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;

  cout << "DIJKSTRA   " << flush;
  float T  = used_time();

  DIJKSTRA_T(G,s,cost,dist,pred);

  cout << endl;
  cout << "     ";
  cout << string("time: %6.2f sec",used_time(T)) << endl;
  cout << endl;

  return 0;
}
