#include <LEDA/s_graph.h>
#include <LEDA/dimacs_graph.h>
#include <LEDA/node_slot.h>
#include <LEDA/edge_slot.h>

#include <LEDA/node_slist_slot.h>

#include <LEDA/templates/dijkstra_s.t>


using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::flush;
using std::endl;
#endif

int main(int argc, char** argv)
{
  typedef s_graph<basic_graph_traits<3,1> > static_graph;
  typedef static_graph::node node;
  typedef static_graph::edge edge;

  static_graph G;
  node s=0;

  edge_slot<int, 0,static_graph> cost(G);
  node_slot<int, 1,static_graph> dist(G);
  node_slot<edge,2,static_graph> pred(G);

/*
  assert(cost.use_edge_data(G));
  assert(dist.use_node_data(G));
  assert(pred.use_node_data(G));
*/


  cout << endl;
  cout << "reading DIMACS shortest paths problem from std input ... ";
  cout << flush;

  read_dimacs_sp(cin,G,s,cost);
  cout << endl;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;

  node_slist_slot<1,static_graph> L;


  dijkstra<int,static_graph> D;

  cout << "DIJKSTRA   " << flush;
  float T  = used_time();

  D(G,s,cost,dist,pred);

  cout << endl;
  cout << "     ";
  cout << string("time: %6.2f sec",used_time(T)) << endl;
  cout << endl;

  return 0;
}
