
#include <LEDA/graph.h>
#include <LEDA/templates/dijkstra1.t>
#include <LEDA/s_graph.h>
#include <LEDA/dimacs_graph.h>

#if defined(LEDA_STD_IO_HEADERS)
using namespace std;
#endif
 

int main(int argc, char** argv)
{
  typedef s_graph<basic_graph,2,2> static_graph;

  static_graph G;
  static_graph::node s=0;

  edge_array<int,static_graph>  cost;
  assert(cost.use_edge_data(G));

  node_array<int,static_graph>  dist;
  assert(dist.use_node_data(G));

  node_array<static_graph::edge,static_graph> pred;
  assert(pred.use_node_data(G));


  cout << endl;
  cout << "reading DIMACS shortest paths problem from std input ... ";
  cout << flush;

  read_dimacs_sp(cin,G,s,cost);
  cout << endl;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;


  cout << "DIJKSTRA   " << flush;
  float T  = used_time();

  DIJKSTRA_T(G,s,cost,dist,pred);

  cout << endl;
  cout << "     ";
  cout << string("time: %6.2f sec",used_time(T)) << endl;
  cout << endl;

  return 0;
}
