#include <LEDA/window.h>
#include <LEDA/file.h>
#include <fstream.h>
#include <time.h>
#include <math.h>

static color bg_col;
static color fg_col;
static color dt_col;

static time_t clock0;
static time_t clock_sum;

static string sw_fname;
static string sum_fname;

void display_time(window* wp)
{ 

  time_t clock; 
  time(&clock);

  clock -= clock0;

  if (!is_file(sw_fname)) 
  { ofstream out(sum_fname);
    out << (clock+clock_sum+20) << endl;
    exit(0);
   }

  clock_t cl = clock;

  tm* T = localtime(&cl); 
  int seconds = T->tm_sec;
  int minutes = T->tm_min;
  int hours = T->tm_hour - 1;

  clock_t cl1 = cl + clock_sum;

  tm* T1 = localtime(&cl1); 
  int seconds1 = T1->tm_sec;
  int minutes1 = T1->tm_min;
  int hours1 = T1->tm_hour - 1;



  double xc = (wp->xmax() - wp->xmin())/2;
  double yc = (wp->ymax() - wp->ymin())/2;

  double y = yc + wp->pix_to_real(7);

  wp->start_buffering();
  wp->clear();

  wp->set_bold_font();
  wp->draw_ctext(xc,y,string("%2d:%02d:%02d",hours,minutes,seconds),fg_col);

  wp->set_text_font();
  y = yc - wp->pix_to_real(16);
  wp->draw_ctext(xc,y,string("%2d:%02d:%02d",hours1,minutes1,seconds1),dt_col);

  wp->flush_buffer();
  wp->stop_buffering();
}


int main(int argc, char** argv)
{
  bg_col = ivory;
  fg_col = black;
  dt_col = blue;

  int x = 0;
  int y = 0;

  if (argc > 1) x = atoi(argv[1]);
  if (argc > 2) y = atoi(argv[2]);
  if (argc > 3) bg_col = argv[3];
  if (argc > 4) fg_col = argv[4];
  if (argc > 5)
  { cerr << "usage: dclock x y bgcol fgcol" << endl;
    return 1;
   }

/*
  if (x < 0) x = window::screen_width() - 132 + x;
  if (y < 0) y = window::screen_height() - 50 + y;
*/


  window W(132,50);

  W.set_redraw(display_time);
  W.set_bg_color(bg_col);
  //W.set_border_width(0);

  W.load_text_font("lucidasans-12");
  W.load_bold_font("T20");

  W.display(W,x,y);

  char* home = getenv("HOME");
  if (home == 0) home = "/root";

  sw_fname  = string("%s/.stop_watch",home);
  sum_fname = string("%s/.stop_sum",home);

cout << sw_fname << endl;
cout << sum_fname << endl;

  ifstream in(sum_fname);
  in >> clock_sum;

  ofstream out(sw_fname);

  time(&clock0);

  out << clock0 << endl;
  

  display_time(&W);

  W.start_timer(1000,display_time);

  W.read_mouse();

  return 0;
}
