/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  mcf_rand.c
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <LEDA/graph/graph.h>
#include <LEDA/graph/graph_alg.h>
#include <LEDA/graph/dimacs.h>
#include <LEDA/system/assert.h>

using namespace leda;

using std::cout;
using std::cerr;
using std::endl;


int main(int argc, char** argv)
{
  if (argc < 3)
  { cerr << "usage: " << argv[0] << " n m [seed] [U] [C]" << endl;
    return 1;
   }

  graph G;

  int n = string(argv[1]).atoi() + 2; // s and t
  int m = string(argv[2]).atoi();

  int U = 1000;
  int C = 1000;

  int seed = n+m+17;

  if (argc > 3) seed = string(argv[3]).atoi();
  if (argc > 4) U = string(argv[4]).atoi();
  if (argc > 5) C = string(argv[5]).atoi();

  rand_int.set_seed(seed);

  random_graph(G,n,m);
  Delete_Loops(G);
  Make_Simple(G);

  m = G.number_of_edges();

  edge_array<int> cap(G);
  edge_array<int> cost(G);
  edge_array<int> lcap(G);
  edge_array<int> flow(G);

  node_array<int> supply(G,0);

  node s = 0;
  node t = 0;

  int max_out = 0;
  int max_in = 0;

  node v;
  forall_nodes(v,G)
  { int out = G.outdeg(v);
    if (out > max_out) { s = v; max_out = out; }  
   }

  forall_rev_nodes(v,G)
  { if (v == s) continue;
    int in = G.indeg(v);
    if (in > max_in) { t = v; max_in = in; }  
   }

  assert(s != 0);
  assert(t != 0);
  assert(s != t);
  
  
  edge e;
  forall_edges(e,G) 
  { cap[e]  = rand_int(U/2,U);
    cost[e] = rand_int(C/3,C);
   }

   MAX_FLOW(G,s,t,cap,flow) ;

  forall_edges(e,G) lcap[e]  = flow[e]/2;

  forall_out_edges(e,s) 
  { supply[target(e)] += flow[e];
    lcap[e] = 0;
   }

  forall_in_edges(e,t) 
  { supply[source(e)] -= flow[e];
    lcap[e] = 0;
   }

  G.del_node(s);
  G.del_node(t);

  Write_Dimacs_MCF(cout,G,supply,lcap,cap,cost);

  return 0;
}

