#include <LEDA/graph/graph.h>
#include <LEDA/graph/graph_alg.h>
#include <LEDA/graph/dimacs_graph.h>

using namespace leda;

using std::cin;
using std::cout;
using std::cerr;
using std::flush;
using std::endl;

int main(int argc, char** argv)
{

  graph G;
  node_array<int> supply(G);
  edge_array<int> lcap(G);
  edge_array<int> ucap(G);
  edge_array<int> cost(G);

  dimacs_mcf D(cin);
  D.translate(G,supply,lcap,ucap,cost);

  edge_array<int> flow(G);
  double n = G.number_of_nodes();
  double m = G.number_of_edges();

  cout << endl;
  cout << string("%6.0f nodes",n) << endl;
  cout << string("%6.0f edges",m) << endl;
  cout << endl;

  float T = used_time();

  bool feasible = MIN_COST_FLOW(G,lcap,ucap,cost,supply,flow);

  double c = 0;
  edge e;
  forall_edges(e,G) c += flow[e]*cost[e];

  if (feasible)
    cout << string("time: %5.2f cost: %12.0f", used_time(T), c) <<endl;
  else
    cout << "infeasible MCF-problem." << endl;

  return 0;
}

