#define LEDA_CHECKING_OFF

#define SWAP_ARRAYS
#define SORT_SMALLER

#include <LEDA/graph/graph.h>
#include <LEDA/graph/dimacs_graph.h>

class opposite_graph {};

#include "incl/mcf_cost_scaling.h"

using namespace leda;

using std::cin;
using std::cout;
using std::cerr;
using std::flush;
using std::endl;

int main(int argc, char** argv)
{
  graph G;

  node_array<int,graph>  supply(G);
  edge_array<int,graph>  lcap(G);

  edge_array<int,graph> ucap(G);
  edge_array<int,graph> cost(G);

  typedef node_array<int,   graph> excess_array;
  typedef node_array<double,graph> pot_array; 
  typedef node_array<int,   graph> dist_array;
  typedef node_array<node,  graph> succ_array; 
  typedef node_array<node,  graph> pred_array; 

  dimacs_mcf D(cin);
  D.translate(G,supply,lcap,ucap,cost);

  edge_array<int,graph> flow(G);

  double n = G.number_of_nodes();
  double m = G.number_of_edges();

  cout << endl;
  cout << string("%6.0f nodes",n) << endl;
  cout << string("%6.0f edges",m) << endl;
  cout << endl;


  mcf_cost_scaling<int,graph,pot_array,excess_array,
                                dist_array,succ_array,pred_array> mcf;



  float f = float(0.1);
 
  // a = log(n^2/m) - 1

  int a = 0;
  for(int d = int(n*n/m); d > 1; d /= 2) a++;  
  if (a < 3) a = 12;

  int   b = a;

  bool do_check = false;
  bool dimacs_statistics = false;


  int i = 1;
  while (i < argc)
  { string s = argv[i];
    if (s[0] != '-') break;
    if (s == "-l") mcf.set_write_log(true);
    if (s == "-c") do_check = true;
    if (s == "-s") dimacs_statistics = true;
    i++;
   }

  if (i < argc) a = b = string(argv[i++]).atoi();
  if (i < argc) f = (float)string(argv[i++]).atof();
  if (i < argc) b = string(argv[i++]).atoi();


  if (dimacs_statistics)
  { D.print_statistics(cout);
    cout << endl;
   }

  
  bool feasible = mcf.run(G,lcap,ucap,cost,supply,flow,f,a,b) ;
     
  double c = mcf.total_cost(G,cost,flow);
     
  if (feasible)
  { cout << string("time: %6.2f cost: %14.0f  a = %d  af = %.3f (%.2f) rs = %d",
                                            mcf.cpu_time(), c,a,f,
                                            (mcf.af()/(1<<mcf.restarts())),
                                            mcf.restarts()) << endl;
     mcf.statistics(cout);

     if (do_check)
     { float T = used_time();
       string msg;
       bool b = mcf.check(G,lcap,ucap,cost,supply,flow,msg);
       cout << string("checking time: %.2f sec",used_time(T)) << endl;
       if (!b) 
       { cerr << "mcf::check failed: " + msg << endl;
         return 2;
        }
      }
     return 0;
   }

  cout << endl;
  cout << "Infeasible MCF-Problem." << endl;
  cout << endl;
  return 1;
}

