#include <LEDA/graph/min_cut.h>

#include <LEDA/graph/node_pq.h>
#include <LEDA/core/p_queue.h>
#include <LEDA/core/map.h>
#include <LEDA/system/assert.h>

using namespace leda;
using std::cout;
using std::endl;
using std::flush;


  
static void combine_s_and_t(GRAPH<list<node>,int>& G, 
                            node_array<edge>& s_edge, node s, node t)
{ G[s].conc(G[t]);

  edge e;
  forall_adj_edges(e,s) s_edge[G.opposite(s,e)] =  e; 

  //assert(s_edge[t] != nil);


  forall_adj_edges(e,t)
  { node v = G.opposite(t,e);
    if ( v == s) continue;
    edge x = s_edge[v];
    if (x == nil)
       G.move_edge(e,s,v);
    else 
       G[x] += G[e]; 
  }

  node u;
  forall_adj_nodes(u,s)  s_edge[u] =  nil; 
 
  G.del_node(t); 
}

  
int min_cut(const graph& G0, const edge_array<int>& weight, 
                             list<node>& C, bool use_heuristic) 
{ 
  edge e;
  forall_edges(e,G0) 
    if ( weight[e] < 0 ) 
      LEDA_EXCEPTION(1,"MIN_CUT: no negative weights");

  GRAPH<list<node>,int> G;
  G.make_undirected();
    
  node_array<node> partner(G0);
  node v; 
  forall_nodes(v,G0)
  { node u = G.new_node(); 
    partner[v] = u;
    G[u].append(v); 
  }

  forall_edges(e,G0) 
  { node v = source(e);
    node w = target(e); 
    if (v != w) G.new_edge(partner[v], partner[w],weight[e]); 
   }

  node_array<edge> s_edge(G,nil);

  node a = G.first_node();   
  int best_value = MAXINT; 
  int cut_weight = MAXINT; 

  int hcount1 = 0;
  int hcount2 = 0;

  while ( G.number_of_nodes() >= 2 ) 
  { node t = a; 
    node s = nil; 

    node_pq<int> PQ(G); 

    node v;
    forall_nodes(v,G) 
    { if (v == a) continue; 
      PQ.insert(v,0); 
     }
                                        
    edge e;
    forall_adj_edges(e,a)
    { node b = G.opposite(a,e);
      int p = PQ.prio(b) - G[e];
      PQ.decrease_inf(b,p); 
     }
                                           
    while (!PQ.empty())
    { s = t; 
      t = PQ.find_min();
      cut_weight =  -PQ.prio(t); 
      PQ.del_min(); 
      forall_adj_edges(e,t)
      { node v = G.opposite(t,e); 
        if (PQ.member(v)) PQ.decrease_p(v,PQ.prio(v) - G[e]); 
       }
     }

    bool new_best_cut = false;
    if ( cut_weight < best_value )
    { C = G[t]; 
      best_value = cut_weight; 
      new_best_cut = true;
cout << string("new_best_cut: n = %4d c = %4d  ",
                       G.number_of_nodes(),best_value) << flush;
     }

    combine_s_and_t(G,s_edge,s,t); 

    if (use_heuristic)
    { 
      //if (!new_best_cut)
      { edge e = G.first_adj_edge(s); 
        while (e) 
        { node t = G.opposite(s,e);    
          if (G[e] >= best_value) 
          { combine_s_and_t(G,s_edge,s,t); 
            e = G.first_adj_edge(s);
            hcount1++;
           }  
          else e = G.adj_succ(e,s);
        }
      }
      //else
      if (new_best_cut)
      { edge e = G.first_edge();
        while (e)
        { node s = G.source(e);
          node t = G.target(e);    
          if (G[e] >= best_value) 
          { combine_s_and_t(G,s_edge,s,t); 
            e = G.first_edge();
            hcount2++;
           }  
          else e = G.succ_edge(e);
         }
       }

     }// heuristic

  }

  cout << string("(%d/%d)",hcount1,hcount2) << endl;

  return best_value; 
}




static void combine_s_and_t1(GRAPH<list<node>,int>& G, list<edge>& L,
                             node s, node t)
{ G[s].conc(G[t]);

  node_array<edge> s_edge(G,nil); 
  edge e;
  forall_adj_edges(e,s)  s_edge[G.opposite(s,e)] =  e; 
  forall_adj_edges(e,t)
  { node v = G.opposite(t,e);
    if ( v == s) { /* G.del_edge(e); */ continue; }
    edge x = s_edge[v];
    if (x == nil)
       G.move_edge(e,s,v);
    else 
      { G[x] += G[e]; 
        // G.del_edge(e);
        L.append(x);
       }
  }
 
  //assert(G.degree(t) == 0);
  //G.del_node(t); 
}

  
int min_cut1(const graph& G0, const edge_array<int>& weight, 
                             list<node>& C, bool use_heuristic) 
{ 

  edge e;
  forall_edges(e,G0) 
    if ( weight[e] < 0 ) 
      LEDA_EXCEPTION(1,"MIN_CUT: no negative weights");

  GRAPH<list<node>,int> G;
  G.make_undirected();
    
  node_array<node> partner(G0);

  node v; 
  forall_nodes(v,G0)
  { node u = G.new_node(); 
    partner[v] = u; 
    G[u].append(v); 
  }

  forall_edges(e,G0) 
  { node v = source(e);
    node w = target(e);
    if (v == w)  continue;
    G.new_edge(partner[v], partner[w],weight[e]); 
  }

  node a = G.first_node();   
  int best_value = MAXINT; 
  int cut_weight = MAXINT; 

  //while ( G.number_of_nodes() >= 2 ) 

  int vnum = G.number_of_nodes();

  while (vnum >= 2)
  { node t = a; 
    node s = nil; 

    node_array<bool> in_PQ(G,false); 
    node_pq<int> PQ(G); 

    node v;
    forall_nodes(v,G) 
    { if (v == a) continue; 
      //if (G.degree(v)) continue; 
      if (G[v].empty()) continue; 
      PQ.insert(v,0); 
      in_PQ[v] = true; 
     }
                                        
    edge e;
    forall_adj_edges(e,a)
    { node b = G.opposite(a,e);
if (G[b].empty()) continue;
      PQ.decrease_inf(b,PQ.prio(b) - G[e]); 
     }
                                           
    while (!PQ.empty())
    { s = t; 
      int p=0;
      t = PQ.del_min(p); 
      in_PQ[t] = false; 
      cut_weight = -p;
      edge e;
      forall_adj_edges(e,t)
      { node v = G.opposite(t,e);
if (G[v].empty()) continue;
        if (in_PQ[v]) PQ.decrease_p(v,PQ.prio(v) - G[e]); 
       }
     }

    bool new_best_cut = false;
    if ( cut_weight < best_value )
    { C = G[t]; 
      best_value = cut_weight; 
      new_best_cut = true;
     }


    list<edge> L;
    combine_s_and_t1(G,L,s,t); 
    vnum--;

    if ( use_heuristic ) 
    { 
      list<edge> L = G.adj_edges(s);
      while (!L.empty())
      { edge e = L.pop();
        node t = G.opposite(s,e);    
/*
if (G[s].empty() || G[t].empty()) continue;
*/
        if ( G[e] >= best_value ) combine_s_and_t1(G,L,s,t); 
       }  

      if ( new_best_cut )
      { 
/*
        list<edge> L = G.all_edges();
        while (!L.empty())
        { edge e = L.pop();
          node s = G.source(e);    
          node t = G.target(e);    
if (G[s].empty() || G[t].empty()) continue;
          if ( G[e] >= best_value ) combine_s_and_t1(G,L,s,t); 
         }  
*/

/*
        bool one_more_round = true;
        while ( one_more_round )
        { one_more_round = false;
          forall_edges(e,G) 
          { node s = G.source(e);
            node t = G.target(e);    
            if ( G[e] >= best_value ) 
            { combine_s_and_t1(G,s,t); 
              one_more_round = true; 
              break; 
             }
           }  
         }
*/
       }
     }
   }

  return best_value; 
}



static void combine_s_and_t2(GRAPH<list<node>,int>& G, 
                             node_array<edge>& s_edge, 
                             list<edge>& L, node s, node t)
{ G[s].conc(G[t]);

  edge e;
  forall_adj_edges(e,s)  
  { node u = G.opposite(s,e);;
    s_edge[u] =  e; 
   }

  forall_adj_edges(e,t)
  { node v = G.opposite(t,e);
    if ( v == s) continue;
    edge x = s_edge[v];
    if (x == nil) 
       G.new_edge(s,v,G[e]); 
    else 
       { G[x] += G[e]; 
         L.append(x);
         assert(s == source(x) || s == target(x));
       }
  }

  forall_adj_edges(e,s)  
  { node u = G.opposite(s,e);
    s_edge[u] =  nil; 
   }
 
  G.del_node(t); 
}

  
int min_cut2(const graph& G0, const edge_array<int>& weight, 
                             list<node>& C, bool use_heuristic) 
{ 

  edge e;
  forall_edges(e,G0) 
    if ( weight[e] < 0 ) 
      LEDA_EXCEPTION(1,"MIN_CUT: no negative weights");

  GRAPH<list<node>,int> G;
  G.make_undirected();
    
  node_array<node> partner(G0);
  node v; 
  forall_nodes(v,G0)
  { partner[v] = G.new_node(); 
    G[partner[v]].append(v); 
  }

  forall_edges(e,G0) 
  if ( source(e) != target(e) ) 
    G.new_edge(partner[source(e)], partner[target(e)],weight[e]); 

  node_array<edge> s_edge(G,nil);

  node a = G.first_node();   
  int best_value = MAXINT; 
  int cut_weight = MAXINT; 


  while ( G.number_of_nodes() >= 2 ) 
  { node t = a; 
    node s = nil; 
    node_array<bool> in_PQ(G,false); 

    node_pq<int> PQ(G); 

    node v;
    forall_nodes(v,G) 
    { if (v == a) continue; 
      PQ.insert(v,0); 
      in_PQ[v] = true; 
     }
                                        
    edge e;
    forall_adj_edges(e,a)
    { node b = G.opposite(a,e);
      int p = PQ.prio(b) - G[e];
      PQ.decrease_inf(b,p); 
     }
                                           
    while (!PQ.empty())
    { s = t; 
      cut_weight =  -PQ.prio(PQ.find_min()); 
      t = PQ.del_min(); 
      in_PQ[t] = false; 
      forall_adj_edges(e,t)
      { if (in_PQ[v = G.opposite(t,e)]) 
           PQ.decrease_p(v,PQ.prio(v) - G[e]); 
       }
     }

    bool new_best_cut = false;
    if ( cut_weight < best_value )
    { C = G[t]; 
      best_value = cut_weight; 
      new_best_cut = true;
     }

    list<edge> L;
    combine_s_and_t2(G,s_edge,L,s,t); 

    if ( use_heuristic ) 
    { 
      while (!L.empty())
      { edge e = L.pop();
        node v = source(e);
        node w = target(e);    
        if (v != s) swap(v,w);
        assert(v == s);
        if ( G[e] >= best_value ) combine_s_and_t2(G,s_edge,L,v,w); 
       }
/*
      if ( new_best_cut )
      { L = G.all_edges();
        while (!L.empty())
        { edge e = L.pop();
          node s = G.source(e);
          node t = G.target(e);    
          if ( G[e] >= best_value ) combine_s_and_t2(G,s_edge,L,s,t); 
         }
       }
*/
     }
   }

  return best_value; 
}



int main()
{
  cout << endl;
  int nn = read_int("n = ");

  int c_min = read_int("c_min = ");
  int c_max = read_int("c_max = ");

  int j_min = 2;
  int j_max = 16;
  int j_step = 2;

  cout << endl;
  for (int j = j_min; j <= j_max; j+=j_step) cout << string("%10d",j);
  cout << endl; 
  cout << endl; 

  for (int i = 1; i <= 8; i++)
  { 
    cout << string("%d  ",i);
cout << endl;

    int n = i*nn;
    
    for (int j = j_min; j <= j_max; j+=j_step)
    { 
      int m = j*n;
      graph G;
      random_simple_undirected_graph(G,n,m);

      edge_array<int> weight(G);
      edge e;
      forall_edges(e,G) weight[e] = rand_int(c_min,c_max);
  
      float T = used_time(); 
  
      list<node> C;
      int cf = MIN_CUT(G,weight,C,true);
      cout << string("%5.2f/",used_time(T)) << flush;
  
/*
      int ct = MIN_CUT(G,weight,C,true);
      cout << string("%.2f/",used_time(T)) << flush;
*/

      int ct = min_cut(G,weight,C,true);
      //cout << string("%.2f ",used_time(T)) << flush;
  
      assert(ct == cf);
    }
    cout << endl;

  }

return 0;
}


