/*******************************************************************************
+
+               phase 1 parallel mit jeweils einer fifo_set für jeden Thread
+ 
*******************************************************************************/


#include <LEDA/graph/graph_alg.h>
#include <LEDA/graph/edge_array.h>
#include <LEDA/graph/graph.h>
#include <LEDA/core/queue.h>
#include <LEDA/core/string.h>
#include <LEDA/system/assert.h>
#include <LEDA/graph/dimacs_graph.h>
#include <LEDA/graph/max_flow.h>
#include <LEDA/graph/templates/max_flow.h> 
#include <LEDA/system/multi_thread.h>
#include <LEDA/system/thread.h>
#include <LEDA/system/timer.h>
#include <LEDA/core/b_queue.h>
#include <LEDA/core/array.h>
#include <LEDA/core/list.h>
#include <iostream>
#include <fstream>


using namespace leda;
using std::cout;
using std::cerr;
using std::cin;
using std::endl;


class fifo_set{

  list<node> L;

public:

  fifo_set(){}

  node del() { if (!L.empty()) return L.pop(); else return nil; }
 
  void insert(node v, int)  { L.append(v); }
  void insert0(node v, int) { L.append(v); }                            
        void split0(int& id, array<fifo_set>& F, node_array<int>& dist)
        {
                node v;
                forall(v,L)
                {
                        F[id].insert(v, dist[v]);
                }
        }
        void split(int& thr_num, array<fifo_set>& F, node_array<int>& dist)
        {
                int g(0);
                node v;
                forall(v,L)
                {
                        int number = g % thr_num;
                        F[number].insert(v, dist[v]);
                        g++;
                }
        } 

  int size() { return L.size(); }

  bool empty() { return L.empty(); }

  void clear() { L.clear(); }

  ~fifo_set(){}
};


class hl_set{

  int max, max_lev;
  array<list<node> > A;

public:

  hl_set(int max_level):A(max_level+1)
  { max = -1; max_lev = max_level;}

  node del()
  { while (max >= 0 && A[max].empty()) max--;
    if (max >= 0) return A[max].pop(); else return nil; 
  }
 
  void insert(node v, int d)
  { A[d].push(v);
    if (d > max) max = d;
  }

  void insert0(node v, int d) { A[d].append(v); }

  bool empty()
  { while (max >= 0 && A[max].empty()) max--;
    return ( max < 0 ); 
  }

  ~hl_set(){}

  void clear()
  { for (int i = 0; i <= max_lev; i++) A[i].clear();  
    max = -1;
  }
};



struct thread_data
{
                graph* G;
                node* s;
                node* t;
                edge_array<int>* cap;
                edge_array<int>* flow;
                node_array<int>* dist;
                node_array<int>* excess;
    array<fifo_set>* liste;
    int* num_pushes; 
    int* num_edge_inspections; 
    int* num_relabels; 
    int* num_global_relabels; 
    float* h; 
                int* heuristic;
                int* limit_heur;
                b_queue<node>* Q;
                node_array<leda_mutex*>* lock;
                int id; 
                bool local_relabel;
                bool knappe_locks;
                node_array<string>* name;
};

bool getlock(node v, node_array<leda_mutex*>& lock)
{
        if (lock[v]->trylock()==0) return true;
        else return false;
} 

bool getlocks(node v, node w, node_array<leda_mutex*>& lock)
{
        bool b(0);
        if (lock[v]->trylock()==0)
        {
                if (lock[w]->trylock()==0)      b=1;
                else lock[v]->unlock();
        }
        return b;
} 

inline void compute_dist_t(const graph& G, node t, const edge_array<int>& flow, 
                           const edge_array<int>& cap, 
                           const node_array<int>& excess, node_array<int>& dist, 
                           fifo_set& U, b_queue<node>& Q)
{ 
  int n = G.number_of_nodes();
  Q.append(t);

  dist[t] = 0;  

  while ( !Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    if ( d >= n) break;
    edge e;

    for(e = G.first_adj_edge(v); e; e = G.adj_succ(e))
    { if ( flow[e] == 0 ) continue;
      node u = target(e); 
      int& du = dist[u];
      if ( du >= n ) 
      { du = d;
        Q.append(u); 
        if ( excess[u] > 0 ) U.insert(u,d); 
       }
     }

    for(e = G.first_in_edge(v); e; e = G.in_succ(e))
    { if ( cap[e] == flow[e] ) continue;
      node u = source(e); 
      int& du = dist[u];
      if ( du >= n )  
      { du = d;  
        Q.append(u); 
        if (excess[u] > 0) U.insert(u,d); 
      }
    }
  }
}



void maxflow(const graph& G, node s, node t, const node_array<string>& name,
                                                                        const edge_array<int>& cap, edge_array<int>& flow, node_array<int>& dist, node_array<int>& excess,
                  array<fifo_set>& U, int& num_pushes, int& num_edge_inspections,
                  int& num_relabels, int& num_global_relabels, float h, int& heuristic, int limit_heur, b_queue<node>& Q, 
                                                                        node_array<leda_mutex*>& lock, int id, bool knappe_locks, bool local_relabel)
{   
        int n = G.number_of_nodes();  

  for(;;) 
  {
    node v = U[id].del();

    if (v == nil) break;
    if (dist[v] >= n) continue;
    if (v == t) continue;
    if ( excess[v] <= 0 ) continue;

    if(knappe_locks==0)
    { if(getlock(v, lock)==0) 
      { U[id].insert(v, dist[v]);
        // wenn Knoten v gelockt, fuege v in die Liste ein und 
        // betrachte ihn spaeer
        continue;
       }

        int  ev   = excess[v];
        int dv   = dist[v];   
        int dmin = MAXINT;
        edge e; 

        if ( ev > 0 )
        { 
          for (e = G.first_adj_edge(v); e; e = G.adj_succ(e))
          { Fetch_And_Increment(&num_edge_inspections);
            int& fe = flow[e];
            int  rc = cap[e] - fe;
            if (rc == 0) continue;
                                        if(getlock(target(e), lock)==0) continue;
            node w = target(e);
            int dw = dist[w];
            if ( dw == dv - 1 ) 
            { Fetch_And_Increment(&num_pushes);
              int& ew = excess[w];
              if (ew == 0) U[id].insert0(w,dw); 
              if (ev <= rc) 
              { ew += ev; fe += ev;
                ev = 0; 
                                                        lock[w]->unlock(); 
                break;
              }
              else
              { ew += rc; fe += rc;
                ev -= rc;
              }
            }
            else { if ( dw < dmin ) dmin = dw; }
                                        lock[w]->unlock();
          }
                        }
        if ( ev > 0 )
        { 
          for (e = G.first_in_edge(v); e; e = G.in_succ(e))
          { Fetch_And_Increment(&num_edge_inspections);
            int& fe = flow[e];
            if (fe == 0) continue;
                                        if(getlock(source(e), lock)==0) continue;
            node w = source(e);
            int dw = dist[w];
            if ( dw == dv - 1 ) 
            { num_pushes++;
              int& ew = excess[w];
              if (ew == 0) U[id].insert0(w,dw); 
              if (ev <= fe) 
              { fe -= ev; ew += ev;
                ev = 0; 
                                                        lock[w]->unlock();
                break;
              }
              else
              { ew += fe; ev -= fe;
                fe = 0;
              }
            }
            else { if ( dw < dmin ) dmin = dw; }
                                        lock[w]->unlock();
          }
                        }
        
        excess[v] = ev;
        
        if (ev > 0 && local_relabel) 
        { 
          if (num_edge_inspections <= limit_heur)
          { 
                dmin++; Fetch_And_Increment(&num_relabels);
            if ( dmin < n)
            { dist[v] = dmin;
              U[id].insert(v,dmin);
            }
                                }
          else 
                                {
                                        lock[v]->unlock();
                                        break;
                                }
                        }
                        lock[v]->unlock();
                }
                else
                {
        int dmin = MAXINT;
        edge e; 

        if ( excess[v] > 0 )
        { 
          for (e = G.first_adj_edge(v); e; e = G.adj_succ(e))
          { Fetch_And_Increment(&num_edge_inspections);
            if (cap[e] - flow[e] == 0) continue;
                                        if(getlocks(v, target(e), lock)==0) continue;
            node w = target(e);
            int dv = dist[v];
            int dw = dist[w];
            if ( dw == dv - 1 ) 
            { Fetch_And_Increment(&num_pushes);
                int& fe = flow[e];
              int& ev = excess[v];
                int  rc = cap[e] - flow[e];
              int& ew = excess[w];
              if (ew == 0) U[id].insert0(w,dw); 
              if (ev <= rc) 
              { ew += ev; fe += ev;
                ev = 0; 
                                                        lock[v]->unlock();
                                                        lock[w]->unlock(); 
                break;
              }
              else
              { ew += rc; fe += rc;
                ev -= rc;
              }
            }
            else { if ( dw < dmin ) dmin = dw; }
                                        lock[v]->unlock();
                                        lock[w]->unlock();
          }
                        }
        if ( excess[v] > 0 )
        { 
          for (e = G.first_in_edge(v); e; e = G.in_succ(e))
          { Fetch_And_Increment(&num_edge_inspections);
            if (flow[e] == 0) continue;
                                        if(getlocks(v, source(e), lock)==0) continue;
            node w = source(e);
            int dv = dist[v];
            int dw = dist[w];
            if ( dw == dv - 1 ) 
            { num_pushes++;
                int& fe = flow[e];
              int& ev = excess[v];
              int& ew = excess[w];
              if (ew == 0) U[id].insert0(w,dw); 
              if (ev <= fe) 
              { fe -= ev; ew += ev;
                ev = 0; 
                                                        lock[v]->unlock();
                                                        lock[w]->unlock();
                break;
              }
              else
              { ew += fe; ev -= fe;
                fe = 0;
              }
            }
            else { if ( dw < dmin ) dmin = dw; }
                                        lock[v]->unlock();
                                        lock[w]->unlock();
          }
                        }
        
        
        if (excess[v] > 0 && local_relabel) 
        { 
          if (num_edge_inspections <= limit_heur)
          { 
                dmin++; Fetch_And_Increment(&num_relabels);
            if ( dmin < n)
            { 
                                                lock[v]->lock();
                                                dist[v] = dmin;
                                                lock[v]->unlock();
              U[id].insert(v,dmin);
            }
                                }
          else 
                                {
                                        break;
                                }
                        }
                }
  }     
}


void* maxflow_thread(void *threadarg)
{ 
        thread_data* my_data = (thread_data *) threadarg;       
        graph& G = *(my_data->G);       
        node& s = *(my_data->s);        
        node& t = *(my_data->t);                
                edge_array<int>& cap = *(my_data->cap);
                edge_array<int>& flow = *(my_data->flow);
                node_array<int>& dist = *(my_data->dist);
                node_array<int>& excess = *(my_data->excess);
        array<fifo_set>& U = *(my_data->liste);         
        int& num_pushes = *(my_data->num_pushes);
        int& num_edge_inspections = *(my_data->num_edge_inspections);
        int& num_relabels = *(my_data->num_relabels);
        int& num_global_relabels = *(my_data->num_global_relabels);
        float& h = *(my_data->h);
        int& heuristic = *(my_data->heuristic);
        int& limit_heur = *(my_data->limit_heur);
        b_queue<node>& Q = *(my_data->Q);
                node_array<leda_mutex*>& lock = *(my_data->lock);
                int& id = my_data->id;
                bool& local_relabel = my_data->local_relabel;
                bool& knappe_locks = my_data->knappe_locks;
        node_array<string>& name = *(my_data->name);    

                maxflow(G, s, t, name, cap, flow, dist, excess, U, num_pushes, num_edge_inspections, 
                                                        num_relabels, num_global_relabels, h, heuristic, limit_heur, Q, lock, id, knappe_locks, local_relabel);
                return 0;
}

int main(int argc, char** argv)
{
  if (argc < 2)
  { cerr << "usage: mfc thrnum" << endl;
    return 1;
   }

  int thr_num = atoi(argv[1]);

  graph G;
  node s;
  node t;
  edge_array<int> cap(G);
  edge_array<int> flow(G);
  int num_pushes, num_edge_inspections, num_relabels, num_global_relabels;

  //int thr_num = 4;

  // Graph einlesen
  read_dimacs_mf(cin,G,s,t,cap);
  if (s == t) LEDA_EXCEPTION(1,"MAXFLOW: source == sink");

  node_array<string> name(G);
  int knoten(0);

  node v;
  forall_nodes(v,G)
  {
    string s("v_%d",knoten);
    name[v] = s;
    knoten++;
  }

/******************* Bitte hier Möglichkeiten wählen!!! **********************/

  bool local_relabel(1); // mit local relabel -> local_relabel = 1; 
  bool knappe_locks(0);  // für knappere locks -> knappe_locks = 1;

  float h = 2;           // Heuristikparameter; 
                         // falls global relabel nur erwünscht,
                         // wenn Listen der Threads leer sind, 
                         // dann setze h = MAXINT;

/*****************************************************************************/

  //for(thr_num=1; thr_num <= 4; thr_num *= 2)
  {
  float t3 = used_time(); 
  float t4 = real_time();

  int n = G.number_of_nodes();
  int m = G.number_of_edges(); 
  fifo_set U; 

  flow.init(G,0);
  node_array<int>  excess(G,0);
  if (G.first_adj_edge(s) == nil) return 0;
 
  node_array<leda_mutex*> lock(G);
  forall_nodes(v,G) lock[v] = new leda_mutex;

  // saturate all edges leaving s

  edge e;
  forall_out_edges(e,s) 
  { int c = cap[e];
    if (c == 0) continue;
    node v = target(e);
    flow[e] = c;
    excess[s] -= c;
    excess[v] += c;
   } 

 
  b_queue<node> Q(n);
  list<node> S;


  int heuristic  = (int) (h*m);
  int limit_heur = heuristic;
  
  node_array<int> dist(G);
  dist.init(G,n);

  compute_dist_t(G,t,flow,cap,excess,dist,U,Q);

  num_relabels = num_pushes = num_edge_inspections = 0;
  num_global_relabels  = 0;


  while(1)
  {

    if(U.size() <= thr_num)
    {
      array<fifo_set> F(thr_num);
      int id = 0;                       
      U.split0(id, F, dist);
      U.clear();
      maxflow(G, s, t, name, cap, flow, dist, excess, F, num_pushes, 
              num_edge_inspections, num_relabels, num_global_relabels, h, 
              heuristic, limit_heur, Q, lock, id, knappe_locks, local_relabel);
                
     }
     else
     {
       array<fifo_set> F(thr_num);
       U.split(thr_num, F, dist);
       U.clear();

       // Threads
       // thread_data thread_data_array[thr_num];

       array<thread_data> thread_data_array(thr_num);
       thread_t* thr = new thread_t[thr_num];


       // Threadargumente
       for(int i=0; i<thr_num; i++) {
        thread_data_array[i].G = &G;
        thread_data_array[i].s = &s;
        thread_data_array[i].t = &t;
        thread_data_array[i].cap = &cap;
        thread_data_array[i].flow = &flow;
        thread_data_array[i].dist = &dist;
        thread_data_array[i].excess = &excess;
        thread_data_array[i].liste = &F;
        thread_data_array[i].num_pushes = &num_pushes;
        thread_data_array[i].num_edge_inspections = &num_edge_inspections;
        thread_data_array[i].num_relabels = &num_relabels;
        thread_data_array[i].num_global_relabels = &num_global_relabels;
        thread_data_array[i].h = &h;
        thread_data_array[i].heuristic = &heuristic;
        thread_data_array[i].limit_heur = &limit_heur;
        thread_data_array[i].Q = &Q;
        thread_data_array[i].lock = &lock;
        thread_data_array[i].id = i;
        thread_data_array[i].local_relabel = local_relabel;
        thread_data_array[i].knappe_locks = knappe_locks;
        thread_data_array[i].name = &name;
       }


      for(int i=0; i<thr_num; i++) 
          thr[i] = thread_create(maxflow_thread,(void*)&thread_data_array[i]);

      for(int i=0; i<thr_num; i++) thread_join(thr[i]);
    }

    if (local_relabel)
    {
      if (num_edge_inspections > limit_heur)
      {
        limit_heur += heuristic;
        U.clear();
        num_global_relabels++;
        dist.init(G,n);
        compute_dist_t(G,t,flow,cap,excess,dist,U,Q);
       }
    }
    else
    {
      if(num_edge_inspections > limit_heur)
      { limit_heur += heuristic;
        U.clear();
       }

     num_global_relabels++;
     dist.init(G,n);
     compute_dist_t(G,t,flow,cap,excess,dist,U,Q);
    }   
                
    if ( U.empty() ) break;
  }


   float time_complete = real_time(t4);

   cout << endl;
   cout << string("threads: %2d h = %d ",thr_num,h) << endl;
   cout << string("flow: %6d ",excess[t]);
   cout << string ("time: %5.2f (%5.2f) sec", time_complete, used_time(t3));
   cout << endl;
   
   int ops = num_relabels + num_pushes + num_edge_inspections;
   
   cout << string("pushes: %7d ",  num_pushes);
   cout << string("relabels: %4d ", num_relabels);
   cout << string("inspects: %7d ", num_edge_inspections);
   cout << string("updates: %2d ",  num_global_relabels);
   cout << string("ops/sec: %6d ",  int(ops/time_complete)) << endl;
   
   
   forall_nodes(v,G) delete lock[v]; 
  }

  return 0;
}
