#define LEDA_CHECKING_OFF

#include <LEDA/graph/graph.h>
#include <LEDA/graph/static_graph.h>
#include <LEDA/graph/node_slot.h>
#include <LEDA/core/b_stack.h>
#include <LEDA/core/array.h>
#include <LEDA/core/tuple.h>
#include <assert.h>

#if !defined(forall_rev_out_edges)
#define forall_rev_out_edges(e,v) \
for(void* loop_stop = ((e = (v+1)->first_out_edge()-1), v->first_out_edge()-1);\
    e != loop_stop; e = e-1)
#endif



using namespace leda;

#include "bicomp.h"



using std::cout;
using std::flush;
using std::endl;


template <class st_graph>
void construct_static_graph(st_graph& Gs, const graph& G) 
{ 
  // constructs a static copy of G

  typedef typename st_graph::node st_node;
  typedef typename st_graph::edge st_edge;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  Gs.start_construction(n,m);
  node_array<st_node> N(G);

  node v;
  forall_nodes(v,G) N[v] = Gs.new_node();

  forall_nodes(v,G)
  { st_node x = N[v];
    for (edge e = G.first_out_edge(v); e; e = G.next_out_edge(e))
    { st_node y = N[G.target(e)]; 
      Gs.new_edge(x,y);
     }
   }
  Gs.finish_construction();
}


void acyclic_dense_graph(graph& G, int n)
{ array<node> V(n);
  for(int i=0; i<n; i++) V[i] = G.new_node();

  for(int i=0; i<n; i++) 
     for(int j=i+1; j<n; j++) 
        G.new_edge(V[i],V[j]);
}
  

void SCC_generator1(graph& G, int n, int m)
{ 
  array<node> V(n);

  for (int i = 0; i < n; i++) V[i] = G.new_node();

  int k = m/n;

  for (int i = 1; i < n; i++) G.new_edge(V[0],V[i]);

  for (int i = 1; i <= k+1; i++) G.new_edge(V[i],V[0]);

  for (int i = k+2; i < n; i++) 
    for (int j = 0; j < k; j++) 
       G.new_edge(V[i],V[j + 1]);
}

void SCC_generator2(graph& G, int n, int m)
{ 
  array<node> V(n);

  for (int i = 0; i < n; i++) V[i] = G.new_node();

  int k = m/n;

  for (int i = 1; i < n; i++) G.new_edge(V[0],V[i]);

  for (int i = 1; i <= k+1; i++) G.new_edge(V[i],V[0]);

  for (int i = k+2; i < n; i++) 
    for (int j = 0; j < k; j++) 
       G.new_edge(V[i],V[k - j]);
}



int main(int argc, char** argv) 
{

  int d = 10;

  if (argc > 1) d = atoi(argv[1]);



/*
int m = 1 << 23;   // 8 Mio

if (argc > 1) m = atoi(argv[1]);

cout << endl;
cout << "m = " << m << endl;
cout << endl;
*/

cout << "  m       n      Bi0   Bi1   Bi2   Bi3   Bi4" << endl;

/*
for (double f = m;  f*(f-1) >= m;  f/=2) 
{
    int n = int(f);
*/


for(int n = 100000; n < 2000000; n += 100000)
{ 
  int m = d*n;

  graph GG;
  random_graph(GG,n,m);


  typedef static_graph<opposite_graph,data_slots<1>,data_slots<1> > st_graph;
  typedef st_graph::node node;
  typedef st_graph::edge edge;

  st_graph G;
  construct_static_graph(G,GG);

  typedef node_slot<int,st_graph,0> node_int_array;


  edge_array<int,st_graph> compnum0(G);   // for comparing results
  BICONNECTED_COMPONENTS0<st_graph,node_int_array> bcc0;
  int bcc_num = bcc0(G,compnum0);

  edge_slot<int,st_graph,0> compnum(G);

  cout << string("%7d %7d ",m,n) << flush;

{
  BICONNECTED_COMPONENTS0<st_graph,node_int_array> bcc;
  int s = bcc(G,compnum);
  cout << string("%5.2f ", bcc.cpu_time()) << flush;
  edge e;
  forall_edges(e,G) assert(compnum0[e] == compnum[e]);
  assert(s == bcc_num);
}

{
  BICONNECTED_COMPONENTS1<st_graph,node_int_array>     bcc;
  int s = bcc(G,compnum);
  cout << string("%5.2f ", bcc.cpu_time()) << flush;
/*
  edge e;
  forall_edges(e,G) assert(compnum0[e] == compnum[e]);
*/
  //cout << "bcc_num = " << bcc_num << "   " << "s = " << s << endl;
  assert(s == bcc_num);
}

{
  BICONNECTED_COMPONENTS2<st_graph,node_int_array>     bcc;
  int s = bcc(G,compnum);
  cout << string("%5.2f ", bcc.cpu_time()) << flush;
/*
  edge e;
  forall_edges(e,G) assert(compnum0[e] == compnum[e]);
*/
  //cout << "bcc_num = " << bcc_num << "   " << "s = " << s << endl;
  assert(s == bcc_num);
}

{
  BICONNECTED_COMPONENTS3<st_graph,node_int_array>  bcc;
  int s = bcc(G,compnum);
  cout << string("%5.2f ", bcc.cpu_time()) << flush;
/*
  edge e;
  forall_edges(e,G) assert(compnum0[e] == compnum[e]);
*/
  //cout << "bcc_num = " << bcc_num << "   " << "s = " << s << endl;
  assert(s == bcc_num);
}


/*
{
  BICONNECTED_COMPONENTS4<st_graph,node_int_array>    bcc;
  int s = bcc(G,compnum);
  cout << string("%5.2f ", bcc.cpu_time()) << flush;
  edge e;
  forall_edges(e,G) assert(compnum0[e] == compnum[e]);
  assert(s == bcc_num);
}
*/

  cout << endl;

}

  return 0;
}


