#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/strong_components.hpp>
#include <iostream>
#include <time.h>

#define BOOST

using namespace boost;


#define forall_nodes(v, G) \
for (std::pair<node,node>  __p = vertices(G); \
__p.first != __p.second ? (v = __p.first, true) : false;  ++__p.first)

#define forall_out_edges(e,v) \
for (std::pair<edge,edge> __p = out_edges(*v,G); \
__p.first != __p.second ? (e = __p.first, true) : false;  ++__p.first)


#define forall_rev_out_edges(e,v) \
for (std::pair<edge,edge> pp = out_edges(*v,G); \
pp.first != pp.second; pp.first = pp.second ) \
for (edge p = pp.second-1, p_stop = pp.first-1; \
p != p_stop ? (e = p, true) : false;  --p)



float used_time() { return clock() * 1e-6; }

float used_time(float& T)
{ float t = T;
  T = used_time();
  t = T-t;
  return t > 0 ? t : 0;
}






template <typename Graph>
void acyclic_dense_graph(Graph& G, int n)
{
  typedef graph_traits<Graph>  traits;
  typedef typename traits::vertex_descriptor   vertex;

  std::vector<vertex> V(n);

  int m = n * (n - 1) / 2;

  for (int i = 0; i < n; ++i) V[i] = add_vertex(G);

  for (int s = 0; s < n - 1; ++s)
    for (int t = s + 1; t < n; ++t)
       add_edge(V[s], V[t], G);
} 


template <typename Graph>
void random_graph(Graph& G, int n, int m)
{
  typedef graph_traits<Graph> traits;
  typedef typename traits::vertex_descriptor vertex;

  std::vector<vertex> V(n);
  std::vector<int>  deg(n);

  G.clear();

  for (int i = 0; i < n; i++)
  { V[i] = add_vertex(G);
    deg[i] = 0;
   }

  for (int i = 0; i < m; i++) deg[random()%n]++;

  for (int i = 0; i < n; i++) 
    for(int d = deg[i]; d > 0; d--)
       { int j = random() % n;
         /*if (i != j) */ add_edge(V[i], V[random()%n], G);
        }
} 


//node and edge arrays

template <class T, class graph_t> 
class node_array {
typedef typename graph_t::node node;

std::vector<T>  vec;

public:

node_array(const graph_t& G) { vec.resize(num_vertices(G)); }

node_array(const graph_t& G, const T& x) 
{ int n = num_vertices(G);
  //vec = new T[n];
  vec.resize(n);
  for(int i=0; i<n; i++) vec[i] = x; 
}

~node_array() { }

typename std::vector<T>::iterator begin() { return vec.begin(); }

T& operator[](node v) { return vec[*v]; }
const T& operator[](node v) const { return vec[*v]; }

};


template <class T, class graph_t> 
class edge_array {
typedef typename graph_t::edge edge;

std::vector<T>  vec;

public:

edge_array(const graph_t& G) { vec.resize(num_edges(G)); }

edge_array(const graph_t& G, const T& x) 
{ int n = num_vertices(G);
  vec = new T[n];
  for(int i=0; i<n; i++) vec[i] = x; 
}

~edge_array() { }

typename std::vector<T>::iterator begin() { return vec.begin(); }

T& operator[](edge e) { return vec[*e]; }
const T& operator[](edge e) const { return vec[*e]; }

};



//bounded stacks

template<class E> class b_stack
{
  E* first;
  E* stop;
  E* free;

public:

  b_stack(int n)
  { free = first = new E[n];
    stop = first + n;
   }
  
 ~b_stack() { delete [] first; }
  
  E&       top()       { return *(free-1); }
  const E& top() const { return *(free-1); }
  const E& pop()       { return *--free; }
  
  void  push(const E& x) { *free++ = x; }
  
  void  clear()       { free = first; }
  int   size()  const { return int(free - first); }
  bool  empty() const { return free == first; }
};




template <class A, class B>
class two_tuple
{ A a;
  B b;

public:

  two_tuple() {}
  two_tuple(const A& u, const B& v): a(u), b(v) { }
  
  A& first() { return a;} 
  B& second() { return b;}

  const A&  first() const { return a;} 
  const B&  second() const { return b;} 
};





template <class Graph>
class boost_graph : public Graph
{
public:

 typedef graph_traits <Graph> traits;
 typedef typename traits::vertex_iterator   node;
 typedef typename traits::out_edge_iterator edge;

 node target(edge e)   const { return ::target(*e,*this);   }
 node stop_node()      const { return node(-1);            }

 edge first_out_edge(node v) const { return  out_edges(*v,*this).first; }
 edge stop_out_edge(node v) const { return  out_edges(*v,*this).second; }
 edge next_out_edge(edge e) const { return e+1; }

 int number_of_nodes() const { return num_vertices(*this); }
 int number_of_edges() const { return num_edges(*this);    }
};

namespace boost {

template<typename Graph>
class graph_traits<boost_graph<Graph> >: public graph_traits<Graph> {};

};




